Package: progressr
Version: 0.19.0
Title: An Inclusive, Unifying API for Progress Updates
Description: A minimal, unifying API for scripts and packages to report progress updates from anywhere including when using parallel processing.  The package is designed such that the developer can to focus on what progress should be reported on without having to worry about how to present it.  The end user has full control of how, where, and when to render these progress updates, e.g. in the terminal using utils::txtProgressBar(), cli::cli_progress_bar(), in a graphical user interface using utils::winProgressBar(), tcltk::tkProgressBar() or shiny::withProgress(), via the speakers using beepr::beep(), or on a file system via the size of a file. Anyone can add additional, customized, progression handlers. The 'progressr' package uses R's condition framework for signaling progress updated. Because of this, progress can be reported from almost anywhere in R, e.g. from classical for and while loops, from map-reduce API:s like the lapply() family of functions, 'purrr', 'plyr', and 'foreach'. It will also work with parallel processing via the 'future' framework, e.g. 'lapply(...) |> futurize()' and 'purrr::map(...) |> futurize()', which uses future.apply::future_lapply() and furrr::future_map() internally. The package is compatible with Shiny applications.
Authors@R: c(person("Henrik", "Bengtsson",
                    role = c("aut", "cre", "cph"),
                    email = "henrikb@braju.com",
                    comment = c(ORCID = "0000-0002-7579-5165")))
License: GPL (>= 3)
Depends: R (>= 3.5.0)
Imports: digest, utils
Suggests: graphics, tcltk, beepr, cli, crayon, pbmcapply, progress,
        purrr, foreach, plyr, doFuture, future, future.apply, furrr,
        ntfy, RPushbullet, rstudioapi, shiny, commonmark, base64enc,
        tools
VignetteBuilder: progressr
Language: en-US
Encoding: UTF-8
URL: https://progressr.futureverse.org,
        https://github.com/futureverse/progressr
BugReports: https://github.com/futureverse/progressr/issues
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-03-31 06:41:07 UTC; hb
Author: Henrik Bengtsson [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-7579-5165>)
Maintainer: Henrik Bengtsson <henrikb@braju.com>
Repository: CRAN
Date/Publication: 2026-03-31 08:10:02 UTC
Built: R 4.6.0; ; 2026-04-25 16:00:55 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(conditionMessage,progression)
S3method(length,progressor)
S3method(print,progression)
S3method(print,progression_handler)
S3method(print,progressor)
export(handler_ascii_alert)
export(handler_beepr)
export(handler_brrr)
export(handler_cli)
export(handler_debug)
export(handler_filesize)
export(handler_newline)
export(handler_notifier)
export(handler_ntfy)
export(handler_pbcol)
export(handler_pbmcapply)
export(handler_progress)
export(handler_rpushbullet)
export(handler_rstudio)
export(handler_shiny)
export(handler_slowdown)
export(handler_tkprogressbar)
export(handler_txtprogressbar)
export(handler_void)
export(handler_winprogressbar)
export(handlers)
export(make_progression_handler)
export(progress_aggregator)
export(progress_progressr)
export(progression)
export(progressor)
export(slow_sum)
export(withProgressShiny)
export(with_progress)
export(without_progress)
importFrom(digest,digest)
importFrom(utils,capture.output)
importFrom(utils,file_test)
importFrom(utils,flush.console)
importFrom(utils,object.size)
importFrom(utils,packageVersion)
importFrom(utils,setTxtProgressBar)
importFrom(utils,str)
importFrom(utils,txtProgressBar)

library(progressr)

options(progressr.clear = FALSE)

message("txtprogressbar ...")

for (style in 1:3) {
  message(sprintf("- style = %d ...", style))
  handlers(handler_txtprogressbar(style = style))
  
  for (x in list(integer(0), 1:10, 1L)) {
    message("length(x): ", length(x))
    with_progress({
      progress <- progressor(along = x)
      for (ii in x) {
        Sys.sleep(getOption("progressr.demo.delay", 0.1))
        progress(message = sprintf("(%s)", paste(letters[1:ii], collapse="")))
      }
    })
  }
  
  message(sprintf("- style = %d ... done", style))
}

message("txtprogressbar ... done")

library(progressr)

options(progressr.debug = FALSE)

message("*** with_progress() - partial-line stdout buffering ...")

progressr::handlers("txtprogressbar")

output <- local({
  options(width = 10L)
  bfr <- character(0L)
  con <- textConnection("bfr", open = "w+", local = TRUE)
  on.exit(close(con))
  utils::capture.output({
    utils::capture.output({
      with_progress({
        p <- progressor(1)
        cat("abc\n")
        cat("DEF")
        p()
        cat("ghi\n")
      })
    }, file = con, type = "output")
  }, file = con, type = "message")
  textConnectionValue(con)
})

## Render combined stdout + stderr output
writeLines(output)

print(output)

## Assert that all stdout is in there and at the expected lines
stopifnot(
  grep("\babc\b", output) == 1L,
  grep("\bDEFghi\b", output) == 2L
)

## Emulate what is rendered with CR-wiping progress bar
visual <- strsplit(output, split = "\r")
visual <- vapply(visual, FUN = tail, n = 1L, FUN.VALUE = NA_character_)
print(visual)
stopifnot(
  length(visual) == 2L,
  visual[1] == "abc",
  visual[2] == "DEFghi"
)

message("*** with_progress() - partial-line stdout buffering ... DONE")

## Settings
options(progressr.clear = TRUE)
options(progressr.debug = FALSE)
options(progressr.demo.delay = 0.0)
options(progressr.enable = TRUE)
options(progressr.enable_after = 0.0)
options(progressr.interval = 0.1)
options(progressr.times = +Inf)


options(progressr.tests.fake_handlers = c(non_supported_progression_handlers(), "handler_beepr", "handler_brrr", "handler_notifier", "handler_progress"))

library(progressr)
library(future)
library(graphics)

p <- function(...) NULL

plot_what_is_done <- function(counts) {
  done <- 0L
  
  for (kk in seq_along(counts)) {
    f <- counts[[kk]]

    ## Already plotted?
    if (!inherits(f, "Future")) {
      done <- done + 1L
      next
    }

    ## Not resolved?
    ## NOTE: This will block, if all workers are busy!
    if (runif(1) < 0.8*(1-(done/length(counts))) || !resolved(f)) next

    message(sprintf("Plotting tile #%d of %d ...", kk, n))
    counts[[kk]] <- value(f)
    screen(kk)
    plot(counts[[kk]])

    done <- done + 1L
  }
  
  counts
}


## Options
region <- getOption("future.demo.mandelbrot.region", 1L)
if (!is.list(region)) {
  if (region == 1L) {
    region <- list(xmid = -0.75, ymid = 0.0, side = 3.0)
  } else if (region == 2L) {
    region <- list(xmid = 0.283, ymid = -0.0095, side = 0.00026)
  } else if (region == 3L) {
    region <- list(xmid = 0.282989, ymid = -0.01, side = 3e-8)
  }
}
nrow <- getOption("future.demo.mandelbrot.nrow", 5L)
resolution <- getOption("future.demo.mandelbrot.resolution", 1024L)
delay <- getOption("future.demo.mandelbrot.delay", interactive())
if (isTRUE(delay)) {
  delay <- function(counts) Sys.sleep(runif(1L, min=0.5, max=5))
} else if (!is.function(delay)) {
  delay <- function(counts) {}
}

## Generate Mandelbrot tiles to be computed
Cs <- mandelbrot_tiles(xmid = region$xmid, ymid = region$ymid,
                       side = region$side, nrow = nrow,
                       resolution = resolution)
message("Tiles: ", paste(dim(Cs), collapse = " by "))		       
if (interactive()) {
  dev.new()
  plot.new()
  split.screen(dim(Cs))
  for (ii in seq_along(Cs)) {
    screen(ii)
    par(mar = c(0, 0, 0, 0))
    text(x = 1 / 2, y = 1 / 2, sprintf("Future #%d\nunresolved", ii), cex = 2)
  }
} else {
  split.screen(dim(Cs))
}

## Create all Mandelbrot tiles via lazy futures
n <- length(Cs)
message(sprintf("* Creating %d Mandelbrot tiles", n))
with_progress({
  p <- progressor(along = Cs)
  counts <- lapply(seq_along(Cs), FUN=function(ii) {
    C <- Cs[[ii]]
    future({
      message(sprintf("Calculating tile #%d of %d ...", ii, n), appendLF = FALSE)
      fit <- mandelbrot(C)
  
      ## Emulate slowness
      delay(fit)
  
      p(sprintf("Tile #%d (PID %d)", ii, Sys.getpid()))
      
      message(" done")
      fit
    }, lazy = TRUE)
  })
  str(counts)

  pp <- 0L
  while (any(sapply(counts, FUN = inherits, "Future"))) {
    counts <- plot_what_is_done(counts)
  }
})

  


close.screen()


message("SUGGESTION: Try to rerun this demo after changing strategy for how futures are resolved, e.g. plan(multisession).\n")

library(progressr)

library(future)

supportedStrategies <- function(cores = NA_integer_, excl = "cluster", ...) {
  strategies <- future:::supportedStrategies(...)
  strategies <- setdiff(strategies, excl)
  if (!is.na(cores)) {
    if (cores == 1L) {
      strategies <- setdiff(strategies, c("multicore", "multisession"))
    } else if (cores > 1L) {
      strategies <- setdiff(strategies, "sequential")
    }
  }
  
  strategies
}

isWin32 <- FALSE
availCores <- 2L

message("*** Demos ...")

message("*** Mandelbrot demo ...")

if (!isWin32) {
  options(future.demo.mandelbrot.nrow = 2L)
  options(future.demo.mandelbrot.resolution = 50L)
  options(future.demo.mandelbrot.delay = FALSE)
  
  for (cores in 1:availCores) {
    message(sprintf("Testing with %d cores ...", cores))
    options(mc.cores = cores)
  
    for (strategy in supportedStrategies(cores)) {
      message(sprintf("- plan('%s') ...", strategy))
      plan(strategy)
      demo("mandelbrot", package = "progressr", ask = FALSE)
      
      ## Explicitly close any PSOCK clusters to avoid 'R CMD check' NOTE
      ## on "detritus in the temp directory" on MS Windows
      plan(sequential)
      
      message(sprintf("- plan('%s') ... DONE", strategy))
    }
  
    message(sprintf("Testing with %d cores ... DONE", cores))
  } ## for (cores ...)
} else {
  message(" - This demo requires R (>= 3.2.0). Skipping test. (Skipping also on Win32 i386 for speed)")
}

message("*** Mandelbrot demo ... DONE")

message("*** Demos ... DONE")

library(progressr)

options(progressr.clear = TRUE)

delay <- getOption("progressr.demo.delay", 0.1)
message("- delay: ", delay, " seconds")

handlers("txtprogressbar")

handlers <- supported_progress_handlers()


message("with_progress() - standard output, messages, warnings ...")

n <- 5L
for (kk in seq_along(handlers)) {
  handler <- handlers[[kk]]
  name <- names(handlers)[kk]
  message(sprintf("* Handler %d ('%s') of %d ...", kk, name, length(handlers)))

  for (type in c("message", "warning")) {
    message(sprintf("  - stdout + %ss", type))
    for (delta in c(0L, +1L, -1L)) {
      message(sprintf("    - delta = %+d", delta))
      truth <- c()
      relay <- record_relay({
        with_progress({
          p <- progressor(n)
          for (ii in seq_len(n + delta)) {
            ## Zero-amount progress with empty message
            p(amount = 0)
            msg <- sprintf("ii = %d", ii)
            ## Zero-amount progress with non-empty message
            p(message = msg, amount = 0)
            truth <<- c(truth, msg)
            cat(msg, "\n", sep = "")
            ## Signal condition
            do.call(type, args = list(msg))
            Sys.sleep(delay)
            ## One-step progress with non-empty message
            p(message = sprintf("(%s)", paste(letters[1:ii], collapse=",")))
          }
        })
      }, classes = type)
      output <- relay$stdout
      message <- gsub("\n$", "", relay$msgs)
      if (delta > 0) {
        message <- grep("no longer listening to this progressor", message, invert = TRUE, value = TRUE)
      }
      stopifnot(
        identical(output, truth),
        identical(message, truth)
      )
    } ## for (delta ...)
  } ## for (signal ...)

  message(sprintf("* Handler %d ('%s') of %d ... done", kk, name, length(handlers)))
}


message("with_progress() - standard output, messages, warnings ... done")

library(progressr)

if (requireNamespace("pbmcapply", quietly = TRUE)) {
  handlers("pbmcapply")
  for (x in list(integer(0), 1:10, 1L)) {
    message("length(x): ", length(x))
    with_progress({ y <- slow_sum(x) })
    print(y)
  }
}

library(progressr)

options(progressr.demo.delay = 0.001)
options(progressr.interval = 0.0)
options(progressr.clear = FALSE)

message("with_progress() ...")

x <- 1:100
truth <- sum(x)

message("with_progress() - default ...")

if (requireNamespace("utils")) {
  with_progress({
    sum <- slow_sum(x)
  })
  print(sum)
  stopifnot(sum == truth)

  with_progress({
    cat("This stdout output will be delayed")
    message("This message will be delayed")
    warning("This warning will be delayed")
    signalCondition(simpleCondition("This simpleCondition will be delayed"))
    sum <- slow_sum(x)
  }, interval = 0.1, enable = TRUE, delay_conditions = "condition")
  print(sum)
  stopifnot(sum == truth)
}

message("with_progress() - default ... done")

message("with_progress() - filesize ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_filesize())
print(sum)
stopifnot(sum == truth)

message("with_progress() - filesize ... done")


message("with_progress() - utils::txtProgressBar() ...")

if (requireNamespace("utils")) {
  with_progress({
    sum <- slow_sum(x)
  }, handlers = handler_txtprogressbar(style = 2L))
  print(sum)
  stopifnot(sum == truth)
}

message("with_progress() - utils::txtProgressBar() ... done")


message("with_progress() - tcltk::tkProgressBar() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_tkprogressbar)

message("with_progress() - tcltk::tkProgressBar() ... done")


message("with_progress() - utils::winProgressBar() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_winprogressbar)

message("with_progress() - utils::winProgressBar() ... done")


message("with_progress() - progress::progress_bar() ...")

if (requireNamespace("progress")) {
  ## Display progress using default handler
  with_progress({
    sum <- slow_sum(x)
  }, handlers = handler_progress(clear = FALSE))
  print(sum)
  stopifnot(sum == truth)
}

message("with_progress() - progress::progress_bar() ... done")


message("with_progress() - pbmcapply::progressBar() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_pbmcapply)

message("with_progress() - pbmcapply::progressBar() ... done")


message("with_progress() - ascii_alert ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_ascii_alert())
print(sum)
stopifnot(sum == truth)

message("with_progress() - ascii_alert ... done")


message("with_progress() - beepr::beep() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_beepr)
print(sum)
stopifnot(sum == truth)

message("with_progress() - beepr::beep() ... done")


message("with_progress() - BRRR::skrrrahh() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_brrr)
print(sum)
stopifnot(sum == truth)

message("with_progress() - BRRR::skrrrahh() ... done")


message("with_progress() - notifier::notify() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_notifier)
print(sum)
stopifnot(sum == truth)

message("with_progress() - notifier::notify() ... done")


message("with_progress() - void ...")

## Mute progress updates
with_progress({
  sum <- slow_sum(x)
}, handlers = NULL)
print(sum)
stopifnot(sum == truth)

message(" - via option")

## NOTE: Set it to NULL, will use the default utils::txtProgressBar()
options(progressr.handlers = list())
with_progress({
  sum <- slow_sum(x)
})
print(sum)
stopifnot(sum == truth)

message("with_progress() - void ... done")


message("with_progress() - multiple handlers ...")

if (requireNamespace("utils", quietly = TRUE)) {
  handlers <- list(handler_txtprogressbar, handler_newline, handler_debug)
  options(progressr.handlers = handlers)
  
  with_progress({
    sum <- slow_sum(x)
  })
  print(sum)
  stopifnot(sum == truth)
}

message("with_progress() - multiple handlers ... done")


message("with_progress() - return value and visibility ...")

res <- with_progress(x)
stopifnot(identical(x, res))

res <- withVisible(with_progress(x))
stopifnot(identical(res$visible, TRUE))

res <- withVisible(with_progress(y <- x))
stopifnot(identical(res$visible, FALSE))

message("with_progress() - return value and visibility ... done")


message("with_progress() ... done")

library(progressr)

if (require("foreach", quietly = TRUE)) {
  message("* with_progress()")
    
  with_progress({
    p <- progressor(4)
    y <- foreach(n = 3:6) %do% {
      p()
      slow_sum(1:n, stdout=TRUE, message=TRUE)
    }
  })

  message("* global progression handler")

  handlers(global = TRUE)
    
  local({
    p <- progressor(4)
    y <- foreach(n = 3:6) %do% {
      p()
      slow_sum(1:n, stdout=TRUE, message=TRUE)
    }
  })
    
  handlers(global = FALSE)
}

source("incl/start.R")

if (requireNamespace("purrr", quietly = TRUE)) {
  message("* with_progress()")

  with_progress({
    p <- progressor(4)
    y <- purrr::map(3:6, function(n) {
      p()
      slow_sum(1:n, stdout=TRUE, message=TRUE)
    })
  })


  message("* global progression handler")

  handlers(global = TRUE)
   
  local({
    p <- progressor(4)
    y <- purrr::map(3:6, function(n) {
      p()
      slow_sum(1:n, stdout=TRUE, message=TRUE)
    })
  })
    
  handlers(global = FALSE)
}

source("incl/end.R")

library(progressr)

if (requireNamespace("future.apply", quietly = TRUE) && !covr) {
  for (strategy in future_strategies) {
    future::plan(strategy)
    print(future::plan())

    message("* with_progress()")
    
    with_progress({
      p <- progressor(4)
      y <- future.apply::future_lapply(3:6, function(n) {
        p()
        slow_sum(1:n, stdout=TRUE, message=TRUE)
      })
    })


    message("* global progression handler")

    handlers(global = TRUE)
    
    local({
      p <- progressor(4)
      y <- future.apply::future_lapply(3:6, function(n) {
        p()
        slow_sum(1:n, stdout=TRUE, message=TRUE)
      })
    })
    
    handlers(global = FALSE)
    
    ## Explicitly close any PSOCK clusters to avoid 'R CMD check' NOTE
    ## on "detritus in the temp directory" on MS Windows
    future::plan("sequential")
  } ## for (strategy ...)
}

library(progressr)

if (requireNamespace("plyr", quietly = TRUE)) {
  message("* with_progress()")

  with_progress({
    y <- plyr::llply(3:6, function(n, ...) {
      slow_sum(1:n, stdout=TRUE, message=TRUE)
    }, .progress = "progressr")
  })



  message("* global progression handler")

  handlers(global = TRUE)
    
  local({
    y <- plyr::llply(3:6, function(n, ...) {
      slow_sum(1:n, stdout=TRUE, message=TRUE)
    }, .progress = "progressr")
  })
    
  handlers(global = FALSE)
}

library(progressr)

if (requireNamespace("purrr", quietly = TRUE)) {
  message("* with_progress()")

  with_progress({
    p <- progressor(4)
    y <- purrr::map(3:6, function(n) {
      p()
      slow_sum(1:n, stdout=TRUE, message=TRUE)
    })
  })


  message("* global progression handler")

  handlers(global = TRUE)
   
  local({
    p <- progressor(4)
    y <- purrr::map(3:6, function(n) {
      p()
      slow_sum(1:n, stdout=TRUE, message=TRUE)
    })
  })
    
  handlers(global = FALSE)
}

library(progressr)

options(progressr.tests.fake_handlers = c("handler_beepr", "handler_brrr", "handler_notifier", "handler_pbmcapply", "handler_tkprogressbar", "handler_winprogressbar"))
options(progressr.clear = FALSE)
options(progressr.enable_after = NULL)
options(progressr.debug = FALSE)
options(progressr.times = NULL)

record_output <- function(expr, envir = parent.frame()) {
  conditions <- list()
  stdout <- utils::capture.output({
    withCallingHandlers(
      expr,
      condition = function(c) {
        if (inherits(c, c("progression", "error"))) return()
        conditions[[length(conditions) + 1L]] <<- c
      }
    )
  }, split = TRUE)
  list(stdout = stdout, conditions = conditions)
}


message("*** with_progress() - delaying output ...")

x <- 1:10

## Record truth
output_truth <- record_output({
  y_truth <- slow_sum(x, stdout = TRUE, message = TRUE, sticky = TRUE)
})

for (delay in c(FALSE, TRUE)) {
  message(sprintf("- with_progress() - delay = %s ...", delay))
  output <- record_output({
    with_progress({
      y <- slow_sum(x, stdout = TRUE, message = TRUE, sticky = TRUE)
    }, delay_stdout = delay,
      delay_conditions = if (delay) "condition" else character(0L))
  })

  stopifnot(identical(output$stdout, output_truth$stdout))
  stopifnot(identical(output$conditions, output_truth$conditions))
  stopifnot(identical(y, y_truth))
  message(sprintf("- with_progress() - delay = %s ... DONE", delay))
}

message("*** with_progress() - delaying output ... DONE")

library(progressr)

message("without_progress() ...")

x <- 1:10
y0 <- slow_sum(x)

with_progress(y <- slow_sum(x))

without_progress(y <- slow_sum(x))

with_progress(without_progress(y <- slow_sum(x)))

message("without_progress() ... done")


message("without_progress() - return value and visibility ...")

res <- without_progress(x)
stopifnot(identical(x, res))

res <- withVisible(without_progress(x))
stopifnot(identical(res$visible, TRUE))

res <- withVisible(without_progress(y <- x))
stopifnot(identical(res$visible, FALSE))

message("without_progress() - return value and visibility ... done")

source("incl/start.R")


message("Exceptions ...")

message("- with_progress()")

invalid <- progression(type = "unknown", session_uuid = "dummy", progressor_uuid = "dummy", progression_index = 0L)
print(invalid)
res <- tryCatch(with_progress({
  signalCondition(invalid)
}, handlers = handler_debug), error = identity)
str(res)
stopifnot(inherits(res, "error"))


message("- progress_aggregator()")

invalid <- progression(type = "unknown", session_uuid = "dummy", progressor_uuid = "dummy", progression_index = 0L)
print(invalid)
local({
  progress <- progress_aggregator(progressor(2L))
  res <- tryCatch(progress({
    signalCondition(invalid)
  }), error = identity)
  str(res)
  stopifnot(inherits(res, "error"))
})


message("Exceptions ... done")

source("incl/end.R")

library(progressr)


message("Exceptions ...")

message("- with_progress()")

invalid <- progression(type = "unknown", session_uuid = "dummy", progressor_uuid = "dummy", progression_index = 0L)
print(invalid)
res <- tryCatch(with_progress({
  signalCondition(invalid)
}, handlers = handler_debug), error = identity)
str(res)
stopifnot(inherits(res, "error"))


message("- progress_aggregator()")

invalid <- progression(type = "unknown", session_uuid = "dummy", progressor_uuid = "dummy", progression_index = 0L)
print(invalid)
local({
  progress <- progress_aggregator(progressor(2L))
  res <- tryCatch(progress({
    signalCondition(invalid)
  }), error = identity)
  str(res)
  stopifnot(inherits(res, "error"))
})


message("Exceptions ... done")

library(progressr)

options(progressr.debug = TRUE)

message("make_progression_handler() ...")

my_handler <- function(symbol = "*", file = stderr(), target = "terminal", ...) {
  reporter <- local({
    list(
      update = function(config, state, progression, ...) {
        if (state$enabled && progression$amount != 0) cat(file = file, symbol)
      }
    )
  })

  make_progression_handler("my_handler", reporter, ...)
}

h1 <- my_handler()
print(h1)

h2 <- my_handler(enable = FALSE)
print(h2)

message("make_progression_handler() ... done")

source("incl/start.R")

options(progressr.clear = FALSE)

if (capabilities("tcltk") && requireNamespace("tcltk", quietly = TRUE)) {
  options(progressr.handlers = handler_tkprogressbar)
}  


message("handler_progress() ...")

for (x in list(integer(0), 1:10, 1L)) {
  message("length(x): ", length(x))
  with_progress({
    progress <- progressor(along = x)
    for (ii in x) {
      Sys.sleep(getOption("progressr.demo.delay", 0.1))
      progress(message = sprintf("(%s)", paste(letters[1:ii], collapse="")))
    }
  })
}

message("handler_progress() ... done")

source("incl/end.R")

library(progressr)

message("progress_aggregator() ...")

with_progress({
  progress <- progressor(steps = 1 + 3 + 10 + 1)
  relay_progress <- progress_aggregator(progress)
  progress()
  relay_progress(slow_sum(1:3))
  relay_progress(slow_sum(1:10))
  progress()
})



message("- Stray progressions from unknown sources")

slow_prod <- function(x, delay = getOption("progressr.demo.delay", 0.05)) {
  progress <- progressor(2*length(x))
  res <- 0
  for (kk in seq_along(x)) {
    progress(message = sprintf("Multiplying %g", kk))
    Sys.sleep(0.8*delay)
    res <- res * x[kk]
    progress(message = "...")
    Sys.sleep(0.2*delay)
  }
  res
}

## This will only show progress for the *first* of the three
## functions that report progress. Any progression updates from
## the second and third will be ignored, because they are from
## a different source
with_progress({
  x <- 1:10
  a <- slow_sum(x)
  b <- slow_prod(x)
  c <- slow_sum(-x)
})


## To get progression from all of them, we need to know how many
## steps they report on and then use a gather-and-relay handler
with_progress({
  x <- 1:10
  progress <- progressor(3*length(x))
  relay_progress <- progress_aggregator(progress)
  relay_progress({
    a <- slow_sum(x)
    b <- slow_prod(x)
    c <- slow_sum(-x)
  })
})


message("progress_aggregator() ... done")

library(progressr)

if (requireNamespace("shiny", quietly = TRUE)) {
  ## This will generate:
  ## Error in shiny::withProgress(expr, ..., env = env, quoted = TRUE) : 
  ## 'session' is not a ShinySession object.
  res <- tryCatch({
    withProgressShiny({
      p <- progressor(3L)
      y <- lapply(1:3, function(n) {
        p()
        slow_sum(1:n, stdout=TRUE, message=TRUE)
      })
      y
    })
  }, error = identity)
  print(res)
}

library(progressr)

options(progressr.clear = FALSE)

options(progressr.handlers = handler_filesize)

message("handler_filesize() ...")

for (x in list(integer(0), 1:10, 1L)) {
  message("length(x): ", length(x))
  with_progress({
    progress <- progressor(along = x)
    for (ii in x) {
      Sys.sleep(getOption("progressr.demo.delay", 0.1))
      progress(message = sprintf("(%s)", paste(letters[1:ii], collapse="")))
    }
  })
}

message("handler_filesize() ... done")

source("incl/start.R")

message("progression() ...")

p <- progression()
print(p)
stopifnot(inherits(p, "progression"))

p <- progress()
print(p)
stopifnot(inherits(p, "progression"))

p2 <- progress(p)
print(p2)
stopifnot(identical(p2, p))

res <- tryCatch(progress(p), progression = function(p) TRUE)
print(res)
stopifnot(isTRUE(res))

res <- FALSE
withCallingHandlers(progress(p), progression = function(p) { res <<- TRUE })
print(res)
stopifnot(isTRUE(res))

message("progression() ... done")

source("incl/end.R")

non_supported_progression_handlers <- function() {
  names <- character(0L)
  for (pkg in c("beepr", "BRRR", "notifier", "pbmcapply", "progress", "shiny")) {
    if (!requireNamespace(pkg, quietly = TRUE))
      names <- c(names, pkg)
  }
  if (!"tcltk" %in% capabilities()) {
    names <- c(names, "tkprogressbar")
  }
  if (.Platform$OS.type != "windows") {
    names <- c(names, "winprogressbar")
  }
  if (!is_rstudio_console()) {
    names <- c(names, "rstudio")
  }
  if (!check_full) {
    names <- c(names, "notifier")
    names <- c(names, "shiny")
  }
  names <- unique(names)
  sprintf("handler_%s", names)
}


supported_progress_handlers <- function(exclude = non_supported_progression_handlers()) {
  handlers <- known_progression_handlers()
  drop <- na.omit(match(exclude, names(handlers)))
  if (length(drop) > 0L) handlers <- handlers[-drop]
  handlers
}


future_strategies <- c("multisession", "sequential")
if (.Platform$OS.type != "windows") {
  future_strategies <- c(future_strategies, "multicore")
}


capture_output <- function(..., split = FALSE, collapse = NULL) {
  bfr <- capture.output(..., split = split)
  if (!is.null(collapse)) bfr <- paste(c(bfr, ""), collapse = "\n")
  bfr
}

record_conditions <- function(expr, ..., classes = "condition", split = FALSE) {
  conditions <- list()
  withCallingHandlers(expr, condition = function(c) {
    if (inherits(c, classes)) {
      attr(c, "received") <- Sys.time()
      conditions[[length(conditions) + 1L]] <<- c
      if (!split) muffle_condition(c)
    }
  })
  conditions
}

record_relay <- function(..., all = FALSE, split = FALSE) {
  stdout <- capture_output(conditions <- record_conditions(...), split = split)
  msgs <- sapply(conditions, FUN = conditionMessage)
  res <- list(stdout = stdout, msgs = msgs)
  if (all) res$conditions <- conditions
  res
}

muffle_condition <- function(cond) {
  muffled <- FALSE
  if (inherits(cond, "message")) {
    invokeRestart("muffleMessage")
    muffled <- TRUE
  } else if (inherits(cond, "warning")) {
    invokeRestart("muffleWarning")
    muffled <- TRUE
  } else if (inherits(cond, "condition")) {
    restarts <- computeRestarts(cond)
    for (restart in restarts) {
      name <- restart$name
      if (is.null(name)) 
          next
      if (!grepl("^muffle", name)) 
          next
      invokeRestart(restart)
      muffled <- TRUE
      break
    }
  }
  invisible(muffled)
}

## Adopted from R.utils::cmsg()
console_msg <- function(..., collapse = "\n", sep = "\n", appendLF = TRUE) {
  fh <- tempfile()
  on.exit(file.remove(fh))
  cat(..., collapse = sep, sep = sep, file = fh)
  if (appendLF) 
    cat("\n", file = fh, append = TRUE)
  if (.Platform$OS.type == "windows") {
    file.show(fh, pager = "console", header = "", title = "",
              delete.file = FALSE)
  } else {
    system(sprintf("cat %s", fh))
  }
  invisible()
}

library(progressr)

message("with_progress() - progressr.debug = TRUE ...")

options(progressr.debug = TRUE)

with_progress({
  y <- slow_sum(1:10)
})


with_progress({
  p <- progressor(steps = 1 + 2 + 1)
  relay_progress <- progress_aggregator(p)
  cond <- p()
  relay_progress(slow_sum(1:2))
  p(type = "finish")
  p() ## one too many - will be ignored
})


message("with_progress() - progressr.debug = TRUE ... done")

library(progressr)

message("progressor() ...")

message("- basic")
local({
  p <- progressor(3L)
  print(p)
  p()
  p("A message")
})

message("- default message")
local({
  p <- progressor(along = 1:3, message = "A default message")
  print(p)
  p()
  p("A message")
})

message("- zero length")
local({
  p <- progressor(0L)
  print(p)
  p()
  p("A message")
})

message("- multiple consequtive progressors")
local({
  message("Progressor #1")
  p <- progressor(2L)
  for (kk in 1:2) p()

  message("Progressor #2")
  p <- progressor(3L)
  for (kk in 1:3) p()

  message("Done")
})

message("progressor() ... DONE")

library(progressr)

message("*** utils ...")


message("*** hpaste() ...")

# Some vectors
x <- 1:6
y <- 10:1
z <- LETTERS[x]

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Abbreviation of output vector
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
printf("x = %s.\n", hpaste(x))
## x = 1, 2, 3, ..., 6.

printf("x = %s.\n", hpaste(x, maxHead = 2))
## x = 1, 2, ..., 6.

printf("x = %s.\n", hpaste(x, maxHead = 3)) # Default
## x = 1, 2, 3, ..., 6.

# It will never output 1, 2, 3, 4, ..., 6
printf("x = %s.\n", hpaste(x, maxHead = 4))
## x = 1, 2, 3, 4, 5 and 6.

# Showing the tail
printf("x = %s.\n", hpaste(x, maxHead = 1, maxTail = 2))
## x = 1, ..., 5, 6.

# Turning off abbreviation
printf("y = %s.\n", hpaste(y, maxHead = Inf))
## y = 10, 9, 8, 7, 6, 5, 4, 3, 2, 1

## ...or simply
printf("y = %s.\n", paste(y, collapse = ", "))
## y = 10, 9, 8, 7, 6, 5, 4, 3, 2, 1

# Change last separator
printf("x = %s.\n", hpaste(x, lastCollapse = " and "))
## x = 1, 2, 3, 4, 5 and 6.

# No collapse
stopifnot(all(hpaste(x, collapse = NULL) == x))

# Empty input
stopifnot(identical(hpaste(character(0)), character(0)))

message("*** hpaste() ... DONE")


message("*** mdebug() ...")

mdebug("Hello #", 1)
mprint(1:3)
mprintf("Hello #%d", 1)
mstr(1:3)

options(progressr.debug = TRUE)
mdebug("Hello #", 2)
mprint(1:3)
mprintf("Hello #%d", 2)
mstr(1:3)

options(progressr.debug = FALSE)
mdebug("Hello #", 3)
mprint(1:3)
mprintf("Hello #%d", 3)
mstr(1:3)

message("*** mdebug() ... DONE")


message("*** stop_if_not() ...")

stop_if_not()
tryCatch(stop_if_not(c(1,2,3,4,5,6,7,8,9,10,1,2,3,4,5,6,7,8,9,10,1,2,3,4,5,6,7,8,9,10,1,2,3,4,5,6,7,8,9,10)), error = identity)

message("*** stop_if_not() ... done")


message("*** %||% ...")

print(NULL %||% TRUE)

print(TRUE %||% FALSE)

message("*** %||% ... done")

message("*** query_r_cmd_check() ...")

print(query_r_cmd_check())

cat("Command line arguments:\n")
args <- commandArgs()
print(args)

cat("Working directory:\n")
pwd <- getwd()
print(pwd)

message("*** query_r_cmd_check() ... done")

message("*** in_r_cmd_check() ...")

print(in_r_cmd_check())

message("*** in_r_cmd_check() ... done")


message("*** .onLoad() ...")

progressr:::.onLoad(pkgname = "progressr")

message("*** .onLoad() ... done")

message("*** known_progression_handlers() ...")

res <- known_progression_handlers()
str(res)

message("*** known_progression_handlers() ... done")

message("*** utils ... DONE")

source("incl/start.R")

if (requireNamespace("furrr", quietly = TRUE)) {
  for (strategy in future_strategies) {
    future::plan(strategy)
    print(future::plan())

    message("* with_progress()")

    with_progress({
      p <- progressor(4)
      y <- furrr::future_map(3:6, function(n) {
        p()
        slow_sum(1:n, stdout=TRUE, message=TRUE)
      })
    })


    message("* global progression handler")

    handlers(global = TRUE)
    
    local({
      p <- progressor(4)
      y <- furrr::future_map(3:6, function(n) {
        p()
        slow_sum(1:n, stdout=TRUE, message=TRUE)
      })
    })
    
    handlers(global = FALSE)
    
    ## Explicitly close any PSOCK clusters to avoid 'R CMD check' NOTE
    ## on "detritus in the temp directory" on MS Windows
    future::plan("sequential")
  } ## for (strategy ...)
}

source("incl/end.R")

if (getRversion() >= "4.0.0") {

library(progressr)

nsinks0 <- sink.number(type = "output")

options(progressr.clear = FALSE)

delay <- getOption("progressr.demo.delay", 0.1)
message("- delay: ", delay, " seconds")

handlers("txtprogressbar")

handlers <- supported_progress_handlers()

handlers(global = FALSE)
stopifnot(sink.number(type = "output") == nsinks0)

handlers(global = TRUE)
stopifnot(sink.number(type = "output") == nsinks0)

message("global progress handlers - standard output, messages, warnings ...")

n <- 5L
for (kk in seq_along(handlers)) {
  handler <- handlers[[kk]]
  name <- names(handlers)[kk]
  message(sprintf("* Handler %d ('%s') of %d ...", kk, name, length(handlers)))

  for (type in c("message", "warning")) {
    message(sprintf("  - stdout + %ss", type))
    for (delta in c(0L, +1L, -1L)) {
      message(sprintf("    - delta = %+d", delta))

      handlers(global = FALSE)
      stopifnot(sink.number(type = "output") == nsinks0)
      handlers(global = TRUE)
      stopifnot(sink.number(type = "output") == nsinks0)

      status <- progressr:::register_global_progression_handler("status")
      stopifnot(
        is.null(status$current_progressor_uuid),
        is.null(status$delays),
        is.null(status$stdout_file),
         length(status$conditions) == 0L,
         is.na(status$capture_conditions)
      )

      nsinks <- sink.number(type = "output")
      stopifnot(nsinks == nsinks0)
      
      truth <- c()
      relay <- record_relay(local({
        p <- progressor(n)
        for (ii in seq_len(n + delta)) {
          ## Zero-amount progress with empty message
          p(amount = 0)
          msg <- sprintf("ii = %d", ii)
          ## Zero-amount progress with non-empty message
          p(message = msg, amount = 0)
          truth <<- c(truth, msg)
          cat(msg, "\n", sep = "")
          ## Signal condition
          do.call(type, args = list(msg))
          Sys.sleep(delay)
          ## One-step progress with non-empty message
          p(message = sprintf("(%s)", paste(letters[1:ii], collapse=",")))
        }
      }), classes = type)
      stopifnot(
        identical(relay$stdout, truth),
        identical(gsub("\n$", "", relay$msgs), truth)
      )
      
      ## Assert sinks are balanced
      stopifnot(sink.number(type = "output") == nsinks)
      
      cat(paste(c(relay$stdout, ""), collapse = "\n"))
      message(relay$message, append = FALSE)
      status <- progressr:::register_global_progression_handler("status")
      console_msg(capture.output(utils::str(status)))
      if (delta == 0L) {
        withCallingHandlers({
          stopifnot(
            is.null(status$current_progressor_uuid),
            is.null(status$delays),
            is.null(status$stdout_file),
            length(status$conditions) == 0L,
            is.na(status$capture_conditions)
          )
        }, error = function(ex) {
          console_msg(paste("An error occurred:", conditionMessage(ex)))
          console_msg(capture.output(utils::str(status)))
        })
      }

    } ## for (delta ...)
  } ## for (signal ...)

  message(sprintf("* Handler %d ('%s') of %d ... done", kk, name, length(handlers)))
}


message("global progress handlers - standard output, messages, warnings ... done")

handlers(global = FALSE)


} ## if (getRversion() >= "4.0.0")

source("incl/start.R")

message("without_progress() ...")

x <- 1:10
y0 <- slow_sum(x)

with_progress(y <- slow_sum(x))

without_progress(y <- slow_sum(x))

with_progress(without_progress(y <- slow_sum(x)))

message("without_progress() ... done")


message("without_progress() - return value and visibility ...")

res <- without_progress(x)
stopifnot(identical(x, res))

res <- withVisible(without_progress(x))
stopifnot(identical(res$visible, TRUE))

res <- withVisible(without_progress(y <- x))
stopifnot(identical(res$visible, FALSE))

message("without_progress() - return value and visibility ... done")


source("incl/end.R")

#' Aggregate Progression Conditions
#'
#' @param progress A [progressor] function.
#'
#' @return A function of class `progress_aggregator`.
#'
#' @example incl/progress_aggregator.R
#'
#' @keywords internal
#' @export
progress_aggregator <- function(progress) {
  stop_if_not(inherits(progress, "progressor"))

  ## Here we can find out how many steps the progressor function wants
  max_steps <- environment(progress)$steps
  
  handler <- function(p) {
    stop_if_not(inherits(p, "progression"))
    type <- p$type
    debug <- getOption("progressr.debug", FALSE)
    if (debug) {
      mprintf("Progression handler %s ...", sQuote(type))
      on.exit(mprintf("Progression handler %s ... done", sQuote(type)))
      mprintf("- progression:")
      mstr(p)
      mprintf("- progressor_uuid: %s", p$progressor_uuid)
      mprintf("- progression_index: %d", p$progression_index)
    }
    if (type == "initiate") {
    } else if (type == "finish") {
    } else if (type == "reset") {
    } else if (type == "shutdown") {
    } else if (type == "update") {
      progress(child = p)
    } else {
      ## Was this meant to be a 'control_progression' condition?
      if (type %in% c("reset", "shutdown", "hide", "unhide", "interrupt")) {
        stop("Unsupported 'progression' type. Was it meant to be a 'control_progression' condition?: ", sQuote(type))
      } else {
        stop("Unknown 'progression' type: ", sQuote(type))
      }
    }
    
    ## Prevent upstream calling handlers to receive progression 'p'.
    ## Restart 'muffleProgression' is not guaranteed to exist, e.g. the
    ## 'progression' condition might be resignaled by a handler that
    ## does not define it.
    tryInvokeRestart("muffleProgression")
  }

  handler <- make_progression_handler("progress_aggregator", handler = handler)
  
  fcn <- function(...) {
    with_progress(..., handlers = handler)
  }
  class(fcn) <- c("progress_aggregator", class(fcn))

  fcn
}

library(progressr)

message("handlers() ...")

hs <- handlers()
print(hs)

for (kk in seq_along(hs)) {
  h <- hs[[kk]]
  print(h)
  handler <- h()
  print(handler)
}


hs <- handlers("txtprogressbar")
print(hs)

for (kk in seq_along(hs)) {
  h <- hs[[kk]]
  print(h)
  handler <- h()
  print(handler)
}

hs <- handlers("handler_txtprogressbar")
print(hs)

message("handlers() - exceptions ...")

## Will as a side-effect set an empty list of handlers()
res <- handlers("non-existing-handler", on_missing = "ignore")
res <- handlers()
stopifnot(is.list(res), length(res) == 0L)

res <- tryCatch({
  handlers("non-existing-handler", on_missing = "warning")
}, warning = identity)
stopifnot(inherits(res, "warning"))

res <- tryCatch({
  handlers("non-existing-handler", on_missing = "error")
}, error = identity)
stopifnot(inherits(res, "error"))


message("handlers() - exceptions ... DONE")

message("handlers(..., global = TRUE) ...")
if (getRversion() >= "4.0.0") {
  old_global <- handlers(global = NA)
  on.exit(handlers(global = old_global), add = TRUE)
  
  handlers(global = FALSE)
  stopifnot(handlers(global = NA) == FALSE)
  
  handlers("txtprogressbar", global = TRUE)
  stopifnot(handlers(global = NA) == TRUE)
  
  handlers("txtprogressbar", global = FALSE)
  stopifnot(handlers(global = NA) == FALSE)

  handlers(global = TRUE)
  handlers("txtprogressbar", global = FALSE)
  stopifnot(handlers(global = NA) == FALSE)
}
message("handlers(..., global = TRUE) ... DONE")

message("handlers() ... DONE")

## Private package functions
`%||%` <- progressr:::`%||%`
hpaste <- progressr:::hpaste
mdebug <- progressr:::mdebug
mprint <- progressr:::mprint
mprintf <- progressr:::mprintf
mstr <- progressr:::mstr
query_r_cmd_check <- progressr:::query_r_cmd_check
in_r_cmd_check <- progressr:::in_r_cmd_check
stop_if_not <- progressr:::stop_if_not
printf <- function(...) cat(sprintf(...))
known_progression_handlers <- progressr:::known_progression_handlers

#' Slowly Calculate Sum of Elements
#'
#' @param x Numeric vector to sum
#'
#' @param delay Delay in seconds after each addition.
#'
#' @param stdout If TRUE, then a text is outputted to the standard output
#' per element.
#'
#' @param message If TRUE, then a message is outputted per element.
#'
#' @param sticky If TRUE, then a "sticky" message is outputted every
#' ten element.
#'
#' @return The sum of all elements in `x`.
#'
#' @section Progress updates:
#' This function signals [progression] conditions as it progresses.
#'
#' @keywords internal
#' @export
slow_sum <- function(x, delay = getOption("progressr.demo.delay", 1.0), stdout = FALSE, message = TRUE, sticky = TRUE) {
  ## Hidden options to simplify help asciicast examples
  if (missing(stdout)) stdout <- getOption("progressr.slow_sum.stdout", FALSE)
  if (missing(message)) message <- getOption("progressr.slow_sum.message", TRUE)
  if (missing(sticky)) sticky <- getOption("progressr.slow_sum.sticky", TRUE)

  p <- progressor(along = x)

  sum <- 0
  for (kk in seq_along(x)) {
    p(amount = 0)   ## "I'm alive" progression update
    Sys.sleep(0.2*delay)
    if (stdout) cat(sprintf("O: Element #%d\n", kk))
    Sys.sleep(0.2*delay)
    p(amount = 0)
    Sys.sleep(0.2*delay)
    sum <- sum + x[kk]
    p(message = sprintf("P: Adding %g", kk))
    Sys.sleep(0.2*delay)
    if (message) message(sprintf("M: Added value %g", x[kk]))
    p(amount = 0)
    Sys.sleep(0.2*delay)
    if (sticky && kk %% 10 == 0) {
      p(
        amount = 0,
        message = sprintf("P: %d elements done", kk),
        class = "sticky"
      )
    }
  }

  p(amount = 0)

  sum
}

#' Progression Handler: Progress Reported as an ANSI Background Color in the Terminal
#'
#' @inheritParams make_progression_handler 
#'
#' @return A function of class `progression_handler` that takes a
#' [progression] condition as its first and only argument.
#'
#' @param adjust (numeric) The adjustment of the progress update,
#' where `adjust = 0` positions the message to the very left, and
#' `adjust = 1` positions the message to the very right.
#'
#' @param pad (integer) Amount of padding on each side of the message,
#' where padding is done by spaces.
#'
#' @param complete,incomplete (function) Functions that take "complete" and
#' "incomplete" strings that comprise the progress bar as input and annotate
#' them to reflect their two different parts.  The default is to annotation
#' them with two different background colors and the same foreground color
#' using the \pkg{cli} package.
#' 
#' @param \ldots Additional arguments passed to [make_progression_handler()].
#'
#' @section Requirements:
#' This progression handler requires the \pkg{cli} package.
#'
#' @section Appearance:
#' Below are a few examples on how to use and customize this progress handler.
#' In all cases, we use `handlers(global = TRUE)`.
#'
#' ```{asciicast handler_pbcol-default}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' library(progressr)
#' handlers("pbcol")
#' y <- slow_sum_p(1:25)
#' ```
#'
#' ```{asciicast handler_pbcol-adjust-mid}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' library(progressr)
#' handlers(handler_pbcol(adjust = 0.5))
#' y <- slow_sum_p(1:25)
#' ```
#'
#' ```{asciicast handler_pbcol-adjust-right-complete}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' library(progressr)
#' handlers(handler_pbcol(
#'   adjust = 1,
#'   complete = function(s) cli::bg_red(cli::col_black(s)),
#'   incomplete = function(s) cli::bg_cyan(cli::col_black(s))
#' ))
#' y <- slow_sum_p(1:25)
#' ```
#'
#' @example incl/handler_pbcol.R
#'
#' @importFrom utils flush.console
#' @export
handler_pbcol <- function(adjust = 0.0, pad = 1L, complete = function(s) cli::bg_blue(cli::col_white(s)), incomplete = function(s) cli::bg_cyan(cli::col_white(s)), intrusiveness = getOption("progressr.intrusiveness.terminal", 1), target = "terminal", ...) {
  crayon_enabled <- getOption("crayon.enabled", NULL)
  if (is.null(crayon_enabled)) crayon_enabled <- (cli::num_ansi_colors() > 1L)

  cat_ <- function(...) {
    cat(..., sep = "", collapse = "", file = stderr())
    flush.console()
  }
  
  erase_progress_bar <- function() {
    cat_(c("\r", rep(" ", times = getOption("width")), "\r"))
  }
  
  redraw_progress_bar <- function(ratio, message, spin = " ") {
    stop_if_not(ratio >= 0, ratio <= 1)
    if (crayon_enabled) {
      options(crayon.enabled = TRUE)
      on.exit(options(crayon.enabled = TRUE), add = TRUE)
    }
    pbstr <- pbcol(
      fraction = ratio,
      msg = message,
      adjust = adjust,
      pad = pad,
      complete = complete,
      incomplete = incomplete,
      spin = spin,
    )
    cat_("\r", pbstr)
  }
  
  reporter <- local({
    spin_state <- 0L
    spinner <- c("-", "\\", "|", "/", "-", "\\", "|", "/")
    
    list(
      initiate = function(config, state, ...) {
        if (!state$enabled || config$times <= 2L) return()
        ratio <- if (config$max_steps == 0) 1 else state$step / config$max_steps
        redraw_progress_bar(ratio = ratio, message = state$message, spin = spinner[spin_state+1L])
      },
      
      reset = function(...) {
        erase_progress_bar()
      },
      
      hide = function(...) {
        erase_progress_bar()
      },

      unhide = function(config, state, ...) {
        if (!state$enabled || config$times <= 2L) return()
        ratio <- if (config$max_steps == 0) 1 else state$step / config$max_steps
        redraw_progress_bar(ratio = ratio, message = state$message, spin = spinner[spin_state+1L])
      },

      interrupt = function(config, state, progression, ...) {
        if (!state$enabled || config$times <= 2L) return()
        erase_progress_bar()
        ratio <- if (config$max_steps == 0) 1 else state$step / config$max_steps
        redraw_progress_bar(ratio = ratio, message = state$message, spin = spinner[spin_state+1L])
        msg <- conditionMessage(progression)
        msg <- paste(c("", msg, ""), collapse = "\n")
        cat_(msg)
      },

      update = function(config, state, progression, ...) {
        if (!state$enabled || config$times <= 2L) return()
        if (state$delta < 0) return()
        spin_state <<- (spin_state+1L) %% length(spinner)
        ratio <- if (config$max_steps == 0) 1 else state$step / config$max_steps
        redraw_progress_bar(ratio = ratio, message = state$message, spin = spinner[spin_state+1L])
      },

      finish = function(config, state, progression, ...) {
        if (config$clear) {
          erase_progress_bar()
        } else {
          redraw_progress_bar(ratio = 1, message = state$message, spin = " ")
          cat("\n", file = stderr())
        }
      }
    )
  })

  make_progression_handler("pbcol", reporter, intrusiveness = intrusiveness, target = target, ...)
}



pbcol <- function(fraction = 0.0, msg = "", adjust = 0, pad = 1L, width = getOption("width") - 1L, complete = function(s) cli::bg_blue(cli::col_white(s)), incomplete = function(s) cli::bg_cyan(cli::col_white(s)), spin = " ") {
  if (length(msg) == 0L) msg <- ""
  stop_if_not(length(msg) == 1L, is.character(msg))

  fraction <- as.numeric(fraction)
  stop_if_not(length(fraction) == 1L, !is.na(fraction),
            fraction >= 0, fraction <= 1)
            
  width <- as.integer(width)
  stop_if_not(length(width) == 1L, !is.na(width), width > 0L)

  msgfraction <- sprintf(" %3.0f%%", 100 * fraction)
  
  ## Pad 'fullmsg' to align horizontally
  nmsg <- nchar(msg) + nchar(msgfraction)
  msgpad <- (width - 2 * pad) - nmsg

  ## Truncate 'msg'?
  if (msgpad < 0) {
    msg <- substr(msg, start = pad, stop = nchar(msg) + msgpad - pad)
    msg <- substr(msg, start = 1L, stop = nchar(msg) - 3L)
    msg <- paste(msg, "...", sep = "")
    msgpad <- (width - 2 * pad) - nchar(msg) - nchar(msgfraction)
    stop_if_not(msgpad >= 0)
  }

  ## Pad 'msg'
  lpad <- floor(   adjust  * msgpad) + pad
  rpad <- floor((1-adjust) * msgpad)
  stop_if_not(lpad >= 0L, rpad >= 0L)
  pmsg <- sprintf("%*s%s%*s%s%s%*s", lpad, "", msg, rpad, "", msgfraction, spin, pad, "")

  ## Make progress bar
  len <- round(fraction * nchar(pmsg), digits = 0L)
  lmsg <- substr(pmsg, start = 1L, stop = len)
  rmsg <- substr(pmsg, start = len + 1L, stop = nchar(pmsg))
  if (!is.null(complete)) lmsg <- complete(lmsg)
  if (!is.null(incomplete)) rmsg <- incomplete(rmsg)
  bar <- paste(lmsg, rmsg, sep = "")
  
  bar
}

source("incl/start.R")

options(progressr.clear = TRUE)

delay <- getOption("progressr.demo.delay", 0.1)
message("- delay: ", delay, " seconds")

handlers("txtprogressbar")

handlers <- supported_progress_handlers()


message("with_progress() - standard output, messages, warnings ...")

n <- 5L
for (kk in seq_along(handlers)) {
  handler <- handlers[[kk]]
  name <- names(handlers)[kk]
  message(sprintf("* Handler %d ('%s') of %d ...", kk, name, length(handlers)))

  for (type in c("message", "warning")) {
    message(sprintf("  - stdout + %ss", type))
    for (delta in c(0L, +1L, -1L)) {
      message(sprintf("    - delta = %+d", delta))
      truth <- c()
      relay <- record_relay({
        with_progress({
          p <- progressor(n)
          for (ii in seq_len(n + delta)) {
            ## Zero-amount progress with empty message
            p(amount = 0)
            msg <- sprintf("ii = %d", ii)
            ## Zero-amount progress with non-empty message
            p(message = msg, amount = 0)
            truth <<- c(truth, msg)
            cat(msg, "\n", sep = "")
            ## Signal condition
            do.call(type, args = list(msg))
            Sys.sleep(delay)
            ## One-step progress with non-empty message
            p(message = sprintf("(%s)", paste(letters[1:ii], collapse=",")))
          }
        })
      }, classes = type)
      stopifnot(
        identical(relay$stdout, truth),
        identical(gsub("\n$", "", relay$msgs), truth)
      )
    } ## for (delta ...)
  } ## for (signal ...)

  message(sprintf("* Handler %d ('%s') of %d ... done", kk, name, length(handlers)))
}


message("with_progress() - standard output, messages, warnings ... done")

source("incl/end.R")

#' Progression Handler: Progress Reported as Debug Information (Text) in the Terminal
#'
#' @inheritParams make_progression_handler
#'
#' @param uuid If TRUE, then the progressor UUID and the owner UUID are shown,
#' otherwise not (default).
#'
#' @param \ldots Additional arguments passed to [make_progression_handler()].
#'
#' @example incl/handler_debug.R
#'
#' @section Appearance:
#' Below is how this progress handler renders by default at 0%, 30% and 99%
#' progress:
#' 
#' With `handlers(handler_debug())`:
#' ```r
#' [21:27:11.236] (0.000s => +0.001s) initiate: 0/100 (+0) '' {clear=TRUE, enabled=TRUE, status=}
#' [21:27:11.237] (0.001s => +0.000s) update: 0/100 (+0) 'Starting' {clear=TRUE, enabled=TRUE, status=}
#' [21:27:14.240] (3.004s => +0.002s) update: 30/100 (+30) 'Importing' {clear=TRUE, enabled=TRUE, status=}
#' [21:27:16.245] (5.009s => +0.001s) update: 100/100 (+70) 'Summarizing' {clear=TRUE, enabled=TRUE, status=}
#' [21:27:16.246] (5.010s => +0.003s) update: 100/100 (+0) 'Summarizing' {clear=TRIE, enabled=TRUE, status=}
#' ```
#' @export
handler_debug <- function(interval = getOption("progressr.interval", 0), intrusiveness = getOption("progressr.intrusiveness.debug", 0), target = "terminal", uuid = FALSE, ...) {
  reporter <- local({
    t_init <- NULL
    
    add_to_log <- function(config, state, progression, ...) {
      t <- Sys.time()
      if (is.null(t_init)) t_init <<- t
      dt <- difftime(t, t_init, units = "secs")
      delay <- difftime(t, progression$time, units = "secs")
      message <- paste(c(state$message, ""), collapse = "")
      entry <- list(now(t), dt, delay, progression$type, state$step, config$max_steps, state$delta, message, config$clear, state$enabled, paste0(progression$status, ""))
      
      msg <- do.call(sprintf, args = c(list("%s(%.3fs => +%.3fs) %s: %.0f/%.0f (%+g) '%s' {clear=%s, enabled=%s, status=%s}"), entry))
      if (uuid) {
        msg <- sprintf("%s [progressor=%s, owner=%s]", msg, progression$progressor_uuid, progression$owner_session_uuid)
      }
      message(msg)
    }

    list(
      reset = function(...) {
        t_init <<- NULL
      },

      hide = function(...) NULL,
      
      unhide = function(...) NULL,

      interrupt = function(...) {
        add_to_log("interrupt", ...)
      },

      initiate = function(...) {
        add_to_log("initiate", ...)
      },
        
      update = function(...) {
        add_to_log("update", ...)
      },
        
      finish = function(...) {
        add_to_log("finish", ...)
      }
    )
  })
  
  make_progression_handler("debug", reporter, interval = interval, intrusiveness = intrusiveness, target = target, ...)
}

source("incl/start.R")

library(future)

supportedStrategies <- function(cores = NA_integer_, excl = "cluster", ...) {
  strategies <- future:::supportedStrategies(...)
  strategies <- setdiff(strategies, excl)
  if (!is.na(cores)) {
    if (cores == 1L) {
      strategies <- setdiff(strategies, c("multicore", "multisession"))
    } else if (cores > 1L) {
      strategies <- setdiff(strategies, "sequential")
    }
  }
  
  strategies
}

isWin32 <- FALSE
availCores <- 2L

message("*** Demos ...")

message("*** Mandelbrot demo ...")

if (!isWin32) {
  options(future.demo.mandelbrot.nrow = 2L)
  options(future.demo.mandelbrot.resolution = 50L)
  options(future.demo.mandelbrot.delay = FALSE)
  
  for (cores in 1:availCores) {
    message(sprintf("Testing with %d cores ...", cores))
    options(mc.cores = cores)
  
    for (strategy in supportedStrategies(cores)) {
      message(sprintf("- plan('%s') ...", strategy))
      plan(strategy)
      demo("mandelbrot", package = "progressr", ask = FALSE)
      
      ## Explicitly close any PSOCK clusters to avoid 'R CMD check' NOTE
      ## on "detritus in the temp directory" on MS Windows
      plan(sequential)
      
      message(sprintf("- plan('%s') ... DONE", strategy))
    }
  
    message(sprintf("Testing with %d cores ... DONE", cores))
  } ## for (cores ...)
} else {
  message(" - This demo requires R (>= 3.2.0). Skipping test. (Skipping also on Win32 i386 for speed)")
}

message("*** Mandelbrot demo ... DONE")

message("*** Demos ... DONE")

source("incl/end.R")

#' Progression Handler: Progress Reported via the Ntfy.sh Messaging Service
#'
#' A progression handler for [ntfy::ntfy_send()] of the \pkg{ntfy} package,
#' which sends notifications via the <https://ntfy.sh> framework.
#'
#' @inheritParams make_progression_handler
#' @inheritParams ntfy::ntfy_send
#'
#' @param \ldots Additional arguments passed to [make_progression_handler()].
#'
#' @example incl/handler_ntfy.R
#'
#' @section Requirements:
#' This progression handler requires the \pkg{ntfy} package.
#'
#' @export
handler_ntfy <- function(intrusiveness = getOption("progressr.intrusiveness.ntfy", 5), target = "gui", ..., title = "Progress update from R") {
  ## Used for package testing purposes only when we want to perform
  ## everything except the last part where the backend is called
  ntfy_send <- function(...) NULL
  if (!is_fake("handler_ntfy")) {
    pkg <- "ntfy"
    if (!requireNamespace(pkg, quietly = TRUE)) {
      stop("Package 'ntfy' is not available. See ?progressr::handler_ntfy() for installation instructions")
    }
    ntfy_send <- get("ntfy_send", mode = "function", envir = getNamespace(pkg))
  }

  notifier <- function(title, message) {
    args <- list(
      message = message,
      title   = title
    )

    do.call(ntfy_send, args = args)
  }

  reporter <- local({
    finished <- FALSE
    
    list(
      reset = function(...) {
        finished <<- FALSE
      },
      
      initiate = function(config, state, progression, ...) {
        if (!state$enabled || config$times == 1L) return()
        progress_notify(title = title, step = state$step, max_steps = config$max_steps, message = state$message, notifier = notifier)
      },
        
      interrupt = function(config, state, progression, ...) {
        msg <- conditionMessage(progression)
        progress_notify(title = title, step = state$step, max_steps = config$max_steps, message = msg, notifier = notifier)
      },

      update = function(config, state, progression, ...) {
        if (!state$enabled || progression$amount == 0 || config$times <= 2L) return()
        progress_notify(title = title, step = state$step, max_steps = config$max_steps, message = state$message, notifier = notifier)
      },
        
      finish = function(config, state, progression, ...) {
        if (finished) return()
        if (!state$enabled) return()
        if (state$delta > 0) progress_notify(title = title, step = state$step, max_steps = config$max_steps, message = state$message, notifier = notifier)
	finished <<- TRUE
      }
    )
  })
  
  make_progression_handler("ntfy", reporter, intrusiveness = intrusiveness, target = target, ...)
}

#' Options and environment variables used by the 'progressr' packages
#'
#' Below are environment variables and \R options that are used by the
#' \pkg{progressr} package.
#
#' Below are all \R options that are currently used by the \pkg{progressr} package.\cr
#' \cr
#' \emph{WARNING: Note that the names and the default values of these options may change in future versions of the package.  Please use with care until further notice.}
#'
#'
#' @section Options for controlling progression reporting:
#'
#' \describe{
#'   \item{\option{progressr.handlers}:}{
#'     (function or list of functions)
#'     Zero or more progression handlers that will report on any progression updates.  If empty list, progress updates are ignored.  If NULL, the default (`handler_txtprogressbar`) progression handlers is used.  The recommended way to set this option is via [progressr::handlers()]. (Default: NULL)
#'   }
#' }
#'
#'
#' @section Options for controlling progression handlers:
#'
#' \describe{
#'   \item{\option{progressr.clear}:}{
#'     (logical)
#'     If TRUE, any output, typically visual, produced by a reporter will be cleared/removed upon completion, if possible. (Default: TRUE)
#'   }
#'
#'   \item{\option{progressr.enable}:}{
#'     (logical)
#'     If FALSE, then progress is not reported.
#'     (Default: TRUE in interactive mode, otherwise FALSE)
#'   }
#'
#'   \item{\option{progressr.enable_after}:}{
#'     (numeric)
#'     Delay (in seconds) before progression updates are reported.
#'     (Default: `0.0`)
#'   }
#'
#'   \item{\option{progressr.times}:}{
#'     (numeric)
#'     The maximum number of times a handler should report progression updates. If zero, then progress is not reported.
#'     (Default: `+Inf`)
#'   }
#'
#'   \item{\option{progressr.interval}:}{
#'     (numeric)
#'     The minimum time (in seconds) between successive progression updates from this handler.
#'     (Default: `0.0`)
#'   }
#'
#'   \item{\option{progressr.intrusiveness}:}{
#'     (numeric)
#'     A non-negative scalar on how intrusive (disruptive) the reporter to the user. This multiplicative scalar applies to the _interval_ and _times_ parameters. (Default: `1.0`)\cr
#'   
#'    \describe{
#'      \item{\option{progressr.intrusiveness.audio}:}{(numeric) intrusiveness for auditory progress handlers (Default: `5.0`)}
#'      \item{\option{progressr.intrusiveness.file}:}{(numeric) intrusiveness for file-based progress handlers (Default: `5.0`)}
#'      \item{\option{progressr.intrusiveness.gui}:}{(numeric) intrusiveness for graphical-user-interface progress handlers (Default: `1.0`)}
#'      \item{\option{progressr.intrusiveness.notifier}:}{(numeric) intrusiveness for progress handlers that creates notifications (Default: `10.0`)}
#'      \item{\option{progressr.intrusiveness.terminal}:}{(numeric) intrusiveness for progress handlers that outputs to the terminal (Default: `1.0`)}
#'      \item{\option{progressr.intrusiveness.debug}:}{(numeric) intrusiveness for "debug" progress handlers (Default: `0.0`)}
#'    }
#'   }
#' }
#'
#' @section Options for controlling how standard output and conditions are relayed:
#'
#' \describe{
#'   \item{\option{progressr.delay_conditions}:}{
#'     (character vector)
#'     condition classes to be captured and relayed at the end after any captured standard output is relayed. (Default: `c("condition")`)
#'   }
#'
#'   \item{\option{progressr.delay_stdout}:}{
#'     (logical)
#'     If TRUE, standard output is captured and relayed at the end just before any captured conditions are relayed. (Default: TRUE)
#'   }
#' }
#'
#' @section Options for controlling interrupts:
#'
#' \describe{
#'   \item{\option{progressr.interrupts}:}{
#'     (logical)
#'     Controls whether interrupts should be detected or not.
#'     If FALSE, then interrupts are not detected and progress information
#'     is generated. (Default: `TRUE`)
#'   }
#'
#'   \item{\option{progressr.delay_stdout}:}{
#'     (logical)
#'     If TRUE, standard output is captured and relayed at the end just before any captured conditions are relayed. (Default: TRUE)
#'   }
#' }
#'
#'
#' @section Options for debugging progression updates:
#'
#' \describe{
#'  \item{\option{progressr.debug}:}{(logical) If TRUE, extensive debug messages are generated. (Default: FALSE)}
#' }
#'
#'
#' @section Options for progressr examples and demos:
#'
#' \describe{
#'  \item{\option{progressr.demo.delay}:}{(numeric) Delay (in seconds) between each iteration of [slow_sum()]. (Default: `1.0`)}
#' }
#'
#' @section Environment variables that set R options:
#' Some of the above \R \option{progressr.*} options can be set by corresponding
#' environment variable \env{R_PROGRESSR_*} _when the \pkg{progressr} package
#' is loaded_.
#' For example, if `R_PROGRESSR_ENABLE = "true"`, then option
#' \option{progressr.enable} is set to `TRUE` (logical).
#' For example, if `R_PROGRESSR_ENABLE_AFTER = "2.0"`, then option
#' \option{progressr.enable_after} is set to `2.0` (numeric).
#'
#' @seealso
#' To set \R options when \R starts (even before the \pkg{progressr} package is loaded), see the \link[base]{Startup} help page.  The \href{https://cran.r-project.org/package=startup}{\pkg{startup}} package provides a friendly mechanism for configuring \R at startup.
#'
#' @aliases
#' progressr.clear
#' progressr.debug
#' progressr.demo.delay
#' progressr.delay_stdout progressr.delay_conditions
#' progressr.enable progressr.enable_after
#' progressr.interrupts
#' progressr.interval
#' progressr.intrusiveness
#' progressr.intrusiveness.audio
#' progressr.intrusiveness.debug
#' progressr.intrusiveness.file
#' progressr.intrusiveness.gui
#' progressr.intrusiveness.notifier
#' progressr.intrusiveness.terminal
#' progressr.handlers
#' progressr.times
#'
#' @name progressr.options
NULL


get_package_option <- function(name, default = NULL, package = .packageName) {
  if (!is.null(package)) {
    name <- paste(package, name, sep = ".")
  }
  getOption(name, default = default)
}

# Set an R option from an environment variable
update_package_option <- function(name, mode = "character", default = NULL, package = .packageName, split = NULL, trim = TRUE, disallow = c("NA"), force = FALSE, debug = FALSE) {
  if (!is.null(package)) {
    name <- paste(package, name, sep = ".")
  }

  mdebugf("Set package option %s", sQuote(name))

  ## Already set? Nothing to do?
  value <- getOption(name, NULL)
  if (!force && !is.null(value)) {
    mdebugf("Already set: %s", sQuote(value))
    return(getOption(name))
  }

  ## name="Pkg.foo.Bar" => env="R_PKG_FOO_BAR"
  env <- gsub(".", "_", toupper(name), fixed = TRUE)
  env <- paste("R_", env, sep = "")

  env_value <- value <- Sys.getenv(env, unset = NA_character_)
  if (is.na(value)) {  
    if (debug) mdebugf("Environment variable %s not set", sQuote(env))
    
    ## Nothing more to do?
    if (is.null(default)) return(getOption(name))

    if (debug) mdebugf("Use argument 'default': ", sQuote(default))
    value <- default
  }

  if (debug) mdebugf("%s=%s", env, sQuote(value))

  ## Trim?
  if (trim) value <- trim(value)

  ## Nothing to do?
  if (!nzchar(value)) return(getOption(name, default = default))

  ## Split?
  if (!is.null(split)) {
    value <- strsplit(value, split = split, fixed = TRUE)
    value <- unlist(value, use.names = FALSE)
    if (trim) value <- trim(value)
  }

  ## Coerce?
  mode0 <- storage.mode(value)
  if (mode0 != mode) {
    suppressWarnings({
      storage.mode(value) <- mode
    })
    if (debug) {
      mdebugf("Coercing from %s to %s: %s", mode0, mode, commaq(value))
    }
  }

  if (length(disallow) > 0) {
    if ("NA" %in% disallow) {
      if (any(is.na(value))) {
        stop(sprintf("Coercing environment variable %s=%s to %s would result in missing values for option %s: %s", sQuote(env), sQuote(env_value), sQuote(mode), sQuote(name), commaq(value)))
      }
    }
    if (is.numeric(value)) {
      if ("non-positive" %in% disallow) {
        if (any(value <= 0, na.rm = TRUE)) {
          stop(sprintf("Environment variable %s=%s specifies a non-positive value for option %s: %s", sQuote(env), sQuote(env_value), sQuote(name), commaq(value)))
        }
      }
      if ("negative" %in% disallow) {
        if (any(value < 0, na.rm = TRUE)) {
          stop(sprintf("Environment variable %s=%s specifies a negative value for option %s: %s", sQuote(env), sQuote(env_value), sQuote(name), commaq(value)))
        }
      }
    }
  }
  
  if (debug) {
    mdebugf("=> options(%s = %s) [n=%d, mode=%s]",
            dQuote(name), commaq(value),
            length(value), storage.mode(value))
  }

  do.call(options, args = structure(list(value), names = name))
  
  getOption(name, default = default)
}


## Set package options based on environment variables
update_package_options <- function(debug = FALSE) {
  update_package_option("demo.delay", mode = "numeric", debug = debug)

  ## make_progression_handler() arguments
  update_package_option("clear", mode = "logical", default = TRUE, debug = debug)

  ## Special case: Support R_PROGRESSR_ENABLE=interactive
  value <- Sys.getenv("R_PROGRESSR_ENABLE", "")
  if (value == "interactive") {
    options(progressr.enable = interactive())
  } else {
    update_package_option("enable", mode = "logical", default = interactive(), debug = debug)
  }

  update_package_option("enable_after", mode = "numeric", default = 0.0, debug = debug)
  update_package_option("interval", mode = "numeric", default = 0.0, debug = debug)
  update_package_option("times", mode = "numeric", default = +Inf, debug = debug)
  update_package_option("interrupts", mode = "logical", default = TRUE, debug = debug)

  ## Life-cycle, e.g. deprecation an defunct
  update_package_option("lifecycle.progress", mode = "character", default = "deprecated", debug = debug)

  ## However, not used
  update_package_option("global.handler", mode = "logical", debug = debug)

  ## For RStudio users
  update_package_option("rstudio.patch", mode = "logical", debug = debug)
}

source("incl/start.R")

if (requireNamespace("shiny", quietly = TRUE)) {
  ## This will generate:
  ## Error in shiny::withProgress(expr, ..., env = env, quoted = TRUE) : 
  ## 'session' is not a ShinySession object.
  res <- tryCatch({
    withProgressShiny({
      p <- progressor(3L)
      y <- lapply(1:3, function(n) {
        p()
        slow_sum(1:n, stdout=TRUE, message=TRUE)
      })
      y
    })
  }, error = identity)
  print(res)
}

source("incl/end.R")

#!/usr/bin/env Rscript

#' Run a 'testme' Test Script
#'
#' R usage:
#' future::testme("<name>")
#'
#' Command-line usage:
#' Rscript tests/test-<name>.R
#'
#' Command-line usage without package re-install:
#' Rscript inst/testme/run.R --name=<name>

cmd_args <- commandArgs(trailingOnly = TRUE)

pattern <- "--package=([[:alpha:][:alnum:]]+)"
idx <- grep(pattern, cmd_args)
if (length(idx) > 0L) {
  stopifnot(length(idx) == 1L)
  testme_package <- gsub(pattern, "\\1", cmd_args[idx])
  cmd_args <- cmd_args[-idx]
} else {
  testme_package <- Sys.getenv("R_TESTME_PACKAGE", NA_character_)
  if (is.na(testme_package)) {
    desc <- read.dcf("DESCRIPTION")
    testme_package <- desc[1, "Package"]
  }
}

pattern <- "--path=([[:alpha:][:alnum:]]+)"
idx <- grep(pattern, cmd_args)
if (length(idx) > 0L) {
  stopifnot(length(idx) == 1L)
  path <- gsub(pattern, "\\1", cmd_args[idx])
  cmd_args <- cmd_args[-idx]
} else {
  path <- Sys.getenv("R_TESTME_PATH", NA_character_)
  if (is.na(path)) {
    path <- file.path("inst", "testme")
  }
  if (!utils::file_test("-d", path)) {
    stop("There exist no such 'R_TESTME_PATH' folder: ", sQuote(path))
  }
}
Sys.setenv(R_TESTME_PATH = path)


pattern <- "--name=([[:alpha:][:alnum:]]+)"
idx <- grep(pattern, cmd_args)
if (length(idx) > 0L) {
  stopifnot(length(idx) == 1L)
  testme_name <- gsub(pattern, "\\1", cmd_args[idx])
  cmd_args <- cmd_args[-idx]
} else {
  testme_name <- NULL
}

## Fallback for 'testme_name'?
if (is.null(testme_name)) {
  if (length(cmd_args) > 0) {
    stopifnot(length(cmd_args) == 1L)
    file <- cmd_args[1]
    if (utils::file_test("-f", file)) {
      testme_name <- gsub("(^test-|[.]R$)", "", basename(file))
    } else {
      stop("No such file: ", file)
    }
  } else {
    testme_name <- Sys.getenv("R_TESTME_NAME", NA_character_)
    if (is.na(testme_name)) {
      stop("testme: Environment variable 'R_TESTME_NAME' is not set")
    }
  }
}


testme_file <- file.path(path, sprintf("test-%s.R", testme_name))
if (!utils::file_test("-f", testme_file)) {
  stop("There exist no such 'testme' file: ", sQuote(testme_file))
}


## -----------------------------------------------------------------
## testme environment
## -----------------------------------------------------------------
on_cran <- function() {
  not_cran <- Sys.getenv("NOT_CRAN", NA_character_)
  if (is.na(not_cran)) {
    not_cran <- FALSE
  } else {
    not_cran <- isTRUE(as.logical(not_cran))
  }
  !interactive() && !not_cran
} ## on_cran()


## Get test script tags
tags <- local({
  lines <- readLines(testme_file, warn = FALSE)
  pattern <- "^#'[[:blank:]]+@tags[[:blank:]]+"
  lines <- grep(pattern, lines, value = TRUE)
  tags <- sub(pattern, "", lines)
  tags
})
if (length(tags) > 0) {
  tags <- sub("[[:blank:]]*$", "", tags)
  tags <- unlist(strsplit(tags, split = "[[:blank:]]+"))
  tags <- sort(unique(tags))
} else {
  tags <- character(0L)
}

## Create 'testme' environment on the search() path
testme_config <- list(
  package = testme_package,
     name = testme_name,
     tags = tags,
   status = "created",
    start = proc.time(),
   script = testme_file,
  on_cran = on_cran(),
    debug = isTRUE(as.logical(Sys.getenv("R_TESTME_DEBUG")))
)
if ("testme" %in% search()) detach(name = "testme")
testme <- attach(testme_config, name = "testme", warn.conflicts = FALSE)
rm(list = c("tags", "testme_package", "testme_name", "testme_file"))


## -----------------------------------------------------------------
## Filters
## -----------------------------------------------------------------
## Skip on CRAN? To run these tests, set env var NOT_CRAN=true
if ("skip_on_cran" %in% tags && on_cran()) {
  testme[["status"]] <- "skipped"
}

code <- Sys.getenv("R_TESTME_FILTER_NAME", NA_character_)
if (!is.na(code)) {
  expr <- tryCatch(parse(text = code), error = identity)
  if (inherits(expr, "error")) {
    stop("Syntax error in R_TESTME_FILTER_NAME: ", sQuote(code))
  }
  
  keep <- tryCatch(eval(expr, envir = testme), error = identity)
  if (inherits(keep, "error")) {
    stop("Evaluation of R_TESTME_FILTER_NAME=%s produced an error: %s",
         sQuote(code), conditionMessage(keep))
  }
  if (!isTRUE(keep)) testme[["status"]] <- "skipped"
}

code <- Sys.getenv("R_TESTME_FILTER_TAGS", NA_character_)
if (!is.na(code)) {
  expr <- tryCatch(parse(text = code), error = identity)
  if (inherits(expr, "error")) {
    stop("Syntax error in R_TESTME_FILTER_TAGS: ", sQuote(code))
  }
  keep <- tryCatch(eval(expr, envir = testme), error = identity)
  if (inherits(keep, "error")) {
    stop("Evaluation of R_TESTME_FILTER_TAGS=%s produced an error: %s",
         sQuote(code), conditionMessage(keep))
  }
  if (!isTRUE(keep)) testme[["status"]] <- "skipped"
}


message(sprintf("Test %s ...", sQuote(testme[["name"]])))

if (testme[["debug"]]) {
  message("testme:")
  message(paste(utils::capture.output(utils::str(as.list(testme))), collapse = "\n"))
}

## Process prologue scripts, if they exist
if (testme[["status"]] != "skipped" &&
    utils::file_test("-d", file.path(path, "_prologue"))) {
  testme[["status"]] <- "prologue"
  local({
    ## Find all prologue scripts
    files <- dir(file.path(path, "_prologue"), pattern = "*[.]R$", full.names = TRUE)
    files <- sort(files)
    testme[["prologue_scripts"]] <- files
  
    ## Source all prologue scripts inside the 'testme' environment
    expr <- quote({
      files <- prologue_scripts
      message(sprintf("Sourcing %d prologue scripts ...", length(files)))
      for (kk in seq_along(files)) {
        file <- files[kk]
        message(sprintf("%02d/%02d prologue script %s", kk, length(files), sQuote(file)))
        source(file, local = TRUE)
      }
      message(sprintf("Sourcing %d prologue scripts ... done", length(files)))
      rm(list = c("kk", "file", "files"))
    })
    eval(expr, envir = testme)
  })

#  ## In case prologue scripts overwrote some elements in 'testme'
#  for (name in names(testme_config)) {
#    testme[[name]] <- testme_config[[name]]
#  }
}


## Run test script
## Note, prologue scripts may trigger test to be skipped
if (testme[["status"]] != "skipped") {
  message("Running test script: ", sQuote(testme[["script"]]))
  testme[["status"]] <- "failed"
  source(testme[["script"]], echo = TRUE)
  testme[["status"]] <- "success"
  
#  ## In case test script overwrote some elements in 'testme'
#  for (name in names(testme_config)) {
#    testme[[name]] <- testme_config[[name]]
#  }
}


## Process epilogue scripts, if they exist
## Note, epilogue scripts may change status or produce check errors
if (testme[["status"]] == "success" &&
    utils::file_test("-d", file.path(path, "_epilogue"))) {
  testme[["status"]] <- "epilogue"
  local({
    ## Find all epilogue scripts
    files <- dir(file.path(path, "_epilogue"), pattern = "*[.]R$", full.names = TRUE)
    files <- sort(files)
    testme[["epilogue_scripts"]] <- files
  
    ## Source all epilogue scripts inside the 'testme' environment
    expr <- quote({
      files <- epilogue_scripts
      message(sprintf("Sourcing %d epilogue scripts ...", length(files)))
      for (kk in seq_along(files)) {
        file <- files[kk]
        message(sprintf("%02d/%02d epilogue script %s", kk, length(files), sQuote(file)))
        source(file, local = TRUE)
      }
      message(sprintf("Sourcing %d epilogue scripts ... done", length(files)))
      rm(list = c("kk", "file", "files"))
    })
    eval(expr, envir = testme)
  })
  testme[["status"]] <- "success"
}

testme[["stop"]] <- proc.time()
dt <- testme[["stop"]] - testme[["start"]]
dt_str <- sprintf("%s=%.1gs", names(dt), dt)
message("Test time: ", paste(dt_str, collapse = ", "))

message(sprintf("Test %s ... %s", sQuote(testme[["name"]]), testme[["status"]]))

if ("testme" %in% search()) detach(name = "testme")

.onLoad <- function(libname, pkgname) {
  debug <- isTRUE(as.logical(Sys.getenv("R_PROGRESSR_DEBUG", "FALSE")))
  if (debug) options(progressr.debug = TRUE)
  debug <- getOption("progressr.debug", debug)

  ## Set package options based on environment variables
  update_package_options(debug = debug)

  ## Record the process ID (PID) when the package is loaded
  is_fork_child()
  
  ## R CMD check
  if (in_r_cmd_check()) {
    options(progressr.demo.delay = 0.0)
  }

  ## R CMD build
  register_vignette_engine_during_build_only(pkgname)

  ## Register a global progression handler on load?
  if (isTRUE(getOption("progressr.global.handler", FALSE))) {
    ## UPDATE It is not possible to register a global calling handler when
    ## there is already an active condition handler as it is here because
    ## loadNamespace()/library() uses tryCatch() internally.  If attempted,
    ## we'll get an error "should not be called with handlers on the stack".
    ## /HB 2020-11-19
#    register_global_progression_handler()
  }

  ## RStudio (>= 2025.05.0) introduced a custom global message handler
  ## for the RStudio Console. It causes conflicts with some of the
  ## progressr handlers.
  ## https://github.com/futureverse/progressr/issues/179
  warn_about_rstudio_condition_handlers()
}

#' Progression Handler: Progress Reported in the RStudio Console
#'
#' @inheritParams make_progression_handler 
#'
#' @param title (character or a function) The "name" of the progressor, which
#' is displayed in front of the progress bar.  If a function, then the name
#' is created dynamically by calling the function when the progressor is
#' created.
#'
#' @param \ldots Additional arguments passed to [make_progression_handler()].
#'
#' @section Requirements:
#' This progression handler works only in the RStudio Console.
#'
#' @section Use this progression handler by default:
#' To use this handler by default whenever using the RStudio Console, add
#' the following to your \file{~/.Rprofile} startup file:
#'
#' ```r
#' if (requireNamespace("progressr", quietly = TRUE)) {
#'   if (Sys.getenv("RSTUDIO") == "1" && !nzchar(Sys.getenv("RSTUDIO_TERM"))) {
#'     options(progressr.handlers = progressr::handler_rstudio)
#'   }
#' }
#' ```
#'
#' @example incl/handler_rstudio.R
#'
#' @export
handler_rstudio <- function(intrusiveness = getOption("progressr.intrusiveness.gui", 1), target = "gui", title = function() format(Sys.time(), "Console %X"), ...) {
  reporter <- local({
    job_id <- NULL
    list(
      initiate = function(config, state, ...) {
        if (!state$enabled || config$times <= 2L) return()
        name <- state$message
        if (length(name) == 0L) {
          if (is.null(title)) {
            name <- "Console"
          } else if (is.character(title)) {
            name <- title
          } else if (is.function(title)) {
            name <- title()
          }
        }
        stop_if_not(
          is.null(job_id),
          is.character(name),
          length(name) == 1L
        )
        job_id <<- rstudioapi::jobAdd(
          name          = name,
          progressUnits = as.integer(config$max_steps),
          status        = "running",
          autoRemove    = FALSE,
          show          = FALSE
        )
      },
      
      update = function(config, state, progression, ...) {
        if (!state$enabled || config$times <= 2L) return()
        ## The RStudio Job progress bar cannot go backwards
        if (state$delta < 0) return()
        ## The RStudio Job progress bar does not have a "spinner"
        if (state$delta == 0) return()
        stop_if_not(!is.null(job_id))
        rstudioapi::jobSetProgress(job_id, units = state$step)
      },

      finish = function(...) {
        if (!is.null(job_id)) rstudioapi::jobRemove(job_id)
        job_id <<- NULL
      }
    )
  })

  make_progression_handler("rstudio", reporter, intrusiveness = intrusiveness, target = target, ...)
}

source("incl/start.R")

options(progressr.demo.delay = 0.001)
options(progressr.interval = 0.0)
options(progressr.clear = FALSE)

message("with_progress() ...")

x <- 1:100
truth <- sum(x)

message("with_progress() - default ...")

if (requireNamespace("utils")) {
  with_progress({
    sum <- slow_sum(x)
  })
  print(sum)
  stopifnot(sum == truth)

  with_progress({
    cat("This stdout output will be delayed")
    message("This message will be delayed")
    warning("This warning will be delayed")
    signalCondition(simpleCondition("This simpleCondition will be delayed"))
    sum <- slow_sum(x)
  }, interval = 0.1, enable = TRUE, delay_conditions = "condition")
  print(sum)
  stopifnot(sum == truth)
}

message("with_progress() - default ... done")

message("with_progress() - filesize ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_filesize())
print(sum)
stopifnot(sum == truth)

message("with_progress() - filesize ... done")


message("with_progress() - utils::txtProgressBar() ...")

if (requireNamespace("utils")) {
  with_progress({
    sum <- slow_sum(x)
  }, handlers = handler_txtprogressbar(style = 2L))
  print(sum)
  stopifnot(sum == truth)
}

message("with_progress() - utils::txtProgressBar() ... done")


message("with_progress() - tcltk::tkProgressBar() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_tkprogressbar)

message("with_progress() - tcltk::tkProgressBar() ... done")


message("with_progress() - utils::winProgressBar() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_winprogressbar)

message("with_progress() - utils::winProgressBar() ... done")


message("with_progress() - progress::progress_bar() ...")

if (requireNamespace("progress")) {
  ## Display progress using default handler
  with_progress({
    sum <- slow_sum(x)
  }, handlers = handler_progress(clear = FALSE))
  print(sum)
  stopifnot(sum == truth)
}

message("with_progress() - progress::progress_bar() ... done")


message("with_progress() - pbmcapply::progressBar() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_pbmcapply)

message("with_progress() - pbmcapply::progressBar() ... done")


message("with_progress() - ascii_alert ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_ascii_alert())
print(sum)
stopifnot(sum == truth)

message("with_progress() - ascii_alert ... done")


message("with_progress() - beepr::beep() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_beepr)
print(sum)
stopifnot(sum == truth)

message("with_progress() - beepr::beep() ... done")


message("with_progress() - notifier::notify() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_notifier)
print(sum)
stopifnot(sum == truth)

message("with_progress() - notifier::notify() ... done")


message("with_progress() - void ...")

## Mute progress updates
with_progress({
  sum <- slow_sum(x)
}, handlers = NULL)
print(sum)
stopifnot(sum == truth)

message(" - via option")

## NOTE: Set it to NULL, will use the default utils::txtProgressBar()
options(progressr.handlers = list())
with_progress({
  sum <- slow_sum(x)
})
print(sum)
stopifnot(sum == truth)

message("with_progress() - void ... done")


message("with_progress() - multiple handlers ...")

if (requireNamespace("utils", quietly = TRUE)) {
  handlers <- list(handler_txtprogressbar, handler_newline, handler_debug)
  options(progressr.handlers = handlers)
  
  with_progress({
    sum <- slow_sum(x)
  })
  print(sum)
  stopifnot(sum == truth)
}

message("with_progress() - multiple handlers ... done")


message("with_progress() - return value and visibility ...")

res <- with_progress(x)
stopifnot(identical(x, res))

res <- withVisible(with_progress(x))
stopifnot(identical(res$visible, TRUE))

res <- withVisible(with_progress(y <- x))
stopifnot(identical(res$visible, FALSE))

message("with_progress() - return value and visibility ... done")


message("with_progress() ... done")

source("incl/end.R")

source("incl/start.R")

message("*** utils ...")


message("*** hpaste() ...")

# Some vectors
x <- 1:6
y <- 10:1
z <- LETTERS[x]

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Abbreviation of output vector
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
printf("x = %s.\n", hpaste(x))
## x = 1, 2, 3, ..., 6.

printf("x = %s.\n", hpaste(x, maxHead = 2))
## x = 1, 2, ..., 6.

printf("x = %s.\n", hpaste(x, maxHead = 3)) # Default
## x = 1, 2, 3, ..., 6.

# It will never output 1, 2, 3, 4, ..., 6
printf("x = %s.\n", hpaste(x, maxHead = 4))
## x = 1, 2, 3, 4, 5 and 6.

# Showing the tail
printf("x = %s.\n", hpaste(x, maxHead = 1, maxTail = 2))
## x = 1, ..., 5, 6.

# Turning off abbreviation
printf("y = %s.\n", hpaste(y, maxHead = Inf))
## y = 10, 9, 8, 7, 6, 5, 4, 3, 2, 1

## ...or simply
printf("y = %s.\n", paste(y, collapse = ", "))
## y = 10, 9, 8, 7, 6, 5, 4, 3, 2, 1

# Change last separator
printf("x = %s.\n", hpaste(x, lastCollapse = " and "))
## x = 1, 2, 3, 4, 5 and 6.

# No collapse
stopifnot(all(hpaste(x, collapse = NULL) == x))

# Empty input
stopifnot(identical(hpaste(character(0)), character(0)))

message("*** hpaste() ... DONE")


message("*** mdebug() ...")

mdebug("Hello #", 1)
mprint(1:3)
mprintf("Hello #%d", 1)
mstr(1:3)

options(progressr.debug = TRUE)
mdebug("Hello #", 2)
mprint(1:3)
mprintf("Hello #%d", 2)
mstr(1:3)

options(progressr.debug = FALSE)
mdebug("Hello #", 3)
mprint(1:3)
mprintf("Hello #%d", 3)
mstr(1:3)

message("*** mdebug() ... DONE")


message("*** stop_if_not() ...")

stop_if_not()
tryCatch(stop_if_not(c(1,2,3,4,5,6,7,8,9,10,1,2,3,4,5,6,7,8,9,10,1,2,3,4,5,6,7,8,9,10,1,2,3,4,5,6,7,8,9,10)), error = identity)

message("*** stop_if_not() ... done")


message("*** %||% ...")

print(NULL %||% TRUE)

print(TRUE %||% FALSE)

message("*** %||% ... done")

message("*** query_r_cmd_check() ...")

print(query_r_cmd_check())

cat("Command line arguments:\n")
args <- commandArgs()
print(args)

cat("Working directory:\n")
pwd <- getwd()
print(pwd)

message("*** query_r_cmd_check() ... done")

message("*** in_r_cmd_check() ...")

print(in_r_cmd_check())

message("*** in_r_cmd_check() ... done")


message("*** .onLoad() ...")

progressr:::.onLoad(pkgname = "progressr")

message("*** .onLoad() ... done")

message("*** known_progression_handlers() ...")

res <- known_progression_handlers()
str(res)

message("*** known_progression_handlers() ... done")

message("*** utils ... DONE")

source("incl/end.R")

#' Progression Handler: Progress Reported as a MS Windows Progress Bars in the GUI
#'
#' A progression handler for `winProgressBar()` in the \pkg{utils} package.
#'
#' @inheritParams make_progression_handler
#'
#' @return A function of class `progression_handler` that takes a
#' [progression] condition as its first and only argument.
#' 
#' @param inputs (named list) Specifies from what sources the MS Windows
#' progress elements 'title' and 'label' should be updated. Valid sources are
#' `"message"`, `"sticky_message"` and `"non_sticky_message"`, where
#' `"message"` is short for `c("non_sticky_message", "sticky_message")`. For
#' example, `inputs = list(title = "sticky_message", label = "message")`
#' will update the 'title' component from sticky messages only,
#' whereas the 'label' component is updated using any message.
#'
#' @param \ldots Additional arguments passed to `tcltk::winProgressBar()`
#' and [make_progression_handler()].
#'
#' @examples
#' \donttest{\dontrun{
#' handlers(handler_winprogressbar())
#' with_progress(y <- slow_sum_p(1:100))
#' }}
#' 
#' @section Requirements:
#' This progression handler requires MS Windows.
#'
#' @export
handler_winprogressbar <- function(intrusiveness = getOption("progressr.intrusiveness.gui", 1), target = "gui", inputs = list(title = NULL, label = "message"), ...) {
  ## Additional arguments passed to the progress-handler backend
  backend_args <- handler_backend_args(...)

  ## Additional arguments passed to the progress-handler backend
  ## Used for package testing purposes only when we want to perform
  ## everything except the last part where the backend is called
  if (!is_fake("handler_winprogressbar")) {
    if (.Platform$OS.type != "windows") {
      stop("handler_winprogressbar requires MS Windows: ",
           sQuote(.Platform$OS.type))
    }
    ## Import functions
    winProgressBar <- utils::winProgressBar
    setWinProgressBar <- utils::setWinProgressBar
  } else {
    winProgressBar <- function(title = "R progress bar", label = "", min = 0, max = 1, initial = 0, width = 300) rawConnection(raw(0L))
    setWinProgressBar <- function(pb, value, title = NULL, label = NULL) NULL
  }

  stop_if_not(
    is.list(inputs),
    !is.null(names(inputs)),
    all(names(inputs) %in% c("title", "label")),
    all(vapply(inputs, FUN = function(x) {
      if (is.null(x)) return(TRUE)
      if (!is.character(x)) return(FALSE)
      x %in% c("message", "non_sticky_message", "sticky_message")
    }, FUN.VALUE = FALSE))
  )
  
  ## Expand 'message' => c("non_sticky_message", "sticky_message")
  for (name in names(inputs)) {
    input <- inputs[[name]]
    if ("message" %in% input) {
      input <- setdiff(input, "message")
      input <- c(input, "non_sticky_message", "sticky_message")
    }
    inputs[[name]] <- unique(input)
  }

  reporter <- local({
    pb_config <- NULL
    
    ## Update winProgressBar
    update_pb <- function(state, progression, ...) {
      ## Update 'title' and 'label' (optional)
      args <- message_to_backend_targets(progression, inputs = inputs, ...)
      for (name in names(args)) pb_config[[name]] <<- args[[name]]

      ## Update progress bar
      args <- pb_config
      args$value <- state$step
      do.call(what = setWinProgressBar, args = args)
    }
    
    list(
      reset = function(...) {
        pb_config <<- NULL
      },
      
      initiate = function(config, state, progression, ...) {
        if (!state$enabled || config$times == 1L) return()
        ## NOTE: 'pb_config' may be re-used for winProgressBar:s
        if (config$clear) stop_if_not(is.null(pb_config))
        args <- c(
          backend_args,
          list(max = config$max_steps, initial = state$step),
          list(...)
        )

        ## winProgressBar() arguments 'title' and 'label' must not be NULL;
        ## use the defaults
        for (name in c("title", "label")) {
          if (is.null(args[[name]])) {
            args[[name]] <- formals(winProgressBar)[[name]]
          }
        }
        
        ## WORKAROUND: If the progress bar is created with label = "", then
        ## it will *not* be possible to modify it with winSetProgressBar()
        ## afterward, cf. "Space will be allocated for the label only if
        ## it is non-empty" in help("winProgressBar", package = "utils").
        if (args$label == "") args$label <- " "

        ## Create progress bar
        args <- args[names(args) %in% names(formals(winProgressBar))]
        pb <- do.call(winProgressBar, args = args)

        ## Record arguments used by setWinProgressBar() later on
        args$pb <- pb
        args <- args[names(args) %in% names(formals(setWinProgressBar))]
        pb_config <<- args
        
        pb_config
      },
        
      interrupt = function(config, state, progression, ...) {
        if (!state$enabled) return()
        msg <- conditionMessage(progression)
        update_pb(state, progression, message = msg)
      },

      update = function(config, state, progression, ...) {
        if (!state$enabled || config$times <= 2L) return()
        update_pb(state, progression)
      },
        
      finish = function(config, state, progression, ...) {
        ## Already finished?
        if (is.null(pb_config)) return()
        if (!state$enabled) return()
        if (config$clear) {
          close(pb_config$pb)
          pb_config <<- NULL
        } else {
          update_pb(state, progression)
        }
      }
    )
  })
  
  make_progression_handler("winprogressbar", reporter, intrusiveness = intrusiveness, target = target, ...)
}

#' Use Progressr with Plyr Map-Reduce Functions
#'
#' A "progress bar" for \pkg{plyr}'s `.progress` argument.
#'
#' @param \ldots Not used.
#'
#' @return A named [base::list] that can be passed as argument `.progress`
#' to any of \pkg{plyr} function accepting that argument.
#'
#' @example incl/plyr_progress_progressr.R
#'
#' @section Limitations:
#' One can use use [doFuture::registerDoFuture()] to run \pkg{plyr} functions
#' in parallel, e.g. `plyr::l_ply(..., .parallel = TRUE)`.  Unfortunately,
#' using `.parallel = TRUE` disables progress updates because, internally,
#' \pkg{plyr} forces `.progress = "none"` whenever `.parallel = TRUE`.
#' Thus, despite the \pkg{future} ecosystem and \pkg{progressr} would support
#' it, it is not possible to run \pkg{dplyr} in parallel _and_ get progress
#' updates at the same time.
#'
#' @export
progress_progressr <- function(...) {
  ## Progressor
  p <- NULL

  ## List of plyr-recognized progress functions
  list(
    init = function(x, ...) {
      p <<- progressor(x, on_exit = FALSE)
    },
    
    step = function() {
      p()
    },
    
    term = function() {
      p(type = "finish")
    }
  )
}

source("incl/start.R")

options(progressr.debug = TRUE)

message("make_progression_handler() ...")

my_handler <- function(symbol = "*", file = stderr(), target = "terminal", ...) {
  reporter <- local({
    list(
      update = function(config, state, progression, ...) {
        if (state$enabled && progression$amount != 0) cat(file = file, symbol)
      }
    )
  })

  make_progression_handler("my_handler", reporter, ...)
}

h1 <- my_handler()
print(h1)

h2 <- my_handler(enable = FALSE)
print(h2)

message("make_progression_handler() ... done")

source("incl/end.R")

#' Progression Handler: Progress Reported as 'beepr' Sounds (Audio)
#'
#' A progression handler for [beepr::beep()].
#'
#' @inheritParams make_progression_handler
#'
#' @return A function of class `progression_handler` that takes a
#' [progression] condition as its first and only argument.
#'
#' @param initiate,update,finish,interrupt (integer) Indices of [beepr::beep()]
#' sounds to play when progress starts, is updated, completes, or is
#' interrupted.  For silence, use `NA_integer_`.
#'
#' @param \ldots Additional arguments passed to [make_progression_handler()].
#'
#' @example incl/handler_beepr.R
#'
#' @section Requirements:
#' This progression handler requires the \pkg{beepr} package.
#'
#' @export
handler_beepr <- function(initiate = 2L, update = 10L,  finish = 11L, interrupt = 9L, intrusiveness = getOption("progressr.intrusiveness.audio", 5.0), target = "audio", ...) {
  ## Used for package testing purposes only when we want to perform
  ## everything except the last part where the backend is called
  if (!is_fake("handler_beepr")) {
    beepr_beep <- beepr::beep
  } else {
    beepr_beep <- function(sound, expr) NULL
  }

  beep <- function(sound) {
    ## Silence?
    if (is.na(sound)) return()
    beepr_beep(sound)
  }

  ## Reporter state
  reporter <- local({
    list(
      initiate = function(config, state, progression, ...) {
        if (!state$enabled || config$times == 1L) return()
	beep(initiate)
      },
        
      interrupt = function(config, state, progression, ...) {
        beep(interrupt)
      },
      
      update = function(config, state, progression, ...) {
        if (!state$enabled || progression$amount == 0 || config$times <= 2L) return()
        beep(update)
      },
        
      finish = function(config, state, progression, ...) {
        if (!state$enabled) return()
        beep(finish)
      }
    )
  })
  
  make_progression_handler("beepr", reporter, intrusiveness = intrusiveness, target = target, ...)
}

#' Progression Handler: No Progress Report
#'
#' @inheritParams make_progression_handler
#'
#' @return A function of class `progression_handler` that takes a
#' [progression] condition as its first and only argument.
#'
#' @param \ldots Additional arguments passed to [make_progression_handler()].
#'
#' @examples
#' \donttest{\dontrun{
#' handlers(handler_void())
#' with_progress(y <- slow_sum_p(1:100))
#' print(y)
#' }}
#'
#' @details
#' This progression handler gives not output - it is invisible and silent.
#'
#' @export
handler_void <- function(intrusiveness = 0, target = "void", enable = FALSE, ...) {
  reporter <- local({
    list(
      initiate = function(config, state, progression, ...) NULL,
      update = function(config, state, progression, ...) NULL,
      finish = function(config, state, progression, ...) NULL
    )
  })
  
  make_progression_handler("void", reporter, intrusiveness = intrusiveness, target = target, enable = enable, ...)
}

#' Progression Handler: Progress Reported via 'progress' Progress Bars (Text) in the Terminal
#'
#' A progression handler for [progress::progress_bar()].
#'
#' @inheritParams make_progression_handler
#'
#' @param format (character string) The format of the progress bar.
#'
#' @param show_after (numeric) Number of seconds to wait before displaying
#' the progress bar.
#'
#' @param \ldots Additional arguments passed to [progress::progress_bar()]
#' and [make_progression_handler()].
#'
#' @section Requirements:
#' This progression handler requires the \pkg{progress} package.
#'
#' @section Appearance:
#' Below are a few examples on how to use and customize this progress handler.
#' In all cases, we use `handlers(global = TRUE)`.
#'
#' ```{asciicast handler_progress-default}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' handlers("progress")
#' y <- slow_sum(1:25)
#' ```
#'
#' ```{asciicast handler_progress-complete}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' handlers(handler_progress(complete = "#"))
#' y <- slow_sum(1:25)
#' ```
#'
#' ```{asciicast handler_progress-format-1}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' handlers(handler_progress(format = ":spin [:bar] :percent :message"))
#' y <- slow_sum(1:25)
#' ```
#'
#' ```{asciicast handler_progress-format-2}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' handlers(handler_progress(format = ":percent [:bar] :eta :message"))
#' y <- slow_sum(1:25)
#' ```
#'
#' @example incl/handler_progress.R
#'
#' @export
handler_progress <- function(format = ":spin [:bar] :percent :message", show_after = 0.0, intrusiveness = getOption("progressr.intrusiveness.terminal", 1), target = "terminal", ...) {
  ## Additional arguments passed to the progress-handler backend
  backend_args <- handler_backend_args(...)

  ## Force evaluation for 'format' here in case 'crayon' is used.  This
  ## works around the https://github.com/r-lib/crayon/issues/48 problem
  stop_if_not(is.character(format), length(format) == 1L, !is.na(format))
  
  if (!is_fake("handler_progress")) {
    progress_bar <- progress::progress_bar
    get_private <- function(pb) {
      pb$.__enclos_env__$private
    }
    erase_progress_bar <- function(pb) {
      if (pb$finished) return()
      private <- get_private(pb)
      private$clear_line(private$width)
      private$cursor_to_start()
    }
    redraw_progress_bar <- function(pb, tokens = list()) {
      if (pb$finished) return()
      private <- get_private(pb)
      private$last_draw <- ""
      private$render(tokens)
    }
  } else {
    progress_bar <- list(
      new = function(...) list(
        finished = FALSE,
        tick = function(...) NULL,
        update = function(...) NULL,
        message = function(...) NULL
      )
    )
    get_private <- function(pb) NULL
    erase_progress_bar <- function(pb) NULL
    redraw_progress_bar <- function(pb, tokens = list()) NULL
  }

  reporter <- local({
    pb <- NULL
    
    make_pb <- function(format, total, clear, show_after, ...) {
      if (!is.null(pb)) return(pb)
      stop_if_not(
        is.character(format), length(format) == 1L,
        is.numeric(total), length(total) == 1L,
        is.logical(clear), length(clear) == 1L,
        is.numeric(show_after), length(show_after) == 1L
      )
      args <- c(list(format = format, total = total, clear = clear,
                     show_after = show_after, force = TRUE, ...),
                backend_args)
      pb <<- do.call(progress_bar$new, args = args)
      pb
    }

    last_tokens <- list()
    pb_tick <- function(pb, delta = 0, message = NULL, ...) {
      if (isTRUE(pb$finished)) return()
      
      ## WORKAROUND: https://github.com/r-lib/progress/issues/119
      private <- get_private(pb)
      if (!is.null(private) && private$total == 0) return()
      
      tokens <- list(message = paste0(message, ""))
      last_tokens <<- tokens
      if (delta < 0) return()
      pb$tick(delta, tokens = tokens)
    }

    pb_update <- function(pb, ratio, ...) {
      if (isTRUE(pb$finished)) return()
      
      ## WORKAROUND: https://github.com/r-lib/progress/issues/119
      private <- get_private(pb)
      if (!is.null(private) && private$total == 0) return()
      
      pb$update(ratio = ratio, ...)
    }

    list(
      reset = function(...) {
        pb <<- NULL
      },

      hide = function(...) {
        if (is.null(pb)) return()
        erase_progress_bar(pb)
      },

      unhide = function(...) {
        if (is.null(pb)) return()
        redraw_progress_bar(pb, tokens = last_tokens)
      },

      interrupt = function(config, state, progression, ...) {
        if (is.null(pb)) return()

        erase_progress_bar(pb)
        redraw_progress_bar(pb)
        
        msg <- conditionMessage(progression)
        msg <- paste(c("", msg, ""), collapse = "\n")
        cat(msg, file = stderr())
      },

      initiate = function(config, state, progression, ...) {
        if (!state$enabled || config$times == 1L) return()
        stop_if_not(is.null(pb))
        make_pb(format = format, total = config$max_steps,
                clear = config$clear, show_after = config$enable_after)
        pb_tick(pb, 0, message = state$message)
      },
        
      update = function(config, state, progression, ...) {
        if (!state$enabled || config$times <= 2L) return()
        make_pb(format = format, total = config$max_steps,
                clear = config$clear, show_after = config$enable_after)
        if (inherits(progression, "sticky") && length(state$message) > 0) {
          pb$message(state$message)
        }
        pb_tick(pb, state$delta, message = state$message)
      },
        
      finish = function(config, state, progression, ...) {
        ## Already finished?
        if (is.null(pb)) return()
        if (!pb$finished) {
          make_pb(format = format, total = config$max_steps,
                  clear = config$clear, show_after = config$enable_after)
          reporter$update(config = config, state = state, progression = progression, ...)
          if (config$clear) pb_update(pb, ratio = 1.0)
        }
        pb <<- NULL
      }
    )
  })

  make_progression_handler("progress", reporter, intrusiveness = intrusiveness, target = target, ...)
}

library(progressr)

message("progression() ...")

p <- progression()
print(p)
stopifnot(inherits(p, "progression"))

p <- progress()
print(p)
stopifnot(inherits(p, "progression"))

p2 <- progress(p)
print(p2)
stopifnot(identical(p2, p))

res <- tryCatch(progress(p), progression = function(p) TRUE)
print(res)
stopifnot(isTRUE(res))

res <- FALSE
withCallingHandlers(progress(p), progression = function(p) { res <<- TRUE })
print(res)
stopifnot(isTRUE(res))

message("progression() ... done")

#' Report on Progress while Evaluating an R Expression
#'
#' @param expr An \R expression to evaluate.
#'
#' @param handlers A progression handler or a list of them.
#' If NULL or an empty list, progress updates are ignored.
#'
#' @param cleanup If TRUE, all progression handlers will be shutdown
#' at the end regardless of the progression is complete or not.
#'
#' @param delay_terminal If TRUE, output and conditions that may end up in
#' the terminal will delayed.
#'
#' @param delay_stdout If TRUE, standard output is captured and relayed
#' at the end just before any captured conditions are relayed.
#'
#' @param delay_conditions A character vector specifying [base::condition]
#' classes to be captured and relayed at the end after any captured
#' standard output is relayed.
#'
#' @param interrupts Controls whether interrupts should be detected or not.
#' If TRUE and a interrupt is signaled, progress handlers are asked to
#' report on the current amount progress when the evaluation was terminated
#' by the interrupt, e.g. when a user pressed Ctrl-C in an interactive session,
#' or a batch process was interrupted because it ran out of time.
#' Note that it's optional for a progress handler to support this and only
#' some do.
#'
#' @param interval (numeric) The minimum time (in seconds) between
#' successive progression updates from handlers.
#'
#' @param enable (logical) If FALSE, then progress is not reported.  The
#' default is to report progress in interactive mode but not batch mode.
#' See below for more details.
#'
#' @return Returns the value of the expression.
#'
#' @example incl/with_progress.R
#'
#' @details
#' If you are writing a Shiny app, use the [withProgressShiny()] function
#' instead of this one.
#'
#' If the global progression handler is enabled, it is temporarily disabled
#' while evaluating the `expr` expression.
#'
#' If a `progression` condition causes the progressor to be completed
#' ("reaches 100%"), then `with_progress()` will muffle the `progression`
#' condition preventing it from reaching, say, any global condition
#' handlers.
#'
#' **IMPORTANT: This function is meant for end users only.  It should not
#' be used by R packages, which only task is to _signal_ progress updates,
#' not to decide if, when, and how progress should be reported.**
#'
#' @section Progression handler functions:
#' Formally, progression handlers are calling handlers that are called
#' when a [progression] condition is signaled.  These handlers are functions
#' that takes one argument which is the [progression] condition.
#'
#' @section Progress updates in batch mode:
#' When running R from the command line, R runs in a non-interactive mode
#' (`interactive()` returns `FALSE`).  The default behavior of
#' `with_progress()` is to _not_ report on progress in non-interactive mode.
#' To have progress being reported on also then, set R options
#' \option{progressr.enable} or environment variable \env{R_PROGRESSR_ENABLE}
#' to `TRUE`.  Alternatively, one can set argument `enable=TRUE` when calling
#' `with_progress()`.  For example,
#' ```sh
#' $ Rscript -e "library(progressr)" -e "with_progress(slow_sum(1:5))"
#' ```
#' will _not_ report on progress, whereas:
#' ```sh
#' $ export R_PROGRESSR_ENABLE=TRUE
#' $ Rscript -e "library(progressr)" -e "with_progress(slow_sum(1:5))"
#' ```
#' will.
#'
#' @seealso
#' For Shiny apps, use [withProgressShiny()] instead of this function.
#' Internally, this function is built around [base::withCallingHandlers()].
#'
#' @export
with_progress <- function(expr, handlers = progressr::handlers(), cleanup = TRUE, delay_terminal = NULL, delay_stdout = NULL, delay_conditions = NULL, interrupts = getOption("progressr.interrupts", TRUE), interval = NULL, enable = NULL) {
    stop_if_not(is.logical(cleanup), length(cleanup) == 1L, !is.na(cleanup))  
    stop_if_not(is.logical(interrupts), length(interrupts) == 1L, !is.na(interrupts))
  
    debug <- getOption("progressr.debug", FALSE)
    if (debug) {
      message("with_progress() ...")
      on.exit(message("with_progress() ... done"), add = TRUE)
    }
  
    ## Deactive global progression handler while using with_progress()
    global_progression_handler(FALSE)
    on.exit(global_progression_handler(TRUE), add = TRUE)
  
    ## FIXME: With zero handlers, progression conditions will be
    ##        passed on upstream just as without with_progress().
    ##        Is that what we want? /HB 2019-05-17
  
    # Nothing to do?
    if (length(handlers) == 0L) {
      if (debug) message("No progress handlers - skipping")
      return(expr)
    }
  
    ## Temporarily set progressr options
    options <- list()
    
    ## Enabled or not?
    if (!is.null(enable)) {
      stop_if_not(is.logical(enable), length(enable) == 1L, !is.na(enable))
      
      # Nothing to do?
      if (!enable) {
        if (debug) message("Progress disabled - skipping")
        return(expr)
      }
  
      options[["progressr.enable"]] <- enable
    }
  
    if (!is.null(interval)) {
      stop_if_not(is.numeric(interval), length(interval) == 1L, !is.na(interval))
      options[["progressr.interval"]] <- interval
    }
  
    if (length(options) > 0L) {  
      oopts <- options(options)
      on.exit(options(oopts), add = TRUE)
    }
  
    progressr_in_globalenv("allow")
    on.exit(progressr_in_globalenv("disallow"), add = TRUE)
  
    handlers <- as_progression_handler(handlers, debug = debug)
  
    ## Nothing to do?
    if (length(handlers) == 0L) {
      if (debug) message("No remaining progress handlers - skipping")
      return(expr)
    }
    
    ## Do we need to buffer?
    delays <- use_delays(handlers,
      terminal   = delay_terminal,
      stdout     = delay_stdout,
      conditions = delay_conditions
    )
    if (debug) {
      what <- c(
        if (delays$terminal) "terminal",
        if (delays$stdout) "stdout",
        delays$conditions
      )
      message("- Buffering: ", paste(sQuote(what), collapse = ", "))
    }
  
    calling_handler <- make_calling_handler(handlers)
    
    ## Flag indicating whether nor not with_progress() exited due to an error
    status <- "incomplete"
  
    ## Tell all progression handlers to shut down at the end and
    ## the status of the evaluation.
    if (cleanup) {
      on.exit({
        if (debug) message("- signaling 'shutdown' to all handlers")
        withCallingHandlers({
          withRestarts({
            signalCondition(control_progression("shutdown", status = status))
          }, muffleProgression = function(p) NULL)
        }, progression = calling_handler)
      }, add = TRUE)
    }
  
    ## Delay standard output?
    stdout_file <- delay_stdout(delays, stdout_file = NULL)
    on.exit(flush_stdout(stdout_file, must_work = TRUE), add = TRUE)
    
    ## Delay conditions?
    conditions <- list()
    if (length(delays$conditions) > 0) {
      on.exit(flush_conditions(conditions), add = TRUE)
    }

    ## Reset all handlers upfront
    if (debug) message("- signaling 'reset' to all handlers")
    withCallingHandlers({
      withRestarts({
        signalCondition(control_progression("reset"))
      }, muffleProgression = function(p) NULL)
    }, progression = calling_handler)

    handle_interrupt_or_error <- function(condition) {
      suspendInterrupts({
        ## Don't capture conditions that are produced by progression handlers
        capture_conditions <<- FALSE
        on.exit(capture_conditions <<- TRUE)
    
        ## Any buffered output to flush?
        if (isTRUE(attr(delays$terminal, "flush"))) {
          if (length(conditions) > 0L || has_buffered_stdout(stdout_file)) {
            calling_handler(control_progression("hide"))
            stdout_file <<- flush_stdout(stdout_file, close = FALSE)
            conditions <<- flush_conditions(conditions)
          }
        }
  
        if (isTRUE(getOption("progressr.interrupts", TRUE))) {
          ## Create progress message saying why the progress was interrupted
          msg <- sprintf("Progress interrupted by %s condition", class(condition)[1])
          msg <- paste(c(msg, conditionMessage(condition)), collapse = ": ")
          calling_handler(control_progression("interrupt", message = msg))
        }
      }) ## suspendInterrupts()
    } ## handle_interrupt_or_error()
  
    ## Just for debugging purposes
    progression_counter <- 0
    
    ## Evaluate expression
    capture_conditions <- TRUE
    finished <- FALSE
    
    withCallingHandlers({
      res <- withVisible(expr)
    },
    
    progression = function(p) {
      ## Reset, because a new progression was created?
      if (p$type == "initiate") {
        if (!finished) calling_handler(control_progression("shutdown"))
        progression_counter <<- 0
        finished <<- FALSE
      } else {
        progression_counter <<- progression_counter + 1
        if (debug) message(sprintf("- received a %s (n=%g)", sQuote(class(p)[1]), progression_counter))
        if (finished) {
          warn_about_too_many_progressions(p)
          return()
        }
      }
      
      ## Don't capture conditions that are produced by progression handlers
      capture_conditions <<- FALSE
      on.exit(capture_conditions <<- TRUE)
  
      ## Any buffered output to flush?
      if (isTRUE(attr(delays$terminal, "flush"))) {
        if (length(conditions) > 0L || has_buffered_stdout(stdout_file)) {
          calling_handler(control_progression("hide"))
          stdout_file <<- flush_stdout(stdout_file, close = FALSE)
          conditions <<- flush_conditions(conditions)
          calling_handler(control_progression("unhide"))
        }
      }

      ## Let the registered 'progressr' calling handlers process
      ## the 'progression' condition. If this resulted in the
      ## progressor being completed, then 'finished' is TRUE
      finished <<- calling_handler(p)
    },
  
    interrupt = handle_interrupt_or_error,
    
    error = handle_interrupt_or_error,
  
    condition = function(c) {
      if (!capture_conditions || finished) return()
      
      if (debug) message("- received a ", sQuote(class(c)[1]))
  
      if (inherits(c, delays$conditions)) {
        ## Record
        conditions[[length(conditions) + 1L]] <<- c
        ## Muffle
        if (inherits(c, "message")) {
          tryInvokeRestart("muffleMessage")
        } else if (inherits(c, "warning")) {
          tryInvokeRestart("muffleWarning")
        } else if (inherits(c, "condition")) {
          ## If there is a "muffle" restart for this condition,
          ## then invoke that restart, i.e. "muffle" the condition
          restarts <- computeRestarts(c)
          for (restart in restarts) {
            name <- restart$name
            if (is.null(name)) next
            if (!grepl("^muffle", name)) next
            tryInvokeRestart(restart)
            break
          }
        }
      }
    })
    
    ## Success
    status <- "ok"
  
    if (isTRUE(res$visible)) {
      res$value
    } else {
      invisible(res$value)
    }
  }

#' Slowly Calculate Sum or Square Roots of Elements
#'
#' @param x Numeric vector to sum or calculate square roots for.
#'
#' @param delay Delay in seconds after each addition/calculation.
#'
#' @param stdout If TRUE, then a text is outputted to the standard output
#' per element.
#'
#' @param message If TRUE, then a message is outputted per element.
#'
#' @param sticky If TRUE, then a "sticky" message is outputted every
#' ten element.
#'
#' @param progress If TRUE, then a progressor is created, otherwise a void
#' progressor that does nothing is used.
#'
#' @return The sum or square roots of all elements in `x`.
#'
#' @section Progress updates:
#' These functions signal [progression] conditions as they progress.
#'
#' @keywords internal
#' @export
slow_sum_p <- function(x, delay = getOption("progressr.demo.delay", 1.0), stdout = getOption("progressr.demo.stdout", FALSE), message = getOption("progressr.demo.message", TRUE), sticky = getOption("progressr.demo.sticky", TRUE), progress = getOption("progressr.demo.progress", TRUE)) {
  ## Hidden options to simplify help asciicast examples
  if (missing(stdout)) stdout <- getOption("progressr.slow_sum_p.stdout", stdout)
  if (missing(message)) message <- getOption("progressr.slow_sum_p.message", message)
  if (missing(sticky)) sticky <- getOption("progressr.slow_sum_p.sticky", sticky)
  if (missing(progress)) progress <- getOption("progressr.slow_sum_p.progress", progress)

  if (progress) {
    p <- progressor(along = x)
  } else {
    p <- function(...) NULL
  }

  sum <- 0
  storage.mode(sum) <- storage.mode(x)
  
  for (kk in seq_along(x)) {
    p(amount = 0)   ## "I'm alive" progression update
    Sys.sleep(0.2*delay)
    if (stdout) cat(sprintf("O: Element #%d\n", kk))
    Sys.sleep(0.2*delay)
    p(amount = 0)
    Sys.sleep(0.2*delay)
    sum <- sum + x[kk]
    p(message = sprintf("P: Adding %g", kk))
    Sys.sleep(0.2*delay)
    if (message) message(sprintf("M: Added value %g", x[kk]))
    p(amount = 0)
    Sys.sleep(0.2*delay)
    if (sticky && kk %% 10 == 0) {
      p(
        amount = 0,
        message = sprintf("P: %d elements done", kk),
        class = "sticky"
      )
    }
  }

  p(amount = 0)

  sum
}

#' @rdname slow_sum_p
#' @export
slow_sum <- function(x, delay = getOption("progressr.demo.delay", 1.0), stdout = FALSE, message = FALSE, sticky = FALSE, progress = FALSE) {
  slow_sum_p(x = x, delay = delay, stdout = stdout, message = message, sticky = sticky, progress = progress)
}

#' @rdname slow_sum_p
#' @export
slow_sqrt_p <- function(x, delay = getOption("progressr.demo.delay", 0.5), stdout = getOption("progressr.demo.stdout", FALSE), message = getOption("progressr.demo.message", TRUE), sticky = getOption("progressr.demo.sticky", TRUE), progress = getOption("progressr.demo.progress", TRUE)) {
  ## Hidden options to simplify help asciicast examples
  if (missing(stdout)) stdout <- getOption("progressr.slow_sqrt_p.stdout", stdout)
  if (missing(message)) message <- getOption("progressr.slow_sqrt_p.message", message)
  if (missing(sticky)) sticky <- getOption("progressr.slow_sqrt_p.sticky", sticky)
  if (missing(progress)) progress <- getOption("progressr.slow_sqrt_p.progress", progress)

  if (progress) {
    p <- progressor(along = x)
  } else {
    p <- function(...) NULL
  }

  res <- numeric(length(x))
  for (kk in seq_along(x)) {
    p(amount = 0)   ## "I'm alive" progression update
    Sys.sleep(0.2*delay)
    if (stdout) cat(sprintf("O: Element #%d\n", kk))
    Sys.sleep(0.2*delay)
    p(amount = 0)
    Sys.sleep(0.2*delay)
    res[kk] <- sqrt(x[kk])
    p(message = sprintf("P: Calculating sqrt(%g)", kk))
    Sys.sleep(0.2*delay)
    if (message) message(sprintf("M: Calculated value %g", res[kk]))
    p(amount = 0)
    Sys.sleep(0.2*delay)
    if (sticky && kk %% 10 == 0) {
      p(
        amount = 0,
        message = sprintf("P: %d elements done", kk),
        class = "sticky"
      )
    }
  }

  p(amount = 0)

  res
}

#' @rdname slow_sum_p
#' @export
slow_sqrt <- function(x, delay = getOption("progressr.demo.delay", 0.5), stdout = FALSE, message = FALSE, sticky = FALSE, progress = FALSE) {
  slow_sqrt_p(x = x, delay = delay, stdout = stdout, message = message, sticky = sticky, progress = progress)
}

source("incl/start.R")

options(progressr.tests.fake_handlers = c("handler_beepr", "handler_notifier", "handler_pbmcapply", "handler_tkprogressbar", "handler_winprogressbar"))
options(progressr.clear = FALSE)
options(progressr.enable_after = NULL)
options(progressr.debug = FALSE)
options(progressr.times = NULL)

record_output <- function(expr, envir = parent.frame()) {
  conditions <- list()
  stdout <- utils::capture.output({
    withCallingHandlers(
      expr,
      condition = function(c) {
        if (inherits(c, c("progression", "error"))) return()
        conditions[[length(conditions) + 1L]] <<- c
      }
    )
  }, split = TRUE)
  list(stdout = stdout, conditions = conditions)
}


message("*** with_progress() - delaying output ...")

x <- 1:10

## Record truth
output_truth <- record_output({
  y_truth <- slow_sum(x, stdout = TRUE, message = TRUE, sticky = TRUE)
})

for (delay in c(FALSE, TRUE)) {
  message(sprintf("- with_progress() - delay = %s ...", delay))
  output <- record_output({
    with_progress({
      y <- slow_sum(x, stdout = TRUE, message = TRUE, sticky = TRUE)
    }, delay_stdout = delay,
      delay_conditions = if (delay) "condition" else character(0L))
  })
  stopifnot(identical(output$stdout, output_truth$stdout))
  stopifnot(identical(output$conditions, output_truth$conditions))
  stopifnot(identical(y, y_truth))
  message(sprintf("- with_progress() - delay = %s ... DONE", delay))
}

message("*** with_progress() - delaying output ... DONE")

source("incl/end.R")

#' Progression Handler: Progress Reported via 'pbmcapply' Progress Bars (Text) in the Terminal
#'
#' A progression handler for [pbmcapply::progressBar()].
#'
#' @inheritParams make_progression_handler
#'
#' @inheritParams handler_txtprogressbar
#'
#' @param char (character) The symbols to form the progress bar for
#' [utils::txtProgressBar()].
#'
#' @param style (character) The progress-bar style according to
#" [pbmcapply::progressBar()].
#'
#' @param substyle (integer) The progress-bar substyle according to
#' [pbmcapply::progressBar()].
#'
#' @param \ldots Additional arguments passed to [pbmcapply::progressBar()]
#' and [make_progression_handler()].
#'
#' @section Requirements:
#' This progression handler requires the \pkg{pbmcapply} package.
#'
#' @section Appearance:
#' Below are a few examples on how to use and customize this progress handler.
#' In all cases, we use `handlers(global = TRUE)`.
#' Since `style = "txt"` corresponds to using [handler_txtprogressbar()]
#' with `style = substyle`, the main usage of this handler is with
#' `style = "ETA"` (default) for which `substyle` is ignored.
#'
#' ```{asciicast handler_pbmcapply-default}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' handlers("pbmcapply")
#' y <- slow_sum(1:25)
#' ```
#'
#' @example incl/handler_pbmcapply.R
#'
#' @importFrom utils file_test flush.console txtProgressBar setTxtProgressBar
#' @export
handler_pbmcapply <- function(char = "=", substyle = 3L, style = "ETA", file = stderr(), intrusiveness = getOption("progressr.intrusiveness.terminal", 1), target = "terminal", ...) {
  ## Additional arguments passed to the progress-handler backend
  backend_args <- handler_backend_args(char = char, substyle = substyle, style = style, ...)

  if (!is_fake("handler_pbmcapply")) {
    progressBar <- pbmcapply::progressBar
    eraseTxtProgressBar <- function(pb) {
      pb_env <- environment(pb$getVal)
      with(pb_env, {
        style_eta <- exists(".time0", inherits = FALSE)
        if (!style_eta) {
          if (style == 1L || style == 2L) {
            n <- .nb
          } else if (style == 3L) {
            n <- 3L + nw * width + 6L
          }
        } else {
          ## FIXME: Seems to work; if not, see pbmcapply:::txtProgressBarETA()
          n <- width
        }
        cat("\r", strrep(" ", times = n), "\r", sep = "", file = file)
        .nb <- 0L
        flush.console()
      })
    }
  } else {
    progressBar <- function(..., style, substyle) txtProgressBar(..., style = substyle)
    setTxtProgressBar <- function(...) NULL
    eraseTxtProgressBar <- function(pb) NULL
    redrawTxtProgressBar <- function(pb) NULL
  }
  
  reporter <- local({
    ## Import functions

    pb <- NULL
    
    make_pb <- function(max, ...) {
      if (!is.null(pb)) return(pb)
      
      ## SPECIAL CASE: pbmcapply::progressBar() does not support max == min
      ## (its 'min' argument defaults to 0)
      if (max == 0) {
        pb_tmp <- txtProgressBar()
        class(pb_tmp) <- c("voidProgressBar", class(pb_tmp))
      } else {
        args <- c(list(max = max, ...), backend_args)
        pb_tmp <- do.call(progressBar, args = args)
      }
      pb <<- pb_tmp
      
      pb
    }

    list(
      reset = function(...) {
        pb <<- NULL
      },

      hide = function(...) {
        if (is.null(pb)) return()
        eraseTxtProgressBar(pb)
      },

      unhide = function(...) {
        if (is.null(pb)) return()
        redrawTxtProgressBar(pb)
      },

      interrupt = function(config, state, progression, ...) {
        eraseTxtProgressBar(pb)
        redrawTxtProgressBar(pb)
        msg <- conditionMessage(progression)
        msg <- paste(c("", msg, ""), collapse = "\n")
        cat(msg, file = file)
      },

      initiate = function(config, state, progression, ...) {
        if (!state$enabled || config$times == 1L) return()
        stop_if_not(is.null(pb))
        make_pb(max = config$max_steps, file = file)
      },
        
      update = function(config, state, progression, ...) {
        if (!state$enabled || config$times <= 2L) return()
        make_pb(max = config$max_steps, file = file)
        stop_if_not(!is.null(pb))
        if (inherits(progression, "sticky")) {
          eraseTxtProgressBar(pb)
          message(paste0(state$message, ""))
          redrawTxtProgressBar(pb)
        }
        if (progression$amount == 0) return()
        setTxtProgressBar(pb, value = state$step)
      },
        
      finish = function(config, state, progression, ...) {
        ## Already finished?
        if (is.null(pb)) return()
        if (!state$enabled) return()
        if (config$clear) {
          eraseTxtProgressBar(pb)
          ## Suppress newline outputted by close()
          pb_env <- environment(pb$getVal)
          file <- pb_env$file
          pb_env$file <- tempfile()
          on.exit({
            if (file_test("-f", pb_env$file)) file.remove(pb_env$file)
            pb_env$file <- file
          })
        } else {
          setTxtProgressBar(pb, value = state$step)
        }
        close(pb)
	pb <<- NULL
      }
    )
  })
  
  make_progression_handler("pbmcapply", reporter, intrusiveness = intrusiveness, target = target, ...)
}

## Record original state
ovars <- ls()
oenvs <- oenvs0 <- Sys.getenv()
oopts0 <- options()

## Default options for tests
oopts <- options()

check_full <- isTRUE(as.logical(Sys.getenv("R_CHECK_FULL", "FALSE")))

## Private
`%||%` <- progressr:::`%||%`
hpaste <- progressr:::hpaste
mdebug <- progressr:::mdebug
mprint <- progressr:::mprint
mprintf <- progressr:::mprintf
mstr <- progressr:::mstr
query_r_cmd_check <- progressr:::query_r_cmd_check
in_r_cmd_check <- progressr:::in_r_cmd_check
stop_if_not <- progressr:::stop_if_not
printf <- function(...) cat(sprintf(...))
known_progression_handlers <- progressr:::known_progression_handlers

is_rstudio_console <- function() {
  (Sys.getenv("RSTUDIO") == "1") && !nzchar(Sys.getenv("RSTUDIO_TERM"))
}

non_supported_progression_handlers <- function() {
  names <- character(0L)
  for (pkg in c("beepr", "notifier", "pbmcapply", "progress", "shiny")) {
    if (!requireNamespace(pkg, quietly = TRUE))
      names <- c(names, pkg)
  }
  if (!"tcltk" %in% capabilities()) {
    names <- c(names, "tkprogressbar")
  }
  if (.Platform$OS.type != "windows") {
    names <- c(names, "winprogressbar")
  }
  if (!is_rstudio_console()) {
    names <- c(names, "rstudio")
  }
  if (!check_full) {
    names <- c(names, "notifier")
    names <- c(names, "shiny")
  }
  names <- unique(names)
  sprintf("handler_%s", names)
}


supported_progress_handlers <- function(exclude = non_supported_progression_handlers()) {
  handlers <- known_progression_handlers()
  drop <- na.omit(match(exclude, names(handlers)))
  if (length(drop) > 0L) handlers <- handlers[-drop]
  handlers
}

## Settings
options(progressr.clear = TRUE)
options(progressr.debug = FALSE)
options(progressr.demo.delay = 0.0)
options(progressr.enable = TRUE)
options(progressr.enable_after = 0.0)
options(progressr.interval = 0.1)
options(progressr.times = +Inf)


options(progressr.tests.fake_handlers = c(non_supported_progression_handlers(), "handler_beepr", "handler_notifier", "handler_progress"))


future_strategies <- c("multisession", "sequential")
if (.Platform$OS.type != "windows") {
  future_strategies <- c(future_strategies, "multicore")
}


## WORKAROUND: Make sure tests also work with 'covr' package
covr <- ("covr" %in% loadedNamespaces())
if (covr) {
  globalenv <- function() parent.frame()
  baseenv <- function() environment(base::sample)
}



capture_output <- function(..., split = FALSE, collapse = NULL) {
  bfr <- capture.output(..., split = split)
  if (!is.null(collapse)) bfr <- paste(c(bfr, ""), collapse = "\n")
  bfr
}

record_conditions <- function(expr, ..., classes = "condition", split = FALSE) {
  conditions <- list()
  withCallingHandlers(expr, condition = function(c) {
    if (inherits(c, classes)) {
      attr(c, "received") <- Sys.time()
      conditions[[length(conditions) + 1L]] <<- c
      if (!split) muffle_condition(c)
    }
  })
  conditions
}

record_relay <- function(..., all = FALSE, split = FALSE) {
  stdout <- capture_output(conditions <- record_conditions(...), split = split)
  msgs <- sapply(conditions, FUN = conditionMessage)
  res <- list(stdout = stdout, msgs = msgs)
  if (all) res$conditions <- conditions
  res
}

muffle_condition <- function(cond) {
  muffled <- FALSE
  if (inherits(cond, "message")) {
    invokeRestart("muffleMessage")
    muffled <- TRUE
  } else if (inherits(cond, "warning")) {
    invokeRestart("muffleWarning")
    muffled <- TRUE
  } else if (inherits(cond, "condition")) {
    restarts <- computeRestarts(cond)
    for (restart in restarts) {
      name <- restart$name
      if (is.null(name)) 
          next
      if (!grepl("^muffle", name)) 
          next
      invokeRestart(restart)
      muffled <- TRUE
      break
    }
  }
  invisible(muffled)
}

## Adopted from R.utils::cmsg()
console_msg <- function(..., collapse = "\n", sep = "\n", appendLF = TRUE) {
  fh <- tempfile()
  on.exit(file.remove(fh))
  cat(..., collapse = sep, sep = sep, file = fh)
  if (appendLF) 
    cat("\n", file = fh, append = TRUE)
  if (.Platform$OS.type == "windows") {
    file.show(fh, pager = "console", header = "", title = "",
              delete.file = FALSE)
  } else {
    system(sprintf("cat %s", fh))
  }
  invisible()
}

#' Use Progressr in Shiny Apps: Plug-in Backward-Compatible Replacement for shiny::withProgress()
#'
#' A plug-in, backward-compatible replacement for [shiny::withProgress()].
#'
#' @inheritParams handler_shiny
#'
#' @param expr,\ldots,env,quoted Arguments passed to [shiny::withProgress()] as is.
#'
#' @param message,detail (character string) The message and the detail message to be passed to [shiny::withProgress()].
#' 
#' @param handlers Zero or more progression handlers used to report on progress.
#'
#' @return The value of [shiny::withProgress].
#'
#' @example incl/withProgressShiny.R
#'
#' @section Requirements:
#' This function requires the \pkg{shiny} package and will use the
#' [handler_shiny()] **progressr** handler internally to report on updates.
#'
#' @export
withProgressShiny <- function(expr, ..., message = NULL, detail = NULL, inputs = list(message = NULL, detail = "message"), env = parent.frame(), quoted = FALSE, handlers = c(shiny = handler_shiny, progressr::handlers(default = NULL))) {
  if (!quoted) expr <- substitute(expr)

  stop_if_not(is.list(inputs), all(names(inputs) %in% c("message", "detail")))

  stop_if_not("shiny" %in% names(handlers))
  if (sum(names(handlers) == "shiny") > 1) {
    warning("Detected a 'shiny' handler set via progressr::handlers()")
  }

  ## Optional, configure 'inputs' from attribute 'input' of arguments
  ## 'message' and 'detail', if and only if that attribute is available.
  args <- list(message = message, detail = detail)
  for (name in names(args)) {
    input <- unique(attr(args[[name]], "input"))
    if (is.null(input)) next
    unknown <- setdiff(input, c("message", "sticky_message", "non_sticky_message"))
    if (length(unknown) > 0) {
      stop(sprintf("Unknown value of attribute %s on argument %s: %s",
           sQuote("input"), sQuote(name), commaq(unknown)))
    }
    inputs[[name]] <- input
  }

  stop_if_not(
    is.list(inputs),
    !is.null(names(inputs)),
    all(names(inputs) %in% c("message", "detail")),
    all(vapply(inputs, FUN = function(x) {
      if (is.null(x)) return(TRUE)
      if (!is.character(x)) return(FALSE)
      x %in% c("message", "non_sticky_message", "sticky_message")
    }, FUN.VALUE = FALSE))
  )

  ## Customize the shiny 'message' target?
  if (is.function(handlers$shiny) &&
      !inherits(handlers$shiny, "progression_handler")) {
    tweaked_handler_shiny <- handlers$shiny
    if (!identical(inputs, formals(tweaked_handler_shiny)$inputs)) {
      formals(tweaked_handler_shiny)$inputs <- inputs
      handlers$shiny <- tweaked_handler_shiny
    }
  }
  
  expr <- bquote(progressr::with_progress({.(expr)}, handlers = .(handlers)))
  res <- withVisible(shiny::withProgress(expr, ..., message = message, detail = detail, env = env, quoted = TRUE))
  if (res$visible) res$value else invisible(res$value)
}

source("incl/start.R")

options(progressr.clear = FALSE)

message("txtprogressbar ...")

for (style in 1:3) {
  message(sprintf("- style = %d ...", style))
  handlers(handler_txtprogressbar(style = style))
  
  for (x in list(integer(0), 1:10, 1L)) {
    message("length(x): ", length(x))
    with_progress({
      progress <- progressor(along = x)
      for (ii in x) {
        Sys.sleep(getOption("progressr.demo.delay", 0.1))
        progress(message = sprintf("(%s)", paste(letters[1:ii], collapse="")))
      }
    })
  }
  
  message(sprintf("- style = %d ... done", style))
}

message("txtprogressbar ... done")

source("incl/end.R")

if (requireNamespace("cli", quietly = TRUE)) {
  library(progressr)
  options(progressr.clear = FALSE)

  message("handler_cli() ...")

  options(progressr.handlers = handler_cli)
  
  # Test 'type' argument defaults for format and format_done
  for (type in c("default", "steps", "percent", "time")) {
    h <- handler_cli(type = type)
    env <- environment(environment(h)$reporter$initiate)
    backend_args <- parent.env(env)$backend_args
    if (type == "default") {
      stopifnot(is.null(backend_args$format), is.null(backend_args$format_done))
    } else if (type == "steps") {
      stopifnot(
        backend_args$format == "{cli::pb_spin} {cli::pb_bar} {cli::pb_current}/{cli::pb_total} {cli::pb_status}",
        backend_args$format_done == "{cli::pb_bar} {cli::pb_current}/{cli::pb_total} {cli::pb_status}"
      )
    } else if (type == "percent") {
      stopifnot(
        backend_args$format == "{cli::pb_spin} {cli::pb_bar} {cli::pb_percent} {cli::pb_status}",
        backend_args$format_done == "{cli::pb_bar} {cli::pb_percent} {cli::pb_status}"
      )
    } else if (type == "time") {
      stopifnot(
        backend_args$format == "[{cli::pb_elapsed}] {cli::pb_spin} {cli::pb_bar} {cli::pb_percent} [{cli::pb_current}/{cli::pb_total}] (ETA: {cli::pb_eta}) {cli::pb_status}",
        backend_args$format_done == "[{cli::pb_elapsed}] {cli::pb_bar} {cli::pb_percent} [{cli::pb_current}/{cli::pb_total}] {cli::pb_status}"
      )
    }
  }

  for (x in list(integer(0), 1:10, 1L)) {
    message("length(x): ", length(x))
    with_progress({
      progress <- progressor(along = x)
      for (ii in x) {
        Sys.sleep(getOption("progressr.demo.delay", 0.1))
        progress(message = sprintf("(%s)", paste(letters[1:ii], collapse="")))
      }
    })
  }

  message("handler_cli() ... done")
} ## if (requireNamespace("cli", ...))

#' Progression Handler: Progress Reported as Plain Progress Bars (Text) in the Terminal
#'
#' A progression handler for [utils::txtProgressBar()].
#'
#' @inheritParams make_progression_handler
#'
#' @param char (character) The symbols to form the progress bar for
#' [utils::txtProgressBar()]. Contrary to `txtProgressBar()`, this handler
#' supports also ANSI-colored symbols.
#'
#' @param style (integer) The progress-bar style according to
#' [utils::txtProgressBar()].
#'
#' @param file (connection) A [base::connection] to where output should be sent.
#'
#' @param \ldots Additional arguments passed to [utils::txtProgressBar()]
#' and [make_progression_handler()].
#'
#' @section Appearance:
#' Below are a few examples on how to use and customize this progress handler.
#' In all cases, we use `handlers(global = TRUE)`.
#'
#' ```{asciicast handler_txtprogressbar-default}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' handlers("txtprogressbar")
#' y <- slow_sum(1:25)
#' ```
#'
#' ```{asciicast handler_txtprogressbar-style-1}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' handlers(handler_txtprogressbar(style = 1L))
#' y <- slow_sum(1:25)
#' ```
#'
#' ```{asciicast handler_txtprogressbar-style-3}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' handlers(handler_txtprogressbar(style = 3L))
#' y <- slow_sum(1:25)
#' ```
#'
#' ```{asciicast handler_txtprogressbar-char}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' handlers(handler_txtprogressbar(char = "#"))
#' y <- slow_sum(1:25)
#' ```
#'
#' ```{asciicast handler_txtprogressbar-char-width-2}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' handlers(handler_txtprogressbar(char = "<>"))
#' y <- slow_sum(1:25)
#' ```
#'
#' ```{asciicast handler_txtprogressbar-char-ansi}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' handlers(handler_txtprogressbar(char = cli::col_red(cli::symbol$heart)))
#' y <- slow_sum(1:25)
#' ```
#'
#' @example incl/handler_txtprogressbar.R
#'
#' @importFrom utils file_test flush.console setTxtProgressBar
#' @export
handler_txtprogressbar <- function(char = "=", style = 3L, file = stderr(), intrusiveness = getOption("progressr.intrusiveness.terminal", 1), target = "terminal", ...) {
  ## Additional arguments passed to the progress-handler backend
  backend_args <- handler_backend_args(char = char, style = style, ...)

  reporter <- local({
    pb <- NULL

    make_pb <- function(max, ...) {
      if (!is.null(pb)) return(pb)
      args <- c(list(max = max, ...), backend_args)
      pb <<- do.call(txtProgressBar2, args = args)
      pb
    }

    list(
      reset = function(...) {
        pb <<- NULL
      },

      hide = function(...) {
        if (is.null(pb)) return()
        eraseTxtProgressBar(pb)
      },

      unhide = function(...) {
        if (is.null(pb)) return()
        redrawTxtProgressBar(pb)
      },

      interrupt = function(config, state, progression, ...) {
        if (is.null(pb)) return()
        
        eraseTxtProgressBar(pb)
        redrawTxtProgressBar(pb)
        
        msg <- conditionMessage(progression)
        msg <- paste(c("", msg, ""), collapse = "\n")
        cat(msg, file = file)
      },

      initiate = function(config, state, progression, ...) {
        if (!state$enabled || config$times == 1L) return()
        stop_if_not(is.null(pb))
        make_pb(max = config$max_steps, file = file)
      },

      update = function(config, state, progression, ...) {
        if (!state$enabled || config$times == 1L) return()
        make_pb(max = config$max_steps, file = file)
        if (inherits(progression, "sticky")) {
          eraseTxtProgressBar(pb)
          message(paste0(state$message, ""))
          redrawTxtProgressBar(pb)
        }
        if (progression$amount == 0) return()
        setTxtProgressBar(pb, value = state$step)
      },
        
      finish = function(config, state, progression, ...) {
        ## Already finished?
        if (is.null(pb)) return()
        if (!state$enabled) return()
        if (config$clear) {
          eraseTxtProgressBar(pb)
          ## Suppress newline outputted by close()
          pb_env <- environment(pb$getVal)
          file <- pb_env$file
          pb_env$file <- tempfile()
          on.exit({
            if (file_test("-f", pb_env$file)) file.remove(pb_env$file)
            pb_env$file <- file
          })
        } else {
          setTxtProgressBar(pb, value = config$max_steps)
        }
        close(pb)
        pb <<- NULL
      }
    )
  })

  
  make_progression_handler("txtprogressbar", reporter, intrusiveness = intrusiveness, target = target, ...)
}


## Erase a utils::txtProgressBar()
eraseTxtProgressBar <- function(pb) {
  if (inherits(pb, "voidProgressBar")) return()
  pb_env <- environment(pb$getVal)
  with(pb_env, {
    if (style == 1L || style == 2L) {
      n <- .nb
    } else if (style == 3L) {
      n <- 3L + nw * width + 6L
    }
    cat("\r", strrep(" ", times = n), "\r", sep = "", file = file)        
    flush.console()

    ## Reset internal counter, cf. utils::txtProgressBar()
    .nb <- 0L
    .pc <- -1L
  })
}

## Redraw a utils::txtProgressBar()
redrawTxtProgressBar <- function(pb) {
  if (inherits(pb, "voidProgressBar")) return()
  setTxtProgressBar(pb, value = pb$getVal())
}


#' @importFrom utils txtProgressBar
txtProgressBar2 <- function(min = 0, max = 1, ..., char) {
  ## SPECIAL CASE: utils::txtProgressBar() does not support max == min
  if (max == min) {
    pb <- txtProgressBar()
    class(pb) <- c("voidProgressBar", class(pb))
    return(pb)
  }

  ## SPECIAL CASE: Support ANSI-colored 'char' strings
  clean_char <- drop_ansi(char)
  pb <- txtProgressBar(min = min, max = max, ..., char = clean_char)
  if (clean_char != char) {
    env <- environment(pb$up)
    env$char <- char
    env$nw <- nchar(clean_char)
  }
  
  pb
}
    

source("incl/start.R")

if (requireNamespace("plyr", quietly = TRUE)) {
  message("* with_progress()")

  with_progress({
    y <- plyr::llply(3:6, function(n, ...) {
      slow_sum(1:n, stdout=TRUE, message=TRUE)
    }, .progress = "progressr")
  })



  message("* global progression handler")

  handlers(global = TRUE)
    
  local({
    y <- plyr::llply(3:6, function(n, ...) {
      slow_sum(1:n, stdout=TRUE, message=TRUE)
    }, .progress = "progressr")
  })
    
  handlers(global = FALSE)
}

source("incl/end.R")

source("incl/start.R")

message("progress_aggregator() ...")

with_progress({
  progress <- progressor(steps = 1 + 3 + 10 + 1)
  relay_progress <- progress_aggregator(progress)
  progress()
  relay_progress(slow_sum(1:3))
  relay_progress(slow_sum(1:10))
  progress()
})



message("- Stray progressions from unknown sources")

slow_prod <- function(x, delay = getOption("progressr.demo.delay", 0.05)) {
  progress <- progressor(2*length(x))
  res <- 0
  for (kk in seq_along(x)) {
    progress(message = sprintf("Multiplying %g", kk))
    Sys.sleep(0.8*delay)
    res <- res * x[kk]
    progress(message = "...")
    Sys.sleep(0.2*delay)
  }
  res
}

## This will only show progress for the *first* of the three
## functions that report progress. Any progression updates from
## the second and third will be ignored, because they are from
## a different source
with_progress({
  x <- 1:10
  a <- slow_sum(x)
  b <- slow_prod(x)
  c <- slow_sum(-x)
})


## To get progression from all of them, we need to know how many
## steps they report on and then use a gather-and-relay handler
with_progress({
  x <- 1:10
  progress <- progressor(3*length(x))
  relay_progress <- progress_aggregator(progress)
  relay_progress({
    a <- slow_sum(x)
    b <- slow_prod(x)
    c <- slow_sum(-x)
  })
})


message("progress_aggregator() ... done")

source("incl/end.R")

#' Progression Handler: Progress Reported via the Pushbullet Messaging Service
#'
#' A progression handler for [RPushbullet::pbPost()] of the \pkg{RPushbullet}
#' package.
#'
#' @inheritParams make_progression_handler
#'
#' @return A function of class `progression_handler` that takes a
#' [progression] condition as its first and only argument.
#' @inheritParams RPushbullet::pbPost
#'
#' @param \ldots Additional arguments passed to [make_progression_handler()].
#'
#' @example incl/handler_rpushbullet.R
#'
#' @section Requirements:
#' This progression handler requires the \pkg{RPushbullet} package, a
#' Pushbullet account, and configuration according to the instructions
#' of the \pkg{RPushbullet} package.  It also requires internet access
#' from the computer where this progress handler is registered.
#'
#' @export
handler_rpushbullet <- function(intrusiveness = getOption("progressr.intrusiveness.rpushbullet", 5), target = "gui", ..., title = "Progress update from R", recipients = NULL, email = NULL, channel = NULL, apikey = NULL, devices = NULL) {
  ## Used for package testing purposes only when we want to perform
  ## everything except the last part where the backend is called
  pbPost <- function(...) NULL
  if (!is_fake("handler_rpushbullet")) {
    pkg <- "RPushbullet"
    if (!requireNamespace(pkg, quietly = TRUE)) {
      stop("Package RPushbullet is not available")
    }
    if (is_rpushbullet_working()) {
      pbPost <- get("pbPost", mode = "function", envir = getNamespace(pkg))
    }
  }

  notifier <- function(title, message) {
    args <- list(
      type  = "note",
      title = title,
      body  = paste(c("", message), collapse = "")
    )

    if (!is.null(recipients)) args$recipients <- recipients
    if (!is.null(email)) args$email <- email
    if (!is.null(channel)) args$channel <- channel
    if (!is.null(apikey)) args$apikey <- apikey
    if (!is.null(devices)) args$devices <- devices

    do.call(pbPost, args = args)
  }

  reporter <- local({
    finished <- FALSE
    
    list(
      reset = function(...) {
        finished <<- FALSE
      },
      
      initiate = function(config, state, progression, ...) {
        if (!state$enabled || config$times == 1L) return()
        progress_notify(title = title, step = state$step, max_steps = config$max_steps, message = state$message, notifier = notifier)
      },
        
      interrupt = function(config, state, progression, ...) {
        msg <- conditionMessage(progression)
        progress_notify(title = title, step = state$step, max_steps = config$max_steps, message = msg, notifier = notifier)
      },

      update = function(config, state, progression, ...) {
        if (!state$enabled || progression$amount == 0 || config$times <= 2L) return()
        progress_notify(title = title, step = state$step, max_steps = config$max_steps, message = state$message, notifier = notifier)
      },
        
      finish = function(config, state, progression, ...) {
        if (finished) return()
        if (!state$enabled) return()
        if (state$delta > 0) progress_notify(title = title, step = state$step, max_steps = config$max_steps, message = state$message, notifier = notifier)
	finished <<- TRUE
      }
    )
  })
  
  make_progression_handler("rpushbullet", reporter, intrusiveness = intrusiveness, target = target, ...)
}
attr(handler_rpushbullet, "validator") <- function() is_rpushbullet_working()


is_rpushbullet_working <- local({
  res <- NA
  pkg <- "RPushbullet"
  
  function(quiet = TRUE) {
    if (!is.na(res)) return(res)

    ## Assert RPushbullet package
    if (!requireNamespace(pkg, quietly = TRUE)) {
      warning(sprintf("Package %s is not installed", sQuote(pkg)), immediate. = TRUE)
      return(FALSE)
    }

    ## Assert internet access
    curl <- "curl"
    if (!requireNamespace(curl, quietly = TRUE)) {
      warning(sprintf("Package %s is not installed", sQuote(curl)), immediate. = TRUE)
      return(FALSE)
    }
    has_internet <- get("has_internet", mode = "function", envir = getNamespace(curl))
    if (!has_internet()) {
      warning("No internet access. The 'rpushbullet' progress handler requires working internet.", immediate. = TRUE)
      return(FALSE)
    }

    ## Validate RPushbullet configuration with 10-second timeout
    timeout <- 10.0
    setTimeLimit(cpu = timeout, elapsed = timeout, transient = TRUE)
    on.exit({
      setTimeLimit(cpu = Inf, elapsed = Inf, transient = FALSE)
    })

    conds <- list()
    withCallingHandlers({
      tryCatch({
        res <<- RPushbullet::pbValidateConf()
      }, error = function(ex) {
        conds <<- c(conds, list(ex))
      })
    }, message = function(cond) {
       conds <<- c(conds, list(cond))
       if (quiet) tryInvokeRestart("muffleMessage")
    }, warning = function(cond) {
       conds <<- c(conds, list(cond))
       if (quiet) tryInvokeRestart("muffleWarning")
    })

    if (is.na(res)) res <- FALSE
    if (!res) {
      msg <- vapply(conds, FUN.VALUE = NA_character_, FUN = conditionMessage)
      msg <- gsub("\n$", "", msg)
      msg <- gsub("^", "    ", msg)
      warning(paste(c("The 'rpushbullet' progress handler will not work, because RPushbullet is not properly configured. See help(\"pbSetup\", package = \"RPushbullet\") for instructions. RPushbullet::pbValidateConf() reported:", msg), collapse = "\n"), immediate. = TRUE)
    }

    res
  }
})

if (getRversion() >= "4.0.0") {

source("incl/start.R")

nsinks0 <- sink.number(type = "output")

options(progressr.clear = FALSE)

delay <- getOption("progressr.demo.delay", 0.1)
message("- delay: ", delay, " seconds")

handlers("txtprogressbar")

handlers <- supported_progress_handlers()

handlers(global = FALSE)
stopifnot(sink.number(type = "output") == nsinks0)

handlers(global = TRUE)
stopifnot(sink.number(type = "output") == nsinks0)

message("global progress handlers - standard output, messages, warnings ...")

n <- 5L
for (kk in seq_along(handlers)) {
  handler <- handlers[[kk]]
  name <- names(handlers)[kk]
  message(sprintf("* Handler %d ('%s') of %d ...", kk, name, length(handlers)))

  for (type in c("message", "warning")) {
    message(sprintf("  - stdout + %ss", type))
    for (delta in c(0L, +1L, -1L)) {
      message(sprintf("    - delta = %+d", delta))

      handlers(global = FALSE)
      stopifnot(sink.number(type = "output") == nsinks0)
      handlers(global = TRUE)
      stopifnot(sink.number(type = "output") == nsinks0)

      status <- progressr:::register_global_progression_handler("status")
      stopifnot(
        is.null(status$current_progressor_uuid),
        is.null(status$delays),
        is.null(status$stdout_file),
         length(status$conditions) == 0L,
         is.na(status$capture_conditions)
      )

      nsinks <- sink.number(type = "output")
      stopifnot(nsinks == nsinks0)
      
      truth <- c()
      relay <- record_relay(local({
        p <- progressor(n)
        for (ii in seq_len(n + delta)) {
          ## Zero-amount progress with empty message
          p(amount = 0)
          msg <- sprintf("ii = %d", ii)
          ## Zero-amount progress with non-empty message
          p(message = msg, amount = 0)
          truth <<- c(truth, msg)
          cat(msg, "\n", sep = "")
          ## Signal condition
          do.call(type, args = list(msg))
          Sys.sleep(delay)
          ## One-step progress with non-empty message
          p(message = sprintf("(%s)", paste(letters[1:ii], collapse=",")))
        }
      }), classes = type)
      stopifnot(
        identical(relay$stdout, truth),
        identical(gsub("\n$", "", relay$msgs), truth)
      )
      
      ## Assert sinks are balanced
      stopifnot(sink.number(type = "output") == nsinks)
      
      cat(paste(c(relay$stdout, ""), collapse = "\n"))
      message(relay$message, append = FALSE)
      status <- progressr:::register_global_progression_handler("status")
      console_msg(capture.output(utils::str(status)))
      if (delta == 0L) {
        withCallingHandlers({
          stopifnot(
            is.null(status$current_progressor_uuid),
            is.null(status$delays),
            is.null(status$stdout_file),
            length(status$conditions) == 0L,
            is.na(status$capture_conditions)
          )
        }, error = function(ex) {
          console_msg(paste("An error occurred:", conditionMessage(ex)))
          console_msg(capture.output(utils::str(status)))
        })
      }

    } ## for (delta ...)
  } ## for (signal ...)

  message(sprintf("* Handler %d ('%s') of %d ... done", kk, name, length(handlers)))
}


message("global progress handlers - standard output, messages, warnings ... done")

handlers(global = FALSE)

source("incl/end.R")

} ## if (getRversion() >= "4.0.0")

if (requireNamespace("cli", quietly = TRUE)) {
  library(progressr)
  options(progressr.clear = FALSE)
  
  message("handler_pbcol ...")
  
  # 1. Test the internal pbcol() function directly
  pbcol <- progressr:::pbcol
  
  message("Testing pbcol() internals ...")
  
  # Test default parameters and basic outputs
  res <- pbcol(fraction = 0.0)
  stopifnot(is.character(res))
  
  res <- pbcol(fraction = 0.5, msg = "Hello")
  stopifnot(is.character(res))
  
  res <- pbcol(fraction = 1.0, msg = "Hello")
  stopifnot(is.character(res))
  
  # Test adjust parameter
  for (adjust in c(0.0, 0.5, 1.0)) {
    res <- pbcol(fraction = 0.5, msg = "Hello", adjust = adjust)
    stopifnot(is.character(res))
  }
  
  # Test msg = NULL / length 0
  res <- pbcol(fraction = 0.5, msg = character(0))
  stopifnot(is.character(res))
  
  res <- pbcol(fraction = 0.5, msg = NULL)
  stopifnot(is.character(res))
  
  # Test truncation when message is too long
  res <- pbcol(fraction = 0.5, msg = paste(rep("x", 200), collapse = ""), width = 40)
  stopifnot(is.character(res))
  
  # Test custom complete / incomplete functions
  res <- pbcol(fraction = 0.5, msg = "Hello", complete = NULL, incomplete = NULL)
  stopifnot(is.character(res))
  
  res <- pbcol(fraction = 0.5, msg = "Hello", complete = function(s) paste0("[C:", s, "]"), incomplete = function(s) paste0("[I:", s, "]"))
  stopifnot(is.character(res))
  
  # Test invalid inputs (should throw errors)
  assert_error <- function(expr) {
    res <- tryCatch(eval(expr), error = identity)
    stopifnot(inherits(res, "error"))
  }
  
  assert_error(pbcol(msg = 123))
  assert_error(pbcol(msg = c("a", "b")))
  assert_error(pbcol(fraction = -0.1))
  assert_error(pbcol(fraction = 1.1))
  assert_error(pbcol(fraction = NA_real_))
  assert_error(pbcol(fraction = c(0.1, 0.2)))
  assert_error(pbcol(width = 0L))
  assert_error(pbcol(width = -5L))
  assert_error(pbcol(width = NA_integer_))
  assert_error(pbcol(adjust = -0.1))
  assert_error(pbcol(adjust = 1.1))
  
  
  # 2. Test the progression handler directly on various progression conditions
  message("Testing handler_pbcol() on progression conditions ...")
  
  h <- handler_pbcol(clear = FALSE)
  stopifnot(inherits(h, "pbcol_progression_handler"))
  
  # Run standard progression lifecycle
  h(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-1", progression_index = 1L))
  h(progression(type = "update", amount = 1, progressor_uuid = "uuid-1", progression_index = 2L))
  h(progression(type = "update", amount = 0, progressor_uuid = "uuid-1", progression_index = 3L))
  # Negative amount to trigger state$delta < 0 branch
  h(progression(type = "update", amount = -1, progressor_uuid = "uuid-1", progression_index = 4L))
  h(progressr:::control_progression("hide"))
  h(progressr:::control_progression("unhide"))
  h(progressr:::control_progression("interrupt", message = "Interrupt message"))
  h(progressr:::control_progression("shutdown"))
  
  # Run with clear = TRUE
  h_clear <- handler_pbcol(clear = TRUE)
  h_clear(progression(type = "initiate", steps = 5, auto_finish = TRUE, progressor_uuid = "uuid-2", progression_index = 1L))
  h_clear(progression(type = "update", amount = 1, progressor_uuid = "uuid-2", progression_index = 2L))
  h_clear(progression(type = "finish", progressor_uuid = "uuid-2", progression_index = 3L))
  
  # Run with times <= 2L (returns early in reporter functions)
  h_times <- handler_pbcol(times = 2L)
  h_times(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-3", progression_index = 1L))
  h_times(progression(type = "update", amount = 1, progressor_uuid = "uuid-3", progression_index = 2L))
  h_times(progressr:::control_progression("unhide"))
  h_times(progressr:::control_progression("interrupt", message = "Interrupt message"))
  h_times(progressr:::control_progression("shutdown"))
  
  # Run with max_steps = 0
  h_zero <- handler_pbcol()
  h_zero(progression(type = "initiate", steps = 0, auto_finish = TRUE, progressor_uuid = "uuid-4", progression_index = 1L))
  h_zero(progressr:::control_progression("shutdown"))
  
  # Run with enable_after > 0 (enabled = FALSE initially)
  h_enable <- handler_pbcol(enable_after = 100.0)
  h_enable(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-5", progression_index = 1L))
  h_enable(progression(type = "update", amount = 1, progressor_uuid = "uuid-5", progression_index = 2L))
  h_enable(progressr:::control_progression("unhide"))
  h_enable(progressr:::control_progression("interrupt", message = "Interrupt message"))
  h_enable(progressr:::control_progression("shutdown"))
  
  # Test different crayon.enabled settings
  options(crayon.enabled = FALSE)
  h_crayon_false <- handler_pbcol()
  h_crayon_false(progression(type = "initiate", steps = 5, auto_finish = TRUE, progressor_uuid = "uuid-6", progression_index = 1L))
  h_crayon_false(progression(type = "update", amount = 1, progressor_uuid = "uuid-6", progression_index = 2L))
  h_crayon_false(progressr:::control_progression("shutdown"))
  
  options(crayon.enabled = TRUE)
  h_crayon_true <- handler_pbcol()
  h_crayon_true(progression(type = "initiate", steps = 5, auto_finish = TRUE, progressor_uuid = "uuid-7", progression_index = 1L))
  h_crayon_true(progression(type = "update", amount = 1, progressor_uuid = "uuid-7", progression_index = 2L))
  h_crayon_true(progressr:::control_progression("shutdown"))
  
  
  # 3. Test using standard with_progress() flow
  message("Testing handler_pbcol() with standard with_progress() flow ...")
  
  for (x in list(integer(0), 1:10, 1L)) {
    message("length(x): ", length(x))
    with_progress({
      progress <- progressor(along = x)
      for (ii in x) {
        Sys.sleep(getOption("progressr.demo.delay", 0.1))
        progress(message = sprintf("(%s)", paste(letters[1:ii], collapse="")))
      }
    }, handlers = handler_pbcol())
  }
  
  message("handler_pbcol ... done")
} ## if (requireNamespace("cli", ...))

#' Creates and Signals a Progression Condition
#'
#' _WARNING:_ `progress()` is defunct - don't use.
#'
#' @param \ldots Arguments pass to [progression()].
#'
#' @param call (expression) A call expression.
#'
#' @return A [base::condition] of class `progression`.
#'
#' @seealso
#' To create a progression condition, use [progression()].
#' To signal a progression condition, use [base::signalCondition()].
#'
#' @keywords internal
#' @export
progress <- function(..., call = sys.call()) {
  action <- getOption("progressr.lifecycle.progress", "defunct")
  signal <- switch(action, deprecated = .Deprecated, defunct = .Defunct)
  signal(msg = sprintf("progress() is %s", action), package = .packageName)
 
  args <- list(...)
  if (length(args) == 1L && inherits(args[[1L]], "condition")) {
    cond <- args[[1L]]
    stop_if_not(inherits(cond, "progression"))
  } else {
    cond <- progression(..., call = call)
  }
  
  withRestarts(
    signalCondition(cond),
    muffleProgression = function(p) NULL
  )
  
  invisible(cond)
}

source("incl/start.R")

message("handlers() ...")

hs <- handlers()
print(hs)

for (kk in seq_along(hs)) {
  h <- hs[[kk]]
  print(h)
  handler <- h()
  print(handler)
}


hs <- handlers("txtprogressbar")
print(hs)

for (kk in seq_along(hs)) {
  h <- hs[[kk]]
  print(h)
  handler <- h()
  print(handler)
}

hs <- handlers("handler_txtprogressbar")
print(hs)

message("handlers() - exceptions ...")

## Will as a side-effect set an empty list of handlers()
res <- handlers("non-existing-handler", on_missing = "ignore")
res <- handlers()
stopifnot(is.list(res), length(res) == 0L)

res <- tryCatch({
  handlers("non-existing-handler", on_missing = "warning")
}, warning = identity)
stopifnot(inherits(res, "warning"))

res <- tryCatch({
  handlers("non-existing-handler", on_missing = "error")
}, error = identity)
stopifnot(inherits(res, "error"))


message("handlers() - exceptions ... DONE")

message("handlers() ... DONE")

source("incl/end.R")

library(progressr)

options(progressr.clear = FALSE)

message("handler_slowdown() ...")

# 1. Check invalid inputs
assert_error <- function(expr) {
  res <- tryCatch(eval(expr), error = identity)
  stopifnot(inherits(res, "error"))
}

assert_error(handler_slowdown(duration = -1))
assert_error(handler_slowdown(duration = Inf))
assert_error(handler_slowdown(duration = NA_real_))
assert_error(handler_slowdown(duration = c(1, 2)))
assert_error(handler_slowdown(duration = "foo"))

# 2. Test manual progression calls when enable = FALSE
h_disabled <- handler_slowdown(enable = FALSE)
stopifnot(inherits(h_disabled, "slowdown_progression_handler"))
h_disabled(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-1", progression_index = 1L))
h_disabled(progression(type = "update", amount = 1, progressor_uuid = "uuid-1", progression_index = 2L))
h_disabled(progression(type = "finish", progressor_uuid = "uuid-1", progression_index = 3L))

# 3. Test manual progression calls when enable = TRUE
h_enabled <- handler_slowdown(duration = 0.1, enable = TRUE)
stopifnot(inherits(h_enabled, "slowdown_progression_handler"))

# Initiate
h_enabled(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-2", progression_index = 1L))

# Update (state$enabled is TRUE)
h_enabled(progression(type = "update", amount = 1, progressor_uuid = "uuid-2", progression_index = 2L))

# Finish
h_enabled(progression(type = "finish", progressor_uuid = "uuid-2", progression_index = 3L))

# 4. Test when state$enabled is FALSE (returns early in update)
h_not_enabled <- handler_slowdown(duration = 0.1, enable_after = 100.0, enable = TRUE)
h_not_enabled(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-3", progression_index = 1L))
# Update (state$enabled is FALSE, returns early)
h_not_enabled(progression(type = "update", amount = 1, progressor_uuid = "uuid-3", progression_index = 2L))
h_not_enabled(progression(type = "finish", progressor_uuid = "uuid-3", progression_index = 3L))

# 5. Run with_progress
message("with_progress(..., handlers = handler_slowdown(duration = 0.05))")
with_progress({
  p <- progressor(along = 1:2)
  for (i in 1:2) {
    p()
  }
}, handlers = handler_slowdown(duration = 0.05))

message("handler_slowdown() ... done")

#' Creates a Progression Calling Handler
#'
#' A progression calling handler is a function that takes a [base::condition]
#' as its first argument and that can be use together with
#' [base::withCallingHandlers()].  This function helps creating such
#' progression calling handler functions.
#'
#' @param name (character) Name of progression handler.
#'
#' @param reporter (list) A named list of reporter functions.
#'
#' @param handler (function) Function take a [progression] condition
#'   as the first argument.
#'
#' @param intrusiveness (numeric) A non-negative scalar on how intrusive
#'   (disruptive) the reporter to the user.
#'
#' @param enable (logical) If FALSE, then progress is not reported.
#'
#' @param enable_after (numeric) Delay (in seconds) before progression
#'   updates are reported.
#'
#' @param times (numeric) The maximum number of times this handler
#'   should report progression updates.
#'   If zero, then progress is not reported.
#'
#' @param interval (numeric) The minimum time (in seconds) between
#'   successive progression updates from this handler.
#'
#' @param clear (logical) If TRUE, any output, typically visual, produced
#'   by a reporter will be cleared/removed upon completion, if possible.
#'
#' @param target (character vector) Specifies where progression updates are
#'   rendered.
#'
#' @param \ldots Not used.
#'
#' @return A function of class `progression_handler` that takes a
#' [progression] condition as its first and only argument.
#'
#' @details
#' The inner details of progression handlers and how to use this function
#' are still to be documented.  Until then, see the source code of existing
#' handlers for how it is used, e.g. `progressr::handler_txtprogressbar`.
#' Please use with care as things might change.
#'
#' @section Reporter functions:
#' The `reporter` argument should be a named list of zero or more of the
#' following functions:
#'
#' * `initiate`
#' * `update`
#' * `finish`
#'
#' These functions are called whenever a [progression] condition of type
#' `"initiate"`, `"update"`, or `"finish"` are received, but only if the
#' condition is for the progression that is currently handled.
#' These functions are called with the following arguments (in order):
#'
#' * `config` - a named list of the configuration of the progression handler:
#'              `max_steps` (integer),
#"              `times` (integer),
#'              `interval` (numeric),
#'              `enable_after` (numeric),
#'              `auto_finish` (logical),
#'              `clear` (logical),
#'              `target` (character vector)
#'
#' * `state` - a named list of the current progress state after accounting
#'             for the most recent `progression` condition:
#'             `step` (integer), `message` (character),
#'             `delta` (integer),
#'             `enabled` (logical),
#'             `timestamps` (POSIXct vector)
#'
#' * `progression` - a [progression] condition
#'
#' * \ldots - not used (reserved for future needs)
#' 
#' In addition to the above functions, the following functions:
#'
#' * `hide`
#' * `unhide`
#' * `reset`
#' * `interrupt`
#'
#' are called whenever the handler "should" hide or unhide the rendered
#' progress, or reset it, or when an interrupt is detected.  In these cases,
#' the `progression` argument is of class `control_progression`.
#'
#' @example incl/make_progression_handler.R
#'
#' @seealso
#' [base::withCallingHandlers()].
#'
#' @keywords internal
#' @export
make_progression_handler <- function(name, reporter = list(), handler = NULL, enable = getOption("progressr.enable", interactive()), enable_after = getOption("progressr.enable_after", 0.0), times = getOption("progressr.times", +Inf), interval = getOption("progressr.interval", 0.0), intrusiveness = 1.0, clear = getOption("progressr.clear", TRUE), target = "terminal", ...) {
  enable <- as.logical(enable)
  stop_if_not(is.logical(enable), length(enable) == 1L, !is.na(enable))
  if (!enable) times <- 0
  name <- as.character(name)
  stop_if_not(length(name) == 1L, !is.na(name), nzchar(name))
#  stop_if_not(is.function(handler))
#  formals <- formals(handler)
#  stop_if_not(length(formals) == 1L)
  stop_if_not(is.list(reporter))
  enable_after <- as.numeric(enable_after)
  stop_if_not(is.numeric(enable_after), length(enable_after),
              !is.na(enable_after), enable_after >= 0)
  times <- as.numeric(times)
  stop_if_not(length(times) == 1L, is.numeric(times), !is.na(times),
              times >= 0)
  interval <- as.numeric(interval)
  stop_if_not(length(interval) == 1L, is.numeric(interval),
              !is.na(interval), interval >= 0)
  clear <- as.logical(clear)
  stop_if_not(is.logical(clear), length(clear) == 1L, !is.na(clear))
  stop_if_not(is.character(target))
  
  ## Disable progress updates?
  if (times == 0 || is.infinite(interval) || is.infinite(intrusiveness)) {
    handler <- function(p) NULL
  }

  ## Set default reporter functions
  for (key in setdiff(c("reset", "initiate", "update", "finish", "hide", "unhide", "interrupt"), names(reporter))) {
    reporter[[key]] <- structure(function(...) NULL, class = "null_function")
  }

  ## Progress state
  active <- FALSE
  max_steps <- NULL
  step <- NULL
  message <- NULL
  auto_finish <- TRUE
  timestamps <- NULL
  milestones <- NULL
  prev_milestone <- NULL
  finished <- FALSE
  enabled <- FALSE

  ## Progress cache
  owner <- NULL
  done <- list()

  ## Sanity checks
  .validate_internal_state <- function(label = "<no-label>") {
    error <- function(...) {
      msg <- sprintf(...)
      stop(sprintf(".validate_internal_state(%s): %s", sQuote(label), msg))
    }
    if (!is.null(timestamps)) {
      if (length(timestamps) == 0L) {
        error(paste("length(timestamps) == 0L but not is.null(timestamps):",
                    sQuote(deparse(timestamps))))
      }
    }
  }
  
  reporter_args <- function(progression) {
    .validate_internal_state("reporter_args() ... begin")
    
    if (!enabled && !is.null(timestamps)) {
      dt <- difftime(Sys.time(), timestamps[1L], units = "secs")
      enabled <<- (dt >= enable_after)
    }

    config <- list(
      max_steps = max_steps,
      times = times,
      interval = interval,
      enable_after = enable_after,
      auto_finish = auto_finish,
      clear = clear,
      target = target
    )

    state <- list(
      step = step,
      message = message,
      timestamps = timestamps,
      delta = step - prev_milestone,
      enabled = enabled
    )
    if (length(state$delta) == 0L) state$delta <- 0L

    .validate_internal_state("reporter_args() ... end")

    c(config, state, list(
      config = config,
      state = state,
      progression = progression
    ))    
  }

  reset_internal_state <- function() {
    ## Progress state
    active <<- FALSE
    max_steps <<- NULL
    step <<- NULL
    message <<- NULL
    auto_finish <<- TRUE
    timestamps <<- NULL
    milestones <<- NULL
    prev_milestone <<- NULL
    finished <<- FALSE
    enabled <<- FALSE
  
    ## Progress cache
    owner <<- NULL
    done <<- list()
  }

  reset_reporter <- function(p) {
    args <- reporter_args(progression = p)
    debug <- getOption("progressr.debug", FALSE)
    if (debug) {
      mprintf("reset_reporter() ...")
      mstr(args)
    }
    nsinks <- sink.number()
    do.call(reporter$reset, args = args)
    stop_if_not(sink.number() == nsinks)
    .validate_internal_state("reset_reporter() ... done")
    if (debug) mprintf("reset_reporter() ... done")
  }

  initiate_reporter <- function(p) {
    args <- reporter_args(progression = p)
    debug <- getOption("progressr.debug", FALSE)
    if (debug) {
      mprintf("initiate_reporter() ...")
      mstr(args)
    }
    stop_if_not(!isTRUE(active))
    stop_if_not(is.null(prev_milestone), length(milestones) > 0L)
    nsinks <- sink.number()
    do.call(reporter$initiate, args = args)
    stop_if_not(sink.number() == nsinks)
    active <<- TRUE
    finished <<- FALSE
    .validate_internal_state("initiate_reporter() ... done")
    if (debug) mprintf("initiate_reporter() ... done")
  }

  update_reporter <- function(p) {
    args <- reporter_args(progression = p)
    debug <- getOption("progressr.debug", FALSE)
    if (debug) {
      mprintf("update_reporter() ...")
      mstr(args)
    }
    stop_if_not(isTRUE(active))
    stop_if_not(!is.null(step), length(milestones) > 0L)
    nsinks <- sink.number()
    do.call(reporter$update, args = args)
    stop_if_not(sink.number() == nsinks)
    .validate_internal_state("update_reporter() ... done")
    if (debug) mprintf("update_reporter() ... done")
  }

  hide_reporter <- function(p) {
    args <- reporter_args(progression = p)
    debug <- getOption("progressr.debug", FALSE)
    if (debug) {
      mprintf("hide_reporter() ...")
      mstr(args)
    }
#    stop_if_not(isTRUE(active))
    if (is.null(reporter$hide)) {
      if (debug) mprintf("hide_reporter() ... skipping; not supported")
      return()
    }
    nsinks <- sink.number()
    do.call(reporter$hide, args = args)
    stop_if_not(sink.number() == nsinks)
    .validate_internal_state("hide_reporter() ... done")
    if (debug) mprintf("hide_reporter() ... done")
  }

  unhide_reporter <- function(p) {
    args <- reporter_args(progression = p)
    debug <- getOption("progressr.debug", FALSE)
    if (debug) {
      mprintf("unhide_reporter() ...")
      mstr(args)
    }
#    stop_if_not(isTRUE(active))
    if (is.null(reporter$unhide)) {
      if (debug) mprintf("unhide_reporter() ... skipping; not supported")
      return()
    }
    nsinks <- sink.number()
    do.call(reporter$unhide, args = args)
    stop_if_not(sink.number() == nsinks)
    .validate_internal_state("unhide_reporter() ... done")
    if (debug) mprintf("unhide_reporter() ... done")
  }

  interrupt_reporter <- function(p) {
    args <- reporter_args(progression = p)
    debug <- getOption("progressr.debug", FALSE)
    if (debug) {
      mprintf("interrupt_reporter() ...")
      mstr(args)
    }
#    stop_if_not(isTRUE(active))
    if (is.null(reporter$interrupt)) {
      if (debug) mprintf("interrupt_reporter() ... skipping; not supported")
      return()
    }
    nsinks <- sink.number()
    do.call(reporter$interrupt, args = args)
    stop_if_not(sink.number() == nsinks)
    .validate_internal_state("interrupt_reporter() ... done")
    if (debug) mprintf("interrupt_reporter() ... done")
  }

  finish_reporter <- function(p) {
    debug <- getOption("progressr.debug", FALSE)

    if (active && !finished) update_progress(p, debug = debug)

    args <- reporter_args(progression = p)
    if (debug) {
      mprintf("finish_reporter() ...")
      mstr(args)
    }

    ## Signal 'finish' if active and not already finished
    ## because it could already have been auto-finished before
    if (active && !finished) {
      update_progress(p, debug = debug)
      nsinks <- sink.number()
      do.call(reporter$finish, args = args)
      stop_if_not(sink.number() == nsinks)
    } else {
      if (debug) {
        why <- if (!active && !finished) {
          "not active"
        } else if (!active && finished) {
          "not active and already finished"
        } else if (active && finished) {
          "already finished"
        }
        message(sprintf("- Hmm ... got a request to 'finish' handler, but it's %s. Oh well, will finish it then", why))
      }
    }
    
    reset_internal_state()
    finished <<- TRUE
    if (debug) message("- owner: ", deparse(owner))
    .validate_internal_state("finish_reporter() ... done")
    if (debug) mprintf("finish_reporter() ... done")
  }

  is_owner <- function(p) {
    progressor_uuid <- p[["progressor_uuid"]]
    if (is.null(owner)) owner <<- progressor_uuid
    (owner == progressor_uuid)
  }

  is_duplicated <- function(p) {
    progressor_uuid <- p[["progressor_uuid"]]
    session_uuid <- p[["session_uuid"]]
    progression_index <- p[["progression_index"]]
    progression_time <- p[["progression_time"]]
    progression_id <- sprintf("%s-%d-%s", session_uuid, progression_index, progression_time)
    db <- done[["progressor_uuid"]]
    res <- is.element(progression_id, db)
    if (!res) {
      db <- c(db, progression_id)
      done[["progressor_uuid"]] <<- db
    }
    res
  }

  ## Used by type == "update" and type == "finish"
  update_progress <- function(p, debug = FALSE) {
    if (!active) {
      if (debug) message("- cannot 'update' handler, because it is not active")
      return(invisible(finished))
    }
    if (debug) mstr(list(step=step, "p$amount"=p[["amount"]], "p$step"=p[["step"]], max_steps=max_steps))
    if (!is.null(p[["step"]])) {
      ## Infer effective 'amount' from previous 'step' and p$step
      p[["amount"]] <- p[["step"]] - step
    }
    step <<- min(max(step + p[["amount"]], 0L), max_steps)
    stop_if_not(step >= 0L)
    msg <- conditionMessage(p)
    if (length(msg) > 0) message <<- msg
    if (step > 0) timestamps[step] <<- Sys.time()
    if (debug) mstr(list(finished = finished, step = step, milestones = milestones, prev_milestone = prev_milestone, interval = interval))
    .validate_internal_state("type=update")

    ## Only update if a new milestone step has been reached ...
    ## ... or if we want to send a zero-amount update
    if ((length(milestones) > 0L && step >= milestones[1]) ||
        p[["amount"]] == 0) {
      skip <- FALSE
      if (interval > 0 && step > 0) {
        dt <- difftime(timestamps[step], timestamps[max(prev_milestone, 1L)], units = "secs")
        skip <- (dt < interval)
        if (debug) mstr(list(dt = dt, timestamps[step], timestamps[prev_milestone], skip = skip))
      }
      if (!skip) {
        if (debug) mstr(list(milestones = milestones))
        update_reporter(p)
        if (p[["amount"]] > 0) prev_milestone <<- step
      }
      if (p[["amount"]] > 0) {
        milestones <<- milestones[milestones > step]
        if (auto_finish && step == max_steps) {
          if (debug) mstr(list(type = "finish (auto)", milestones = milestones))
          finish_reporter(p)
        }
      }
    }
  } ## update_progress()


  if (is.null(handler)) {
    handler <- function(p) {
      stop_if_not(inherits(p, "progression"))

      ## Ignore if running in a forked child process
      if (is_fork_child()) return(invisible(FALSE))
      
      if (inherits(p, "control_progression")) {
        type <- p[["type"]]
        if (type == "reset") {
          reset_internal_state()
          reset_reporter(p)
          .validate_internal_state(sprintf("handler(type=%s) ... end", type))
        } else if (type == "shutdown") {
          finish_reporter(p)
          .validate_internal_state(sprintf("handler(type=%s) ... end", type))
        } else if (type == "hide") {
          hide_reporter(p)
          .validate_internal_state(sprintf("handler(type=%s) ... end", type))
        } else if (type == "unhide") {
          unhide_reporter(p)
          .validate_internal_state(sprintf("handler(type=%s) ... end", type))
        } else if (type == "interrupt") {
          interrupt_reporter(p)
          .validate_internal_state(sprintf("handler(type=%s) ... end", type))
        } else {
          stop("Unknown 'control_progression' type: ", sQuote(type))
        }
        .validate_internal_state(sprintf("control_progression ... end", type))
        return(invisible(finished))
      }        

      debug <- getOption("progressr.debug", FALSE)

      ## Ignore stray progressions coming from other sources, e.g.
      ## a function of a package that started to report on progression.
      if (!is_owner(p)) {
        if (debug) message("- not owner of this progression. Skipping")
        return(invisible(finished))
      }
      
      duplicated <- is_duplicated(p)
      
      type <- p[["type"]]
      if (debug) {
        mprintf("Progression calling handler %s ...", sQuote(type))
        mprintf("- progression:")
        mstr(p)
        mprintf("- progressor_uuid: %s", p[["progressor_uuid"]])
        mprintf("- progression_index: %s", p[["progression_index"]])
        mprintf("- duplicated: %s", duplicated)
      }

      if (duplicated) {
        if (debug) mprintf("Progression calling handler %s ... condition already done", sQuote(type))
        return(invisible(finished))
      } else if (active && finished) {
        if (debug) mprintf("Progression calling handler %s ... active but already finished", sQuote(type))
        return(invisible(finished))
      }

      if (type == "initiate") {
        if (active) {
          if (debug) message("- cannot 'initiate' handler, because it is already active")
          return(invisible(finished))
        }
        max_steps <<- p[["steps"]]
        if (debug) mstr(list(max_steps=max_steps))
        stop_if_not(!is.null(max_steps), is.numeric(max_steps), length(max_steps) == 1L, max_steps >= 0)
        auto_finish <<- p[["auto_finish"]]
        times <- min(times, max_steps)
        if (debug) mstr(list(auto_finish = auto_finish, times = times, interval = interval, intrusiveness = intrusiveness))
        
        ## Adjust 'times' and 'interval' according to 'intrusiveness'
        times <- min(c(times / intrusiveness, max_steps), na.rm = TRUE)
        times <- max(times, 1L)
        interval <- interval * intrusiveness
        if (debug) mstr(list(times = times, interval = interval))

        ## Milestone steps that need to be reach in order to trigger an
        ## update of the reporter
        milestones <<- if (times == 1L) {
          c(max_steps)
        } else if (times == 2L) {
          c(0L, max_steps)
        } else {
          seq(from = 0L, to = max_steps, length.out = times + 1L)[-1]
        }

        ## Timestamps for when steps where reached
        ## Note that they will remain NA for "skipped" steps
        timestamps <<- rep(as.POSIXct(NA), times = max_steps)
        timestamps[1] <<- Sys.time()
        
        step <<- 0L
        message <<- character(0L)
        if (debug) mstr(list(finished = finished, milestones = milestones))
        initiate_reporter(p)
        prev_milestone <<- step
        .validate_internal_state(sprintf("handler(type=%s) ... end", type))
      } else if (type == "finish") {
        if (debug) mstr(list(finished = finished, milestones = milestones))
        finish_reporter(p)
        .validate_internal_state("type=finish")
      } else if (type == "update") {
        update_progress(p, debug = debug)
        .validate_internal_state(sprintf("handler(type=%s) ... end", type))
      } else {
        ## Was this meant to be a 'control_progression' condition?
        if (type %in% c("reset", "shutdown", "hide", "unhide", "interrupt")) {
          stop("Unsupported 'progression' type. Was it meant to be a 'control_progression' condition?: ", sQuote(type))
        } else {
          stop("Unknown 'progression' type: ", sQuote(type))
        }
      }

      ## Sanity checks
      .validate_internal_state(sprintf("handler() ... end", type))

      if (debug) mprintf("Progression calling handler %s ... done", sQuote(type))
      invisible(finished)
    } ## handler()
  }

  class(handler) <- c(sprintf("%s_progression_handler", name),
                      "progression_handler", "calling_handler",
                      class(handler))
      
  handler
}


#' @export
print.progression_handler <- function(x, ...) {
  s <- sprintf("Progression calling handler of class %s:", sQuote(class(x)[1]))
  
  env <- environment(x)
  s <- c(s, " * configuration:")
  s <- c(s, sprintf("   - name: %s", sQuote(env$name %||% "<NULL>")))
  s <- c(s, sprintf("   - max_steps: %s", env$max_steps %||% "<NULL>"))
  s <- c(s, sprintf("   - enable: %s", env$enable))
  s <- c(s, sprintf("   - enable_after: %g seconds", env$enable_after))
  s <- c(s, sprintf("   - times: %g", env$times))
  s <- c(s, sprintf("   - interval: %g seconds", env$interval))
  s <- c(s, sprintf("   - intrusiveness: %g", env$intrusiveness))
  s <- c(s, sprintf("   - auto_finish: %s", env$auto_finish))
  s <- c(s, sprintf("   - clear: %s", env$clear))
  s <- c(s, sprintf("   - target: %s", paste(sQuote(env$target), collapse = ", ")))
  s <- c(s, sprintf("   - milestones: %s", hpaste(env$milestones %||% "<NULL>")))
  s <- c(s, sprintf("   - owner: %s", hpaste(env$owner %||% "<NULL>")))

  s <- c(s, " * state:")
  s <- c(s, sprintf("   - enabled: %s", env$enabled))
  s <- c(s, sprintf("   - finished: %s", env$finished))
  s <- c(s, sprintf("   - step: %s", env$step %||% "<NULL>"))
  s <- c(s, sprintf("   - message: %s", env$message %||% "<NULL>"))
  s <- c(s, sprintf("   - prev_milestone: %s", env$prev_milestone %||% "<NULL>"))
  s <- c(s, sprintf("   - delta: %g", (env$step - env$prev_milestone) %||% 0L))
  s <- c(s, sprintf("   - timestamps: %s", hpaste(env$timestamps %||% "<NULL>")))

  s <- paste(s, collapse = "\n")
  cat(s, "\n", sep = "")
}


# Additional arguments passed to the progress backend
handler_backend_args <- function(...) {
  args <- list(...)
  if (length(args) == 0L) return(list())
  
  names <- names(args)
  if (is.null(names) || !all(nzchar(names))) {
    stop("Additional arguments must be named")
  }
  
  ## Drop arguments passed to make_progression_handler()
  names <- setdiff(names, names(formals(make_progression_handler)))
  args[names]
}


# Map progression message to optional target arguments
# Currently used by handler_shiny() and handler_winprogressbar()
message_to_backend_targets <- function(progression, inputs, message = progression$message) {
  if (length(message) == 0) return(list())

  args <- list()
  for (target in names(inputs)) {
    if (inherits(progression, "sticky")) {
      if ("sticky_message" %in% inputs[[target]])
        args[[target]] <- message
    } else {
      if ("non_sticky_message" %in% inputs[[target]])
        args[[target]] <- message
    }
  }

  args
}
  

#' Create a Progressor Function that Signals Progress Updates
#'
#' @inheritParams progression
#'
#' @param steps (integer) Number of progressing steps.
#'
#' @param along (vector; alternative) Alternative that sets
#' `steps = length(along)`.
#'
#' @param offset,scale (integer; optional) scale and offset applying transform
#' `steps <- scale * steps + offset`.
#'
#' @param transform (function; optional) A function that takes the effective
#' number of `steps` as input and returns another finite and non-negative
#' number of steps.
#'
#' @param label (character) A label.
#'
#' @param trace (logical) If TRUE, then the call stack is recorded, otherwise
#' not.
#'
#' @param initiate (logical) If TRUE, the progressor will signal a
#' [progression] 'initiate' condition when created.
#'
#' @param finalize (logical) If TRUE, the progressor signals a [progression]
#' 'shutdown' condition when finalized by the garbage collector.
#'
#' @param auto_finish (logical) If TRUE, then the progressor will signal a
#' [progression] 'finish' condition as soon as the last step has been reached.
#'
#' @param enable (logical) If TRUE, [progression] conditions are signaled when
#' calling the progressor function created by this function.
#' If FALSE, no [progression] conditions is signaled because the progressor
#' function is an empty function that does nothing.
#'
#' @param on_exit,envir (logical) If TRUE, then the created progressor will
#' signal a [progression] 'finish' condition when the calling frame exits.
#' This is ignored if the calling frame (`envir`) is the global environment.
#'
#' @return A function of class `progressor`.
#'
#' @details
#' A `progressor` function can only be created inside a local environment,
#' e.g. inside a function, within a `local()` call, or within a
#' `with_progress()` call.  Notably, it _cannot_ be create at the top level,
#' e.g. immediately at the R prompt or outside a local environment in an
#' R script.  If attempted, an informative error message is produced, e.g.
#'
#' ```r
#' > p <- progressr::progressor(100)
#' Error in progressr::progressor(100) : 
#'   A progressor must not be created in the global environment unless
#' wrapped in a with_progress() or without_progress() call. Alternatively,
#' create it inside a function or in a local() environment to make sure
#' there is a finite life span of the progressor
#' ```
#'
#' @export
progressor <- local({
  progressor_count <- 0L

  void_progressor <- function(...) NULL
  environment(void_progressor)$enable <- FALSE
  class(void_progressor) <- c("progressor", class(void_progressor))

  function(steps = length(along), along = NULL, offset = 0L, scale = 1L, transform = function(steps) scale * steps + offset, message = character(0L), label = NA_character_, trace = FALSE, initiate = TRUE, finalize = FALSE, auto_finish = TRUE, on_exit = !identical(envir, globalenv()), enable = getOption("progressr.enable", TRUE), envir = parent.frame()) {
    stop_if_not(is.logical(enable), length(enable) == 1L, !is.na(enable))

    ## Quickly return a moot progressor function?
    if (!enable) return(void_progressor)
    
    stop_if_not(!is.null(steps) || !is.null(along))
    stop_if_not(length(steps) == 1L, is.numeric(steps), !is.na(steps),
                steps >= 0)
    stop_if_not(length(offset) == 1L, is.numeric(offset), !is.na(offset))
    stop_if_not(length(scale) == 1L, is.numeric(scale), !is.na(scale))
    stop_if_not(is.function(transform))

    label <- as.character(label)
    stop_if_not(length(label) == 1L)

    steps <- transform(steps)
    stop_if_not(length(steps) == 1L, is.numeric(steps), !is.na(steps),
                steps >= 0)

    stop_if_not(is.logical(on_exit), length(on_exit) == 1L, !is.na(on_exit))

    if (identical(envir, globalenv())) {
      if (!progressr_in_globalenv()) {
        stop("A progressor must not be created in the global environment unless wrapped in a with_progress() or without_progress() call. Alternatively, create it inside a function or in a local() environment to make sure there is a finite life span of the progressor")
      }
      if (on_exit) {
        stop("It is not possible to create a progressor in the global environment with on_exit = TRUE")
      }
    }

    owner_session_uuid <- session_uuid(attributes = TRUE)
    progressor_count <<- progressor_count + 1L
    progressor_uuid <- progressor_uuid(progressor_count)
    progression_index <- 0L

    fcn <- function(message = character(0L), ..., type = "update") {
      progression_index <<- progression_index + 1L
      cond <- progression(
        type = type,
        message = message,
        ...,
        progressor_uuid = progressor_uuid,
        progression_index = progression_index,
        owner_session_uuid = owner_session_uuid,
        call = if (trace) sys.call() else NULL,
        calls = if (trace) sys.calls() else NULL
      )
      withRestarts(
        signalCondition(cond),
        muffleProgression = function(p) NULL
      )
      invisible(cond)
    } ## fcn()
    formals(fcn)$message <- message
    class(fcn) <- c("progressor", class(fcn))

    ## WORKAROUND: Use teeny, custom enviroment for the progressor function.
    ## The default would otherwise be to inherit the parent frame, which
    ## might contain very large objects.
    progressor_envir <- new.env(parent = getNamespace(.packageName))
    for (name in c("progression_index", "progressor_uuid",
                   "owner_session_uuid", "progressor_count",
                   "enable", "initiate", "auto_finish", "trace",
                   "steps", "label", "offset", "scale")) {
      progressor_envir[[name]] <- get(name)
    }
    environment(fcn) <- progressor_envir
    
    ## Is there already be an active '...progressr'?
    ## If so, make sure it is finished and then remove it
    if (exists("...progressor", mode = "function", envir = envir)) {
      ...progressor <- get("...progressor", mode = "function", envir = envir)
      
      ## Ideally, we produce a warning or an error here if the existing
      ## progressor is not finished.  Currently, we don't have a way to
      ## query that, so we leave that for the future. /HB 2021-02-28

      ## Finish it (although it might already have been done via auto-finish)
      ...progressor(type = "finish")

      ## Remove it (while avoiding false 'R CMD check' NOTE)
      do.call(unlockBinding, args = list("...progressor", env = envir))
      rm("...progressor", envir = envir)
    }

    ## Initiate?
    if (initiate) {
      fcn(
        type = "initiate",
        steps = steps,
        auto_finish = auto_finish
      )
    }

    ## Add on.exit(...progressor(type = "finish"))?
    if (on_exit) {
      assign("...progressor", value = fcn, envir = envir)
      lockBinding("...progressor", env = envir)
      call <- call("...progressor", type = "finish")
      do.call(base::on.exit, args = list(call, add = TRUE), envir = envir)
    }

    if (finalize) {
      ## Self-reference progressor function needed by the finalizer function
      progressor_envir$self <- fcn
      reg.finalizer(progressor_envir, function(e) {
        print(utils::ls.str(e))
        if (is.function(e$self)) {
          e$self(type = "shutdown")
          e$self <- NULL
        }
      })
    }

    fcn
  }
})



#' @importFrom utils object.size
#' @export
print.progressor <- function(x, ...) {
  s <- sprintf("%s:", class(x)[1])
  e <- environment(x)
  pe <- parent.env(e)

  s <- c(s, paste("- label:", e$label))
  s <- c(s, paste("- steps:", e$steps))
  s <- c(s, paste("- initiate:", e$initiate))
  s <- c(s, paste("- auto_finish:", e$auto_finish))

  if (is.function(e$message)) {
    message <- "<a function>"
  } else {
    message <- hpaste(deparse(e$message))
  }
  s <- c(s, paste("- default message:", message))

  call <- vapply(e$calls, FUN = function(call) deparse(call[1]), FUN.VALUE = "")
  stack <- if (e$trace) paste(call, collapse = " -> ") else "<disabled>"
  s <- c(s, paste("- call stack:", stack))

  s <- c(s, paste("- progressor_uuid:", e$progressor_uuid))
  s <- c(s, paste("- progressor_count:", pe$progressor_count))
  s <- c(s, paste("- progression_index:", e$progression_index))
  owner_session_uuid <- e$owner_session_uuid
  s <- c(s, paste("- owner_session_uuid:", owner_session_uuid))
  
  s <- c(s, paste("- enable:", e$enable))

  size <- object.size(x)
  size2 <- serialization_size(x)
  s <- c(s, sprintf("- size: %s [%s serialized]", format(size, units = "auto", standard = "SI"), format(size2, units = "auto", standard = "SI")))

  s <- paste(s, collapse = "\n")
  cat(s, "\n", sep = "")
  
  invisible(x)
}


#' @export
length.progressor <- function(x) {
  environment(x)[["steps"]]
}


progressr_in_globalenv <- local({
  state <- FALSE
  
  function(action = c("query", "allow", "disallow")) {
    action <- match.arg(action)
    if (action == "query") return(state)
    old_state <- state
    state <<- switch(action, allow = TRUE, disallow = FALSE)
    invisible(old_state)
  }
})

source("incl/start.R")

message("progressor() ...")

message("- basic")
local({
  p <- progressor(3L)
  print(p)
  p()
  p("A message")
})

message("- default message")
local({
  p <- progressor(along = 1:3, message = "A default message")
  print(p)
  p()
  p("A message")
})

message("- zero length")
local({
  p <- progressor(0L)
  print(p)
  p()
  p("A message")
})

message("- multiple consequtive progressors")
local({
  message("Progressor #1")
  p <- progressor(2L)
  for (kk in 1:2) p()

  message("Progressor #2")
  p <- progressor(3L)
  for (kk in 1:3) p()

  message("Done")
})

message("progressor() ... DONE")

source("incl/end.R")

#' Progression Handler: Progress Reported as the Size of a File on the File System
#'
#' @inheritParams make_progression_handler
#'
#' @param file (character) A filename.
#'
#' @param \ldots Additional arguments passed to [make_progression_handler()].
#'
#' @examples
#' \donttest{\dontrun{
#' handlers(handler_filesize(file = "myscript.progress"))
#' with_progress(y <- slow_sum(1:100))
#' print(y)
#' }}
#'
#' @details
#' This progression handler reports progress by updating the size of a file
#' on the file system. This provides a convenient way for an R script running
#' in batch mode to report on the progress such that the user can peek at the
#' file size (by default in 0-100 bytes) to assess the amount of the progress
#' made, e.g. `ls -l -- *.progress`.
#' If the \file{*.progress} file is accessible via for instance SSH, SFTP,
#' FTPS, HTTPS, etc., then progress can be assessed from a remote location.
#'
#' @importFrom utils file_test
#' @export
handler_filesize <- function(file = "default.progress", intrusiveness = getOption("progressr.intrusiveness.file", 5), target = "file", enable = getOption("progressr.enable", TRUE), ...) {
  reporter <- local({
    set_file_size <- function(config, state, progression, message = state$message) {
      ## Troubleshoot https://github.com/futureverse/progressr/issues/168
      stop_if_not(
        length(config$max_steps) == 1L, is.numeric(config$max_steps),
        !is.na(config$max_steps), is.finite(config$max_steps),
        config$max_steps >= 0
      )
      
      ratio <- if (config$max_steps == 0) 1 else state$step / config$max_steps
      size <- round(100 * ratio)
      current_size <- file.size(file)
      if (is.na(current_size)) file.create(file, showWarnings = FALSE)
      if (size == 0L) return()
      if (progression$amount == 0) return()          

      head <- sprintf("%g/%g: ", state$step, config$max_steps)
      nhead <- nchar(head)
      tail <- sprintf(" [%d%%]", round(100 * ratio))
      ntail <- nchar(tail)
      mid <- paste0(message, "")
      nmid <- nchar(mid)
      padding <- size - (nhead + nmid + ntail)
      if (padding <= 0) {
        msg <- paste(head, mid, tail, sep = "")
        if (padding < 0) msg <- substring(msg, first = 1L, last = size)
      } else if (padding > 0) {
        mid <- paste(c(mid, " ", rep(".", times = padding - 1L)), collapse = "")
        msg <- paste(head, mid, tail, sep = "")
      }
      
      cat(file = file, append = FALSE, msg)
    }
    
    list(
      initiate = function(config, state, progression, ...) {
        if (!state$enabled) return()
        set_file_size(config = config, state = state, progression = progression)
      },
      
      interrupt = function(config, state, progression, ...) {
        if (!state$enabled) return()
        msg <- conditionMessage(progression)
        set_file_size(config = config, state = state, progression = progression, message = msg)
      },
      
      update = function(config, state, progression, ...) {
        if (!state$enabled) return()
        set_file_size(config = config, state = state, progression = progression)
      },
      
      finish = function(config, state, progression, ...) {
        if (!state$enabled) return()
        if (config$clear) {
	  if (file_test("-f", file)) file.remove(file)
	} else {
          set_file_size(config = config, state = state, progression = progression)
	}
      }
    )
  })
  
  make_progression_handler("filesize", reporter, intrusiveness = intrusiveness, target = target, enable = enable, ...)
}

## From R.utils 2.0.2 (2015-05-23)
hpaste <- function(..., sep = "", collapse = ", ", lastCollapse = NULL, maxHead = if (missing(lastCollapse)) 3 else Inf, maxTail = if (is.finite(maxHead)) 1 else Inf, abbreviate = "...") {
  if (is.null(lastCollapse)) lastCollapse <- collapse

  # Build vector 'x'
  x <- paste(..., sep = sep)
  n <- length(x)

  # Nothing todo?
  if (n == 0) return(x)
  if (is.null(collapse)) return(x)

  # Abbreviate?
  if (n > maxHead + maxTail + 1) {
    head <- x[seq_len(maxHead)]
    tail <- rev(rev(x)[seq_len(maxTail)])
    x <- c(head, abbreviate, tail)
    n <- length(x)
  }

  if (!is.null(collapse) && n > 1) {
    if (lastCollapse == collapse) {
      x <- paste(x, collapse = collapse)
    } else {
      xT <- paste(x[1:(n-1)], collapse = collapse)
      x <- paste(xT, x[n], sep = lastCollapse)
    }
  }

  x
} # hpaste()

# More efficient than the default utils::capture.output()
#' @importFrom utils capture.output
capture_output <- function(expr, envir = parent.frame(), ...) {
  nsinks <- sink.number()
  res <- eval({
    file <- rawConnection(raw(0L), open = "w")
    on.exit(close(file))
    capture.output(expr, file = file)
    rawToChar(rawConnectionValue(file))
  }, envir = envir, enclos = baseenv())
  stop_if_not(sink.number() == nsinks)
  unlist(strsplit(res, split = "\n", fixed = TRUE), use.names = FALSE)
}

now <- function(x = Sys.time(), format = "[%H:%M:%OS3] ") {
  format(as.POSIXlt(x, tz = ""), format = format)
}

mdebug <- function(..., debug = getOption("progressr.debug", FALSE)) {
  if (!debug) return()
  message(now(), ...)
}

mprintf <- function(..., appendLF = TRUE, debug = getOption("progressr.debug", FALSE)) {
  if (!debug) return()
  message(now(), sprintf(...), appendLF = appendLF)
}

mdebugf <- mprintf

mprint <- function(..., appendLF = TRUE, debug = getOption("progressr.debug", FALSE)) {
  if (!debug) return()
  message(paste(now(), capture_output(print(...)), sep = "", collapse = "\n"), appendLF = appendLF)
}

#' @importFrom utils str
mstr <- function(..., appendLF = TRUE, debug = getOption("progressr.debug", FALSE)) {
  if (!debug) return()
  message(paste(now(), capture_output(str(...)), sep = "", collapse = "\n"), appendLF = appendLF)
}

commaq <- function(x, sep = ", ") paste(sQuote(x), collapse = sep)

trim <- function(s) sub("[\t\n\f\r ]+$", "", sub("^[\t\n\f\r ]+", "", s))



stop_if_not <- function(..., calls = sys.calls()) {
  res <- list(...)
  n <- length(res)
  if (n == 0L) return()

  for (ii in 1L:n) {
    res_ii <- .subset2(res, ii)
    if (length(res_ii) != 1L || is.na(res_ii) || !res_ii) {
        mc <- match.call()
        call <- deparse(mc[[ii + 1]], width.cutoff = 60L)
        if (length(call) > 1L) call <- paste(call[1L], "...")
        msg <- sprintf("%s is not TRUE", sQuote(call))
        if (FALSE) {
          callstack <- paste(as.character(calls), collapse = " -> ")
          msg <- sprintf("%s [call stack: %s]", msg, callstack)
        }
        stop(msg, call. = FALSE, domain = NA)
    }
  }
}


## Used for package testing purposes only when we want to perform
## everything except the last part where the backend is called
## This allows us to cover more of the code in package tests
is_fake <- local({
  cache <- list()
  function(name) {
    fake <- cache[[name]]
    if (is.null(fake)) {
      fake <- name %in% getOption("progressr.tests.fake_handlers")
      cache[[name]] <<- fake
    }
    fake
  }
})

known_progression_handlers <- function(exclude = NULL) {
  ns <- asNamespace(.packageName)
  handlers <- ls(envir = ns, pattern = "^handler_")
  handlers <- setdiff(handlers, c("handler_backend_args", "make_progression_handler", "print.progression_handler"))
  handlers <- setdiff(handlers, exclude)
  handlers <- mget(handlers, envir = ns, inherits = FALSE)
  handlers
}


`%||%` <- function(lhs, rhs) {
  if (is.null(lhs)) rhs else lhs
}


## From R.utils 2.7.0 (2018-08-26)
query_r_cmd_check <- function(...) {
  evidences <- list()

  # Command line arguments
  args <- commandArgs()

  evidences[["vanilla"]] <- is.element("--vanilla", args)

  # Check the working directory
  pwd <- getwd()
  dirname <- basename(pwd)
  parent <- basename(dirname(pwd))
  pattern <- ".+[.]Rcheck$"

  # Is 'R CMD check' checking tests?
  evidences[["tests"]] <- (
    grepl(pattern, parent) && grepl("^tests(|_.*)$", dirname)
  )

  # Is the current working directory as expected?
  evidences[["pwd"]] <- (evidences[["tests"]] || grepl(pattern, dirname))

  # Is 'R CMD check' checking examples?
  evidences[["examples"]] <- is.element("CheckExEnv", search())
  
  # SPECIAL: win-builder?
  evidences[["win-builder"]] <- (.Platform$OS.type == "windows" && grepl("Rterm[.]exe$", args[1]))

  if (evidences[["win-builder"]]) {
    n <- length(args)
    if (all(c("--no-save", "--no-restore", "--no-site-file", "--no-init-file") %in% args)) {
      evidences[["vanilla"]] <- TRUE
    }

    if (grepl(pattern, parent)) {
      evidences[["pwd"]] <- TRUE
    }
  }


  if (!evidences$vanilla || !evidences$pwd) {
    res <- "notRunning"
  } else if (evidences$tests) {
    res <- "checkingTests"
  } else if (evidences$examples) {
    res <- "checkingExamples"
  } else {
    res <- "notRunning"
  }

  attr(res, "evidences") <- evidences
  
  res
}

in_r_cmd_check <- function() { query_r_cmd_check() != "notRunning" }


## Check whether current R process is running as a forked child
is_fork_child <- local({
  main_pid <- NULL
  function() {
    if (is.null(main_pid)) main_pid <<- Sys.getpid()
    Sys.getpid() != main_pid
  }
})


serialization_size <- function(x) {
  size <- length(serialize(x, connection = NULL, xdr = TRUE))
  class(size) <- "object_size"
  size
}


ansi_pattern <- "(?:(?:\\x{001b}\\[)|\\x{009b})(?:(?:[0-9]{1,3})?(?:(?:;[0-9]{0,3})*)?[A-M|f-m])|\\x{001b}[A-M]"

drop_ansi <- function(x) {
  gsub(ansi_pattern, "", x, perl = TRUE, useBytes = TRUE)
}

#' @details
#' `without_progress()` evaluates an expression while ignoring all
#' progress updates.
#'
#' @rdname with_progress
#' @export
without_progress <- function(expr) {
  progressr_in_globalenv("allow")
  on.exit(progressr_in_globalenv("disallow"))

  ## Deactive global progression handler while using without_progress()
  global_progression_handler(FALSE)
  on.exit(global_progression_handler(TRUE), add = TRUE)

  withCallingHandlers({
    res <- withVisible(expr)
  }, progression = function(p) {
    ## Restart 'muffleProgression' is not guaranteed to exist, e.g. the
    ## 'progression' condition might be resignaled by a handler that
    ## does not define it.
    tryInvokeRestart("muffleProgression")
  })
  
  if (isTRUE(res$visible)) {
    res$value
  } else {
    invisible(res$value)
  }
}

#' Progression Handler: Progress Reported as 'BRRR' Sounds (Audio)
#'
#' A progression handler for `skrrrahh()` of the \pkg{BRRR} package.
#'
#' @inheritParams make_progression_handler
#'
#' @return A function of class `progression_handler` that takes a
#' [progression] condition as its first and only argument.
#'
#' @param initiate,update,finish,interrupt (integer or character) Indices or
#' names of `skrrrahh()` sounds to play when progress starts, is updated,
#' completes, or is interrupted. For silence, use `NA_integer_`.
#'
#' @param \ldots Additional arguments passed to [make_progression_handler()].
#'
#' @example incl/handler_brrr.R
#'
#' @section Requirements:
#' This progression handler requires the \pkg{BRRR} package, which is only
#' available from <https://github.com/brooke-watson/BRRR>.  This can be
#' installed as `remotes::install_github("brooke-watson/BRRR")`.
#'
#' @export
handler_brrr <- function(initiate = "takeoff", update = "bigsean4",  finish = "khaled3", interrupt = "busta", intrusiveness = getOption("progressr.intrusiveness.audio", 5.0), target = "audio", ...) {
  ## Used for package testing purposes only when we want to perform
  ## everything except the last part where the backend is called
  if (!is_fake("handler_brrr")) {
    pkg <- "BRRR"
    if (!requireNamespace(pkg, quietly = TRUE)) {
      stop(sprintf("Package '%s' is not available. See ?progressr::handler_notifier() for installation instructions", pkg))
    }
    BRRR_skrrrahh <- get("skrrrahh", mode = "function", envir = getNamespace(pkg))
  } else {
    BRRR_skrrrahh <- function(sound, expr) NULL
  }

  skrrrahh <- function(sound) {
    ## Silence?
    if (is.na(sound)) return()
    BRRR_skrrrahh(sound)
  }

  ## Reporter state
  reporter <- local({
    list(
      initiate = function(config, state, progression, ...) {
        if (!state$enabled || config$times == 1L) return()
        skrrrahh(initiate)
      },
        
      interrupt = function(config, state, progression, ...) {
        skrrrahh(interrupt)
      },
      
      update = function(config, state, progression, ...) {
        if (!state$enabled || progression$amount == 0 || config$times <= 2L) return()
        skrrrahh(update)
      },
        
      finish = function(config, state, progression, ...) {
        if (!state$enabled) return()
        skrrrahh(finish)
      }
    )
  })
  
  make_progression_handler("brrr", reporter, intrusiveness = intrusiveness, target = target, ...)
}

source("incl/start.R")

message("with_progress() - progressr.debug = TRUE ...")

options(progressr.debug = TRUE)

with_progress({
  y <- slow_sum(1:10)
})


with_progress({
  progress <- progressor(steps = 1 + 2 + 1)
  relay_progress <- progress_aggregator(progress)
  p <- progress()
  progressr::progress(p) ## duplicated - will be ignored
  relay_progress(slow_sum(1:2))
  progress(type = "finish")
  progress() ## one too many - will be ignored
})


message("with_progress() - progressr.debug = TRUE ... done")

source("incl/end.R")

warn_about_too_many_progressions <- function(progression, global = FALSE) {
  ## We might receive progress updates after the fact that the progressor
  ## has been completed

  ## It is okay if it's a zero-amount progression
  amount <- progression$amount
  if (is.numeric(amount) && amount == 0) return()

  ## It is okay if it's a 'finish' progression
  type <- progression$type
  if (is.character(type) && type == "finish") return()

  ## This is okay if it's signaled from within with_progress() and then
  ## caught by the global 'progression' handler
  if (global) {
    ## ... unless the 'progression' signaled within with_progress(), which
    ## in case it might be re-signaled such that it reaches the global
    ## 'progression' handler here
    is_with_progress <- function(call) {
      identical(call, quote(with_progress)) || identical(call, quote(progressr::with_progress))
    }
    calls <- lapply(sys.calls(), FUN = .subset2, 1L)
    if (any(vapply(calls, FUN = is_with_progress, FUN.VALUE = FALSE))) {
      return()
    }
  }

  ## In all cases, this is a mistake that we should warn about
  m <- conditionMessage(progression)
  m <- if (length(m) == 0) "character(0)" else sQuote(m)
  info <- sprintf("amount=%g; msg=%s", amount, m)
  msg <- if (global) "The global progression handlers" else "with_progress()"
  msg <- sprintf("%s received a progression %s request (%s), but is no longer listening to this progressor. This can happen when code signals more progress updates than steps in the progressor. To troubleshoot this, retry with progressr::handlers(\"debug\")", msg, sQuote(type), info)
  warning(msg, call. = FALSE, immediate. = TRUE)
} ## warn_about_too_many_progressions()

source("incl/start.R")

if (requireNamespace("pbmcapply", quietly = TRUE)) {
  handlers("pbmcapply")
  for (x in list(integer(0), 1:10, 1L)) {
    message("length(x): ", length(x))
    with_progress({ y <- slow_sum(x) })
    print(y)
  }
}

source("incl/end.R")

library(progressr)

message("*** toc/tic ...")

tic <- progressr:::tic
toc <- progressr:::toc

# Parse functions for testing
parse_elapsed <- function(msg) {
  as.numeric(gsub("^\\[([0-9.]+)\\s+[a-zA-Z]+\\]\n$", "\\1", msg))
}

parse_expr_elapsed <- function(msg) {
  elapsed_tic <- as.numeric(gsub("^\\[([0-9.]+)\\s+[a-zA-Z]+\\]\\s+\\+([0-9.]+)\\s+[a-zA-Z]+\\s+:\\s+.*\n$", "\\1", msg))
  elapsed_dt <- as.numeric(gsub("^\\[([0-9.]+)\\s+[a-zA-Z]+\\]\\s+\\+([0-9.]+)\\s+[a-zA-Z]+\\s+:\\s+.*\n$", "\\2", msg))
  list(tic = elapsed_tic, dt = elapsed_dt)
}

# 1. Test tic() / toc() existence and basic class / visibility
stopifnot(
  is.function(tic),
  is.function(toc)
)

# 2. Test basic tic() and toc() message outputs and parsing
res <- record_relay({
  tic()
  Sys.sleep(0.1)
  toc()
})

stopifnot(
  length(res$msgs) == 2L,
  grepl("^\\[[0-9.]+\\s+[a-zA-Z]+\\]\n$", res$msgs[1]),
  grepl("^\\[[0-9.]+\\s+[a-zA-Z]+\\]\n$", res$msgs[2]),
  parse_elapsed(res$msgs[1]) == 0,
  parse_elapsed(res$msgs[2]) >= 0.09
)

# 3. Test reset argument
res_reset <- record_relay({
  tic()
  Sys.sleep(0.1)
  toc(reset = TRUE)
  Sys.sleep(0.1)
  toc()
})

stopifnot(
  length(res_reset$msgs) == 3L,
  parse_elapsed(res_reset$msgs[1]) == 0,
  parse_elapsed(res_reset$msgs[2]) == 0,
  parse_elapsed(res_reset$msgs[3]) >= 0.09
)

# 4. Test expression evaluation and return value
res_expr <- record_relay({
  tic()
  val <- toc(1 + 2)
  stopifnot(val == 3)
})

stopifnot(
  length(res_expr$msgs) == 2L,
  grepl("^\\[[0-9.]+\\s+[a-zA-Z]+\\]\\s+\\+[0-9.]+\\s+[a-zA-Z]+\\s+:\\s+.*\n$", res_expr$msgs[2])
)

# 5. Test width parameter
res_width <- record_relay({
  tic()
  toc(Sys.sleep(0.01), width = 10L)
})

stopifnot(
  length(res_width$msgs) == 2L,
  # Code should be truncated to "Sys.sleep( ..."
  grepl("Sys.sleep\\( \\.\\.\\.", res_width$msgs[2])
)

# 6. Test expression with error
res_err <- record_relay({
  tic()
  try(toc(stop("my_error")), silent = TRUE)
})

stopifnot(
  length(res_err$msgs) == 2L,
  grepl("stop\\(\"my_error\"\\)", res_err$msgs[2])
)

message("*** toc/tic ... DONE")

library(progressr)

message("handlers() ...")

hs <- handlers()
print(hs)

for (kk in seq_along(hs)) {
  h <- hs[[kk]]
  print(h)
  handler <- h()
  print(handler)
}


hs <- handlers("txtprogressbar")
print(hs)

for (kk in seq_along(hs)) {
  h <- hs[[kk]]
  print(h)
  handler <- h()
  print(handler)
}

hs <- handlers("handler_txtprogressbar")
print(hs)

message("handlers() - exceptions ...")

## Will as a side-effect set an empty list of handlers()
res <- handlers("non-existing-handler", on_missing = "ignore")
res <- handlers()
stopifnot(is.list(res), length(res) == 0L)

res <- tryCatch({
  handlers("non-existing-handler", on_missing = "warning")
}, warning = identity)
stopifnot(inherits(res, "warning"))

res <- tryCatch({
  handlers("non-existing-handler", on_missing = "error")
}, error = identity)
stopifnot(inherits(res, "error"))


message("handlers() - exceptions ... DONE")

message("handlers() ... DONE")

source("incl/start.R")

if (require("foreach", quietly = TRUE)) {
  message("* with_progress()")
    
  with_progress({
    p <- progressor(4)
    y <- foreach(n = 3:6) %do% {
      p()
      slow_sum(1:n, stdout=TRUE, message=TRUE)
    }
  })

  message("* global progression handler")

  handlers(global = TRUE)
    
  local({
    p <- progressor(4)
    y <- foreach(n = 3:6) %do% {
      p()
      slow_sum(1:n, stdout=TRUE, message=TRUE)
    }
  })
    
  handlers(global = FALSE)
}

source("incl/end.R")

#' Progression Handler: Slow Down Progress Reporting
#'
#' @inheritParams make_progression_handler
#'
#' @return A function of class `progression_handler` that takes a
#' [progression] condition as its first and only argument.
#'
#' @param duration (numeric) Added slowdown (in seconds) across all progress
#' steps, e.g. if `duration = 10.0` and `steps = 100L`, then each step unit is
#' delayed 0.1 seconds, totaling a ten-second slowdown.
#'
#' @param \ldots Additional arguments passed to [make_progression_handler()].
#'
#' @examples
#' \donttest{\dontrun{
#' handlers(handler_slowdown(duration = 30.0)) ## 30-second total slowdown
#' with_progress(y <- slow_sum_p(1:100))
#' print(y)
#' }}
#'
#' @details
#' This progression handler delays each progress step a certain number of
#' seconds. It does not render any output.
#'
#' @export
handler_slowdown <- function(duration = 10.0, intrusiveness = 0, target = "void", enable = getOption("progressr.enable", TRUE), ...) {
  stopifnot(
    length(duration) == 1L, is.numeric(duration),
    is.finite(duration), duration >= 0
  )
  reporter <- local({
    list(
      initiate = function(config, state, progression, ...) NULL,
      
      update = function(config, state, progression, ...) {
        if (!state$enabled) return()
        delta <- duration / config$max_steps ## delay per step unit
        Sys.sleep(progression$amount * delta)
      },
      
      finish = function(config, state, progression, ...) NULL
    )
  })
  
  make_progression_handler("slowdown", reporter, intrusiveness = intrusiveness, target = target, enable = enable, ...)
}

if (exists(".knitr_asciicast_process", envir = .GlobalEnv)) {
  rm(list = ".knitr_asciicast_process", envir = .GlobalEnv)
}

asciicast::init_knitr_engine(
  echo = TRUE,
  echo_input = FALSE,
  interactive = TRUE,
  timeout = as.integer(Sys.getenv("ASCIICAST_TIMEOUT", 10)),
  startup = quote({
    library(progressr)
    options(width = 70)
    options(progressr.enable = TRUE)
    options(progressr.demo.delay = 0.15)
    options(progressr.show_after = 0.0)
    options(progressr.clear = FALSE)
    
    ## To simplify examples
    options(progressr.slow_sum.stdout = FALSE)
    options(progressr.slow_sum.message = FALSE)
    options(progressr.slow_sum.sticky = FALSE)
    
    handlers(global = TRUE)
  })
)

knitr::opts_chunk$set(
  asciicast_knitr_output = "html",
  asciicast_include_style = FALSE,
  asciicast_theme = "pkgdown"
)

list(
  markdown = TRUE,
  knitr_chunk_options = list(
    cache = TRUE,
    cache_lazy = FALSE,
    cache.path = file.path(getwd(), "man/_cache/"),
    fig.path = file.path(getwd(), "man/figures"),
    error = TRUE
  ),
  restrict_image_formats = TRUE
)

#' Add or Remove a Global 'progression' Handler
#'
#' @param action (character string)
#' If `"add"`, a global handler is added.
#' If `"remove"`, it is removed, if it exists.
#' If `"query"`, checks whether a handler is registered or not.
#'
#' @return Returns TRUE if a handler is registered, otherwise FALSE.
#' If `action = "query"`, the value is visible, otherwise invisible.
#'
#' @section Requirements:
#' This function requires R (>= 4.0.0) - the version in which global calling
#' handlers where introduces.
#'
#' @example incl/register_global_progression_handler.R
#'
#' @keywords internal
register_global_progression_handler <- function(action = c("add", "remove", "query")) {
  action <- match.arg(action[1], choices = c("add", "remove", "query", "status"))

  if (getRversion() < "4.0.0") {
    warning("register_global_progression_handler() requires R (>= 4.0.0)")
    return(invisible(FALSE))
  }

  ## All existing handlers
  handlers <- globalCallingHandlers()

  exists <- vapply(handlers, FUN = identical, global_progression_handler, FUN.VALUE = FALSE)
  if (sum(exists) > 1L) {
    warning("Detected more than one registered 'global_progression_handler'. Did you register it manually?")
  }
  
  if (action == "add") {
    if (!any(exists)) {
      globalCallingHandlers(condition = global_progression_handler)
    }
    invisible(TRUE)
  } else if (action == "remove") {
    global_progression_handler(control_progression("shutdown"))
    handlers <- handlers[!exists]
    ## Remove all
    globalCallingHandlers(NULL)
    ## Add back the ones we didn't drop
    globalCallingHandlers(handlers)
    invisible(FALSE)
  } else if (action == "query") {
    any(exists)
  } else if (action == "status") {
    global_progression_handler(control_progression("status"))
  }
}


#' A Global Calling Handler For 'progression':s
#'
#' @param condition A logical scalar or a condition object.
#'
#' @return Nothing.
#'
#' @section Requirements:
#' This function requires R (>= 4.0.0) - the version in which global calling
#' handlers where introduces.
#' 
#' @keywords internal
global_progression_handler <- local({
  active <- TRUE
  
  current_progressor_uuid <- NULL
  calling_handler <- NULL
  delays <- NULL
  stdout_file <- NULL
  capture_conditions <- NA
  conditions <- list()
  genv <- globalenv()

  update_calling_handler <- function() {
    handlers <- handlers()
    # Nothing to do?
    if (length(handlers) == 0L) return(NULL)

    handlers <- as_progression_handler(handlers)

    # Nothing to do?
    if (length(handlers) == 0L) return(NULL)

    ## Do we need to buffer?
    delays <<- use_delays(handlers)

    calling_handler <<- make_calling_handler(handlers)
  }

  interrupt_calling_handler <- function(progression, debug = FALSE) {
    if (is.null(calling_handler)) return()
    
    ## Don't capture conditions that are produced by progression handlers
    capture_conditions <<- FALSE
    on.exit(capture_conditions <<- TRUE)
  
    ## Any buffered output to flush?
    if (isTRUE(attr(delays$terminal, "flush"))) {
      if (length(conditions) > 0L || has_buffered_stdout(stdout_file)) {
        calling_handler(control_progression("hide"))
        stdout_file <<- flush_stdout(stdout_file, close = FALSE)
        conditions <<- flush_conditions(conditions)
      }
    }
  
    calling_handler(progression)
  }

  finish <- function(progression = control_progression("shutdown"), debug = FALSE) {
    finished <- FALSE

    ## Is progress handler active?
    if (!is.null(current_progressor_uuid)) {
      if (debug) message(" - shutdown progression handlers")
      if (!is.null(calling_handler)) {
        stdout_file <<- delay_stdout(delays, stdout_file = stdout_file)
        finished <- calling_handler(progression)
        ## Note that we might not be able to close 'stdout_file' due
        ## to blocking, non-balanced sinks
        stdout_file <<- flush_stdout(stdout_file, close = TRUE, must_work = FALSE)
        conditions <<- flush_conditions(conditions)
        delays <<- NULL
        if (debug) message(" - finished: ", finished)
      } else {
        finished <- TRUE
      }
    } else {
      if (debug) message(" - no active global progression handler")
    }

    ## Note that we might not have been able to close 'stdout_file' in previous
    ## calls to finish() due to blocking, non-balanced sinks. Try again here,
    ## just in case
    if (!is.null(stdout_file)) {
      stdout_file <<- flush_stdout(stdout_file, close = TRUE, must_work = FALSE)
    }

    current_progressor_uuid <<- NULL
    calling_handler <<- NULL
    capture_conditions <<- NA
    finished <- TRUE
    stop_if_not(length(conditions) == 0L, is.null(delays), isTRUE(finished), is.na(capture_conditions))
    
    finished
  }

  handle_progression <- function(progression, debug = getOption("progressr.global.debug", FALSE)) {
    ## To please R CMD check
    calling_handler <- NULL; rm(list = "calling_handler")

    ## Don't capture conditions that are produced by progression handlers
    last_capture_conditions <- capture_conditions
    capture_conditions <<- FALSE
    on.exit({
      if (is.null(current_progressor_uuid)) {
        capture_conditions <<- NA
      } else if (!is.na(capture_conditions)) {
        capture_conditions <<- TRUE
      }
    })

    stop_if_not(inherits(progression, "progression"))
    
    assign(".Last.progression", value = progression, envir = genv, inherits = FALSE)

    if (debug) message(sprintf("*** Caught a %s condition:", sQuote(class(progression)[1])))
    progressor_uuid <- progression[["progressor_uuid"]]
    if (debug) message(" - source: ", progressor_uuid)
    
    ## Listen to this progressor?
    if (!is.null(current_progressor_uuid) &&
        !identical(progressor_uuid, current_progressor_uuid)) {
      if (debug) message(" - action: ignoring, already listening to another")
      return()
    }


    if (!is.null(calling_handler) && !is.null(delays)) {
      ## Any buffered output to flush?
      if (isTRUE(attr(delays$terminal, "flush"))) {
        if (length(conditions) > 0L || has_buffered_stdout(stdout_file)) {
          calling_handler(control_progression("hide"))
          stdout_file <<- flush_stdout(stdout_file, close = FALSE)
          stop_if_not(inherits(stdout_file, "connection"))
          conditions <<- flush_conditions(conditions)
          calling_handler(control_progression("unhide"))
        }
      }
    }

    type <- progression[["type"]]
    if (debug) message(" - type: ", type)
  
    if (type == "initiate") {
      if (identical(progressor_uuid, current_progressor_uuid)) {
        stop(sprintf("INTERNAL ERROR: Already listening to this progressor which just sent another %s request", sQuote(type)))
      }
      if (debug) message(" - start listening")
#      finished <- finish(debug = debug)
#      stop_if_not(is.null(stdout_file), length(conditions) == 0L)
      current_progressor_uuid <<- progressor_uuid
      if (debug) message(" - reset progression handlers")
      update_calling_handler()
      if (!is.null(calling_handler)) {      
        stdout_file <<- delay_stdout(delays, stdout_file = stdout_file)
        calling_handler(control_progression("reset"))
        if (debug) message(" - initiate progression handlers")
        finished <- calling_handler(progression)
        if (debug) message(" - finished: ", finished)
        if (finished) {
          finished <- finish(debug = debug)
          stop_if_not(length(conditions) == 0L, is.na(capture_conditions), isTRUE(finished))
        }
      }
    } else if (type == "update") {
      if (is.null(current_progressor_uuid)) {
        warn_about_too_many_progressions(progression, global = TRUE)
        return()
      }

      if (debug) message(" - update progression handlers")
      if (!is.null(calling_handler)) {
        stdout_file <<- delay_stdout(delays, stdout_file = stdout_file)
        finished <- calling_handler(progression)
        if (debug) message(" - finished: ", finished)
        if (finished) {
          calling_handler(control_progression("shutdown"))
          finished <- finish(debug = debug)
          stop_if_not(length(conditions) == 0L, is.na(capture_conditions), isTRUE(finished))
        }
      }
    } else if (type == "finish") {
      finished <- finish(debug = debug)
      stop_if_not(length(conditions) == 0L, is.na(capture_conditions), isTRUE(finished))
    } else if (type == "status") {
      status <- list(
        current_progressor_uuid = current_progressor_uuid,
        calling_handler = calling_handler,
        delays = delays,
        stdout_file = stdout_file,
        capture_conditions = last_capture_conditions,
        conditions = conditions
      )
      if (debug) message(" - done")
      return(status)
    }

    if (debug) message(" - done")
  } ## handle_progression()


  function(condition) {
    if (is.logical(condition)) {
      stop_if_not(length(condition) == 1L, !is.na(condition))
      active <<- condition
      return()
    }
    
    if (!active) return()
    
    debug <- getOption("progressr.global.debug", FALSE)
    
    ## Shut down progression handling?
    if (inherits(condition, c("interrupt", "error"))) {
      if (isTRUE(getOption("progressr.interrupts", TRUE))) {
        ## Create progress message saying why the progress was interrupted
        msg <- sprintf("Progress interrupted by %s condition", class(condition)[1])
        msg <- paste(c(msg, conditionMessage(condition)), collapse = ": ")
        suspendInterrupts({
          interrupt_calling_handler(control_progression("interrupt", message = msg), debug = debug)
        })
      }

      suspendInterrupts({
        progression <- control_progression("shutdown")
        finished <- finish(debug = debug)
        stop_if_not(length(conditions) == 0L, is.na(capture_conditions), isTRUE(finished))
      })
      return()
    }

    ## A 'progression' update?
    if (inherits(condition, "progression")) {
      suspendInterrupts({
        res <- handle_progression(condition, debug = debug)
      })
      return(res)
    }

    ## Nothing do to?
    if (is.na(capture_conditions) || !isTRUE(capture_conditions)) return()

    ## Nothing do to?
    if (is.null(delays) || !inherits(condition, delays$conditions)) return()

    ## Record non-progression condition to be flushed later
    conditions[[length(conditions) + 1L]] <<- condition
    
    ## Muffle it for now
    if (inherits(condition, "message")) {
      tryInvokeRestart("muffleMessage")
    } else if (inherits(condition, "warning")) {
      tryInvokeRestart("muffleWarning")
    } else if (inherits(condition, "condition")) {
      ## If there is a "muffle" restart for this condition,
      ## then invoke that restart, i.e. "muffle" the condition
      restarts <- computeRestarts(condition)
      for (restart in restarts) {
        name <- restart$name
        if (is.null(name)) next
        if (!grepl("^muffle", name)) next
        tryInvokeRestart(restart)
        break
      }
    }
  }
}) ## global_progression_handler()



if (getRversion() < "4.0.0") {
  globalCallingHandlers <- function(...) {
    stop("register_global_progression_handler() requires R (>= 4.0.0)")
  }
}


as_progression_handler <- function(handlers, drop = TRUE, debug = FALSE) {
  ## FIXME(?)
  if (!is.list(handlers)) handlers <- list(handlers)

  if (debug) {
    message("as_progression_handler() ...")
    message(sprintf("- Number of progression handlers: %d", length(handlers)))
    on.exit({
      message(sprintf("- Number of progression handlers: %d", length(handlers)))
      message("as_progression_handler() ... done")
    })
  }

  for (kk in seq_along(handlers)) {
    handler <- handlers[[kk]]
    stop_if_not(is.function(handler))
    if (!inherits(handler, "progression_handler")) {
      if (debug) message(sprintf("- Handler %d is a function; calling it", kk))
      handler <- handler()
      stop_if_not(is.function(handler),
                  inherits(handler, "progression_handler"))
      handlers[[kk]] <- handler
    }
  }

  ## Keep only enabled handlers?
  if (drop) {
    if (debug) message(sprintf("- drop = TRUE => dropping non-enabled handlers"))
    enabled <- vapply(handlers, FUN = function(h) {
      env <- environment(h)
      value <- env$enable
      isTRUE(value) || is.null(value)
    }, FUN.VALUE = TRUE)
    
    if (debug && !all(enabled)) {
      message(sprintf("- Detected %d non-enabled handlers: %s", sum(!enabled), vapply(handlers[!enabled], FUN = function(h) class(h)[1], FUN.VALUE = NA_character_)))
    }
    
    handlers <- handlers[enabled]
  }

  handlers
}

#' Control How Progress is Reported
#'
#' @param \ldots One or more progression handlers.  Alternatively, this
#' functions accepts also a single vector of progression handlers as input.
#' If this vector is empty, then an empty set of progression handlers will
#' be set.
#'
#' @param append (logical) If FALSE, the specified progression handlers
#' replace the current ones, otherwise appended to them.
#'
#' @param on_missing (character) If `"error"`, an error is thrown if one of
#' the progression handlers does not exists.  If `"warning"`, a warning
#' is produces and the missing handlers is ignored.  If `"ignore"`, the
#' missing handlers is ignored.
#'
#' @param default The default progression calling handler to use if none
#' are set.
#'
#' @param global If TRUE, then the global progression handler is enabled.
#' If FALSE, it is disabled.  If NA, then TRUE is returned if it is enabled,
#' otherwise FALSE.  Argument `global` must not used with other arguments.
#'
#' @return (invisibly) the previous list of progression handlers set.
#' If no arguments are specified, then the current set of progression
#' handlers is returned.
#' If `global` is specified, then TRUE is returned if the global progression
#' handlers is enabled, otherwise false.
#'
#' @details
#' This function provides a convenient alternative for getting and setting
#' option \option{progressr.handlers}.
#'
#' @section For package developers:
#' **IMPORTANT: Setting progression handlers is a privilege that should be
#' left to the end user. It should not be used by R packages, which only task
#' is to _signal_ progress updates, not to decide if, when, and how progress
#' should be reported.**
#'
#' If you have to set or modify the progression handlers inside a function,
#' please make sure to undo the settings afterward.  If not, you will break
#' whatever progression settings the user already has for other purposes
#' used elsewhere.  To undo you settings, you can do:
#'
#' ```r
#' old_handlers <- handlers(c("beepr", "progress"))
#' on.exit(handlers(old_handlers), add = TRUE)
#' ```
#'
#' @section Configuring progression handling during R startup:
#' A convenient place to configure the default progression handler and to
#' enable global progression reporting by default is in the \file{~/.Rprofile}
#' startup file.  For example, the following will (i) cause your interactive
#' R session to use global progression handler by default, and (ii) report
#' progress via the \pkg{progress} package when in the terminal and via the
#' RStudio Jobs progress bar when in the RStudio Console.
#' [handler_txtprogressbar], 
#' other whenever using the RStudio Console, add
#' the following to your \file{~/.Rprofile} startup file:
#'
#' ```r
#' if (interactive() && requireNamespace("progressr", quietly = TRUE)) {
#'   ## Enable global progression updates
#'   if (getRversion() >= 4) progressr::handlers(global = TRUE)
#'
#'   ## In RStudio Console, or not?
#'   if (Sys.getenv("RSTUDIO") == "1" && !nzchar(Sys.getenv("RSTUDIO_TERM"))) {
#'     options(progressr.handlers = progressr::handler_rstudio)
#'   } else {
#'     options(progressr.handlers = progressr::handler_progress)
#'   }
#' }
#' ```
#'
#' @example incl/handlers.R
#'
#' @export
handlers <- function(..., append = FALSE, on_missing = c("error", "warning", "ignore"), default = handler_txtprogressbar, global = NULL) {
  stop_if_not(
    is.null(global) ||
    ( is.logical(global) && length(global) == 1L )
  )
  args <- list(...)
  nargs <- length(args)

  if (nargs == 0L) {
    ## Get the current set of progression handlers?
    if (is.null(global)) {
      if (!is.list(default) && !is.null(default)) default <- list(default)
      return(getOption("progressr.handlers", default))
    }

    ## Check, register, or reset global calling handlers?
    if (is.na(global)) {
      return(register_global_progression_handler(action = "query"))
    }
    action <- if (isTRUE(global)) "add" else "remove"
    return(invisible(register_global_progression_handler(action = action)))
  }

  if (!is.null(global)) {
    stop("Argument 'global' must not be specified when also registering progress handlers")
  }

  on_missing <- match.arg(on_missing)
  
  ## Was a list specified?
  if (nargs == 1L && is.vector(args[[1]])) {
    args <- args[[1]]
  }

  handlers <- list()
  
  names <- names(args)
  for (kk in seq_along(args)) {
    handler <- args[[kk]]
    stop_if_not(length(handler) == 1L)
    
    if (is.character(handler)) {
      name <- handler
      name2 <- sprintf("handler_%s", name)
      
      handler <- NULL
      if (exists(name2, mode = "function")) {
        handler <- get(name2, mode = "function")
      }
      
      if (is.null(handler)) {
        if (exists(name, mode = "function")) {
          handler <- get(name, mode = "function")
        }
      }
      
      if (is.null(handler)) {
        if (on_missing == "error") {
          stop("No such progression handler found: ", sQuote(name))
	} else if (on_missing == "warning") {
          warning("Ignoring non-existing progression handler: ", sQuote(name))
	}
        next
      }
    } else {
      name <- NULL
    }
    stop_if_not(is.function(handler), length(formals(handler)) >= 1L)

    ## Validate?
    validator <- attr(handler, "validator")
    if (is.function(validator)) {
      is_valid <- validator()
      if (!is_valid) next
    }

    if (!is.null(name)) names[kk] <- name
    handlers[[kk]] <- handler
  }
  stop_if_not(is.list(handlers))
  names(handlers) <- names

  ## Drop non-existing handlers
  keep <- vapply(handlers, FUN.VALUE = FALSE, FUN = is.function)
  handlers <- handlers[keep]

  if (append) {
    current <- getOption("progressr.handlers", list())
    if (length(current) > 0L) handlers <- c(current, handlers)
  }

  old_handlers <- options(progressr.handlers = handlers)[[1]]
  if (is.null(old_handlers)) old_handlers <- list()
  
  invisible(old_handlers)
}

library(progressr)

options(progressr.clear = FALSE)

message("handler_void() ...")

# 1. Test creation and properties
h_disabled <- handler_void(enable = FALSE)
stopifnot(inherits(h_disabled, "void_progression_handler"))
print(h_disabled)

h_enabled <- handler_void(enable = TRUE)
stopifnot(inherits(h_enabled, "void_progression_handler"))
print(h_enabled)

# 2. Test manual progression calls when enable = FALSE
h_disabled(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-1", progression_index = 1L))
h_disabled(progression(type = "update", amount = 1, progressor_uuid = "uuid-1", progression_index = 2L))
h_disabled(progression(type = "finish", progressor_uuid = "uuid-1", progression_index = 3L))

# 3. Test manual progression calls when enable = TRUE
h_enabled(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-2", progression_index = 1L))
h_enabled(progression(type = "update", amount = 1, progressor_uuid = "uuid-2", progression_index = 2L))
h_enabled(progression(type = "finish", progressor_uuid = "uuid-2", progression_index = 3L))

# 4. Run inside with_progress
for (enable in c(FALSE, TRUE)) {
  message(sprintf("with_progress(..., handlers = handler_void(enable = %s))", enable))
  with_progress({
    p <- progressor(along = 1:5)
    for (i in 1:5) {
      Sys.sleep(getOption("progressr.demo.delay", 0.1))
      p()
    }
  }, handlers = handler_void(enable = enable))
}

message("handler_void() ... done")

#' A Progression Condition
#'
#' A progression condition represents a progress in an \R program.
#'
#' @param message (character vector or a function) If a character vector, then
#' it is pasted together into a single string using an empty separator.
#' If a function, then the message is constructed by `conditionMessage(p)`
#' calling this function with the progression condition `p` itself as the
#' first argument.
#'
#' @param amount (numeric) The total amount of progress made.
#'
#' @param step (numeric) The step completed. If specified, `amount` is ignored.
#' _WARNING: Argument `step` should only be used when in full control of the
#' order when this progression condition is signaled._ For example, it must not
#' be signaled as one of many parallel progress updates signaled concurrently,
#' because then we cannot guarantee the order these progressions arrive.
#'
#' @param time (POSIXct) A timestamp.
#'
#' @param \ldots Additional named elements.
#'
#' @param type Type of progression made.
#'
#' @param class (character) Zero or more class names to prepend.
#'
#' @param progressor_uuid (character string) A character string that is unique
#' for the current progressor and the current \R session.
#'
#' @param progression_index (integer) A non-negative integer that is
#' incremented by one for each progression condition created.
#'
#' @param progression_time (POSIXct or character string) A timestamp specifying
#' when the progression condition was created.
#'
#' @param owner_session_uuid (character string) A character string that is
#' unique for the \R session where the progressor was created.
#'
#' @param call (expression) A call expression.
#'
#' @param calls (pairlist) The calls that lead up to this progression update.
#'
#' @return A [base::condition] of class `progression`.
#'
#' @seealso
#' To signal a progression condition, use [base::signalCondition()].
#'
#' @keywords internal
#' @export
progression <- function(message = character(0L), amount = 1.0, step = NULL, time = progression_time, ..., type = "update", class = NULL, progressor_uuid = NULL, progression_index = NULL, progression_time = Sys.time(), call = NULL, calls = sys.calls(), owner_session_uuid = NULL) {
  amount <- as.numeric(amount)
  time <- as.POSIXct(time)
  stop_if_not(is.character(type), length(type) == 1L, !is.na(type))
  class <- as.character(class)
  if (inherits(progression_time, "POSIXct")) {
    progression_time <- format(progression_time, format = "%F %H:%M:%OS3 %z")
  }
  stop_if_not(length(progression_time) == 1L, is.character(progression_time))
  args <- list(...)
  nargs <- length(args)
  if (nargs > 0L) {
    names <- names(args)
    stop_if_not(!is.null(names), all(nzchar(names)),
                length(unique(names)) == nargs)
  }
  
  structure(
    list(
      owner_session_uuid = owner_session_uuid,
      progressor_uuid = progressor_uuid,
      session_uuid = session_uuid(),
      progression_index = progression_index,
      progression_time = progression_time,
      type = type,
      message = message,
      amount = amount,
      step = step,
      time = time,
      ...,
      call = call,
      calls = calls
    ),
    class = c(class, "progression", "immediateCondition", "condition")
  )
}


#' @export
conditionMessage.progression <- function(c) {
  message <- NextMethod("conditionMessage")  ## == c$message

  ## Dynamically generate message from the 'progression' condition?
  if (is.function(message)) {
    message_fcn <- message
    message <- message_fcn(c)
  }

  message <- as.character(message)
  if (length(message) > 0L) message <- paste(message, collapse = "")
  
  message
}


#' @importFrom utils object.size
#' @export
print.progression <- function(x, ...) {
  s <- sprintf("%s:", class(x)[1])
  s <- c(s, paste("- call:", deparse(conditionCall(x))))
  s <- c(s, paste("- type:", x$type))
  s <- c(s, sprintf("- message: [%s] %s", class(x$message)[1], sQuote(conditionMessage(x))))
  s <- c(s, paste("- amount:", x$amount))
  s <- c(s, paste("- step:", x$step))
  s <- c(s, paste("- time:", x$time))
  s <- c(s, paste("- progressor_uuid:", x$progressor_uuid))
  s <- c(s, paste("- progression_index:", x$progression_index))
  s <- c(s, paste("- progression_time:", x$progression_time))
  s <- c(s, paste("- session_uuid:", x$session_uuid))
  s <- c(s, paste("- owner_session_uuid:", x$owner_session_uuid))
  s <- c(s, paste("- classes:", paste(sQuote(class(x)), collapse = ", ")))
  size <- object.size(x)
  size2 <- serialization_size(x)
  s <- c(s, sprintf("- size: %s [%s serialized]", format(size, units = "auto", standard = "SI"), format(size2, units = "auto", standard = "SI")))
  s <- paste(s, collapse = "\n")
  cat(s, "\n", sep = "")
  
  invisible(x)
}

## Settings
options(progressr.clear = TRUE)
options(progressr.debug = FALSE)
options(progressr.demo.delay = 0.0)
options(progressr.enable = TRUE)
options(progressr.enable_after = 0.0)
options(progressr.interval = 0.1)
options(progressr.times = +Inf)


options(progressr.tests.fake_handlers = c(non_supported_progression_handlers(), "handler_beepr", "handler_notifier", "handler_progress"))

library(progressr)

if (require("foreach", quietly = TRUE)) {
  message("* with_progress()")
    
  with_progress({
    p <- progressor(4)
    y <- foreach(n = 3:6) %do% {
      p()
      slow_sum_p(1:n, stdout=TRUE, message=TRUE)
    }
  })

  message("* global progression handler")

  handlers(global = TRUE)
    
  local({
    p <- progressor(4)
    y <- foreach(n = 3:6) %do% {
      p()
      slow_sum_p(1:n, stdout=TRUE, message=TRUE)
    }
  })
    
  handlers(global = FALSE)
}

#' A progressr handler useful for Futureverse tutorials
#'
#' @inheritParams make_progression_handler
#' 
#' @param \ldots Arguments passed to the selected progress handler.
#'
#' @param spec The progress handler to use.
#'
#' @param type Controls which predefined handler format string to use.
#'
#' @return A function of class `progression_handler` that takes a
#' [progression] condition as its first and only argument.
#'
#' @details
#' The default is to use the \pkg{progress} package. Contrary to the
#' \pkg{cli} package, the \pkg{progress} package has the capacity to
#' display _all_ progress updates, even when receiving progress updates
#' at a high rate from, for example, parallel processing.
#' 
#' @examples
#' # Display progress and keep the progress output when done
#' handlers("tutorial")
#' 
#' # Display progress and keep every progress reported
#' handlers("tutorial", "newline")
#' 
#' @export
handler_tutorial <- function(..., spec = c("progress", "cli"), type = "time", clear = FALSE) {
  spec <- match.arg(spec)
  if (!requireNamespace(spec, quietly = TRUE)) {
    stop(sprintf("Please install package: %s", sQuote(spec)))
  }
  if (spec == "cli") {
    handler_cli(..., type = type, clear = clear)
  } else if (spec == "progress") {
    handler_progress(..., type = type, clear = clear)
  }
}

use_delays <- function(handlers, terminal = NULL, stdout = NULL, conditions = NULL) {
  ## Do we need to buffer terminal output?
  if (is.null(terminal)) {
    delay <- vapply(handlers, FUN = function(h) {
      env <- environment(h)
      any(env$target == "terminal")
    }, FUN.VALUE = NA)
    terminal <- any(delay, na.rm = TRUE)
    
    ## If buffering output, does all handlers support intermediate flushing?
    if (terminal) {
      flush <- vapply(handlers, FUN = function(h) {
        env <- environment(h)
        if (!any(env$target == "terminal")) return(TRUE)
        !inherits(env$reporter$hide, "null_function")
      }, FUN.VALUE = NA)
      attr(terminal, "flush") <- all(flush, na.rm = TRUE)
    }
  }

  if (is.null(stdout)) {
    stdout <- getOption("progressr.delay_stdout", terminal)
  }

  if (is.null(conditions)) {
    conditions <- getOption("progressr.delay_conditions", {
      if (terminal) c("condition") else character(0L)
    })
  }

  list(terminal = terminal, stdout = stdout, conditions = conditions)
}


delay_stdout <- function(delays, stdout_file) {
  ## Delay standard output?
  if (is.null(stdout_file) && delays$stdout) {
    stdout_file <- buffer_stdout()
  }
  stdout_file
}


buffer_stdout <- function() {
  stdout_file <- rawConnection(raw(0L), open = "w")
  sink(stdout_file, type = "output", split = FALSE)
  attr(stdout_file, "sink_index") <- sink.number(type = "output")
  stdout_file
} ## buffer_stdout()


flush_stdout <- function(stdout_file, close = TRUE, must_work = FALSE) {
  if (is.null(stdout_file)) return(NULL)

  ## Can we close the sink we opened?
  ## It could be that a progressor completes while there is a surrounding
  ## sink active, e.g. an active capture.output(), or when signalled within
  ## a sequential future.  Because of this, we might not be able to flush
  ## close the sink here.
  sink_index <- attr(stdout_file, "sink_index")
  if (sink_index != sink.number("output")) {
    if (must_work) {
      stop(sprintf("[progressr] Cannot flush stdout because the current sink index (%d) is out of sync with the sink we want to close (%d)", sink.number("output"), sink_index))
    }
    return(stdout_file)
  }
  
  sink(split = FALSE, type = "output")
  stdout <- rawToChar(rawConnectionValue(stdout_file))
  ## WORKAROUND: Positron 2025.09.0 appends a newline for cat("") [1], while
  ## it should output nothing. We use nzchar(stdout) here to avoid this
  ## from happening.
  ## [1] https://github.com/posit-dev/positron/issues/9486
  ## cat("") nzchar(stdout) is used to handle the cat("") case
  if (length(stdout) > 0 && nzchar(stdout)) cat(stdout, file = stdout())
  close(stdout_file)
  stdout_file <- NULL
  if (!close) stdout_file <- buffer_stdout()
  stdout_file
} ## flush_stdout()

has_buffered_stdout <- function(stdout_file) {
  !is.null(stdout_file) && (length(rawConnectionValue(stdout_file)) > 0L)
}


flush_conditions <- function(conditions, patch = FALSE) {
  ## Patch https://github.com/futureverse/progressr/issues/179
  patch <- patch_rstudio_console()
  if (patch) {
    message <- function(m) {
      cat(conditionMessage(m), file = stderr())
    }
    warning <- function(w) {
      call <- w[["call"]]
      if (is.null(call)) {
        cat(gettext("Warning:"), file = stderr())
      } else {
        dcall <- deparse(call)
        cat(gettextf("Warning in %s :", dcall), file = stderr())
      }
      cat(conditionMessage(w), "\n", sep = "", file = stderr())
    }
  }
  
  for (c in conditions) {
    if (inherits(c, "message")) {
      message(c)
    } else if (inherits(c, "warning")) {
      warning(c)
    } else if (inherits(c, "condition")) {
      signalCondition(c)
    }
  }
  
  list()
} ## flush_conditions()
 

#' Progression Handler: Progress Reported via 'shiny' Widgets (GUI) in the HTML Browser
#'
#' A progression handler for \pkg{shiny} and [shiny::withProgress()].
#'
#' @inheritParams make_progression_handler
#'
#' @param inputs (named list) Specifies from what sources the Shiny progress
#' elements 'message' and 'detail' should be updated.  Valid sources are
#' `"message"`, `"sticky_message"` and `"non_sticky_message"`, where
#' `"message"` is short for `c("non_sticky_message", "sticky_message")`. For
#' example, `inputs = list(message = "sticky_message", detail = "message")`
#' will update the Shiny 'message' component from sticky messages only,
#' whereas the 'detail' component is updated using any message.
#'
#' @param \ldots Additional arguments passed to [make_progression_handler()].
#'
#' @examples
#' \donttest{\dontrun{
#' handlers(handler_shiny())
#' with_progress(y <- slow_sum_p(1:100))
#' }}
#'
#' @section Requirements:
#' This progression handler requires the \pkg{shiny} package.
#'
#' @details
#' For most Shiny application there is little need to use this Shiny handler
#' directly.  Instead, it is sufficient to use [withProgressShiny()].
#'
#' @keywords internal
#' @export
handler_shiny <- function(intrusiveness = getOption("progressr.intrusiveness.gui", 1), target = "gui", inputs = list(message = NULL, detail = "message"), enable = getOption("progressr.enable", TRUE), ...) {
  stop_if_not(
    is.list(inputs),
    !is.null(names(inputs)),
    all(names(inputs) %in% c("message", "detail")),
    all(vapply(inputs, FUN = function(x) {
      if (is.null(x)) return(TRUE)
      if (!is.character(x)) return(FALSE)
      x %in% c("message", "non_sticky_message", "sticky_message")
    }, FUN.VALUE = FALSE))
  )

  ## Expand 'message' => c("non_sticky_message", "sticky_message")
  for (name in names(inputs)) {
    input <- inputs[[name]]
    if ("message" %in% input) {
      input <- setdiff(input, "message")
      input <- c(input, "non_sticky_message", "sticky_message")
    }
    inputs[[name]] <- unique(input)
  }
  
  reporter <- local({
    list(
      interrupt = function(config, state, progression, ...) {
        if (!state$enabled) return()
        stop_if_not(
          length(config$max_steps) == 1L, is.numeric(config$max_steps),
          !is.na(config$max_steps), is.finite(config$max_steps),
          config$max_steps >= 0
        )
        msg <- conditionMessage(progression)
        amount <- if (config$max_steps == 0) 1 else progression$amount / config$max_steps
        args <- c(
          list(amount = amount),
          message_to_backend_targets(progression, inputs = inputs, message = msg)
        )
        do.call(shiny::incProgress, args = args)
      },
      
      update = function(config, state, progression, ...) {
        if (!state$enabled) return()
        stop_if_not(
          length(config$max_steps) == 1L, is.numeric(config$max_steps),
          !is.na(config$max_steps), is.finite(config$max_steps),
          config$max_steps >= 0
        )
        amount <- if (config$max_steps == 0) 1 else progression$amount / config$max_steps
        args <- c(
          list(amount = amount),
          message_to_backend_targets(progression, inputs = inputs)
        )
        do.call(shiny::incProgress, args = args)
      }
    )
  })
  
  make_progression_handler("shiny", reporter, intrusiveness = intrusiveness, target = target, enable = enable, ...)
}

#' progressr: A Unifying API for Progress Updates
#'
#' The \pkg{progressr} package provides a minimal, unifying API for scripts
#' and packages to report progress updates from anywhere including when
#' using parallel processing.
#'
#' The package is designed such that _the developer_ can to focus on _what_
#' progress should be reported on without having to worry about _how_ to
#' present it.
#'
#' The _end user_ has full control of _how_, _where_, and _when_ to render
#' these progress updates.  For instance, they can chose to report progress
#' in the terminal using [utils::txtProgressBar()] (default) or
#' [cli::cli_progress_bar()], via the R graphical user interface (GUI)
#' using `utils::winProgressBar()` or [tcltk::tkProgressBar()], or
#' via the RStudio GUI using `rstudioapi::jobSetProgress()`.
#' An alternative to above visual rendering, is to report progress as audio
#' using [beepr::beep()].
#' It is also possible to use a combination of above progression handlers,
#' e.g. a progress bar in the terminal together with audio updates.
#' Besides the existing handlers, it is possible to develop custom
#' progression handlers.
#'
#' The \pkg{progressr} package is compatible with \pkg{shiny} applications
#' and \pkg{knitr} rendering.
#'
#' The \pkg{progressr} package uses R's condition framework for signaling
#' progress updated. Because of this, progress can be reported from almost
#' anywhere in R, e.g. from classical for and while loops, from map-reduce
#' APIs like the [lapply()] family of functions, \pkg{purrr}, \pkg{plyr}, and
#' \pkg{foreach}.
#' The \pkg{progressr} package will also work with parallel processing via
#' the \pkg{future} framework, e.g. [future.apply::future_lapply()],
#' [furrr::future_map()], and [foreach::foreach()] with \pkg{doFuture}.
#'
#' @section Progression Handlers:
#'
#' In the terminal:
#' * [handler_txtprogressbar] (built-in; default)
#' * [handler_pbcol] (built-in)
#' * [handler_ascii_alert] (built-in)
#' * [handler_pbmcapply] (requires the \pkg{pbmcapply} package)
#' * [handler_progress] (requires the \pkg{progress} package)
#'
#' In a graphical user interface (GUI):
#' * [handler_rstudio] (only in the RStudio Console)
#' * [handler_tkprogressbar] (built-in)
#' * [handler_winprogressbar] (built-in)
#'
#' As sound:
#' * [handler_ascii_alert] (built-in)
#' * [handler_beepr] (requires the \pkg{beepr} package)
#'
#' Via the file system:
#' * [handler_filesize] (built-in)
#'
#' In Shiny:
#' * [withProgressShiny] (requires the \pkg{shiny} package)
#'
#' Via notification systems:
#' * [handler_ntfy] (requires the \pkg{ntfy} package)
#' * [handler_notifier] (requires the non-CRAN \pkg{notifier} package)
#' * [handler_rpushbullet] (requires the \pkg{RPushBullet} package)
#'
#' Miscellaneous:
#' * [handler_debug] (built-in)
#' * [handler_newline] (built-in)
#' * [handler_slowdown] (built-in)
#' * [handler_void] (built-in)
#'
#' @example incl/progressr-package.R
#'
#' @keywords programming iteration
#'
#' @aliases progressr-package
#' @name progressr
"_PACKAGE"

progress_notify <- function(title = "R 'progressr' notification", message, step, max_steps, notifier) {
  stop_if_not(is.function(notifier))
  
  ratio <- if (max_steps == 0) 1 else step / max_steps
  ratio <- sprintf("%.0f%%", 100*ratio)
  msg <- paste(c("", message), collapse = "")
  body <- sprintf("[%s] %s", ratio, msg)
  
  notifier(title = title, message = body)
} ## progress_notify()

#' Progression Handler: Progress Reported via the Operating-System Notification Framework (GUI, Text)
#'
#' A progression handler for `notify()` of the \pkg{notifier} package.
#'
#' @inheritParams make_progression_handler
#'
#' @return A function of class `progression_handler` that takes a
#' [progression] condition as its first and only argument.
#'
#' @param \ldots Additional arguments passed to [make_progression_handler()].
#'
#' @example incl/handler_notifier.R
#'
#' @section Requirements:
#' This progression handler requires the \pkg{notifier} package, which is only
#' available from <https://github.com/gaborcsardi/notifier>.  This can be
#' installed as `remotes::install_github("gaborcsardi/notifier@62d484")`.
#'
#' @export
handler_notifier <- function(intrusiveness = getOption("progressr.intrusiveness.notifier", 10), target = "gui", ...) {
  ## Used for package testing purposes only when we want to perform
  ## everything except the last part where the backend is called
  if (!is_fake("handler_notifier")) {
    pkg <- "notifier"
    if (!requireNamespace(pkg, quietly = TRUE)) {
      stop("Package 'notifier' is not available. See ?progressr::handler_notifier() for installation instructions")
    }
    notifier_ <- get("notify", mode = "function", envir = getNamespace(pkg))
    notifier <- function(title = "R 'progressr' notification", message, ...) {
      notifier_(msg = message, title = title)
    }
  } else {
    notifier <- function(title = "R 'progressr' notification", message, ...) NULL
  }

  reporter <- local({
    finished <- FALSE
    
    list(
      reset = function(...) {
        finished <<- FALSE
      },
      
      interrupt = function(config, state, progression, ...) {
        msg <- conditionMessage(progression)
        progress_notify(step = state$step, max_steps = config$max_steps, message = msg, notifier = notifier)
      },

      initiate = function(config, state, progression, ...) {
        if (!state$enabled || config$times == 1L) return()
        progress_notify(step = state$step, max_steps = config$max_steps, message = state$message, notifier = notifier)
      },
        
      update = function(config, state, progression, ...) {
        if (!state$enabled || progression$amount == 0 || config$times <= 2L) return()
        progress_notify(step = state$step, max_steps = config$max_steps, message = state$message, notifier = notifier)
      },
        
      finish = function(config, state, progression, ...) {
        if (finished) return()
        if (!state$enabled) return()
        if (state$delta > 0) {
          progress_notify(step = state$step, max_steps = config$max_steps, message = state$message, notifier = notifier)
        }
	finished <<- TRUE
      }
    )
  })
  
  make_progression_handler("notifier", reporter, intrusiveness = intrusiveness, target = target, ...)
}

#' Progression Handler: Progress Reported as a New Line (Text) in the Terminal
#'
#' @inheritParams make_progression_handler
#'
#' @param symbol (character string) The character symbol to be outputted,
#' which by default is the ASCII NL character (`'\n'` = `'\013'`) character.
#'
#' @param file (connection) A [base::connection] to where output should be sent.
#'
#' @param \ldots Additional arguments passed to [make_progression_handler()].
#'
#' @keywords internal
#' @export
handler_newline <- function(symbol = "\n", file = stderr(), intrusiveness = getOption("progressr.intrusiveness.debug", 0), target = "terminal", ...) {
  reporter <- local({
    list(
      hide     = function(...) NULL,
      unhide   = function(...) NULL,
      interrupt = function(config, state, progression, ...) {
        msg <- conditionMessage(progression)
        msg <- paste(c("", msg, ""), collapse = "\n")
        cat(msg, file = file)
      },
      initiate = function(...) cat(file = file, symbol),
      update   = function(...) cat(file = file, symbol),
      finish   = function(...) cat(file = file, symbol)
    )
  })
  
  make_progression_handler("newline", reporter, intrusiveness = intrusiveness, target = target, ...)
}

#' Progression Handler: Progress Reported as a Tcl/Tk Progress Bars in the GUI
#'
#' A progression handler for [tcltk::tkProgressBar()].
#'
#' @inheritParams make_progression_handler
#'
#' @return A function of class `progression_handler` that takes a
#' [progression] condition as its first and only argument.
#' @inheritParams handler_winprogressbar
#'
#' @param \ldots Additional arguments passed to [tcltk::tkProgressBar()]
#' and [make_progression_handler()].
#'
#' @example incl/handler_tkprogressbar.R
#'
#' @section Requirements:
#' This progression handler requires the \pkg{tcltk} package and that the
#' current R session supports Tcl/Tk (`capabilities("tcltk")`).
#'
#' @export
handler_tkprogressbar <- function(intrusiveness = getOption("progressr.intrusiveness.gui", 1), target = "terminal", inputs = list(title = NULL, label = "message"), ...) {
  ## Additional arguments passed to the progress-handler backend
  backend_args <- handler_backend_args(...)
  
  ## Used for package testing purposes only when we want to perform
  ## everything except the last part where the backend is called
  if (!is_fake("handler_tkprogressbar")) {
    if (!capabilities("tcltk")) {
      stop("handler_tkprogressbar requires TclTk support")
    }
    ## Import functions
    tkProgressBar <- tcltk::tkProgressBar
    setTkProgressBar <- tcltk::setTkProgressBar
  } else {
    tkProgressBar <- function(title = "R progress bar", label = "", min = 0, max = 1, initial = 0, width = 300) rawConnection(raw(0L))
    setTkProgressBar <- function(pb, value, title = NULL, label = NULL) NULL
  }

  stop_if_not(
    is.list(inputs),
    !is.null(names(inputs)),
    all(names(inputs) %in% c("title", "label")),
    all(vapply(inputs, FUN = function(x) {
      if (is.null(x)) return(TRUE)
      if (!is.character(x)) return(FALSE)
      x %in% c("message", "non_sticky_message", "sticky_message")
    }, FUN.VALUE = FALSE))
  )
  
  ## Expand 'message' => c("non_sticky_message", "sticky_message")
  for (name in names(inputs)) {
    input <- inputs[[name]]
    if ("message" %in% input) {
      input <- setdiff(input, "message")
      input <- c(input, "non_sticky_message", "sticky_message")
    }
    inputs[[name]] <- unique(input)
  }

  reporter <- local({
    pb_config <- NULL

    ## Update tkProgressBar
    update_pb <- function(state, progression, ...) {
      ## Update 'title' and 'label' (optional)
      args <- message_to_backend_targets(progression, inputs = inputs, ...)
      for (name in names(args)) pb_config[[name]] <<- args[[name]]

      ## Update progress bar
      args <- pb_config
      args$value <- state$step
      do.call(what = setTkProgressBar, args = args)
    }

    list(
      reset = function(...) {
        pb_config <<- NULL
      },
      
      initiate = function(config, state, progression, ...) {
        if (!state$enabled || config$times == 1L) return()
        ## NOTE: 'pb_config' may be re-used for tkProgressBar:s
        if (config$clear) stop_if_not(is.null(pb_config))
        args <- c(
          backend_args,
          list(max = config$max_steps, initial = state$step),
          list(...)
        )

        ## tkProgressBar() arguments 'title' and 'label' must not be NULL;
        ## use the defaults
        for (name in c("title", "label")) {
          if (is.null(args[[name]])) {
            args[[name]] <- formals(tkProgressBar)[[name]]
          }
        }

        ## Create progress bar
        args <- args[names(args) %in% names(formals(tkProgressBar))]
        pb <- do.call(tkProgressBar, args = args)

        ## Record arguments used by setTkProgressBar() later on
        args$pb <- pb
        args <- args[names(args) %in% names(formals(setTkProgressBar))]
        pb_config <<- args
      },
        
      interrupt = function(config, state, progression, ...) {
        if (!state$enabled) return()
        msg <- conditionMessage(progression)
        update_pb(state, progression, message = msg)
      },

      update = function(config, state, progression, ...) {
        if (!state$enabled || config$times <= 2L) return()
        update_pb(state, progression)
      },
        
      finish = function(config, state, progression, ...) {
        ## Already finished?
        if (is.null(pb_config)) return()
        if (!state$enabled) return()
        if (config$clear) {
          close(pb_config$pb)
          pb_config <<- NULL
        } else {
          update_pb(state, progression)
        }
      }
    )
  })
  
  make_progression_handler("tkprogressbar", reporter, intrusiveness = intrusiveness, target = target, ...)
}

get_rstudio_version <- local({
  .version <- NULL
  function() {
    if (is.null(.version)) {
      .version <<- package_version("0.0")
      if ("tools:rstudio" %in% search()) {
        envir <- as.environment("tools:rstudio")
        ## There are cases where 'tools:rstudio' exist, but there is no
        ## RStudio.Version() function. See
        ## https://github.com/futureverse/progressr/issues/183 for
        ## an example. Not sure how that happens. /HB 2025-09-23
        if (exists("RStudio.Version", mode = "function", envir = envir, inherits = FALSE)) {
          RStudio.Version <- get("RStudio.Version", mode = "function", envir = envir, inherits = FALSE)
          .version <<- RStudio.Version()[["version"]]
        }
      }
    }
    .version
  }
})


patch_rstudio_console_evidence <- function() {
  res <- c(
    console = FALSE, ## In RStudio Console?
    version = FALSE, ## Affected RStudio version?
    message = FALSE, ## RStudio use custom message handler?
    warning = FALSE  ## RStudio use custom warning handler?
  )

  ## RStudio?
  if (Sys.getenv("RSTUDIO") == "1") {
    ## Not RStudio Terminal, i.e. RStudio Console?
    if (!nzchar(Sys.getenv("RSTUDIO_TERM"))) {
      res["console"] <- TRUE
    }

    ## RStudio v2025.5.0?
    if (get_rstudio_version() >= "2025.5.0" &&
        get_rstudio_version() <  "2025.5.1") {
      res["version"] <- TRUE
    }
  }

  ## No global message and warning handlers?
  handlers <- globalCallingHandlers()
  
  handler <- handlers[["message"]]
  if (is.null(handler)) {
    res["message"] <- TRUE
  } else {
    code <- deparse(body(handler))
    search <- ".rs.globalCallingHandlers.onMessageImpl"
    if (any(grepl(search, code, fixed = TRUE))) {
      res["message"] <- TRUE
    }
  }

  handler <- handlers[["warning"]]
  if (is.null(handler)) {
    res["warning"] <- TRUE
  } else {
    code <- deparse(body(handler))
    search <- ".rs.globalCallingHandlers.onWarningImpl"
    if (any(grepl(search, code, fixed = TRUE))) {
      res["warning"] <- TRUE
    }
  }

  res
}


patch_rstudio_console <- local({
  patch <- NA
  
  function(reset = FALSE) {
    if (isTRUE(reset)) patch <<- NA
    if (!is.na(patch)) return(patch)

    ## Always respect option
    opt <- getOption("progressr.rstudio.patch")
    if (is.logical(opt) && length(opt) == 1L && !is.na(opt)) {
      return(opt)
    }

    affected <- patch_rstudio_console_evidence()
    if (!all(affected)) {
      patch <<- FALSE
      return(FALSE)
    }

    TRUE
  }
})


#' @importFrom utils packageVersion
warn_about_rstudio_condition_handlers <- function() {
  affected <- patch_rstudio_console_evidence()
  if (!all(affected)) return()
  msg <- sprintf("You are running RStudio %s, which has a custom message and warning handlers that are incompatible with some of the progress handlers in progressr %s, e.g. 'cli'. Until fixed, one workaround is to disable RStudio's custom handlers, by going to 'Tools' -> 'Global Options ...' -> 'Console' and switch 'Use Extended Display' to 'Errors only'. See <https://github.com/futureverse/progressr/issues/179> for further details", get_rstudio_version(), packageVersion("progressr"))
  warning(msg, call. = FALSE, immediate. = FALSE)
}

library(progressr)

message("with_progress() - progressr.debug = TRUE ...")

options(progressr.debug = TRUE)

with_progress({
  y <- slow_sum(1:10)
})


with_progress({
  progress <- progressor(steps = 1 + 2 + 1)
  relay_progress <- progress_aggregator(progress)
  p <- progress()
  progressr::progress(p) ## duplicated - will be ignored
  relay_progress(slow_sum(1:2))
  progress(type = "finish")
  progress() ## one too many - will be ignored
})


message("with_progress() - progressr.debug = TRUE ... done")

library(progressr)

if (requireNamespace("furrr", quietly = TRUE)) {
  for (strategy in future_strategies) {
    future::plan(strategy)
    print(future::plan())

    message("* with_progress()")

    with_progress({
      p <- progressor(4)
      y <- furrr::future_map(3:6, function(n) {
        p()
        slow_sum_p(1:n, stdout=TRUE, message=TRUE)
      })
    })


    message("* global progression handler")

    handlers(global = TRUE)
    
    local({
      p <- progressor(4)
      y <- furrr::future_map(3:6, function(n) {
        p()
        slow_sum_p(1:n, stdout=TRUE, message=TRUE)
      })
    })
    
    handlers(global = FALSE)
    
    ## Explicitly close any PSOCK clusters to avoid 'R CMD check' NOTE
    ## on "detritus in the temp directory" on MS Windows
    future::plan("sequential")
  } ## for (strategy ...)
}

library(progressr)

options(progressr.clear = FALSE)

message("handler_ntfy() ...")

# 1. Force ntfy handler to run in fake mode
old_fake_handlers <- getOption("progressr.tests.fake_handlers")
env <- environment(progressr:::is_fake)
env$cache <- list()
options(progressr.tests.fake_handlers = c(old_fake_handlers, "handler_ntfy"))

# 2. Create handler
h <- handler_ntfy()
stopifnot(inherits(h, "ntfy_progression_handler"))

# Initiate
h(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-1", progression_index = 1L))

# Update
h(progression(type = "update", amount = 1, progressor_uuid = "uuid-1", progression_index = 2L))

# Update with amount = 0 (returns early)
h(progression(type = "update", amount = 0, progressor_uuid = "uuid-1", progression_index = 3L))

# Interrupt
h(progressr:::control_progression("interrupt", message = "Interrupted progression"))

# Shutdown
h(progressr:::control_progression("shutdown"))


# 3. Test when finished = TRUE in finish()
h_finish <- handler_ntfy()
h_finish(progression(type = "initiate", steps = 10, auto_finish = FALSE, progressor_uuid = "uuid-2", progression_index = 1L))
h_finish(progression(type = "update", amount = 1, progressor_uuid = "uuid-2", progression_index = 2L))
h_finish(progressr:::control_progression("shutdown"))
# Second shutdown to check finished == TRUE branch
h_finish(progressr:::control_progression("shutdown"))


# 4. Test times = 1L (returns early in initiate)
h_times1 <- handler_ntfy(times = 1L)
h_times1(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-3", progression_index = 1L))
h_times1(progressr:::control_progression("shutdown"))


# 5. Test times <= 2L (returns early in update)
h_times2 <- handler_ntfy(times = 2L)
h_times2(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-4", progression_index = 1L))
h_times2(progression(type = "update", amount = 1, progressor_uuid = "uuid-4", progression_index = 2L))
h_times2(progressr:::control_progression("shutdown"))


# 6. Test state$enabled is FALSE (returns early in initiate, update, finish)
h_enable <- handler_ntfy(enable_after = 100.0)
h_enable(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-5", progression_index = 1L))
h_enable(progression(type = "update", amount = 1, progressor_uuid = "uuid-5", progression_index = 2L))
h_enable(progressr:::control_progression("shutdown"))


# 7. Test state$delta <= 0 in finish() (should skip progress_notify)
h_delta <- handler_ntfy()
h_delta(progression(type = "initiate", steps = 10, auto_finish = FALSE, progressor_uuid = "uuid-6", progression_index = 1L))
h_delta(progression(type = "update", amount = 0, progressor_uuid = "uuid-6", progression_index = 2L))
h_delta(progressr:::control_progression("shutdown"))


# 8. Test standard with_progress flow
with_progress({
  p <- progressor(along = 1:2)
  for (i in 1:2) p()
}, handlers = handler_ntfy())


# Restore options
options(progressr.tests.fake_handlers = old_fake_handlers)
env$cache <- list()

message("handler_ntfy() ... done")

make_calling_handler <- function(handlers) {
  if (length(handlers) > 1L) {
    calling_handler <- function(p) {
      finished <- FALSE
      for (kk in seq_along(handlers)) {
        handler <- handlers[[kk]]
        if (handler(p)) finished <- TRUE
      }
      invisible(finished)
    }
  } else {
    calling_handler <- handlers[[1]]
  }
  calling_handler
}

library(progressr)

options(progressr.clear = FALSE)

message("handler_tutorial() ...")

# 1. Test basic function properties and invalid arguments
stopifnot(
  is.function(handler_tutorial)
)

err <- tryCatch(handler_tutorial(spec = "invalid"), error = identity)
stopifnot(inherits(err, "error"))

# 2. Test progress spec
if (requireNamespace("progress", quietly = TRUE)) {
  h1 <- handler_tutorial(spec = "progress")
  stopifnot(inherits(h1, "progress_progression_handler"))
  
  with_progress({
    p <- progressor(along = 1:2)
    for (i in 1:2) p()
  }, handlers = h1)
}

# 3. Test cli spec
if (requireNamespace("cli", quietly = TRUE)) {
  h2 <- handler_tutorial(spec = "cli")
  stopifnot(inherits(h2, "cli_progression_handler"))
  
  with_progress({
    p <- progressor(along = 1:2)
    for (i in 1:2) p()
  }, handlers = h2)
}

message("handler_tutorial() ... done")

library(progressr)

if (requireNamespace("pbmcapply", quietly = TRUE)) {
  handlers("pbmcapply")
  for (x in list(integer(0), 1:10, 1L)) {
    message("length(x): ", length(x))
    with_progress({ y <- slow_sum_p(x) })
    print(y)
  }
}

library(progressr)

options(progressr.demo.delay = 0.001)
options(progressr.interval = 0.0)
options(progressr.clear = FALSE)

message("with_progress() ...")

x <- 1:100
truth <- sum(x)

message("with_progress() - default ...")

if (requireNamespace("utils")) {
  with_progress({
    sum <- slow_sum(x)
  })
  print(sum)
  stopifnot(sum == truth)

  with_progress({
    cat("This stdout output will be delayed")
    message("This message will be delayed")
    warning("This warning will be delayed")
    signalCondition(simpleCondition("This simpleCondition will be delayed"))
    sum <- slow_sum(x)
  }, interval = 0.1, enable = TRUE, delay_conditions = "condition")
  print(sum)
  stopifnot(sum == truth)
}

message("with_progress() - default ... done")

message("with_progress() - filesize ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_filesize())
print(sum)
stopifnot(sum == truth)

message("with_progress() - filesize ... done")


message("with_progress() - utils::txtProgressBar() ...")

if (requireNamespace("utils")) {
  with_progress({
    sum <- slow_sum(x)
  }, handlers = handler_txtprogressbar(style = 2L))
  print(sum)
  stopifnot(sum == truth)
}

message("with_progress() - utils::txtProgressBar() ... done")


message("with_progress() - tcltk::tkProgressBar() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_tkprogressbar)

message("with_progress() - tcltk::tkProgressBar() ... done")


message("with_progress() - utils::winProgressBar() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_winprogressbar)

message("with_progress() - utils::winProgressBar() ... done")


message("with_progress() - progress::progress_bar() ...")

if (requireNamespace("progress")) {
  ## Display progress using default handler
  with_progress({
    sum <- slow_sum(x)
  }, handlers = handler_progress(clear = FALSE))
  print(sum)
  stopifnot(sum == truth)
}

message("with_progress() - progress::progress_bar() ... done")


message("with_progress() - pbmcapply::progressBar() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_pbmcapply)

message("with_progress() - pbmcapply::progressBar() ... done")


message("with_progress() - ascii_alert ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_ascii_alert())
print(sum)
stopifnot(sum == truth)

message("with_progress() - ascii_alert ... done")


message("with_progress() - beepr::beep() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_beepr)
print(sum)
stopifnot(sum == truth)

message("with_progress() - beepr::beep() ... done")


message("with_progress() - notifier::notify() ...")

with_progress({
  sum <- slow_sum(x)
}, handlers = handler_notifier)
print(sum)
stopifnot(sum == truth)

message("with_progress() - notifier::notify() ... done")


message("with_progress() - void ...")

## Mute progress updates
with_progress({
  sum <- slow_sum(x)
}, handlers = NULL)
print(sum)
stopifnot(sum == truth)

message(" - via option")

## NOTE: Set it to NULL, will use the default utils::txtProgressBar()
options(progressr.handlers = list())
with_progress({
  sum <- slow_sum(x)
})
print(sum)
stopifnot(sum == truth)

message("with_progress() - void ... done")


message("with_progress() - multiple handlers ...")

if (requireNamespace("utils", quietly = TRUE)) {
  handlers <- list(handler_txtprogressbar, handler_newline, handler_debug)
  options(progressr.handlers = handlers)
  
  with_progress({
    sum <- slow_sum(x)
  })
  print(sum)
  stopifnot(sum == truth)
}

message("with_progress() - multiple handlers ... done")


message("with_progress() - return value and visibility ...")

res <- with_progress(x)
stopifnot(identical(x, res))

res <- withVisible(with_progress(x))
stopifnot(identical(res$visible, TRUE))

res <- withVisible(with_progress(y <- x))
stopifnot(identical(res$visible, FALSE))

message("with_progress() - return value and visibility ... done")


message("with_progress() ... done")

#' Evaluate expression with stop watch reporting
#'
#' @param expr (optional) Expression to be evaluated and timed.
#'
#' @param reset (optional) If TRUE, the internal timer is reset.
#'
#' @return
#' Return the value of `expr`, if specified, otherwise NULL invisible.
#' 
#' @examples
#' tic <- progressr:::tic
#' toc <- progressr:::toc
#'
#' tic()
#' Sys.sleep(1.2) |> toc()
#' Sys.sleep(0.2) |> toc()
#' toc()
#'
#' tic()
#' Sys.sleep(1.2) |> toc()
#'
#' @noRd
toc <- local({
  start <- NULL
             
  function(expr = NULL, reset = FALSE, width = 30L) {
    t0 <- Sys.time()
    if (isTRUE(reset) || is.null(start)) start <<- t0
    if (missing(expr)) {
      dt_start <- round(difftime(t0, start), digits = 1L)
      message(sprintf("[%s]", format(dt_start)))
      return(invisible())
    }

    on.exit({
      code <- deparse(substitute(expr))
      code <- paste(code, collapse = " ")
      code <- gsub(" +", " ", code)
      if (nchar(code) > width) {
        code <- substr(code[1], start = 1L, stop = width)
        code <- sprintf("%s ...", code)
      }
      t1 <- Sys.time()
      dt <- round(difftime(t1, t0), digits = 2L)
      dt_tic <- round(difftime(t1, start), digits = 1L)
      message(sprintf("[%s] +%s : %s", format(dt_tic), format(dt), code))
    })

    expr
  }
})    

#' @noRd
tic <- function() toc(reset = TRUE)

#' Progression Handler: Progress Reported via 'cli' Progress Bars (Text) in the Terminal
#'
#' A progression handler for [cli::cli_progress_bar()].
#'
#' @inheritParams make_progression_handler
#'
#' @return A function of class `progression_handler` that takes a
#' [progression] condition as its first and only argument.
#'
#' @param show_after (numeric) Number of seconds to wait before displaying
#' the progress bar.
#'
#' @param type (character) The type of progress bar to display, which
#' controls the default `format` and `format_done` passed to [cli::cli_progress_bar()].
#' If `"default"`, the \pkg{cli} default format is used (`format = NULL`).
#' If `"steps"`, the progress bar shows the current and total number of
#' steps using the format string
#' `"{cli::pb_spin} {cli::pb_bar} {cli::pb_current}/{cli::pb_total} {cli::pb_status}"`
#' (and `format_done = "{cli::pb_bar} {cli::pb_current}/{cli::pb_total} {cli::pb_status}"`).
#' If `"percent"`, the progress bar shows the percentage completed using
#' the format string
#' `"{cli::pb_spin} {cli::pb_bar} {cli::pb_percent} {cli::pb_status}"`
#' (and `format_done = "{cli::pb_bar} {cli::pb_percent} {cli::pb_status}"`).
#' If `"time"`, the progress bar shows the percentage completed, the
#' current and total number of steps, the estimated time remaining (ETA),
#' and the total elapsed time using the format string
#' `"[{cli::pb_elapsed}] {cli::pb_spin} {cli::pb_bar} {cli::pb_percent} [{cli::pb_current}/{cli::pb_total}] (ETA: {cli::pb_eta}) {cli::pb_status}"`
#' (and `format_done = "[{cli::pb_elapsed}] {cli::pb_bar} {cli::pb_percent} [{cli::pb_current}/{cli::pb_total}] {cli::pb_status}"`).
#' For the meaning of these format variables, see
#' [Progress bar variables][cli::progress-variables] in the \pkg{cli} package.
#' This argument is ignored if `format` is explicitly specified via `...`.
#'
#' @param \ldots Additional arguments passed to [cli::cli_progress_bar()]
#' and [make_progression_handler()].
#'
#' @section Requirements:
#' This progression handler requires the \pkg{cli} package.
#'
#' @section Appearance:
#' Below are a few examples on how to use and customize this progress handler.
#' In all cases, we use `handlers(global = TRUE)`.
#'
#' ```{asciicast handler_cli-default}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' library(progressr)
#' handlers("cli")
#' y <- slow_sum_p(1:25)
#' ```
#'
#' ```{asciicast handler_cli-type-steps}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' library(progressr)
#' handlers(handler_cli(type = "steps"))
#' y <- slow_sum_p(1:25)
#' ```
#'
#' ```{asciicast handler_cli-type-percent}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' library(progressr)
#' handlers(handler_cli(type = "percent"))
#' y <- slow_sum_p(1:25)
#' ```
#'
#' ```{asciicast handler_cli-type-time}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' library(progressr)
#' handlers(handler_cli(type = "time"))
#' y <- slow_sum_p(1:25)
#' ```
#'
#' ```{asciicast handler_cli-format-1}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' library(progressr)
#' handlers(handler_cli(format = "{cli::pb_spin} {cli::pb_bar} {cli::pb_current}/{cli::pb_total} {cli::pb_status}"))
#' y <- slow_sum_p(1:25)
#' ```
#'
#' @example incl/handler_cli.R
#'
#' @export
handler_cli <- function(show_after = 0.0, intrusiveness = getOption("progressr.intrusiveness.terminal", 1), target = "terminal", type = c("default", "steps", "percent", "time"), ...) {
  type <- match.arg(type)

  ## Additional arguments passed to the progress-handler backend
  backend_args <- handler_backend_args(...)

  ## Default 'format' depending on 'type', unless 'format' is already specified
  if (is.null(backend_args[["format"]])) {
    format <- switch(type,
      default = NULL,
      steps   = "{cli::pb_spin} {cli::pb_bar} {cli::pb_current}/{cli::pb_total} {cli::pb_status}",
      percent = "{cli::pb_spin} {cli::pb_bar} {cli::pb_percent} {cli::pb_status}",
      time    = "[{cli::pb_elapsed}] {cli::pb_spin} {cli::pb_bar} {cli::pb_percent} [{cli::pb_current}/{cli::pb_total}] (ETA: {cli::pb_eta}) {cli::pb_status}"
    )
    if (!is.null(format)) backend_args[["format"]] <- format
  }

  ## Default 'format_done' depending on 'type', unless 'format_done' is already specified
  if (is.null(backend_args[["format_done"]])) {
    format_done <- switch(type,
      default = NULL,
      steps   = "{cli::pb_bar} {cli::pb_current}/{cli::pb_total} {cli::pb_status}",
      percent = "{cli::pb_bar} {cli::pb_percent} {cli::pb_status}",
      time    = "[{cli::pb_elapsed}] {cli::pb_bar} {cli::pb_percent} [{cli::pb_current}/{cli::pb_total}] {cli::pb_status}"
    )
    if (!is.null(format_done)) backend_args[["format_done"]] <- format_done
  }

  if (!is_fake("handler_cli")) {
    cli_n_colors <- cli::num_ansi_colors()

    cli_freeze_color_options <- function() {      
      options(
        cli.num_colors = cli_n_colors,
        crayon.enabled = (cli_n_colors > 1L),
        crayon.colors = cli_n_colors
      )
    }

    cli_progress_call <- function(fun, ...) {
      opts <- cli_freeze_color_options()
      on.exit(options(opts), add = TRUE)

      withCallingHandlers({
        res <- fun(...)
      }, cliMessage = function(msg) {
        output <- conditionMessage(msg)
        ## WORKAROUND: cli_progress_done() outputs a newline
        ## at the end. This prevents us from (optionally)
        ## clearing the line afterward
        if (identical(fun, cli::cli_progress_done)) {
          output <- gsub("[\n\r]+$", "", output)
        }
        cat(output, file = stderr())
        tryInvokeRestart("muffleMessage")
      })

      invisible(res)
    }

    cli_progress_bar <- function(total, ...) {
      ## WORKAROUND: Do not involve 'cli' when total = 0
      if (total == 0) return(NA_character_)

      ## Make sure to always use the built-in 'cli' handler, no matter
      ## what 'cli.progress_handlers*' options might be set
      opts2 <- options(cli.progress_handlers_only = "cli")
      on.exit(options(opts2), add = TRUE)

      cli_progress_call(cli::cli_progress_bar, total = total, ...)
    }

    cli_progress_update <- function(id, ..., .envir) {
      ## WORKAROUND: Do not involve 'cli' when total = 0
      if (.envir$total == 0) return(id)
      cli_progress_call(cli::cli_progress_update, id = id, ..., .envir = .envir)
    }

    cli_progress_done <- function(id, ..., .envir) {
      ## WORKAROUND: Do not involve 'cli' when total = 0
      if (.envir$total == 0) return(id)
      cli_progress_call(cli::cli_progress_done, id = id, ..., .envir = .envir)
    }

    erase_progress_bar <- function(pb) {
      if (is.null(pb)) return()
      stopifnot(is.character(pb$id), is.environment(pb$envir))
      pb_width <- local({
        oopts <- options(cli.width = NULL)
        on.exit(options(oopts))
        cli::console_width()
      }) - 1L
      msg <- c("\r", rep(" ", times = pb_width), "\r")
      msg <- paste(msg, collapse = "")
      cat(msg, file = stderr())
    }
    
    redraw_progress_bar <- function(pb) {
      if (is.null(pb)) return()
      stopifnot(is.character(pb$id), is.environment(pb$envir))
      cli_progress_update(id = pb$id, inc = 0, .envir = pb$envir)
    }
  } else {
    cli_progress_bar <- function(total, ...) "id-dummy"
    cli_progress_update <- function(id, ..., .envir) "id-dummy"
    cli_progress_done <- function(id, ..., .envir) "id-dummy"
    erase_progress_bar <- function(pb) NULL
    redraw_progress_bar <- function(pb) NULL
  }

  reporter <- local({
    pb <- NULL
    
    make_pb <- function(total, clear, show_after, ...) {
      if (!is.null(pb)) return(pb)
      stop_if_not(
        is.numeric(total), length(total) == 1L, is.finite(total)
      )
      envir <- new.env()
      envir$total <- total
      args <- c(list(total = total, ..., .envir = envir), backend_args)
      args$clear <- FALSE
      args$auto_terminate <- FALSE
      args$.auto_close <- FALSE

      old_show_after <- getOption("cli.progress_show_after", NULL)
      options(cli.progress_show_after = show_after)
      on.exit(options(cli.progress_show_after = old_show_after))

      id <- do.call(cli_progress_bar, args = args)

      pb <<- list(id = id, total = total, envir = envir, show_after = show_after)
      stopifnot(is.character(pb$id), is.environment(pb$envir))
    }

    pb_tick <- function(pb, delta = 0, message = NULL, ...) {
      if (is.null(pb)) return()
      stopifnot(
        is.character(pb$id), length(pb$id) == 1L,
        is.numeric(delta), length(delta) == 1L, is.finite(delta),
        is.environment(pb$envir)
      )
      if (delta <= 0) return()

      old_show_after <- getOption("cli.progress_show_after", NULL)
      options(cli.progress_show_after = pb$show_after)
      on.exit(options(cli.progress_show_after = old_show_after))

      cli_progress_update(id = pb$id, inc = delta, status = message, .envir = pb$envir)
    }

    pb_update <- function(pb, ratio, message = NULL, ...) {
      if (is.null(pb)) return()
      stopifnot(is.character(pb$id), is.environment(pb$envir))

      old_show_after <- getOption("cli.progress_show_after", NULL)
      options(cli.progress_show_after = pb$show_after)
      on.exit(options(cli.progress_show_after = old_show_after))

      if (ratio >= 1.0) {
        cli_progress_done(id = pb$id, .envir = pb$envir)
      } else {
        set <- ratio * pb$total
        stopifnot(length(set) == 1L, is.numeric(set), is.finite(set),
                  set >= 0, set < pb$total)
        cli_progress_update(id = pb$id, set = set, status = message, .envir = pb$envir)
      }
    }

    list(
      reset = function(...) {
        pb <<- NULL
      },

      hide = function(...) {
        if (is.null(pb)) return()
        erase_progress_bar(pb)
      },

      unhide = function(...) {
        if (is.null(pb)) return()
        redraw_progress_bar(pb)
      },

      interrupt = function(config, state, progression, ...) {
        if (is.null(pb)) return()
        stopifnot(is.character(pb$id), is.environment(pb$envir))

        ## WORKAROUND: At times, we might get 'Error in
        ## del_from:length(app$styles) : NA/NaN argument'.
        tryCatch({
          erase_progress_bar(pb)
          redraw_progress_bar(pb)
        }, error = identity)
        
        msg <- conditionMessage(progression)
        msg <- paste(c("", msg, ""), collapse = "\n")
        cat(msg, file = stderr())
      },

      initiate = function(config, state, progression, ...) {
        if (!state$enabled || config$times == 1L) return()
        stop_if_not(is.null(pb))
        make_pb(total = config$max_steps,
                clear = config$clear, show_after = config$enable_after)
        pb_tick(pb, delta = 0, message = state$message)
      },
        
      update = function(config, state, progression, ...) {
        if (!state$enabled || config$times <= 2L) return()
        if (progression$amount == 0) return()
        
        make_pb(total = config$max_steps,
                clear = config$clear, show_after = config$enable_after)
        if (inherits(progression, "sticky") && length(state$message) > 0) {
          # FIXME: pb$message(state$message)
        }
        old_show_after <- getOption("cli.progress_show_after", NULL)
        options(cli.progress_show_after = pb$show_after)
        on.exit(options(cli.progress_show_after = old_show_after))

        cli_progress_update(id = pb$id, set = state$step, status = state$message, .envir = pb$envir)
      },
        
      finish = function(config, state, progression, ...) {
        ## Already finished?
        if (is.null(pb)) return()
        stopifnot(is.character(pb$id), is.environment(pb$envir))
        make_pb(total = config$max_steps,
                clear = config$clear, show_after = config$enable_after)
        reporter$update(config = config, state = state, progression = progression, ...)
        if (config$clear) {
          pb_update(pb, ratio = 1.0)
          erase_progress_bar(pb)
        } else {
          pb_update(pb, ratio = 1.0)
          cat(file = stderr(), "\n")
        }
        
        pb <<- NULL
      }
    )
  })

  make_progression_handler("cli", reporter, intrusiveness = intrusiveness, target = target, ...)
}

library(progressr)

options(progressr.clear = FALSE)

if (requireNamespace("progress", quietly = TRUE)) {
  
  # Save original fake handlers option
  old_fake_handlers <- getOption("progressr.tests.fake_handlers")
  
  # A helper to run tests for a given mode (fake or real)
  run_tests <- function(mode_label) {
    message(sprintf("Running handler_progress tests in %s mode...", mode_label))
    
    # 1. Test basic with_progress flow for all types
    for (type in c("default", "steps", "percent", "time")) {
      message(sprintf("handler_progress(type = '%s') ...", type))
      options(progressr.handlers = handler_progress(type = type))
      for (x in list(integer(0), 1:10, 1L)) {
        message("length(x): ", length(x))
        with_progress({
          progress <- progressor(along = x)
          for (ii in x) {
            Sys.sleep(getOption("progressr.demo.delay", 0.1))
            progress(message = sprintf("(%s)", paste(letters[1:ii], collapse="")))
          }
        })
      }
      message(sprintf("handler_progress(type = '%s') ... done", type))
    }
    
    # 2. Test manual progression inputs to cover internal states and branches
    # Test normal lifecycle with clear = FALSE
    h <- handler_progress(clear = FALSE)
    h(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-1", progression_index = 1L))
    h(progression(type = "update", amount = 1, progressor_uuid = "uuid-1", progression_index = 2L))
    h(progression(type = "update", amount = 0, progressor_uuid = "uuid-1", progression_index = 3L))
    # Negative amount to trigger delta < 0
    h(progression(type = "update", amount = -1, progressor_uuid = "uuid-1", progression_index = 4L))
    # Sticky progression
    h(progression(type = "update", amount = 1, class = "sticky", message = "sticky message", progressor_uuid = "uuid-1", progression_index = 5L))
    h(progressr:::control_progression("hide"))
    h(progressr:::control_progression("unhide"))
    h(progressr:::control_progression("interrupt", message = "Interrupt message"))
    h(progressr:::control_progression("shutdown"))
    
    # Test lifecycle with clear = TRUE
    h_clear <- handler_progress(clear = TRUE)
    h_clear(progression(type = "initiate", steps = 5, auto_finish = TRUE, progressor_uuid = "uuid-2", progression_index = 1L))
    h_clear(progression(type = "update", amount = 1, progressor_uuid = "uuid-2", progression_index = 2L))
    h_clear(progressr:::control_progression("shutdown"))
    
    # Test times = 1L (returns early in initiate)
    h_times1 <- handler_progress(times = 1L)
    h_times1(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-3", progression_index = 1L))
    h_times1(progressr:::control_progression("shutdown"))
    
    # Test times <= 2L (returns early in update)
    h_times2 <- handler_progress(times = 2L)
    h_times2(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-4", progression_index = 1L))
    h_times2(progression(type = "update", amount = 1, progressor_uuid = "uuid-4", progression_index = 2L))
    h_times2(progressr:::control_progression("shutdown"))
    
    # Test max_steps = 0
    h_zero <- handler_progress()
    h_zero(progression(type = "initiate", steps = 0, auto_finish = TRUE, progressor_uuid = "uuid-5", progression_index = 1L))
    h_zero(progression(type = "update", amount = 0, progressor_uuid = "uuid-5", progression_index = 2L))
    h_zero(progressr:::control_progression("shutdown"))
    
    # Test enabled = FALSE initially (enable_after > 0)
    h_enable <- handler_progress(enable_after = 100.0)
    h_enable(progression(type = "initiate", steps = 10, auto_finish = TRUE, progressor_uuid = "uuid-6", progression_index = 1L))
    h_enable(progression(type = "update", amount = 1, progressor_uuid = "uuid-6", progression_index = 2L))
    h_enable(progressr:::control_progression("shutdown"))
    
    # Test pb$finished branches (auto_finish = FALSE and ticking to completion)
    h_finished <- handler_progress(clear = FALSE)
    h_finished(progression(type = "initiate", steps = 2, auto_finish = FALSE, progressor_uuid = "uuid-7", progression_index = 1L))
    h_finished(progression(type = "update", amount = 2, progressor_uuid = "uuid-7", progression_index = 2L))
    h_finished(progressr:::control_progression("hide"))
    h_finished(progressr:::control_progression("unhide"))
    h_finished(progressr:::control_progression("shutdown"))
  }
  
  # Run in Fake Mode (default config of testme)
  run_tests("fake")
  
  # Run in Real Mode
  # Clear is_fake cache and remove handler_progress from fake_handlers option
  env <- environment(progressr:::is_fake)
  env$cache <- list()
  options(progressr.tests.fake_handlers = setdiff(old_fake_handlers, "handler_progress"))
  
  run_tests("real")
  
  # Restore options
  options(progressr.tests.fake_handlers = old_fake_handlers)
  env$cache <- list()
}

#! /usr/bin/env Rscript
## This runs testme test script inst/testme/test-with_progress,delay_stdout.R
## Don't edit - it was autogenerated by inst/testme/deploy.R
progressr:::testme("with_progress,delay_stdout")

library(progressr)

options(progressr.tests.fake_handlers = c("handler_beepr", "handler_brrr", "handler_notifier", "handler_pbmcapply", "handler_tkprogressbar", "handler_winprogressbar"))
options(progressr.clear = FALSE)
options(progressr.enable_after = NULL)
options(progressr.debug = FALSE)
options(progressr.times = NULL)

record_output <- function(expr, envir = parent.frame()) {
  conditions <- list()
  stdout <- utils::capture.output({
    withCallingHandlers(
      expr,
      condition = function(c) {
        if (inherits(c, c("progression", "error"))) return()
        conditions[[length(conditions) + 1L]] <<- c
      }
    )
  }, split = TRUE)
  list(stdout = stdout, conditions = conditions)
}


message("*** with_progress() - delaying output ...")

x <- 1:10

## Record truth
output_truth <- record_output({
  y_truth <- slow_sum_p(x, stdout = TRUE, message = TRUE, sticky = TRUE)
})

for (delay in c(FALSE, TRUE)) {
  message(sprintf("- with_progress() - delay = %s ...", delay))
  output <- record_output({
    with_progress({
      y <- slow_sum_p(x, stdout = TRUE, message = TRUE, sticky = TRUE)
    }, delay_stdout = delay,
      delay_conditions = if (delay) "condition" else character(0L))
  })

  stopifnot(identical(output$stdout, output_truth$stdout))
  stopifnot(identical(output$conditions, output_truth$conditions))
  stopifnot(identical(y, y_truth))
  message(sprintf("- with_progress() - delay = %s ... DONE", delay))
}

message("*** with_progress() - delaying output ... DONE")

library(progressr)

options(progressr.demo.delay = 0.001)
options(progressr.interval = 0.0)
options(progressr.clear = FALSE)


message("slow_sum_p() ...")

x <- 1:5
truth <- sum(x)

res <- slow_sum_p(x)
stopifnot(identical(res, truth))

res <- with_progress({
  slow_sum_p(x)
})
stopifnot(identical(res, truth))

message("slow_sum_p() ... done")


message("slow_sum() ...")

x <- 1:5
truth <- sum(x)

# slow_sum should not produce any output or messages or progress updates by default
out <- capture.output({
  msgs <- list()
  progressions <- list()
  withCallingHandlers({
    res <- slow_sum(x)
  }, message = function(m) {
    msgs[[length(msgs) + 1L]] <<- m$message
  }, progression = function(p) {
    progressions[[length(progressions) + 1L]] <<- p
  })
})
stopifnot(identical(res, truth))
stopifnot(length(out) == 0L)
stopifnot(length(msgs) == 0L)
stopifnot(length(progressions) == 0L)

# check that we can enable options/arguments on slow_sum
out <- capture.output({
  res <- slow_sum(1:2, stdout = TRUE)
})
stopifnot(any(grepl("O: Element", out)))

message("slow_sum() ... done")


message("slow_sqrt_p() ...")

x <- c(1, 4, 9, 16, 25)
truth <- sqrt(x)

res <- slow_sqrt_p(x)
stopifnot(identical(res, truth))

res <- with_progress({
  slow_sqrt_p(x)
})
stopifnot(identical(res, truth))

## Test progressr.demo.stdout, progressr.demo.message options
# 1. Test progressr.demo.stdout
options(progressr.demo.stdout = TRUE)
out <- capture.output({
  res <- slow_sqrt_p(1:2)
})
stopifnot(any(grepl("O: Element", out)))

options(progressr.demo.stdout = FALSE)
out <- capture.output({
  res <- slow_sqrt_p(1:2)
})
stopifnot(!any(grepl("O: Element", out)))
options(progressr.demo.stdout = NULL)

# 2. Test progressr.demo.message
options(progressr.demo.message = FALSE)
msgs <- list()
withCallingHandlers({
  res <- slow_sqrt_p(1:2)
}, message = function(m) {
  msgs[[length(msgs) + 1L]] <<- m$message
})
stopifnot(length(msgs) == 0L)

options(progressr.demo.message = TRUE)
msgs <- list()
withCallingHandlers({
  res <- slow_sqrt_p(1:2)
}, message = function(m) {
  msgs[[length(msgs) + 1L]] <<- m$message
})
stopifnot(length(msgs) > 0L)
stopifnot(any(grepl("M: Calculated value", unlist(msgs))))
options(progressr.demo.message = NULL)

# 3. Test explicit arguments override options
options(progressr.demo.stdout = TRUE)
out <- capture.output({
  res <- slow_sqrt_p(1:2, stdout = FALSE)
})
stopifnot(!any(grepl("O: Element", out)))
options(progressr.demo.stdout = NULL)

options(progressr.demo.message = FALSE)
msgs <- list()
withCallingHandlers({
  res <- slow_sqrt_p(1:2, message = TRUE)
}, message = function(m) {
  msgs[[length(msgs) + 1L]] <<- m$message
})
stopifnot(length(msgs) > 0L)
options(progressr.demo.message = NULL)

# 4. Test progressr.demo.progress option and parameter
progressions <- list()
withCallingHandlers({
  res <- slow_sqrt_p(1:2, progress = FALSE)
}, progression = function(p) {
  progressions[[length(progressions) + 1L]] <<- p
})
stopifnot(length(progressions) == 0L)

progressions <- list()
withCallingHandlers({
  res <- slow_sqrt_p(1:2, progress = TRUE)
}, progression = function(p) {
  progressions[[length(progressions) + 1L]] <<- p
})
stopifnot(length(progressions) > 0L)

# Test option progressr.demo.progress
options(progressr.demo.progress = FALSE)
progressions <- list()
withCallingHandlers({
  res <- slow_sqrt_p(1:2)
}, progression = function(p) {
  progressions[[length(progressions) + 1L]] <<- p
})
stopifnot(length(progressions) == 0L)
options(progressr.demo.progress = NULL)

message("slow_sqrt_p() ... done")


message("slow_sqrt() ...")

x <- c(1, 4, 9, 16, 25)
truth <- sqrt(x)

# slow_sqrt should not produce any output or messages or progress updates by default
out <- capture.output({
  msgs <- list()
  progressions <- list()
  withCallingHandlers({
    res <- slow_sqrt(x)
  }, message = function(m) {
    msgs[[length(msgs) + 1L]] <<- m$message
  }, progression = function(p) {
    progressions[[length(progressions) + 1L]] <<- p
  })
})
stopifnot(identical(res, truth))
stopifnot(length(out) == 0L)
stopifnot(length(msgs) == 0L)
stopifnot(length(progressions) == 0L)

# check that we can enable options/arguments on slow_sqrt
out <- capture.output({
  res <- slow_sqrt(1:2, stdout = TRUE)
})
stopifnot(any(grepl("O: Element", out)))

message("slow_sqrt() ... done")

## Create a universally unique identifier (UUID) for an R object
#' @importFrom digest digest
uuid <- function(source, keep_source = FALSE) {
  uuid <- digest(source)
  uuid <- strsplit(uuid, split = "")[[1]]
  uuid <- paste(c(uuid[1:8], "-", uuid[9:12], "-", uuid[13:16], "-", uuid[17:20], "-", uuid[21:32]), collapse = "")
  if (keep_source) attr(uuid, "source") <- source
  uuid
} ## uuid()


## A universally unique identifier (UUID) for the current
## R process UUID. Generated only once per process ID 'pid'.
## The 'pid' may differ when using forked processes.
session_uuid <- local({
  uuids <- list()

  function(pid = Sys.getpid(), attributes = FALSE) {
    pidstr <- as.character(pid)
    uuid <- uuids[[pidstr]]
    if (is.null(uuid)) {
      info <- Sys.info()
      host <- Sys.getenv(c("HOST", "HOSTNAME", "COMPUTERNAME"))
      host <- host[nzchar(host)]
      host <- if (length(host) == 0L) info[["nodename"]] else host[1L]
      info <- list(
        host = host,
        info = info,
        time = Sys.time(),
        tempdir = tempdir(),
        pid = pid,
        random = stealth_sample.int(.Machine$integer.max, size = 1L)
      )
      uuid <- uuid(info, keep_source = TRUE)
      uuids[[pidstr]] <<- uuid
    }
    
    if (!attributes) attr(uuid, "source") <- NULL
    
    uuid
  }
})


progressor_uuid <- function(id, attributes = FALSE) {
  uuid(list(session_uuid = session_uuid(), id = id), keep_source = attributes)
}

## A version of base::sample.int() that does not change .Random.seed
stealth_sample.int <- function(...) {
  oseed <- .GlobalEnv$.Random.seed
  on.exit({
    if (is.null(oseed)) {
      rm(list = ".Random.seed", envir = .GlobalEnv, inherits = FALSE)
    } else {
      .GlobalEnv$.Random.seed <- oseed
    }
  })
  suppressWarnings(sample.int(...))
}

#' Progression Handler: Progress Reported as ASCII BEL Symbols (Audio or Blink) in the Terminal
#'
#' A progression handler based on `cat("\a", file=stderr())`.
#'
#' @inheritParams make_progression_handler
#'
#' @return A function of class `progression_handler` that takes a
#' [progression] condition as its first and only argument.
#'
#' @param symbol (character string) The character symbol to be outputted,
#' which by default is the ASCII BEL character (`'\a'` = `'\007'`) character.
#'
#' @param file (connection) A [base::connection] to where output should be sent.
#'
#' @param \ldots Additional arguments passed to [make_progression_handler()].
#'
#' @example incl/handler_ascii_alert.R
#'
#' @export
handler_ascii_alert <- function(symbol = "\a", file = stderr(), intrusiveness = getOption("progressr.intrusiveness.audio", 5.0), target = c("terminal", "audio"), ...) {
  reporter <- local({
    list(
      hide   = function(...) NULL,
      unhide = function(...) NULL,
      interrupt = function(config, state, progression, ...) {
        msg <- conditionMessage(progression)
        msg <- paste(c("", msg, ""), collapse = "\n")
        cat(msg, file = file)
      },
      update = function(config, state, progression, ...) {
        if (state$enabled && progression$amount != 0) cat(file = file, symbol)
      }
    )
  })

  make_progression_handler("ascii_alert", reporter, intrusiveness = intrusiveness, target = target, ...)
}

<!--
%\VignetteIndexEntry{progressr: Appendix}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{packages}
%\VignetteEngine{progressr::selfonly}
-->

## How tos

### Report on progress in non-interactive mode ("batch mode")

When running R from the command line, R runs in a non-interactive mode
(`interactive()` returns `FALSE`).  The default behavior of
**progressr** is to _not_ report on progress in non-interactive mode.
To report on progress also then, set R options `progressr.enable` or
environment variable `R_PROGRESSR_ENABLE` to `TRUE`.  For example,

```sh
$ Rscript -e "library(progressr)" -e "with_progress(y <- slow_sum(1:10))"
```

will _not_ report on progress, whereas

```sh
$ export R_PROGRESSR_ENABLE=TRUE
$ Rscript -e "library(progressr)" -e "with_progress(y <- slow_sum(1:10))"
```

will.


## Notes of caution

### Avoid sending progress updates too frequently

Signaling progress updates comes with some overhead.  In situations
where we use progress updates, this overhead is typically much smaller
than the task we are processing in each step.  However, if the task we
iterate over is quick, then the extra time induced by the progress
updates might end up dominating the overall processing time.  If that
is the case, a simple solution is to only signal progress updates
every nth step.  Here is a version of `slow_sum()` that signals
progress every 10th iteration:

```
slow_sum <- function(x) {
  p <- progressr::progressor(length(x) / 10)
  sum <- 0
  for (kk in seq_along(x)) {
    Sys.sleep(0.1)
    sum <- sum + x[kk]
    if (kk %% 10 == 0) p(message = sprintf("Adding %g", x[kk]))
  }
  sum
}
```

The overhead of progress signaling may depend on context.  For
example, in parallel processing with near-live progress updates via
'multisession' futures, each progress update is communicated via a
socket connections back to the main R session.  These connections
might become clogged up if progress updates are too frequent.


## Known Limitations

### The global progress handler cannot be set during package load

It is _not_ possible to call `handlers(global = TRUE)` in all
circumstances.  For example, it cannot be called within `tryCatch()`
and `withCallingHandlers()`;

```r
> tryCatch(handlers(global = TRUE), error = identity)
Error in globalCallingHandlers(NULL) :
  should not be called with handlers on the stack
```

This is not a bug - neither in **progressr** nor in R itself. It's due
to a conservative design on how _global_ calling handlers should work
in R. If it were allowed, there's a risk we might end up getting weird and
unpredictable behaviors when messages, warnings, errors, and other
types of conditions are signaled.

Because `tryCatch()` and `withCallingHandlers()` are used in many
places throughout base R, this means that we also cannot call
`handlers(global = TRUE)` as part of a package's startup process,
e.g. `.onLoad()` or `.onAttach()`.

Another example of this error is if `handlers(global = TRUE)` is used
inside package vignettes and dynamic documents such as Rmarkdown.  In
such cases, the global progress handler has to be enabled _prior_ to
processing the document, e.g.

```r
> progressr::handlers(global = TRUE)
> rmarkdown::render("input.Rmd")
```


### A progressor cannot be created in the global environment

It is not possible to create a progressor in the global environment,
e.g. in the top-level of a script. It can only be created inside a
function, within `with_progress({ ... })`, `local({ ... })`, or a
similar construct.  For example, the following:

```r
library(progressr)
handlers(global = TRUE)

xs <- 1:5
p <- progressor(along = xs)
y <- lapply(xs, function(x) {
  Sys.sleep(0.1)
  p(sprintf("x=%g", x))
  sqrt(x)
})
```

results in an error if tried:

```
Error in progressor(along = xs) :
  A progressor must not be created in the global environment unless wrapped in a
  with_progress() or without_progress() call. Alternatively, create it inside a
  function or in a local() environment to make sure there is a finite lifespan
  of the progressor
```

The solution is to wrap it in a `local({ ... })` call, or more
explicitly, in a `with_progress({ ... })` call:

```r
library(progressr)
handlers(global = TRUE)

xs <- 1:5
with_progress({
  p <- progressor(along = xs)
  y <- lapply(xs, function(x) {
    Sys.sleep(0.1)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
})
#  |====================                               |  40%
```

The main reason for this is to limit the lifespan of each progressor.
If we created it in the global environment, there is a significant
risk it would never finish and block all of the following progressors.


### No coloring by the 'cli' and 'crayon' packages

The **[cli]** package can be used to output colored text and messages
in the terminal, e.g. `message(cli::col_blue("hey"))`. However,
**cli** disables such coloring by default when there is an active
progressor. The workaround is to set **cli** option `cli.num_colors` to:

```r
options(cli.num_colors = cli::num_ansi_colors())
```

This will force **cli** to use the same number of colors with and
without an active progressor.

The same happens when using the superseeded **[crayon]** package for
colorization. To re-enable coloring for **crayon**, set the same
(sic!)  above R options.

The technical reason for **cli** and **crayon** disabling coloring, is
that the **progressr** package buffers standard output using a sink,
which these packages consider to be mono-color only.


## Known Issues

### RStudio bug #16331: Setting global progressr handlers during startup does not work

Setting the global progressr handler in `~/.Rprofile` does _not_ work
in RStudio 2025.09:

```r
if (requireNamespace("progressr", quietly = TRUE)) {
  progressr::handlers(global = TRUE)
}
```

This is due to a [bug introduced in RStudio
2025.09](https://github.com/rstudio/rstudio/issues/16331), which has
been fixed for the next release RStudio 2025.11. If you are using
RStudio 2025.09, the workaround is to instead use:

```r
if (requireNamespace("progressr", quietly = TRUE)) {
  progressr::handlers(global = TRUE)

  ## Workaround for RStudio 2025.09 console bug #16331
  if (nzchar(Sys.getenv("RSTUDIO")) && !nzchar(Sys.getenv("RSTUDIO_TERM"))) {
    invisible(addTaskCallback(function(...) {
      ver <- RStudio.Version()$version
      if (ver >= "2025.09" && ver < "2025.11") {
        message("Workaround for RStudio 2025.09 bug #16331: Added progressr global handler")
        progressr::handlers(global = TRUE)
      } else {
        warning("Workaround for RStudio 2025.09 bug #16331: Not needed in RStudio v", ver, ". Please remove task callback 'rstudio::progressr::once' in your Rprofile startup file", call. = FALSE, immediate. = TRUE)
      }
      removeTaskCallback("rstudio::progressr::once")
    }, name = "rstudio::progressr::once"))
  }
}
```


### Positron bug #6892: Setting global progressr handlers during startup does not work

Positron does not support setting global calling handlers during R's
startup process, e.g. in `~/.Rprofile`. Even if such handlers are
registered, they have no effect. This is a [bug in
Positron](https://github.com/posit-dev/positron/issues/6892)
(2025-03-20), which was most recently confirmed with Position
2025.09.0 on Linux. Because of this, having something like in your
`~/.Rprofile`:

```r
if (requireNamespace("progressr", quietly = TRUE)) {
  progressr::handlers(global = TRUE)
}
```

will have no effect. If used, the workaround is to manually
re-registering all calling handlers, which can be done as:

```r
if (requireNamespace("progressr", quietly = TRUE)) {
  progressr::handlers(global = TRUE)

  ## Workaround for Positron (>= 2025.09) bug #16331
  if (nzchar(Sys.getenv("POSITRON"))) local({
    ver <- numeric_version(Sys.getenv("POSITRON_VERSION"))
    if (ver >= "2025.09") {
      message("Workaround for Positron (>= 2025.09) bug #6892: progressr global handler will be installed *after* the next call has been completed")
      invisible(addTaskCallback(function(...) {
        message("Workaround for Positron (>= 2025.09) bug #6892: re-installed progressr global handler")
        globalCallingHandlers(globalCallingHandlers(NULL))
        removeTaskCallback("positron::progressr::once")
      }, name = "positron::progressr::once"))
    }
  })
}
```

Alternatively, call:

```r
globalCallingHandlers(globalCallingHandlers(NULL))
```

or

```r
progressr::handlers(global = FALSE)  ## important
progressr::handlers(global = TRUE)
```

at the prompt.


### Positron: Messages add extra newline before the final progress step

One of the features of **progressr** is that messages are buffered
during progress reporting and relayed as soon as possible, which
typically happens just before handlers re-render the progress
output. This way you can use `message()` as usual, regardless whether
progress is reported or not.

Currently, when using Positron (e.g. Positron 2025.09.0), any
`message()` output adds an extra newline before the _final_ progress
step is reported, e.g.

```r
> progressr::handlers(global = TRUE)
> progressr::handlers("cli")
> y <- progressr::slow_sum(1:5, message = TRUE)
M: Added value 1
M: Added value 2
M: Added value 3
M: Added value 4

M: Added value 5
> 
```

I do not fully understand the reason for this, but I hope we can get
to the bottom of it and fix it, either in **progressr** or in
Positron, e.g. [Bug #9486](https://github.com/posit-dev/positron/issues/9486)
(2025-09-18).


### Jupyter: Reporting progress to stderr (default) does not work

The default for most terminal progress renders, including the ones for
**progressr**, display the progress on standard error (stderr). Due to
a limitation in Jupyter, this default does not work there. The reason is
that [Jupyter silently
drops](https://github.com/futureverse/progressr/issues/170) any output
sent to stderr, e.g.

```r
> cat("hello stderr\n", file = stderr())
> cat("hello stdout\n", file = stdout())
hello stdout
>
```

If we try the following

```r
library(progressr)
handlers(globals = TRUE)
handlers("txtprogressbar")
y <- slow_sum(1:20)
```

there will be no progress being reported. This is not specific to
**progressr**, we have the same problem with for instance **cli**. Try
for instance,

```r
void <- cli::cli_progress_demo(delay = 1.0)
```

The workaround is to direct all progress output to the standard output
(stdout) when working in Jupyter. For this to work, we also need to
disable the buffering ("delaying") of any other output to stdout.

```r
library(progressr)
handlers(globals = TRUE)

## Workaround for Jupyter
options(progressr.enable = TRUE, progressr.delay_stdout = FALSE)

## Jupyter requires that progress is rendered to standard output;
## it does not work with the default standard error
handlers(handler_txtprogressbar(file = stdout()))

y <- slow_sum(1:20)
```


### Jupyter bug #732 - handlers("progress") output is messy

Jupyter has other outputting issues. Specifically, Jupyter [injects an
_extra_ newline at the end of every
message](https://github.com/IRkernel/IRkernel/issues/732), e.g.

```r
> message("abc", appendLF = FALSE); message("def", appendLF = FALSE)
abc
def
> message("abc"); message("def")
abc

def

>
```

This causes any progress framework (e.g. the **progress** package)
that reports via messages to render progress output very poorly or not
at all.


## Design and Implementation

### Under the hood

When using the **progressr** package, progression updates are
communicated via R's condition framework, which provides methods for
creating, signaling, capturing, muffling, and relaying conditions.
Progression updates are of classes `progression` and
`immediateCondition`(\*).  The below figure gives an example how
progression conditions are created, signaled, and rendered.

(\*) The `immediateCondition` class of conditions are relayed as soon
as possible by the **[future]** framework, which means that
progression updates produced in parallel workers are reported to the
end user as soon as the main R session have received them.

![](imgs/slow_sum.svg)

_Figure: Sequence diagram illustrating how signaled progression
conditions are captured by `with_progress()`, or the global
progression handler, and relayed to the two progression handlers
'progress' (a progress bar in the terminal) and 'beepr' (auditory)
that the end user has chosen._


### Roadmap

Because this project is under active development, the progressr API is
currently kept at a very minimum.  This will allow for the framework
and the API to evolve while minimizing the risk for breaking code that
depends on it.  The roadmap for developing the API is roughly:

* [x] Provide minimal API for producing progress updates,
      i.e. `progressor()`, `with_progress()`, `handlers()`

* [x] Add support for global progress handlers removing the need for
      the user having to specify `with_progress()`, 
      i.e. `handlers(global = TRUE)` and `handlers(global = FALSE)`

* [ ] Make it possible to create a progressor also in the global
      environment (see 'Known Issues' above)

* [ ] Add support for nested progress updates

* [ ] Add API to allow users and package developers to design
      additional progression handlers

For a more up-to-date view on what features might be added, see
<https://github.com/futureverse/progressr/issues>.


[progressr]: https://progressr.futureverse.org
[future]: https://future.futureverse.org
[cli]: https://cran.r-project.org/package=cli
[crayon]: https://cran.r-project.org/package=crayon

<!--
%\VignetteIndexEntry{progressr: Replace 'cli' Progress Bars with 'progressr'}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{cli}
%\VignetteKeyword{purrr}
%\VignetteEngine{progressr::selfonly}
-->

The **cli** package is used for progress reporting by several
packages, notably tidyverse packages. For instance, in **purrr**, you
can do:

```r
y <- purrr::map(1:100, \(x) Sys.sleep(0.1), .progress = TRUE)
```

to report on progress via the **cli** package as `map()` is iterating
over the elements.  Now, instead of using the default, built-in
**cli** progress bar, we can customize **cli** to report on progress
via **[progressr]** instead.  To do this, set R option
`cli.progress_handlers` as:

```r
options(cli.progress_handlers = "progressr")
```

With this option set, **cli** will now report on progress according to
your `progressr::handlers()` settings.  For example, with:

```r
progressr::handlers(c("beepr", "rstudio"))
```

will report on progress using **beepr** and the RStudio Console
progress panel.

To make **cli** report via **progressr** in all your R session, set
the above R option in your <code>~/.Rprofile</code> startup file, e.g.

```r
if (requireNamespace("progressr", quietly = TRUE)) {
  options(cli.progress_handlers = "progressr")
}
```


_Note:_ A **cli** progress bar can have a "name", which can be
specified in **purrr** function via argument `.progress`,
e.g. `.progress = "processing"`.  This name is then displayed in front
of the progress bar.  However, because the **progressr** framework
does not have a concept of progress "name", this is silently ignored
when using `options(cli.progress_handlers = "progressr")`.


[progressr]: https://progressr.futureverse.org
[beepr]: https://cran.r-project.org/package=beepr
[cli]: https://cran.r-project.org/package=cli
[purrr]: https://cran.r-project.org/package=purrr

<!--
%\VignetteIndexEntry{progressr: Debug}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{debugging}
%\VignetteEngine{progressr::selfonly}
-->

Progress updates generated by **[progressr]** are formally R conditions
of class `progression` that are signaled and handled by R's condition
framework. Sometimes these `progression` conditions are not signaled,
or transmitted, as expected. To troubleshoot such problems, we can use
the built-in "debug" handler. For example,

```r
> library(progressr)
> handlers("debug")
> handlers(global = TRUE)
> 
> y <- slow_sum(1:3)
[21:17:01.786] (0.000s => +0.015s) initiate: 0/3 (+0) '' {clear=TRUE, enabled=TRUE, status=}
[21:17:01.975] (0.189s => +0.179s) update: 0/3 (+0) '' {clear=TRUE, enabled=TRUE, status=}
[21:17:02.378] (0.593s => +0.002s) update: 0/3 (+0) '' {clear=TRUE, enabled=TRUE, status=}
[21:17:02.581] (0.796s => +0.002s) update: 1/3 (+1) 'P: Adding 1' {clear=TRUE, enabled=TRUE, status=}
M: Added value 1
[21:17:02.788] (1.002s => +0.004s) update: 1/3 (+0) 'P: Adding 1' {clear=TRUE, enabled=TRUE, status=}
[21:17:02.990] (1.205s => +0.001s) update: 1/3 (+0) 'P: Adding 1' {clear=TRUE, enabled=TRUE, status=}
[21:17:03.394] (1.608s => +0.002s) update: 1/3 (+0) 'P: Adding 1' {clear=TRUE, enabled=TRUE, status=}
[21:17:03.597] (1.812s => +0.002s) update: 2/3 (+1) 'P: Adding 2' {clear=TRUE, enabled=TRUE, status=}
M: Added value 2
[21:17:03.807] (2.021s => +0.007s) update: 2/3 (+0) 'P: Adding 2' {clear=TRUE, enabled=TRUE, status=}
[21:17:04.010] (2.224s => +0.001s) update: 2/3 (+0) 'P: Adding 2' {clear=TRUE, enabled=TRUE, status=}
[21:17:04.412] (2.627s => +0.001s) update: 2/3 (+0) 'P: Adding 2' {clear=TRUE, enabled=TRUE, status=}
[21:17:04.616] (2.830s => +0.002s) update: 3/3 (+1) 'P: Adding 3' {clear=TRUE, enabled=TRUE, status=}
[21:17:04.618] (2.832s => +0.004s) update: 3/3 (+0) 'P: Adding 3' {clear=TRUE, enabled=TRUE, status=}
M: Added value 3
> 
```

[progressr]: https://progressr.futureverse.org

<!--
%\VignetteIndexEntry{progressr: Render 'knitr' Documents}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{knitr}
%\VignetteEngine{progressr::selfonly}
-->

When compiling ("knitting") a knitr-based vignette, for instance, via
`knitr::knit()`, **[knitr]** shows the progress of code chunks
processed thus far using a progress bar.  We can use **[progressr]**
also for this type of progress reporting.  To have **knitr** report on
progress via **progressr**, set R option `knitr.progress.fun` as:

```r
options(knitr.progress.fun = function(total, labels) {
  p <- progressr::progressor(total, on_exit = FALSE)
  list(
    update = function(i) p(sprintf("chunk: %s", labels[i])),
    done = function() p(type = "finish")
  )
})
```

That's it! With the above `knitr.progress.fun` option set, try to knit
a document, e.g.

```r
progressr::handlers(global = TRUE)
infile <- system.file("examples", "knitr-minimal.Rnw", package = "knitr")
outfile <- knitr::knit(infile)
```

To customize the progress bar, try for instance:

```r
progressr::handlers(global = TRUE)
options(crayon.enabled = TRUE)  ## force use of colors
progressr::handlers("pbcol")
```

This will produce a colored progress bar in the terminal.

_Comment:_ If the above knit example completes too quickly for you to
see the progress updates, you can add the "slowdown" handler to slow
down the knitting process, e.g.

```r
progressr::handlers(c("pbcol", "slowdown"))
```

This will cause the overall process to take at least ten seconds
(default).


To make **knitr** report via **progressr** in all your R session, set
the above R option in your <code>~/.Rprofile</code> startup file, e.g.

```r
if (requireNamespace("progressr", quietly = TRUE)) {
  options(knitr.progress.fun = function(total, labels) {
    p <- progressr::progressor(total, on_exit = FALSE)
    list(
      update = function(i) p(sprintf("chunk: %s", labels[i])),
      done = function() p(type = "finish")
    )
  })
}
```


[progressr]: https://progressr.futureverse.org
[knitr]: https://cran.r-project.org/package=knitr

<!--
%\VignetteIndexEntry{progressr: Map-Reduce Calls, e.g. lapply() and map()}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{lapply}
%\VignetteKeyword{foreach}
%\VignetteKeyword{purrr}
%\VignetteKeyword{plyr}
%\VignetteEngine{progressr::selfonly}
-->

Progress updates by **[progressr]** are designed to work out of the box
for any iterator framework in R, e.g. `lapply()`, **[foreach]**,
**[purrr]**, and **[plyr]**. Below you will find a set of examples that
illustrate how to use **progressr** in common use cases.


## Base R apply functions

```r
library(progressr)
handlers(global = TRUE)

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  lapply(xs, function(x) {
    Sys.sleep(0.1)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}

y <- my_fcn(1:10)
#  |====================                               |  40%
```


## The foreach package

```r
library(foreach)
library(progressr)
handlers(global = TRUE)

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  foreach(x = xs) %do% {
    Sys.sleep(0.1)
    p(sprintf("x=%g", x))
    sqrt(x)
  }
}

y <- my_fcn(1:10)
#  |====================                               |  40%
```


## The purrr package

```r
library(purrr)
library(progressr)
handlers(global = TRUE)

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  map(xs, function(x) {
    Sys.sleep(0.1)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}

y <- my_fcn(1:10)
#  |====================                               |  40%
```


## The plyr package

```r
library(plyr)
library(progressr)
handlers(global = TRUE)

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  llply(xs, function(x, ...) {
    Sys.sleep(0.1)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}

y <- my_fcn(1:10)
#  |====================                               |  40%
```

Note how this solution does not make use of **plyr**'s `.progress`
argument, because the above solution is more powerful and more
flexible, e.g. we have more control over progress updates and their
messages.  However, if you prefer the traditional **plyr** approach,
you can use `.progress = "progressr"`, e.g. `y <- llply(..., .progress
= "progressr")`.


[progressr]: https://progressr.futureverse.org
[purrr]: https://cran.r-project.org/package=purrr
[foreach]: https://cran.r-project.org/package=foreach
[plyr]: https://cran.r-project.org/package=plyr

<!--
%\VignetteIndexEntry{progressr: Customize How Progress is Reported}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{handlers}
%\VignetteEngine{progressr::selfonly}
-->

_Intended audience: end-users only_


## Terminal-based progress bars

The default is to present progress via `utils::txtProgressBar()`,
which is available on all R installations.  It presents itself as an
ASCII-based horizontal progress bar in the R terminal. This is
rendered as:

![SVG animation of the default "txtprogressbar" progress handler](imgs/handler_txtprogressbar-default.svg)

We can tweak this "txtprogressbar" handler to use red hearts for the
bar, e.g.

```r
handlers(handler_txtprogressbar(char = cli::col_red(cli::symbol$heart)))
```

which results in:

![SVG animation of the "txtprogressbar" progress handler with red hearts](imgs/handler_txtprogressbar-char-ansi.svg)

Another example is:

```r
handlers(handler_pbcol(
      adjust = 1.0,
    complete = function(s) cli::bg_red(cli::col_black(s)),
  incomplete = function(s) cli::bg_cyan(cli::col_black(s))
))
```

which results in:

![SVG animation of the "pbcol" progress handler with text aligned to the right](imgs/handler_pbcol-adjust-right-complete.svg)

To change the default, to, say, `cli_progress_bar()` by the **[cli]**
package, set:

```r
handlers("cli")
```

This progress handler will present itself as:

![SVG animation of the default "cli" progress handler](imgs/handler_cli-default.svg)

To instead use `progress_bar()` by the **[progress]** package, set:

```r
handlers("progress")
```
This progress handler will present itself as:

![SVG animation of the default "progress" progress handler](imgs/handler_progress-default.svg)


## Auditory progress updates

Progress updates do not have to be presented visually. They can
equally well be communicated via audio. For example, using:

```r
handlers("beepr")
```

will present itself as sounds played at the beginning, while progressing, and at the end (using different **[beepr]** sounds).  There will be _no_ output written to the terminal;

```r
> y <- slow_sum(1:10)
> y
[1] 55
>
```


## Concurrent auditory and visual progress updates

It is possible to have multiple progress handlers presenting progress
updates at the same time.  For example, to get both visual and
auditory updates, use:

```r
handlers("txtprogressbar", "beepr")
```


## Silence all progress

To silence all progress updates, use:

```r
handlers("void")
```


## Further configuration of progress handlers

Above we have seen examples where `handlers()` takes one or more
strings as input, e.g. `handlers(c("progress", "beepr"))`.  This is
short for a more flexible specification where we can pass a list of
handler functions, e.g.

```r
handlers(list(
  handler_progress(),
  handler_beepr()
))
```

With this construct, we can make adjustments to the default behavior
of these progress handlers.  For example, we can configure the
`format`, `width`, and `complete` arguments of
`progress::progress_bar$new()`, and tell **beepr** to use a different
`finish` sound and generate sounds at most every two seconds by
setting:

```r
handlers(list(
  handler_progress(
    format   = ":spin :current/:total (:message) [:bar] :percent in :elapsed ETA: :eta",
    width    = 60,
    complete = "+"
  ),
  handler_beepr(
    finish   = "wilhelm",
    interval = 2.0
  )
))
```


## Set a default progress handler for all your R sessions

To set the default progress handler, or handlers, in all your R
sessions, call `progressr::handlers(...)` in your
<code>~/.Rprofile</code> startup file.  For example,

```r
if (requireNamespace("progressr", quietly = TRUE)) {
  progressr::handlers(global = TRUE)
  if (requireNamespace("cli", quietly = TRUE)) {
    progressr::handlers("cli")
  }
}
```


[progressr]: https://progressr.futureverse.org
[beepr]: https://cran.r-project.org/package=beepr
[cli]: https://cran.r-project.org/package=cli
[progress]: https://cran.r-project.org/package=progress

<!--
%\VignetteIndexEntry{progressr: An Introduction}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{Rprofile}
%\VignetteKeyword{Renviron}
%\VignetteEngine{progressr::selfonly}
-->
<!-- DO NOT EDIT THIS FILE! Edit 'OVERVIEW.md' instead and then rebuild this file with 'make vignettes' -->
<img src="../man/figures/progressr-logo.png" alt="The hexlogo for the 'progressr' package" style="width: 120px; float: right; margin-right: 1ex; margin-left: 1ex;"/>

The **[progressr]** package provides a minimal API for reporting
progress updates in [R](https://www.r-project.org/).  The design is to
separate the representation of progress updates from how they are
presented.  What type of progress to signal is controlled by the
developer.  How these progress updates are rendered is controlled by
the end user.  For instance, some users may prefer visual feedback
such as a horizontal progress bar in the terminal, whereas others may
prefer auditory feedback.  The **[progressr]** framework is designed
to work out-of-the-box also for parallel and distributed processing,
especially with the **[futureverse]** ecosystem.

Design motto:

> The developer is responsible for providing progress updates but it's
> only the end user who decides if, when, and how progress should be
> presented. No exceptions will be allowed.


## Two Minimal APIs - One For Developers and One For End-Users

<div style="overflow: hidden">
<div style="float: left">
<table style="border: 1px solid #999; box-shadow: 2px 2px 2px #999;">
<tr><th>Developer's API</th></tr>
<tr style="vertical-align: top">
<td>
<p>
1. Set up a progressor with a certain number of steps:
</p>
<pre>
p &lt;- progressor(nsteps)
p &lt;- progressor(along = x)
</pre>

<p>
2. Signal progress:
</p>

<pre>
p()               # one-step progress
p(amount = 0)     # "still alive"
p("loading ...")  # pass on a message
</pre>
</td>
</tr>
</table>
</div>
<div style="float: left">&nbsp;&nbsp;&nbsp;&nbsp;</div>
<div style="float: left">
<table style="border: 1px solid #999; box-shadow: 2px 2px 2px #999;">
<tr><th>End-user's API</th></tr>
<tr style="vertical-align: top">
<td>
<p>
1a. Subscribe to progress updates from everywhere:
</p>

<pre>
handlers(global = TRUE)

y &lt;- slow_sum(1:5)
y &lt;- slow_sum(6:10)
</pre>

<p>
1b. Subscribe to a specific expression:
</p>

<pre>
with_progress({
  y &lt;- slow_sum(1:5)
  y &lt;- slow_sum(6:10)
})
</pre>

<p>
2. Configure how progress is presented:
</p>

<pre>
handlers("progress")
handlers("txtprogressbar", "beepr")
handlers(handler_pbcol(enable_after = 3.0))
handlers(handler_progress(complete = "#"))
</pre>
</td>
</table>
</div>
</div>

## A simple example

Assume that we have a function `slow_sum()` for adding up the values
in a vector.  It is so slow, that we like to provide progress updates
to whoever might be interested in it.  With the **progressr** package,
this can be done as:

```r
slow_sum <- function(x) {
  p <- progressr::progressor(along = x)
  sum <- 0
  for (kk in seq_along(x)) {
    Sys.sleep(0.1)
    sum <- sum + x[kk]
    p(message = sprintf("Adding %g", x[kk]))
  }
  sum
}
```

Note how there are _no_ arguments (e.g. `.progress = TRUE`) in the
code that specify how progress is presented. This is by design and
because the only task for the developer is to decide on where in the
code it makes sense to signal that progress has been made. As we will
see next, it should be up to the end user, and end user only, of this
code to decide whether they want to receive progress updates or not,
and, if so, in what format. Asking them to specify a special
"progress" argument adds a lot of friction, it clutters up the code,
and, importantly, might not even be possible for end users to do
(e.g. they call a package function that in turn calls the progress
reporting function of interest).

Now, if we call this function, without further settings:

```r
> y <- slow_sum(1:10)
> y
[1] 55
>
``` 

the default is that there will be _no_ progress updates. To get
progress updates, we need to request them to be "handled", which we do
by:

```r
> progressr::handlers(global = TRUE)
```

After this, progress will be reported:

```r
> y <- slow_sum(1:10)
  |====================                               |  40%
> y <- slow_sum(10:1)
  |========================================           |  80%
```

To disable reporting again, do:

```r
> handlers(global = FALSE)
```


## Customizing how progress is reported

By default, **[progressr]** presents progress via the built-in
`utils::txtProgressBar()`. It presents itself as a rudimentary
ASCII-based horizontal progress bar in the R terminal. See
`help("handler_txtprogressbar")` for how to customize the look of
"txtprogressbar", e.g. colorization and Unicode.  There are many other
ways to report on progress, including visually, auditory, and via
notification systems. You can also use a mix of these, e.g.

```r
handlers(c("cli", "beepr", "ntfy"))
```

See the 'Customizing How Progress is Reported' vignette for examples.


## Additional Features

### Support for progressr elsewhere

Note that progression updates by **progressr** are designed to work out
of the box for any iterator framework in R. See the different package
vignettes for details. Prominent examples are:

 * `lapply()` etc. in base R
 * `map()` etc. by the **[purrr]** package
 * `llply()` etc. by the **[plyr]** package
 * `foreach()` iterations by the **[foreach]** package

and near-live progress reporting in parallel and distributed
processing via the **[future]** framework. The modern approach is to
use the **[futurize]** package, e.g.

 * `lapply(...) |> futurize()`
 * `map(...) |> futurize()` (**[purrr]**)
 * `llply(...) |> futurize()` (**[plyr]**)
 * `foreach(...) |> futurize()` (**[foreach]**)
 * `bplapply(...) |> futurize()` (**[BiocParallel]**)
 
The traditional counterparts are:

 * `future_lapply()` etc.  by the **[future.apply]** package
 * `future_map()` etc. by the **[furrr]** package
 * `llply()` etc. by the **[plyr]** and **[doFuture]** packages
 * `foreach()` iterations via the **foreach** and **[doFuture]** packages
 * `bplapply()` etc. by the **[BiocParallel]** and **[doFuture]** packages

Other uses of **progressr** are:

 * make packages that report progress via the **[cli]** package
   (e.g. **purrr**) report progress via **progressr**
 * make `knit()` of the **[knitr]** package report via **progressr**


### Use regular output as usual alongside progress updates

In contrast to other progress-bar frameworks, output from `message()`,
`cat()`, `print()` and so on will _not_ interfere with progress
reported via **progressr**.  For example, say we have:

```r
slow_sqrt <- function(xs) {
  p <- progressor(along = xs)
  lapply(xs, function(x) {
    message("Calculating the square root of ", x)
    Sys.sleep(2)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}
```

we will get:

```r
> library(progressr)
> handlers(global = TRUE)
> handlers("progress")
> y <- slow_sqrt(1:8)
Calculating the square root of 1
Calculating the square root of 2
- [===========>-----------------------------------]  25% x=2
```

This works because **progressr** will briefly buffer any output
internally and only release it when the next progress update is
received just before the progress is re-rendered in the terminal.
This is why you see a two second delay when running the above example.
Note that, if we use progress handlers that do not output to the
terminal, such as `handlers("beepr")`, then output does not have to be
buffered and will appear immediately.


_Comment_: When signaling a warning using `warning(msg, immediate. =
TRUE)` the message is immediately outputted to the standard-error
stream.  However, this is not possible to emulate when warnings are
intercepted using calling handlers.  This is a limitation of R that
cannot be worked around.  Because of this, the above call will behave
the same as `warning(msg)` - that is, all warnings will be buffered by
R internally and released only when all computations are done.


### Sticky messages

As seen above, some progress handlers present the progress message as
part of its output, e.g. the "progress" handler will display the
message as part of the progress bar.  It is also possible to "push"
the message up together with other terminal output.  This can be done
by adding class attribute `"sticky"` to the progression signaled.
This works for several progress handlers that output to the terminal.
For example, with:

```r
slow_sum <- function(x) {
  p <- progressr::progressor(along = x)
  sum <- 0
  for (kk in seq_along(x)) {
    Sys.sleep(0.1)
    sum <- sum + x[kk]
    p(sprintf("Step %d", kk), class = if (kk %% 5 == 0) "sticky", amount = 0)
    p(message = sprintf("Adding %g", x[kk]))
  }
  sum
}
```

we get

```r
> handlers("txtprogressbar")
> y <- slow_sum(1:30)
Step 5
Step 10
  |====================                               |  43%
```

and

```r
> handlers("progress")
> y <- slow_sum(1:30)
Step 5
Step 10
/ [===============>-------------------------]  43% Adding 13
```



[futureverse]: https://www.futureverse.org
[progressr]: https://progressr.futureverse.org
[future]: https://future.futureverse.org
[futurize]: https://futurize.futureverse.org
[future.apply]: https://future.apply.futureverse.org
[furrr]: https://furrr.futureverse.org
[doFuture]: https://doFuture.futureverse.org
[beepr]: https://cran.r-project.org/package=beepr
[cli]: https://cran.r-project.org/package=cli
[foreach]: https://cran.r-project.org/package=foreach
[progress]: https://cran.r-project.org/package=progress
[purrr]: https://cran.r-project.org/package=purrr
[doParallel]: https://cran.r-project.org/package=doParallel
[knitr]: https://cran.r-project.org/package=knitr
[pbapply]: https://cran.r-project.org/package=pbapply
[pbmcapply]: https://cran.r-project.org/package=pbmcapply
[plyr]: https://cran.r-project.org/package=plyr
[BiocParallel]: https://www.bioconductor.org/packages/BiocParallel/

<!--
%\VignetteIndexEntry{progressr: Parallel and Distributed Processing}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progress}
%\VignetteKeyword{parallel}
%\VignetteKeyword{distributed}
%\VignetteEngine{progressr::selfonly}
-->

## TL;DR

The **progressr** package works seamlessly with parallel and
distributed processing using **[futureverse]**, and it will also
provide near-live progress updates while the parallel processing is
still running. For example,

```r
library(future)
library(progressr)
plan(multisession, workers = 2)
handlers(global = TRUE)
handlers("progress")

my_fcn <- function(xs) {
  p <- progressr::progressor(along = xs)
  future.apply::future_lapply(xs, function(x, ...) {
    Sys.sleep((10.0-x)/2)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}

y <- my_fcn(1:10)
# / [================>-----------------------------]  40% x=2
```


## Introduction

The **[futureverse]** framework, which provides a unified API for parallel
and distributed processing in R, has built-in support for the kind of
progression updates produced by the **progressr** package.  This means
that you can use it with, for instance, **[future.apply]**, **[furrr]**,
and **[foreach]** with **[doFuture]**, and **[plyr]** or
**[BiocParallel]** with **doFuture**.  In contrast, _non-future_
parallelization methods such as **parallel**'s `mclapply()` and
`parallel::parLapply()`, and **foreach** adapters like **doParallel**
do _not_ support progress reports via **progressr**.


### future_lapply() - parallel lapply()

Here is an example that uses `future_lapply()` of the **[future.apply]** package to parallelize on the local machine while at the same time signaling progression updates:

```r
library(future.apply)
plan(multisession, workers = 2)

library(progressr)
handlers(global = TRUE)

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  future_lapply(xs, function(x, ...) {
    Sys.sleep((10.0-x)/2)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}

y <- my_fcn(1:10)
# / [================>-----------------------------]  40% x=2
```


### foreach() with doFuture

Here is an example that uses `foreach()` of the **[foreach]** package
together with `%dofuture%` of the **[doFuture]** package to
parallelize while reporting on progress.  This example parallelizes on
the local machine; it works also for remote machines:

```r
library(doFuture)    ## %dofuture%
plan(multisession, workers = 2)

library(progressr)
handlers(global = TRUE)
handlers("progress")

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  foreach(x = xs) %dofuture% {
    Sys.sleep((10.0-x)/2)
    p(sprintf("x=%g", x))
    sqrt(x)
  }
}

y <- my_fcn(1:10)
# / [================>-----------------------------]  40% x=2
```


For existing code using the traditional `%dopar%` operators of the
**[foreach]** package, we can register the **[doFuture]** adapter and
use the same **progressr** as above to report progress updates;

```r
library(doFuture)
registerDoFuture()      ## %dopar% parallelizes via future
plan(multisession, workers = 2)

library(progressr)
handlers(global = TRUE)
handlers("progress")

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  foreach(x = xs) %dopar% {
    Sys.sleep((10.0-x)/2)
    p(sprintf("x=%g", x))
    sqrt(x)
  }
}

y <- my_fcn(1:10)
# / [================>-----------------------------]  40% x=2
```


### future_map() - parallel purrr::map()

Here is an example that uses `future_map()` of the **[furrr]** package
to parallelize on the local machine while at the same time signaling
progression updates:

```r
library(furrr)
plan(multisession, workers = 2)

library(progressr)
handlers(global = TRUE)
handlers("progress")

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  future_map(xs, function(x) {
    Sys.sleep((10.0-x)/2)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}

y <- my_fcn(1:10)
# / [================>-----------------------------]  40% x=2
```

_Note:_ This solution does not involve the `.progress = TRUE`
argument that **furrr** implements.  Because **progressr** is more
generic and because `.progress = TRUE` only supports certain future
backends and produces errors on non-supported backends, I recommend
to stop using `.progress = TRUE` and use the **progressr** package
instead.


### BiocParallel::bplapply() - parallel lapply()

Here is an example that uses `bplapply()` of the **[BiocParallel]**
package to parallelize on the local machine while at the same time
signaling progression updates:

```r
library(BiocParallel)
library(doFuture)
register(DoparParam())  ## BiocParallel parallelizes via %dopar%
registerDoFuture()      ## %dopar% parallelizes via future
plan(multisession, workers = 2)

library(progressr)
handlers(global = TRUE)
handlers("progress")

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  bplapply(xs, function(x) {
    Sys.sleep((10.0-x)/2)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}

y <- my_fcn(1:10)
# / [================>-----------------------------]  40% x=2
```


### plyr::llply(..., .parallel = TRUE) with doFuture

Here is an example that uses `llply()` of the **[plyr]** package to
parallelize on the local machine while at the same time signaling
progression updates:

```r
library(plyr)
library(doFuture)
registerDoFuture()      ## %dopar% parallelizes via future
plan(multisession, workers = 2)

library(progressr)
handlers(global = TRUE)
handlers("progress")

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  llply(xs, function(x, ...) {
    Sys.sleep((10.0-x)/2)
    p(sprintf("x=%g", x))
    sqrt(x)
  }, .parallel = TRUE)
}

y <- my_fcn(1:10)
# / [================>-----------------------------]  40% x=2
```

_Note:_ As an alternative to the above, recommended approach, one can
use `.progress = "progressr"` together with `.parallel = TRUE`.  This
requires **plyr** (>= 1.8.7).


### Near-live versus buffered progress updates with futures

As of August 2025, there are six types of **future** backends that are
known(*) to provide near-live progress updates:

 1. `sequential`,
 2. `multicore`,
 3. `multisession`, and
 4. `cluster` (local and remote)
 5. `future.callr::callr`
 6. `future.mirai::mirai_multisession`

Here "near-live" means that the progress handlers will report on
progress almost immediately when the progress is signaled on the
worker. This is because these parallel backends handle the special
condition class `immediateCondition` - they detect when such
conditions are signaled and relay them to the parent R process as soon
as possible. For all other future backends, the progress updates are
only relayed back to the main machine and reported together with the
results of the futures.  For instance, if `future_lapply(X, FUN)`
chunks up the processing of, say, 100 elements in `X` into eight
futures, we will see progress from each of the 100 elements as they
are done when using a future backend supporting "near-live" updates,
whereas we will only see those updates flushed eight times when
using any other types of future backends.

(*) Other future backends may gain support for "near-live" progress
updating later.  Adding support for those is independent of the
**progressr** package.  Feature requests for adding that support
should go to those future-backend packages.

[futureverse]: https://www.futureverse.org
[progressr]: https://progressr.futureverse.org
[future]: https://future.futureverse.org
[future.apply]: https://future.apply.futureverse.org
[furrr]: https://furrr.futureverse.org
[doFuture]: https://doFuture.futureverse.org
[progress]: https://cran.r-project.org/package=progress
[purrr]: https://cran.r-project.org/package=purrr
[foreach]: https://cran.r-project.org/package=foreach
[doParallel]: https://cran.r-project.org/package=doParallel
[knitr]: https://cran.r-project.org/package=knitr
[plyr]: https://cran.r-project.org/package=plyr
[BiocParallel]: https://www.bioconductor.org/packages/BiocParallel/

<!--
%\VignetteIndexEntry{progressr: Appendix}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{packages}
%\VignetteEngine{progressr::selfonly}
-->

## How-tos

### Report on progress in non-interactive mode ("batch mode")

When running R from the command line, R runs in a non-interactive mode
(`interactive()` returns `FALSE`).  The default behavior of
**progressr** is to _not_ report on progress in non-interactive mode.
To report on progress also then, set R options `progressr.enable` or
environment variable `R_PROGRESSR_ENABLE` to `TRUE`.  For example,

```sh
$ Rscript -e "library(progressr)" -e "with_progress(y <- slow_sum_p(1:10))"
```

will _not_ report on progress, whereas

```sh
$ export R_PROGRESSR_ENABLE=TRUE
$ Rscript -e "library(progressr)" -e "with_progress(y <- slow_sum_p(1:10))"
```

will.


## Notes of caution

### Avoid sending progress updates too frequently

Signaling progress updates comes with some overhead.  In situations
where we use progress updates, this overhead is typically much smaller
than the task we are processing in each step.  However, if the task we
iterate over is quick, then the extra time induced by the progress
updates might end up dominating the overall processing time.  If that
is the case, a simple solution is to only signal progress updates
every nth step.  Here is a version of `slow_sum_p()` that signals
progress every 10th iteration:

```
slow_sum_p <- function(x) {
  p <- progressr::progressor(length(x) / 10)
  sum <- 0
  for (kk in seq_along(x)) {
    Sys.sleep(0.1)
    sum <- sum + x[kk]
    if (kk %% 10 == 0) p(message = sprintf("Adding %g", x[kk]))
  }
  sum
}
```

The overhead of progress signaling may depend on context.  For
example, in parallel processing with near-live progress updates via
'multisession' futures, each progress update is communicated via a
socket connections back to the main R session.  These connections
might become clogged up if progress updates are too frequent.


## Known Limitations

### The global progress handler cannot be set during package load

It is _not_ possible to call `handlers(global = TRUE)` in all
circumstances.  For example, it cannot be called within `tryCatch()`
and `withCallingHandlers()`;

```r
> tryCatch(handlers(global = TRUE), error = identity)
Error in globalCallingHandlers(NULL) :
  should not be called with handlers on the stack
```

This is not a bug - neither in **progressr** nor in R itself. It's due
to a conservative design on how _global_ calling handlers should work
in R. If it were allowed, there's a risk we might end up getting weird and
unpredictable behaviors when messages, warnings, errors, and other
types of conditions are signaled.

Because `tryCatch()` and `withCallingHandlers()` are used in many
places throughout base R, this means that we also cannot call
`handlers(global = TRUE)` as part of a package's startup process,
e.g. `.onLoad()` or `.onAttach()`.

Another example of this error is if `handlers(global = TRUE)` is used
inside package vignettes and dynamic documents such as Rmarkdown.  In
such cases, the global progress handler has to be enabled _prior_ to
processing the document, e.g.

```r
> progressr::handlers(global = TRUE)
> rmarkdown::render("input.Rmd")
```


### A progressor cannot be created in the global environment

It is not possible to create a progressor in the global environment,
e.g. in the top-level of a script. It can only be created inside a
function, within `with_progress({ ... })`, `local({ ... })`, or a
similar construct.  For example, the following:

```r
library(progressr)
handlers(global = TRUE)

xs <- 1:5
p <- progressor(along = xs)
y <- lapply(xs, function(x) {
  Sys.sleep(0.1)
  p(sprintf("x=%g", x))
  sqrt(x)
})
```

results in an error if tried:

```
Error in progressor(along = xs) :
  A progressor must not be created in the global environment unless wrapped in a
  with_progress() or without_progress() call. Alternatively, create it inside a
  function or in a local() environment to make sure there is a finite lifespan
  of the progressor
```

The solution is to wrap it in a `local({ ... })` call, or more
explicitly, in a `with_progress({ ... })` call:

```r
library(progressr)
handlers(global = TRUE)

xs <- 1:5
with_progress({
  p <- progressor(along = xs)
  y <- lapply(xs, function(x) {
    Sys.sleep(0.1)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
})
#  |====================                               |  40%
```

The main reason for this is to limit the lifespan of each progressor.
If we created it in the global environment, there is a significant
risk it would never finish and block all of the following progressors.


### No coloring by the 'cli' and 'crayon' packages

The **[cli]** package can be used to output colored text and messages
in the terminal, e.g. `message(cli::col_blue("hey"))`. However,
**cli** disables such coloring by default when there is an active
progressor. The workaround is to set **cli** option `cli.num_colors` to:

```r
options(cli.num_colors = cli::num_ansi_colors())
```

This will force **cli** to use the same number of colors with and
without an active progressor.

The same happens when using the superseded **[crayon]** package for
colorization. To re-enable coloring for **crayon**, set the same
(sic!)  above R options.

The technical reason for **cli** and **crayon** disabling coloring, is
that the **progressr** package buffers standard output using a sink,
which these packages consider to be mono-color only.


## Known Issues

### RStudio bug #16331: Setting global progressr handlers during startup does not work

Setting the global progressr handler in `~/.Rprofile` does _not_ work
in RStudio 2025.09:

```r
if (requireNamespace("progressr", quietly = TRUE)) {
  progressr::handlers(global = TRUE)
}
```

This is due to a [bug introduced in RStudio
2025.09](https://github.com/rstudio/rstudio/issues/16331), which has
been fixed for the next release RStudio 2025.11. If you are using
RStudio 2025.09, the workaround is to instead use:

```r
if (requireNamespace("progressr", quietly = TRUE)) {
  progressr::handlers(global = TRUE)

  ## Workaround for RStudio 2025.09 console bug #16331
  if (nzchar(Sys.getenv("RSTUDIO")) && !nzchar(Sys.getenv("RSTUDIO_TERM"))) {
    invisible(addTaskCallback(function(...) {
      ver <- RStudio.Version()$version
      if (ver >= "2025.09" && ver < "2025.11") {
        message("Workaround for RStudio 2025.09 bug #16331: Added progressr global handler")
        progressr::handlers(global = TRUE)
      } else {
        warning("Workaround for RStudio 2025.09 bug #16331: Not needed in RStudio v", ver, ". Please remove task callback 'rstudio::progressr::once' in your Rprofile startup file", call. = FALSE, immediate. = TRUE)
      }
      removeTaskCallback("rstudio::progressr::once")
    }, name = "rstudio::progressr::once"))
  }
}
```


### Positron bug #6892: Setting global progressr handlers during startup does not work

Positron does not support setting global calling handlers during R's
startup process, e.g. in `~/.Rprofile`. Even if such handlers are
registered, they have no effect. This is a [bug in
Positron](https://github.com/posit-dev/positron/issues/6892)
(2025-03-20), which was most recently confirmed with Position
2025.09.0 on Linux. Because of this, having something like in your
`~/.Rprofile`:

```r
if (requireNamespace("progressr", quietly = TRUE)) {
  progressr::handlers(global = TRUE)
}
```

will have no effect. If used, the workaround is to manually
re-registering all calling handlers, which can be done as:

```r
if (requireNamespace("progressr", quietly = TRUE)) {
  progressr::handlers(global = TRUE)

  ## Workaround for Positron (>= 2025.09) bug #16331
  if (nzchar(Sys.getenv("POSITRON"))) local({
    ver <- numeric_version(Sys.getenv("POSITRON_VERSION"))
    if (ver >= "2025.09") {
      message("Workaround for Positron (>= 2025.09) bug #6892: progressr global handler will be installed *after* the next call has been completed")
      invisible(addTaskCallback(function(...) {
        message("Workaround for Positron (>= 2025.09) bug #6892: re-installed progressr global handler")
        globalCallingHandlers(globalCallingHandlers(NULL))
        removeTaskCallback("positron::progressr::once")
      }, name = "positron::progressr::once"))
    }
  })
}
```

Alternatively, call:

```r
globalCallingHandlers(globalCallingHandlers(NULL))
```

or

```r
progressr::handlers(global = FALSE)  ## important
progressr::handlers(global = TRUE)
```

at the prompt.


### Positron: Messages add extra newline before the final progress step

One of the features of **progressr** is that messages are buffered
during progress reporting and relayed as soon as possible, which
typically happens just before handlers re-render the progress
output. This way you can use `message()` as usual, regardless whether
progress is reported or not.

Currently, when using Positron (e.g. Positron 2025.09.0), any
`message()` output adds an extra newline before the _final_ progress
step is reported, e.g.

```r
> progressr::handlers("cli", global = TRUE)
> y <- progressr::slow_sum_p(1:5, message = TRUE)
M: Added value 1
M: Added value 2
M: Added value 3
M: Added value 4

M: Added value 5
> 
```

I do not fully understand the reason for this, but I hope we can get
to the bottom of it and fix it, either in **progressr** or in
Positron, e.g. [Bug #9486](https://github.com/posit-dev/positron/issues/9486)
(2025-09-18).


### Jupyter: Reporting progress to stderr (default) does not work

The default for most terminal progress renders, including the ones for
**progressr**, display the progress on standard error (stderr). Due to
a limitation in Jupyter, this default does not work there. The reason is
that [Jupyter silently
drops](https://github.com/futureverse/progressr/issues/170) any output
sent to stderr, e.g.

```r
> cat("hello stderr\n", file = stderr())
> cat("hello stdout\n", file = stdout())
hello stdout
>
```

If we try the following

```r
library(progressr)
handlers(globals = TRUE)
handlers("txtprogressbar")
y <- slow_sum_p(1:20)
```

there will be no progress being reported. This is not specific to
**progressr**, we have the same problem with for instance **cli**. Try
for instance,

```r
void <- cli::cli_progress_demo(delay = 1.0)
```

The workaround is to direct all progress output to the standard output
(stdout) when working in Jupyter. For this to work, we also need to
disable the buffering ("delaying") of any other output to stdout.

```r
library(progressr)
handlers(globals = TRUE)

## Workaround for Jupyter
options(progressr.enable = TRUE, progressr.delay_stdout = FALSE)

## Jupyter requires that progress is rendered to standard output;
## it does not work with the default standard error
handlers(handler_txtprogressbar(file = stdout()))

y <- slow_sum_p(1:20)
```


### Jupyter bug #732 - handlers("progress") output is messy

Jupyter has other outputting issues. Specifically, Jupyter [injects an
_extra_ newline at the end of every
message](https://github.com/IRkernel/IRkernel/issues/732), e.g.

```r
> message("abc", appendLF = FALSE); message("def", appendLF = FALSE)
abc
def
> message("abc"); message("def")
abc

def

>
```

This causes any progress framework (e.g. the **progress** package)
that reports via messages to render progress output very poorly or not
at all.


## Design and Implementation

### Under the hood

When using the **progressr** package, progression updates are
communicated via R's condition framework, which provides methods for
creating, signaling, capturing, muffling, and relaying conditions.
Progression updates are of classes `progression` and
`immediateCondition`(\*).  The below figure gives an example how
progression conditions are created, signaled, and rendered.

(\*) The `immediateCondition` class of conditions are relayed as soon
as possible by the **[future]** framework, which means that
progression updates produced in parallel workers are reported to the
end user as soon as the main R session have received them.

![](imgs/slow_sum_p.svg)

_Figure: Sequence diagram illustrating how signaled progression
conditions are captured by `with_progress()`, or the global
progression handler, and relayed to the two progression handlers
'progress' (a progress bar in the terminal) and 'beepr' (auditory)
that the end user has chosen._


### Roadmap

Because this project is under active development, the progressr API is
currently kept at a very minimum.  This will allow for the framework
and the API to evolve while minimizing the risk for breaking code that
depends on it.  The roadmap for developing the API is roughly:

* [x] Provide minimal API for producing progress updates,
      i.e. `progressor()`, `with_progress()`, `handlers()`

* [x] Add support for global progress handlers removing the need for
      the user having to specify `with_progress()`, 
      i.e. `handlers(global = TRUE)` and `handlers(global = FALSE)`

* [ ] Make it possible to create a progressor also in the global
      environment (see 'Known Issues' above)

* [ ] Add support for nested progress updates

* [ ] Add API to allow users and package developers to design
      additional progression handlers

For a more up-to-date view on what features might be added, see
<https://github.com/futureverse/progressr/issues>.


[progressr]: https://progressr.futureverse.org
[future]: https://future.futureverse.org
[cli]: https://cran.r-project.org/package=cli
[crayon]: https://cran.r-project.org/package=crayon

<!--
%\VignetteIndexEntry{progressr: An Introduction}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{Rprofile}
%\VignetteKeyword{Renviron}
%\VignetteEngine{progressr::selfonly}
-->
<!-- DO NOT EDIT THIS FILE! Edit 'OVERVIEW.md' instead and then rebuild this file with 'make vigs' -->
The **[progressr]** package provides a minimal API for reporting
progress updates in [R](https://www.r-project.org/).  The design is to
separate the representation of progress updates from how they are
presented.  What type of progress to signal is controlled by the
developer.  How these progress updates are rendered is controlled by
the end user.  For instance, some users may prefer visual feedback
such as a horizontal progress bar in the terminal, whereas others may
prefer auditory feedback.


<img src="imgs/three_in_chinese.gif" alt="Three strokes writing three in Chinese" style="float: right; margin-right: 1ex; margin-left: 1ex;"/>

Design motto:

> The developer is responsible for providing progress updates but it's
> only the end user who decides if, when, and how progress should be
> presented. No exceptions will be allowed.


## Two Minimal APIs - One For Developers and One For End-Users

<div style="overflow: hidden">
<div style="float: left">
<table style="border: 1px solid #999; box-shadow: 2px 2px 2px #999;">
<tr><th>Developer's API</th></tr>
<tr style="vertical-align: top">
<td>
<p>
1. Set up a progressor with a certain number of steps:
</p>
<pre>
p &lt;- progressor(nsteps)
p &lt;- progressor(along = x)
</pre>

<p>
2. Signal progress:
</p>

<pre>
p()               # one-step progress
p(amount = 0)     # "still alive"
p("loading ...")  # pass on a message
</pre>
</td>
</tr>
</table>
</div>
<div style="float: left">&nbsp;&nbsp;&nbsp;&nbsp;</div>
<div style="float: left">
<table style="border: 1px solid #999; box-shadow: 2px 2px 2px #999;">
<tr><th>End-user's API</th></tr>
<tr style="vertical-align: top">
<td>
<p>
1a. Subscribe to progress updates from everywhere:
</p>

<pre>
handlers(global = TRUE)

y &lt;- slow_sum(1:5)
y &lt;- slow_sum(6:10)
</pre>

<p>
1b. Subscribe to a specific expression:
</p>

<pre>
with_progress({
  y &lt;- slow_sum(1:5)
  y &lt;- slow_sum(6:10)
})
</pre>

<p>
2. Configure how progress is presented:
</p>

<pre>
handlers("progress")
handlers("txtprogressbar", "beepr")
handlers(handler_pbcol(enable_after = 3.0))
handlers(handler_progress(complete = "#"))
</pre>
</td>
</table>
</div>
</div>

## A simple example

Assume that we have a function `slow_sum()` for adding up the values
in a vector.  It is so slow, that we like to provide progress updates
to whoever might be interested in it.  With the **progressr** package,
this can be done as:

```r
slow_sum <- function(x) {
  p <- progressr::progressor(along = x)
  sum <- 0
  for (kk in seq_along(x)) {
    Sys.sleep(0.1)
    sum <- sum + x[kk]
    p(message = sprintf("Adding %g", x[kk]))
  }
  sum
}
```

Note how there are _no_ arguments in the code that specifies how
progress is presented.  The only task for the developer is to decide
on where in the code it makes sense to signal that progress has been
made.  As we will see next, it is up to the end user of this code to
decide whether they want to receive progress updates or not, and, if
so, in what format.


### Without reporting on progress

When calling this function as in:
```r
> y <- slow_sum(1:10)
> y
[1] 55
>
``` 

it will behave as any function and there will be no progress
updates displayed.


### Reporting on progress

If we are only interested in progress for a particular call, we can
do:

```r
> library(progressr)
> with_progress(y <- slow_sum(1:10))
  |====================                               |  40%
```

However, if we want to report on progress from _every_ call, wrapping
the calls in `with_progress()` might become too cumbersome.  If so, we
can enable the global progress handler:

```r
> library(progressr)
> handlers(global = TRUE)
```

so that progress updates are reported on wherever signaled, e.g.

```r
> y <- slow_sum(1:10)
  |====================                               |  40%
> y <- slow_sum(10:1)
  |========================================           |  80%
```

This requires R 4.0.0 or newer.  To disable this again, do:

```r
> handlers(global = FALSE)
```

In the below examples, we will assume `handlers(global = TRUE)` is
already set.


## Customizing how progress is reported

### Terminal-based progress bars

The default is to present progress via `utils::txtProgressBar()`,
which is available on all R installations.  It presents itself as an
ASCII-based horizontal progress bar in the R terminal. This is
rendered as:

![SVG animation of the default "txtprogressbar" progress handler](imgs/handler_txtprogressbar-default.svg)

We can tweak this "txtprogressbar" handler to use red hearts for the
bar, e.g.

```r
handlers(handler_txtprogressbar(char = cli::col_red(cli::symbol$heart)))
```

which results in:

![SVG animation of the "txtprogressbar" progress handler with red hearts](imgs/handler_txtprogressbar-char-ansi.svg)

Another example is:

```r
handlers(handler_pbcol(
      adjust = 1.0,
    complete = function(s) cli::bg_red(cli::col_black(s)),
  incomplete = function(s) cli::bg_cyan(cli::col_black(s))
))
```

which results in:

![SVG animation of the "pbcol" progress handler with text aligned to the right](imgs/handler_pbcol-adjust-right-complete.svg)

To change the default, to, say, `cli_progress_bar()` by the **[cli]**
package, set:

```r
handlers("cli")
```

This progress handler will present itself as:

![SVG animation of the default "cli" progress handler](imgs/handler_cli-default.svg)

To instead use `progress_bar()` by the **[progress]** package, set:

```r
handlers("progress")
```
This progress handler will present itself as:

![SVG animation of the default "progress" progress handler](imgs/handler_progress-default.svg)

To set the default progress handler, or handlers, in all your R
sessions, call `progressr::handlers(...)` in your
<code>~/.Rprofile</code> startup file.



### Auditory progress updates

Progress updates do not have to be presented visually. They can
equally well be communicated via audio. For example, using:

```r
handlers("beepr")
```

will present itself as sounds played at the beginning, while progressing, and at the end (using different **[beepr]** sounds).  There will be _no_ output written to the terminal;

```r
> y <- slow_sum(1:10)
> y
[1] 55
>
```


### Concurrent auditory and visual progress updates

It is possible to have multiple progress handlers presenting progress
updates at the same time.  For example, to get both visual and
auditory updates, use:

```r
handlers("txtprogressbar", "beepr")
```


### Silence all progress

To silence all progress updates, use:

```r
handlers("void")
```


### Further configuration of progress handlers

Above we have seen examples where the `handlers()` takes one or more
strings as input, e.g. `handlers(c("progress", "beepr"))`.  This is
short for a more flexible specification where we can pass a list of
handler functions, e.g.

```r
handlers(list(
  handler_progress(),
  handler_beepr()
))
```

With this construct, we can make adjustments to the default behavior
of these progress handlers.  For example, we can configure the
`format`, `width`, and `complete` arguments of
`progress::progress_bar$new()`, and tell **beepr** to use a different
`finish` sound and generate sounds at most every two seconds by
setting:

```r
handlers(list(
  handler_progress(
    format   = ":spin :current/:total (:message) [:bar] :percent in :elapsed ETA: :eta",
    width    = 60,
    complete = "+"
  ),
  handler_beepr(
    finish   = "wilhelm",
    interval = 2.0
  )
))
```


## Sticky messages

As seen above, some progress handlers present the progress message as
part of its output, e.g. the "progress" handler will display the
message as part of the progress bar.  It is also possible to "push"
the message up together with other terminal output.  This can be done
by adding class attribute `"sticky"` to the progression signaled.
This works for several progress handlers that output to the terminal.
For example, with:

```r
slow_sum <- function(x) {
  p <- progressr::progressor(along = x)
  sum <- 0
  for (kk in seq_along(x)) {
    Sys.sleep(0.1)
    sum <- sum + x[kk]
    p(sprintf("Step %d", kk), class = if (kk %% 5 == 0) "sticky", amount = 0)
    p(message = sprintf("Adding %g", x[kk]))
  }
  sum
}
```

we get

```r
> handlers("txtprogressbar")
> y <- slow_sum(1:30)
Step 5
Step 10
  |====================                               |  43%
```

and

```r
> handlers("progress")
> y <- slow_sum(1:30)
Step 5
Step 10
/ [===============>-------------------------]  43% Adding 13
```


## Use regular output as usual alongside progress updates

In contrast to other progress-bar frameworks, output from `message()`,
`cat()`, `print()` and so on, will _not_ interfere with progress
reported via **progressr**.  For example, say we have:

```r
slow_sqrt <- function(xs) {
  p <- progressor(along = xs)
  lapply(xs, function(x) {
    message("Calculating the square root of ", x)
    Sys.sleep(2)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}
```

we will get:

```r
> library(progressr)
> handlers(global = TRUE)
> handlers("progress")
> y <- slow_sqrt(1:8)
Calculating the square root of 1
Calculating the square root of 2
- [===========>-----------------------------------]  25% x=2
```

This works because **progressr** will briefly buffer any output
internally and only release it when the next progress update is
received just before the progress is re-rendered in the terminal.
This is why you see a two second delay when running the above example.
Note that, if we use progress handlers that do not output to the
terminal, such as `handlers("beepr")`, then output does not have to be
buffered and will appear immediately.


_Comment_: When signaling a warning using `warning(msg, immediate. =
TRUE)` the message is immediately outputted to the standard-error
stream.  However, this is not possible to emulate when warnings are
intercepted using calling handlers, which are used by
`with_progress()`.  This is a limitation of R that cannot be worked
around.  Because of this, the above call will behave the same as
`warning(msg)` - that is, all warnings will be buffered by R
internally and released only when all computations are done.


## Support for progressr elsewhere

Note that progression updates by **progressr** is designed to work out
of the box for any iterator framework in R.  Below is an set of
examples for the most common ones.


### Base R Apply Functions

```r
library(progressr)
handlers(global = TRUE)

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  lapply(xs, function(x) {
    Sys.sleep(0.1)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}

my_fcn(1:5)
#  |====================                               |  40%
```


### The foreach package

```r
library(foreach)
library(progressr)
handlers(global = TRUE)

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  foreach(x = xs) %do% {
    Sys.sleep(0.1)
    p(sprintf("x=%g", x))
    sqrt(x)
  }
}

my_fcn(1:5)
#  |====================                               |  40%
```

### The purrr package

```r
library(purrr)
library(progressr)
handlers(global = TRUE)

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  map(xs, function(x) {
    Sys.sleep(0.1)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}

my_fcn(1:5)
#  |====================                               |  40%
```


### The plyr package

```r
library(plyr)
library(progressr)
handlers(global = TRUE)

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  llply(xs, function(x, ...) {
    Sys.sleep(0.1)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}

my_fcn(1:5)
#  |====================                               |  40%
```

Note how this solution does not make use of **plyr**'s `.progress`
argument, because the above solution is more powerful and more
flexible, e.g. we have more control on progress updates and their
messages.  However, if you prefer the traditional **plyr** approach,
you can use `.progress = "progressr"`, e.g. `y <- llply(..., .progress
= "progressr")`.




### The knitr package

When compiling ("knitting") an knitr-based vignette, for instance, via
`knitr::knit()`, **[knitr]** shows the progress of code chunks
processed thus far using a progress bar.  In **knitr** (>= 1.42) [to
be released as of 2022-12-12], we can use **progressr** for this
progress reporting.  To do this, set R option `knitr.progress.fun` as:

```r
options(knitr.progress.fun = function(total, labels) {
  p <- progressr::progressor(total, on_exit = FALSE)
  list(
    update = function(i) p(sprintf("chunk: %s", labels[i])),
    done = function() p(type = "finish")
  )
})
```

This configures **knitr** to signal progress via the **progressr**
framework.  To report on these, use:

```r
progressr::handlers(global = TRUE)
```


### Replace any cli progress bars with progressr updates

The **cli** package is used for progress reporting by many several
packages, notably tidyverse packages.  For instance, in **purrr**, you
can do:

```r
y <- purrr::map(1:100, \(x) Sys.sleep(0.1), .progress = TRUE)
```

to report on progress via the **cli** package as `map()` is iterating
over the elements.  Now, instead of using the default, built-in
**cli** progress bar, we can customize **cli** to report on progress
via **progressr** instead.  To do this, set R option
`cli.progress_handlers` as:

```r
options(cli.progress_handlers = "progressr")
```

With this option set, **cli** will now report on progress according to
your `progressr::handlers()` settings.  For example, with:

```r
progressr::handlers(c("beepr", "rstudio"))
```

will report on progress using **beepr** and the RStudio Console
progress panel.

To make **cli** report via **progressr** in all your R session, set
the above R option in your <code>~/.Rprofile</code> startup file.

_Note:_ A **cli** progress bar can have a "name", which can be
specfied in **purrr** function via argument `.progress`,
e.g. `.progress = "processing"`.  This name is then displayed in front
of the progress bar.  However, because the **progressr** framework
does not have a concept of progress "name", they are silently ignored
when using `options(cli.progress_handlers = "progressr")`.



## Parallel processing and progress updates

The **[future]** framework, which provides a unified API for parallel
and distributed processing in R, has built-in support for the kind of
progression updates produced by the **progressr** package.  This means
that you can use it with for instance **[future.apply]**, **[furrr]**,
and **[foreach]** with **[doFuture]**, and **[plyr]** or
**[BiocParallel]** with **doFuture**.  In contrast, _non-future_
parallelization methods such as **parallel**'s `mclapply()` and,
`parallel::parLapply()`, and **foreach** adapters like **doParallel**
do _not_ support progress reports via **progressr**.


### future_lapply() - parallel lapply()

Here is an example that uses `future_lapply()` of the **[future.apply]** package to parallelize on the local machine while at the same time signaling progression updates:

```r
library(future.apply)
plan(multisession)

library(progressr)
handlers(global = TRUE)
handlers("progress", "beepr")

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  future_lapply(xs, function(x, ...) {
    Sys.sleep(6.0-x)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}

my_fcn(1:5)
# / [================>-----------------------------]  40% x=2
```


### foreach() with doFuture

Here is an example that uses `foreach()` of the **[foreach]** package
together with `%dofuture%` of the **[doFuture]** package to
parallelize while reporting on progress.  This example parallelizes on
the local machine, it works alsof for remote machines:

```r
library(doFuture)    ## %dofuture%
plan(multisession)

library(progressr)
handlers(global = TRUE)
handlers("progress", "beepr")

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  foreach(x = xs) %dofuture% {
    Sys.sleep(6.0-x)
    p(sprintf("x=%g", x))
    sqrt(x)
  }
}

my_fcn(1:5)
# / [================>-----------------------------]  40% x=2
```


For existing code using the traditional `%dopar%` operators of the
**[foreach]** package, we can register the **[doFuture]** adaptor and
use the same **progressr** as above to progress updates;

```r
library(doFuture)
registerDoFuture()      ## %dopar% parallelizes via future
plan(multisession)

library(progressr)
handlers(global = TRUE)
handlers("progress", "beepr")

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  foreach(x = xs) %dopar% {
    Sys.sleep(6.0-x)
    p(sprintf("x=%g", x))
    sqrt(x)
  }
}

my_fcn(1:5)
# / [================>-----------------------------]  40% x=2
```


### future_map() - parallel purrr::map()

Here is an example that uses `future_map()` of the **[furrr]** package
to parallelize on the local machine while at the same time signaling
progression updates:

```r
library(furrr)
plan(multisession)

library(progressr)
handlers(global = TRUE)
handlers("progress", "beepr")

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  future_map(xs, function(x) {
    Sys.sleep(6.0-x)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}

my_fcn(1:5)
# / [================>-----------------------------]  40% x=2
```

_Note:_ This solution does not involved the `.progress = TRUE`
argument that **furrr** implements.  Because **progressr** is more
generic and because `.progress = TRUE` only supports certain future
backends and produces errors on non-supported backends, I recommended
to stop using `.progress = TRUE` and use the **progressr** package
instead.


### BiocParallel::bplapply() - parallel lapply()

Here is an example that uses `bplapply()` of the **[BiocParallel]**
package to parallelize on the local machine while at the same time
signaling progression updates:

```r
library(BiocParallel)
library(doFuture)
register(DoparParam())  ## BiocParallel parallelizes via %dopar%
registerDoFuture()      ## %dopar% parallelizes via future
plan(multisession)

library(progressr)
handlers(global = TRUE)
handlers("progress", "beepr")

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  bplapply(xs, function(x) {
    Sys.sleep(6.0-x)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}

my_fcn(1:5)
# / [================>-----------------------------]  40% x=2
```


### plyr::llply(..., .parallel = TRUE) with doFuture

Here is an example that uses `llply()` of the **[plyr]** package to
parallelize on the local machine while at the same time signaling
progression updates:

```r
library(plyr)
library(doFuture)
registerDoFuture()      ## %dopar% parallelizes via future
plan(multisession)

library(progressr)
handlers(global = TRUE)
handlers("progress", "beepr")

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  llply(xs, function(x, ...) {
    Sys.sleep(6.0-x)
    p(sprintf("x=%g", x))
    sqrt(x)
  }, .parallel = TRUE)
}

my_fcn(1:5)
# / [================>-----------------------------]  40% x=2
```

_Note:_ As an alternative to the above, recommended approach, one can
use `.progress = "progressr"` together with `.parallel = TRUE`.  This
requires **plyr** (>= 1.8.7).


### Near-live versus buffered progress updates with futures

As of November 2020, there are four types of **future** backends that are known(*) to provide near-live progress updates:

 1. `sequential`,
 2. `multicore`,
 3. `multisession`, and
 4. `cluster` (local and remote)

Here "near-live" means that the progress handlers will report on
progress almost immediately when the progress is signaled on the
worker.  For all other future backends, the progress updates are only
relayed back to the main machine and reported together with the
results of the futures.  For instance, if `future_lapply(X, FUN)`
chunks up the processing of, say, 100 elements in `X` into eight
futures, we will see progress from each of the 100 elements as they
are done when using a future backend supporting "near-live" updates,
whereas we will only see those updated to be flushed eight times when
using any other types of future backends.

(*) Other future backends may gain support for "near-live" progress
updating later.  Adding support for those is independent of the
**progressr** package.  Feature requests for adding that support
should go to those future-backend packages.



## Note of caution - sending progress updates too frequently

Signaling progress updates comes with some overhead.  In situation
where we use progress updates, this overhead is typically much smaller
than the task we are processing in each step.  However, if the task we
iterate over is quick, then the extra time induced by the progress
updates might end up dominating the overall processing time.  If that
is the case, a simple solution is to only signal progress updates
every n:th step.  Here is a version of `slow_sum()` that signals
progress every 10:th iteration:

```
slow_sum <- function(x) {
  p <- progressr::progressor(length(x) / 10)
  sum <- 0
  for (kk in seq_along(x)) {
    Sys.sleep(0.1)
    sum <- sum + x[kk]
    if (kk %% 10 == 0) p(message = sprintf("Adding %g", x[kk]))
  }
  sum
}
```

The overhead of progress signaling may depend on context.  For
example, in parallel processing with near-live progress updates via
'multisession' futures, each progress update is communicated via a
socket connections back to the main R session.  These connections
might become clogged up if progress updates are too frequent.


## Progress updates in non-interactive mode ("batch mode")

When running R from the command line, R runs in a non-interactive mode
(`interactive()` returns `FALSE`).  The default behavior of
**progressr** is to _not_ report on progress in non-interactive mode.
To reported on progress also then, set R options `progressr.enable` or
environment variable `R_PROGRESSR_ENABLE` to `TRUE`.  For example,

```sh
$ Rscript -e "library(progressr)" -e "with_progress(y <- slow_sum(1:10))"
```

will _not_ report on progress, whereas

```sh
$ export R_PROGRESSR_ENABLE=TRUE
$ Rscript -e "library(progressr)" -e "with_progress(y <- slow_sum(1:10))"
```

will.



## Roadmap

Because this project is under active development, the progressr API is
currently kept at a very minimum.  This will allow for the framework
and the API to evolve while minimizing the risk for breaking code that
depends on it.  The roadmap for developing the API is roughly:

* [x] Provide minimal API for producing progress updates,
      i.e. `progressor()`, `with_progress()`, `handlers()`

* [x] Add support for global progress handlers removing the need for
      the user having to specify `with_progress()`,
      i.e. `handlers(global = TRUE)` and `handlers(global = FALSE)`

* [ ] Make it possible to create a progressor also in the global
      environment (see 'Known issues' below)

* [ ] Add support for nested progress updates

* [ ] Add API to allow users and package developers to design
      additional progression handlers

For a more up-to-date view on what features might be added, see
<https://github.com/futureverse/progressr/issues>.


## Appendix

### Known issues

#### A progressor cannot be created in the global environment

It is not possible to create a progressor in the global environment,
e.g. in the the top-level of a script.  It has to be created inside a
function, within `with_progress({ ... })`, `local({ ... })`, or a
similar construct.  For example, the following:

```r
library(progressr)
handlers(global = TRUE)

xs <- 1:5
p <- progressor(along = xs)
y <- lapply(xs, function(x) {
  Sys.sleep(0.1)
  p(sprintf("x=%g", x))
  sqrt(x)
})
```

results in an error if tried:

```
Error in progressor(along = xs) : 
  A progressor must not be created in the global environment unless wrapped in a
  with_progress() or without_progress() call. Alternatively, create it inside a
  function or in a local() environment to make sure there is a finite life span
  of the progressor
```

The solution is to wrap it in a `local({ ... })` call, or more
explicitly, in a `with_progress({ ... })` call:

```r
library(progressr)
handlers(global = TRUE)

xs <- 1:5
with_progress({
  p <- progressor(along = xs)
  y <- lapply(xs, function(x) {
    Sys.sleep(0.1)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
})
#  |====================                               |  40%
```

The main reason for this is to limit the life span of each progressor.
If we created it in the global environment, there is a significant
risk it would never finish and block all of the following progressors.


#### The global progress handler cannot be set everywhere

It is _not_ possible to call `handlers(global = TRUE)` in all
circumstances.  For example, it cannot be called within `tryCatch()`
and `withCallingHandlers()`;

```r
> tryCatch(handlers(global = TRUE), error = identity)
Error in globalCallingHandlers(NULL) : 
  should not be called with handlers on the stack
```

This is not a bug - neither in **progressr** nor in R itself. It's due
to a conservative design on how _global_ calling handlers should work
in R. If it allowed, there's a risk we might end up getting weird and
unpredictable behaviors when messages, warnings, errors, and other
types of conditions are signaled.

Because `tryCatch()` and `withCallingHandlers()` is used in many
places throughout base R, this means that we also cannot call
`handlers(global = TRUE)` as part of a package's startup process,
e.g. `.onLoad()` or `.onAttach()`.

Another example of this error is if `handlers(global = TRUE)` is used
inside package vignettes and dynamic documents such as Rmarkdown.  In
such cases, the global progress handler has to be enabled _prior_ to
processing the document, e.g.

```r
> progressr::handlers(global = TRUE)
> rmarkdown::render("input.Rmd")
```


### Under the hood

When using the **progressr** package, progression updates are
communicated via R's condition framework, which provides methods for
creating, signaling, capturing, muffling, and relaying conditions.
Progression updates are of classes `progression` and
`immediateCondition`(\*).  The below figure gives an example how
progression conditions are created, signaled, and rendered.

(\*) The `immediateCondition` class of conditions are relayed as soon
as possible by the **[future]** framework, which means that
progression updates produced in parallel workers are reported to the
end user as soon as the main R session have received them.




![](imgs/slow_sum.svg)

_Figure: Sequence diagram illustrating how signaled progression
conditions are captured by `with_progress()`, or the global
progression handler, and relayed to the two progression handlers
'progress' (a progress bar in the terminal) and 'beepr' (auditory)
that the end user has chosen._


### Debugging

To debug progress updates, use:

```r
> handlers("debug")
> with_progress(y <- slow_sum(1:3))
[23:19:52.738] (0.000s => +0.002s) initiate: 0/3 (+0) '' {clear=TRUE, enabled=TRUE, status=}
[23:19:52.739] (0.001s => +0.000s) update: 0/3 (+0) '' {clear=TRUE, enabled=TRUE, status=}
[23:19:52.942] (0.203s => +0.002s) update: 0/3 (+0) '' {clear=TRUE, enabled=TRUE, status=}
[23:19:53.145] (0.407s => +0.001s) update: 0/3 (+0) '' {clear=TRUE, enabled=TRUE, status=}
[23:19:53.348] (0.610s => +0.002s) update: 1/3 (+1) 'P: Adding 1' {clear=TRUE, enabled=TRUE, status=}
M: Adding value 1
[23:19:53.555] (0.817s => +0.004s) update: 1/3 (+0) 'P: Adding 1' {clear=TRUE, enabled=TRUE, status=}
[23:19:53.758] (1.020s => +0.001s) update: 1/3 (+0) 'P: Adding 1' {clear=TRUE, enabled=TRUE, status=}
[23:19:53.961] (1.223s => +0.001s) update: 1/3 (+0) 'P: Adding 1' {clear=TRUE, enabled=TRUE, status=}
[23:19:54.165] (1.426s => +0.001s) update: 1/3 (+0) 'P: Adding 1' {clear=TRUE, enabled=TRUE, status=}
[23:19:54.368] (1.630s => +0.001s) update: 2/3 (+1) 'P: Adding 2' {clear=TRUE, enabled=TRUE, status=}
M: Adding value 2
[23:19:54.574] (1.835s => +0.003s) update: 2/3 (+0) 'P: Adding 2' {clear=TRUE, enabled=TRUE, status=}
[23:19:54.777] (2.039s => +0.001s) update: 2/3 (+0) 'P: Adding 2' {clear=TRUE, enabled=TRUE, status=}
[23:19:54.980] (2.242s => +0.001s) update: 2/3 (+0) 'P: Adding 2' {clear=TRUE, enabled=TRUE, status=}
[23:19:55.183] (2.445s => +0.001s) update: 2/3 (+0) 'P: Adding 2' {clear=TRUE, enabled=TRUE, status=}
[23:19:55.387] (2.649s => +0.001s) update: 3/3 (+1) 'P: Adding 3' {clear=TRUE, enabled=TRUE, status=}
[23:19:55.388] (2.650s => +0.003s) update: 3/3 (+0) 'P: Adding 3' {clear=TRUE, enabled=TRUE, status=}
M: Adding value 3
[23:19:55.795] (3.057s => +0.000s) shutdown: 3/3 (+0) 'P: Adding 3' {clear=TRUE, enabled=TRUE, status=ok}
```



[progressr]: https://cran.r-project.org/package=progressr
[beepr]: https://cran.r-project.org/package=beepr
[cli]: https://cran.r-project.org/package=cli
[progress]: https://cran.r-project.org/package=progress
[purrr]: https://cran.r-project.org/package=purrr
[future]: https://cran.r-project.org/package=future
[foreach]: https://cran.r-project.org/package=foreach
[future.apply]: https://cran.r-project.org/package=future.apply
[doParallel]: https://cran.r-project.org/package=doParallel
[doFuture]: https://cran.r-project.org/package=doFuture
[furrr]: https://cran.r-project.org/package=furrr
[knitr]: https://cran.r-project.org/package=knitr
[pbapply]: https://cran.r-project.org/package=pbapply
[pbmcapply]: https://cran.r-project.org/package=pbmcapply
[plyr]: https://cran.r-project.org/package=plyr
[BiocParallel]: https://www.bioconductor.org/packages/BiocParallel/

<!--
%\VignetteIndexEntry{progressr: Debug}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{debugging}
%\VignetteEngine{progressr::selfonly}
-->

Progress updates generated by **[progressr]** are formally R conditions
of class `progression` that are signaled and handled by R's condition
framework. Sometimes these `progression` conditions are not signaled,
or transmitted, as expected. To troubleshoot such problems, we can use
the built-in "debug" handler. For example,

```r
> library(progressr)
> handlers("debug", global = TRUE)
> 
> y <- slow_sum_p(1:3)
[21:17:01.786] (0.000s => +0.015s) initiate: 0/3 (+0) '' {clear=TRUE, enabled=TRUE, status=}
[21:17:01.975] (0.189s => +0.179s) update: 0/3 (+0) '' {clear=TRUE, enabled=TRUE, status=}
[21:17:02.378] (0.593s => +0.002s) update: 0/3 (+0) '' {clear=TRUE, enabled=TRUE, status=}
[21:17:02.581] (0.796s => +0.002s) update: 1/3 (+1) 'P: Adding 1' {clear=TRUE, enabled=TRUE, status=}
M: Added value 1
[21:17:02.788] (1.002s => +0.004s) update: 1/3 (+0) 'P: Adding 1' {clear=TRUE, enabled=TRUE, status=}
[21:17:02.990] (1.205s => +0.001s) update: 1/3 (+0) 'P: Adding 1' {clear=TRUE, enabled=TRUE, status=}
[21:17:03.394] (1.608s => +0.002s) update: 1/3 (+0) 'P: Adding 1' {clear=TRUE, enabled=TRUE, status=}
[21:17:03.597] (1.812s => +0.002s) update: 2/3 (+1) 'P: Adding 2' {clear=TRUE, enabled=TRUE, status=}
M: Added value 2
[21:17:03.807] (2.021s => +0.007s) update: 2/3 (+0) 'P: Adding 2' {clear=TRUE, enabled=TRUE, status=}
[21:17:04.010] (2.224s => +0.001s) update: 2/3 (+0) 'P: Adding 2' {clear=TRUE, enabled=TRUE, status=}
[21:17:04.412] (2.627s => +0.001s) update: 2/3 (+0) 'P: Adding 2' {clear=TRUE, enabled=TRUE, status=}
[21:17:04.616] (2.830s => +0.002s) update: 3/3 (+1) 'P: Adding 3' {clear=TRUE, enabled=TRUE, status=}
[21:17:04.618] (2.832s => +0.004s) update: 3/3 (+0) 'P: Adding 3' {clear=TRUE, enabled=TRUE, status=}
M: Added value 3
> 
```

[progressr]: https://progressr.futureverse.org

<!--
%\VignetteIndexEntry{progressr: An Introduction}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{Rprofile}
%\VignetteKeyword{Renviron}
%\VignetteEngine{progressr::selfonly}
-->
<!-- DO NOT EDIT THIS FILE! Edit 'OVERVIEW.md' instead and then rebuild this file with 'make vignettes' -->
The **[progressr]** package provides a minimal API for reporting
progress updates in [R](https://www.r-project.org/).  The design is to
separate the representation of progress updates from how they are
presented.  What type of progress to signal is controlled by the
developer.  How these progress updates are rendered is controlled by
the end user.  For instance, some users may prefer visual feedback
such as a horizontal progress bar in the terminal, whereas others may
prefer auditory feedback.  The **[progressr]** framework is designed
to work out-of-the-box also with parallel and distributed processing,
especially with the **[futureverse]** ecosystem.

<img src="imgs/three_in_chinese.gif" alt="Three strokes writing three in Chinese" style="float: right; margin-right: 1ex; margin-left: 1ex;"/>

Design motto:

> The developer is responsible for providing progress updates but it's
> only the end user who decides if, when, and how progress should be
> presented. No exceptions will be allowed.


## Two Minimal APIs - One For Developers and One For End-Users

<div style="overflow: hidden">
<div style="float: left">
<table style="border: 1px solid #999; box-shadow: 2px 2px 2px #999;">
<tr><th>Developer's API</th></tr>
<tr style="vertical-align: top">
<td>
<p>
1. Set up a progressor with a certain number of steps:
</p>
<pre>
p &lt;- progressor(nsteps)
p &lt;- progressor(along = x)
</pre>

<p>
2. Signal progress:
</p>

<pre>
p()               # one-step progress
p(amount = 0)     # "still alive"
p("loading ...")  # pass on a message
</pre>
</td>
</tr>
</table>
</div>
<div style="float: left">&nbsp;&nbsp;&nbsp;&nbsp;</div>
<div style="float: left">
<table style="border: 1px solid #999; box-shadow: 2px 2px 2px #999;">
<tr><th>End-user's API</th></tr>
<tr style="vertical-align: top">
<td>
<p>
1a. Subscribe to progress updates from everywhere:
</p>

<pre>
handlers(global = TRUE)

y &lt;- slow_sum(1:5)
y &lt;- slow_sum(6:10)
</pre>

<p>
1b. Subscribe to a specific expression:
</p>

<pre>
with_progress({
  y &lt;- slow_sum(1:5)
  y &lt;- slow_sum(6:10)
})
</pre>

<p>
2. Configure how progress is presented:
</p>

<pre>
handlers("progress")
handlers("txtprogressbar", "beepr")
handlers(handler_pbcol(enable_after = 3.0))
handlers(handler_progress(complete = "#"))
</pre>
</td>
</table>
</div>
</div>

## A simple example

Assume that we have a function `slow_sum()` for adding up the values
in a vector.  It is so slow, that we like to provide progress updates
to whoever might be interested in it.  With the **progressr** package,
this can be done as:

```r
slow_sum <- function(x) {
  p <- progressr::progressor(along = x)
  sum <- 0
  for (kk in seq_along(x)) {
    Sys.sleep(0.1)
    sum <- sum + x[kk]
    p(message = sprintf("Adding %g", x[kk]))
  }
  sum
}
```

Note how there are _no_ arguments (e.g. `.progress = TRUE`) in the
code that specify how progress is presented. This is by design and
because the only task for the developer is to decide on where in the
code it makes sense to signal that progress has been made. As we will
see next, it should be up to the end user, and end user only, of this
code to decide whether they want to receive progress updates or not,
and, if so, in what format. Asking them to specify a special
"progress" argument adds a lot of friction, it clutters up the code,
and, importantly, might not even be possible for end users to do
(e.g. they call a package function that in turn calls the progress
reporting function of interest).

Now, if we call this function, without further settings:

```r
> y <- slow_sum(1:10)
> y
[1] 55
>
``` 

the default is that there will be _no_ progress updates. To get
progress updates, we need to request them to be "handled", which we do
by:

```r
> progressr::handlers(global = TRUE)
```

After this, progress will be reported;

```r
> y <- slow_sum(1:10)
  |====================                               |  40%
> y <- slow_sum(10:1)
  |========================================           |  80%
```

To disable reporting again, do:

```r
> handlers(global = FALSE)
```


## Customizing how progress is reported

By default, **[progressr]** presents progress via the built-in
`utils::txtProgressBar()`. It presents itself as a rudimentary
ASCII-based horizontal progress bar in the R terminal. See
`help("handler_txtprogressbar")` for how to customize the look of
"txtprogressbar", e.g. colorization and Unicode.  There are many other
ways to report on progress, including visually, auditory, and via
notification systems. You can also use a mix of these, e.g.

```r
handlers(c("cli", "beepr", "ntfy"))
```

See the 'Customizing How Progress is Reported' vignette for for examples.


## Additional Features

### Support for progressr elsewhere

Note that progression updates by **progressr** is designed to work out
of the box for any iterator framework in R. See the different package
vignettes for details. Prominent examples are:

 * `lapply()` etc. of base R
 * `map()` etc. by the **[purrr]** package
 * `llply()` etc. by the **[plyr]** package
 * `foreach()` iterations by the **[foreach]** package

and near-live progress reporting in parallel and distributed
processing via the **[future]** framework:

 * `future_lapply()` etc.  by the **[future.apply]** package
 * `future_map()` etc. by the **[furrr]** package
 * `llply()` etc. by the **[plyr]** and **[doFuture]** packages
 * `foreach()` iterations via the **foreach** and **[doFuture]** packages
 * `bplapply()` etc. by the **[BiocParallel]** and **[doFuture]** packages

Other uses of **progressr** are:

 * make packages that report progress via the **[cli]** package
   (e.g. **purrr**) report progress via **progressr**
 * make `knit()` of the **[knitr]** package report via **progressr**


### Use regular output as usual alongside progress updates

In contrast to other progress-bar frameworks, output from `message()`,
`cat()`, `print()` and so on, will _not_ interfere with progress
reported via **progressr**.  For example, say we have:

```r
slow_sqrt <- function(xs) {
  p <- progressor(along = xs)
  lapply(xs, function(x) {
    message("Calculating the square root of ", x)
    Sys.sleep(2)
    p(sprintf("x=%g", x))
    sqrt(x)
  })
}
```

we will get:

```r
> library(progressr)
> handlers(global = TRUE)
> handlers("progress")
> y <- slow_sqrt(1:8)
Calculating the square root of 1
Calculating the square root of 2
- [===========>-----------------------------------]  25% x=2
```

This works because **progressr** will briefly buffer any output
internally and only release it when the next progress update is
received just before the progress is re-rendered in the terminal.
This is why you see a two second delay when running the above example.
Note that, if we use progress handlers that do not output to the
terminal, such as `handlers("beepr")`, then output does not have to be
buffered and will appear immediately.


_Comment_: When signaling a warning using `warning(msg, immediate. =
TRUE)` the message is immediately outputted to the standard-error
stream.  However, this is not possible to emulate when warnings are
intercepted using calling handlers.  This is a limitation of R that
cannot be worked around.  Because of this, the above call will behave
the same as `warning(msg)` - that is, all warnings will be buffered by
R internally and released only when all computations are done.


### Sticky messages

As seen above, some progress handlers present the progress message as
part of its output, e.g. the "progress" handler will display the
message as part of the progress bar.  It is also possible to "push"
the message up together with other terminal output.  This can be done
by adding class attribute `"sticky"` to the progression signaled.
This works for several progress handlers that output to the terminal.
For example, with:

```r
slow_sum <- function(x) {
  p <- progressr::progressor(along = x)
  sum <- 0
  for (kk in seq_along(x)) {
    Sys.sleep(0.1)
    sum <- sum + x[kk]
    p(sprintf("Step %d", kk), class = if (kk %% 5 == 0) "sticky", amount = 0)
    p(message = sprintf("Adding %g", x[kk]))
  }
  sum
}
```

we get

```r
> handlers("txtprogressbar")
> y <- slow_sum(1:30)
Step 5
Step 10
  |====================                               |  43%
```

and

```r
> handlers("progress")
> y <- slow_sum(1:30)
Step 5
Step 10
/ [===============>-------------------------]  43% Adding 13
```



[futureverse]: https://www.futureverse.org
[progressr]: https://progressr.futureverse.org
[future]: https://future.futureverse.org
[future.apply]: https://future.apply.futureverse.org
[furrr]: https://furrr.futureverse.org
[doFuture]: https://doFuture.futureverse.org
[beepr]: https://cran.r-project.org/package=beepr
[cli]: https://cran.r-project.org/package=cli
[foreach]: https://cran.r-project.org/package=foreach
[progress]: https://cran.r-project.org/package=progress
[purrr]: https://cran.r-project.org/package=purrr
[doParallel]: https://cran.r-project.org/package=doParallel
[knitr]: https://cran.r-project.org/package=knitr
[pbapply]: https://cran.r-project.org/package=pbapply
[pbmcapply]: https://cran.r-project.org/package=pbmcapply
[plyr]: https://cran.r-project.org/package=plyr
[BiocParallel]: https://www.bioconductor.org/packages/BiocParallel/

<!--
%\VignetteIndexEntry{progressr: Render 'knitr' Documents}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{knitr}
%\VignetteEngine{progressr::selfonly}
-->

When compiling ("knitting") an knitr-based vignette, for instance, via
`knitr::knit()`, **[knitr]** shows the progress of code chunks
processed thus far using a progress bar.  We can use **[progressr]**
also for this type of progress reporting.  To have **knitr** report on
progress via **progressr**, set R option `knitr.progress.fun` as:

```r
options(knitr.progress.fun = function(total, labels) {
  p <- progressr::progressor(total, on_exit = FALSE)
  list(
    update = function(i) p(sprintf("chunk: %s", labels[i])),
    done = function() p(type = "finish")
  )
})
```

That's it! With the above `knitr.progress.fun` option set, try to knit
a document, e.g.

```r
progressr::handlers(global = TRUE)
infile <- system.file("examples", "knitr-minimal.Rnw", package = "knitr")
outfile <- knitr::knit(infile)
```

To customize the progress bar, try for instance:

```r
progressr::handlers(global = TRUE)
options(crayon.enabled = TRUE)  ## force use of colors
progressr::handlers("pbcol")
```

This will produce a colored progress bar in the terminal.

_Comment:_ If the above knit example completes too quickly for you to
see the progress updates, you can add the "slowdown" handler to slow
down the knitting process, e.g.

```r
progressr::handlers(c("pbcol", "slowdown"))
```

This will add the overall process to take at least ten seconds
(default).


To make **knitr** report via **progressr** in all your R session, set
the above R option in your <code>~/.Rprofile</code> startup file, e.g.

```r
if (requireNamespace("progressr", quietly = TRUE)) {
  options(knitr.progress.fun = function(total, labels) {
    p <- progressr::progressor(total, on_exit = FALSE)
    list(
      update = function(i) p(sprintf("chunk: %s", labels[i])),
      done = function() p(type = "finish")
    )
  })
}
```


[progressr]: https://progressr.futureverse.org
[knitr]: https://cran.r-project.org/package=knitr

<!--
%\VignetteIndexEntry{progressr: Customize How Progress is Reported}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{handlers}
%\VignetteEngine{progressr::selfonly}
-->

_Intended audience: end-users only_


## Terminal-based progress bars

The default is to present progress via `utils::txtProgressBar()`,
which is available on all R installations.  It presents itself as an
ASCII-based horizontal progress bar in the R terminal. This is
rendered as:

![SVG animation of the default "txtprogressbar" progress handler](imgs/handler_txtprogressbar-default.svg)

We can tweak this "txtprogressbar" handler to use red hearts for the
bar, e.g.

```r
handlers(handler_txtprogressbar(char = cli::col_red(cli::symbol$heart)))
```

which results in:

![SVG animation of the "txtprogressbar" progress handler with red hearts](imgs/handler_txtprogressbar-char-ansi.svg)

Another example is:

```r
handlers(handler_pbcol(
      adjust = 1.0,
    complete = function(s) cli::bg_red(cli::col_black(s)),
  incomplete = function(s) cli::bg_cyan(cli::col_black(s))
))
```

which results in:

![SVG animation of the "pbcol" progress handler with text aligned to the right](imgs/handler_pbcol-adjust-right-complete.svg)

To change the default, to, say, `cli_progress_bar()` by the **[cli]**
package, set:

```r
handlers("cli")
```

This progress handler will present itself as:

![SVG animation of the default "cli" progress handler](imgs/handler_cli-default.svg)

To instead use `progress_bar()` by the **[progress]** package, set:

```r
handlers("progress")
```
This progress handler will present itself as:

![SVG animation of the default "progress" progress handler](imgs/handler_progress-default.svg)


## Auditory progress updates

Progress updates do not have to be presented visually. They can
equally well be communicated via audio. For example, using:

```r
handlers("beepr")
```

will present itself as sounds played at the beginning, while progressing, and at the end (using different **[beepr]** sounds).  There will be _no_ output written to the terminal;

```r
> y <- slow_sum(1:10)
> y
[1] 55
>
```


## Concurrent auditory and visual progress updates

It is possible to have multiple progress handlers presenting progress
updates at the same time.  For example, to get both visual and
auditory updates, use:

```r
handlers("txtprogressbar", "beepr")
```


## Silence all progress

To silence all progress updates, use:

```r
handlers("void")
```


## Further configuration of progress handlers

Above we have seen examples where the `handlers()` takes one or more
strings as input, e.g. `handlers(c("progress", "beepr"))`.  This is
short for a more flexible specification where we can pass a list of
handler functions, e.g.

```r
handlers(list(
  handler_progress(),
  handler_beepr()
))
```

With this construct, we can make adjustments to the default behavior
of these progress handlers.  For example, we can configure the
`format`, `width`, and `complete` arguments of
`progress::progress_bar$new()`, and tell **beepr** to use a different
`finish` sound and generate sounds at most every two seconds by
setting:

```r
handlers(list(
  handler_progress(
    format   = ":spin :current/:total (:message) [:bar] :percent in :elapsed ETA: :eta",
    width    = 60,
    complete = "+"
  ),
  handler_beepr(
    finish   = "wilhelm",
    interval = 2.0
  )
))
```


## Set a default progress handler for all your R sessions

To set the default progress handler, or handlers, in all your R
sessions, call `progressr::handlers(...)` in your
<code>~/.Rprofile</code> startup file.  For example,

```r
if (requireNamespace("progressr", quietly = TRUE)) {
  progressr::handlers(global = TRUE)
  if (requireNamespace("cli", quietly = TRUE)) {
    progressr::handlers("cli")
  }
}
```


[progressr]: https://progressr.futureverse.org
[beepr]: https://cran.r-project.org/package=beepr
[cli]: https://cran.r-project.org/package=cli
[progress]: https://cran.r-project.org/package=progress

<!--
%\VignetteIndexEntry{progressr: Parallel and Distributed Processing}
%\VignetteAuthor{Henrik Bengtsson}
%\VignetteKeyword{R}
%\VignetteKeyword{package}
%\VignetteKeyword{vignette}
%\VignetteKeyword{progress}
%\VignetteKeyword{parallel}
%\VignetteKeyword{distributed}
%\VignetteEngine{progressr::selfonly}
-->

## TL;DR

The **progressr** package works seamlessly with parallel and
distributed processing using **[futureverse]**, and it will also
provide near-live progress updates while the parallel processing is
still running. For example,

```r
library(progressr)
handlers("progress", global = TRUE)

library(futurize)
plan(multisession, workers = 2)

my_fcn <- function(xs) {
  p <- progressr::progressor(along = xs)
  lapply(xs, function(x, ...) {
    Sys.sleep((10.0-x)/2)
    p(sprintf("x=%g", x))
    sqrt(x)
  }) |> futurize()
}

y <- my_fcn(1:10)
# / [================>-----------------------------]  40% x=2
```

Alternatively, we can use `progressify()` from the **[progressify]** package
to automatically add progress reporting without modifying the function's
internal code:

```r
library(progressify)
handlers("progress", global = TRUE)

library(futurize)
plan(multisession, workers = 2)

my_fcn <- function(xs) {
  lapply(xs, function(x, ...) {
    Sys.sleep((10.0-x)/2)
    sqrt(x)
  }) |> progressify() |> futurize()
}

y <- my_fcn(1:10)
# / [================>-----------------------------]  40%
```

## Introduction

The **[futureverse]** framework, which provides a unified API for parallel
and distributed processing in R, has built-in support for the kind of
progression updates produced by the **progressr** package.  The modern,
recommended approach to parallelize such code is using `futurize()` from
the **[futurize]** package, which supports common map-reduce and iteration
functions like `lapply()`, `purrr::map()`, `foreach()`, `bplapply()`, and
`llply()`.

Traditional parallelization packages such as **[future.apply]**, **[furrr]**,
and **[foreach]** with **[doFuture]** (specifically, `%dofuture%` or registered
via `registerDoFuture()`) are still fully supported and can be used as alternatives.

In contrast, _non-future_ parallelization methods such as **parallel**'s
`mclapply()` and `parallel::parLapply()`, and **foreach** adapters like
**doParallel** do _not_ support progress reports via **progressr**.

### lapply() with futurize()

Here is an example that uses `futurize()` of the **[futurize]** package to parallelize a standard `lapply()` call on the local machine while at the same time signaling progression updates:

```r
library(progressr)
handlers("progress", global = TRUE)

library(futurize)
plan(multisession, workers = 2)

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  lapply(xs, function(x, ...) {
    Sys.sleep((10.0-x)/2)
    p(sprintf("x=%g", x))
    sqrt(x)
  }) |> futurize()
}

y <- my_fcn(1:10)
# / [================>-----------------------------]  40% x=2
```

Note that using `future_lapply()` of the **[future.apply]** package is still supported as a traditional alternative.

### foreach() with futurize()

Here is an example that uses `foreach()` of the **[foreach]** package
together with `futurize()` of the **[futurize]** package to
parallelize a standard sequential `%do%` loop while reporting on progress.
This example parallelizes on the local machine; it works also for remote machines:

```r
library(progressr)
handlers("progress", global = TRUE)

library(foreach)
library(futurize)
plan(multisession, workers = 2)

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  foreach(x = xs) %do% {
    Sys.sleep((10.0-x)/2)
    p(sprintf("x=%g", x))
    sqrt(x)
  } |> futurize()
}

y <- my_fcn(1:10)
# / [================>-----------------------------]  40% x=2
```

Note that using the `%dofuture%` operator of the **[doFuture]** package,
or traditional `%dopar%` registered via `registerDoFuture()`, is still supported.
For example:

```r
library(progressr)
handlers("progress", global = TRUE)

library(doFuture)
registerDoFuture()      ## %dopar% parallelizes via future
plan(multisession, workers = 2)

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  foreach(x = xs) %dopar% {
    Sys.sleep((10.0-x)/2)
    p(sprintf("x=%g", x))
    sqrt(x)
  }
}

y <- my_fcn(1:10)
# / [================>-----------------------------]  40% x=2
```

### purrr::map() with futurize()

Here is an example that uses `purrr::map()` and `futurize()` to parallelize
on the local machine while at the same time signaling progression updates:

```r
library(progressr)
handlers("progress", global = TRUE)

library(futurize)
plan(multisession, workers = 2)

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  purrr::map(xs, function(x) {
    Sys.sleep((10.0-x)/2)
    p(sprintf("x=%g", x))
    sqrt(x)
  }) |> futurize()
}

y <- my_fcn(1:10)
# / [================>-----------------------------]  40% x=2
```

_Note:_ Using `future_map()` of the **[furrr]** package is still supported.
This solution does not involve the `.progress = TRUE`
argument that **furrr** implements.  Because **progressr** is more
generic and because `.progress = TRUE` only supports certain future
backends and produces errors on non-supported backends, I recommend
to stop using `.progress = TRUE` and use the **progressr** package
instead.

### BiocParallel::bplapply() with futurize()

Here is an example that uses `bplapply()` of the **[BiocParallel]**
package and `futurize()` to parallelize on the local machine while
at the same time signaling progression updates:

```r
library(progressr)
handlers("progress", global = TRUE)

library(BiocParallel)
library(futurize)
plan(multisession, workers = 2)

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  bplapply(xs, function(x) {
    Sys.sleep((10.0-x)/2)
    p(sprintf("x=%g", x))
    sqrt(x)
  }) |> futurize()
}

y <- my_fcn(1:10)
# / [================>-----------------------------]  40% x=2
```

### plyr::llply() with futurize()

Here is an example that uses `llply()` of the **[plyr]** package and
`futurize()` to parallelize on the local machine while at the same
time signaling progression updates:

```r
library(progressr)
handlers("progress", global = TRUE)

library(plyr)
library(futurize)
plan(multisession, workers = 2)

my_fcn <- function(xs) {
  p <- progressor(along = xs)
  llply(xs, function(x, ...) {
    Sys.sleep((10.0-x)/2)
    p(sprintf("x=%g", x))
    sqrt(x)
  }) |> futurize()
}

y <- my_fcn(1:10)
# / [================>-----------------------------]  40% x=2
```

_Note:_ As an alternative to the above, recommended approach, one can
use `.progress = "progressr"` together with `.parallel = TRUE` and the **[doFuture]** package.  This
requires **plyr** (>= 1.8.7).

### Near-live versus buffered progress updates with futures

As of August 2025, there are six types of **future** backends that are
known(*) to provide near-live progress updates:

 1. `sequential`,
 2. `multicore`,
 3. `multisession`, and
 4. `cluster` (local and remote)
 5. `future.callr::callr`
 6. `future.mirai::mirai_multisession`

Here "near-live" means that the progress handlers will report on
progress almost immediately when the progress is signaled on the
worker. This is because these parallel backends handle the special
condition class `immediateCondition` - they detect when such
conditions are signaled and relay them to the parent R process as soon
as possible. For all other future backends, the progress updates are
only relayed back to the main machine and reported together with the
results of the futures.  For instance, if `lapply(X, FUN) |> futurize()`
chunks up the processing of, say, 100 elements in `X` into eight
futures, we will see progress from each of the 100 elements as they
are done when using a future backend supporting "near-live" updates,
whereas we will only see those updates flushed eight times when
using any other types of future backends.

(*) Other future backends may gain support for "near-live" progress
updating later.  Adding support for those is independent of the
**progressr** package.  Feature requests for adding that support
should go to those future-backend packages.

[futureverse]: https://www.futureverse.org
[progressr]: https://progressr.futureverse.org
[future]: https://future.futureverse.org
[futurize]: https://futurize.futureverse.org
[progressify]: https://progressify.futureverse.org
[future.apply]: https://future.apply.futureverse.org
[furrr]: https://furrr.futureverse.org
[doFuture]: https://doFuture.futureverse.org
[progress]: https://cran.r-project.org/package=progress
[purrr]: https://cran.r-project.org/package=purrr
[foreach]: https://cran.r-project.org/package=foreach
[doParallel]: https://cran.r-project.org/package=doParallel
[knitr]: https://cran.r-project.org/package=knitr
[plyr]: https://cran.r-project.org/package=plyr
[BiocParallel]: https://www.bioconductor.org/packages/BiocParallel/

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_slowdown.R
\name{handler_slowdown}
\alias{handler_slowdown}
\title{Progression Handler: Slow Down Progress Reporting}
\usage{
handler_slowdown(
  duration = 10,
  intrusiveness = 0,
  target = "void",
  enable = getOption("progressr.enable", TRUE),
  ...
)
}
\arguments{
\item{duration}{(numeric) Added slowdown (in seconds) across all progress
steps, e.g. if \code{duration = 10.0} and \code{steps = 100L}, then each step unit is
delayed 0.1 seconds, totaling a ten-second slowdown.}

\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{enable}{(logical) If FALSE, then progress is not reported.}

\item{\ldots}{Additional arguments passed to \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\description{
Progression Handler: Slow Down Progress Reporting
}
\details{
This progression handler delays each progress step a certain number of
seconds. It does not render any output.
}
\examples{
\donttest{\dontrun{
handlers(handler_slowdown(duration = 30.0)) ## 30-second total slowdown
with_progress(y <- slow_sum(1:100))
print(y)
}}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progression.R
\name{progression}
\alias{progression}
\title{A Progression Condition}
\usage{
progression(
  message = character(0L),
  amount = 1,
  step = NULL,
  time = progression_time,
  ...,
  type = "update",
  class = NULL,
  progressor_uuid = NULL,
  progression_index = NULL,
  progression_time = Sys.time(),
  call = NULL,
  calls = sys.calls(),
  owner_session_uuid = NULL
)
}
\arguments{
\item{message}{(character vector or a function) If a character vector, then
it is pasted together into a single string using an empty separator.
If a function, then the message is constructed by \code{conditionMessage(p)}
calling this function with the progression condition \code{p} itself as the
first argument.}

\item{amount}{(numeric) The total amount of progress made.}

\item{step}{(numeric) The step completed. If specified, \code{amount} is ignored.
\emph{WARNING: Argument \code{step} should only be used when in full control of the
order when this progression condition is signaled.} For example, it must not
be signaled as one of many parallel progress updates signaled concurrently,
because then we cannot guarantee the order these progressions arrive.}

\item{time}{(POSIXct) A timestamp.}

\item{type}{Type of progression made.}

\item{class}{(character) Zero or more class names to prepend.}

\item{progressor_uuid}{(character string) A character string that is unique
for the current progressor and the current \R session.}

\item{progression_index}{(integer) A non-negative integer that is
incremented by one for each progression condition created.}

\item{progression_time}{(POSIXct or character string) A timestamp specifying
when the progression condition was created.}

\item{call}{(expression) A call expression.}

\item{calls}{(pairlist) The calls that lead up to this progression update.}

\item{owner_session_uuid}{(character string) A character string that is
unique for the \R session where the progressor was created.}

\item{\ldots}{Additional named elements.}
}
\value{
A \link[base:conditions]{base::condition} of class \code{progression}.
}
\description{
A progression condition represents a progress in an \R program.
}
\seealso{
To signal a progression condition, use \code{\link[base:conditions]{base::signalCondition()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/withProgressShiny.R
\name{withProgressShiny}
\alias{withProgressShiny}
\title{Use Progressr in Shiny Apps: Plug-in Backward-Compatible Replacement for shiny::withProgress()}
\usage{
withProgressShiny(
  expr,
  ...,
  message = NULL,
  detail = NULL,
  inputs = list(message = NULL, detail = "message"),
  env = parent.frame(),
  quoted = FALSE,
  handlers = c(shiny = handler_shiny, progressr::handlers(default = NULL))
)
}
\arguments{
\item{expr, \ldots, env, quoted}{Arguments passed to \code{\link[shiny:withProgress]{shiny::withProgress()}} as is.}

\item{message, detail}{(character string) The message and the detail message to be passed to \code{\link[shiny:withProgress]{shiny::withProgress()}}.}

\item{inputs}{(named list) Specifies from what sources the Shiny progress
elements 'message' and 'detail' should be updated.  Valid sources are
\code{"message"}, \code{"sticky_message"} and \code{"non_sticky_message"}, where
\code{"message"} is short for \code{c("non_sticky_message", "sticky_message")}. For
example, \code{inputs = list(message = "sticky_message", detail = "message")}
will update the Shiny 'message' component from sticky messages only,
whereas the 'detail' component is updated using any message.}

\item{handlers}{Zero or more progression handlers used to report on progress.}
}
\value{
The value of \link[shiny:withProgress]{shiny::withProgress}.
}
\description{
A plug-in, backward-compatible replacement for \code{\link[shiny:withProgress]{shiny::withProgress()}}.
}
\section{Requirements}{

This function requires the \pkg{shiny} package and will use the
\code{\link[=handler_shiny]{handler_shiny()}} \strong{progressr} handler internally to report on updates.
}

\examples{
library(shiny)
library(progressr)

app <- shinyApp(
  ui = fluidPage(
    plotOutput("plot")
  ),

  server = function(input, output) {
    output$plot <- renderPlot({
      X <- 1:15
      withProgressShiny(message = "Calculation in progress",
                        detail = "Starting ...",
                        value = 0, {
        p <- progressor(along = X)
        y <- lapply(X, FUN=function(x) {
          Sys.sleep(0.25)
          p(sprintf("x=\%d", x))
        })
      })
      
      plot(cars)

      ## Terminate the Shiny app
      Sys.sleep(1.0)
      stopApp(returnValue = invisible())
    })
  }
)

local({
  oopts <- options(device.ask.default = FALSE)
  on.exit(options(oopts))
  if (interactive()) print(app)
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_rstudio.R
\name{handler_rstudio}
\alias{handler_rstudio}
\title{Progression Handler: Progress Reported in the RStudio Console}
\usage{
handler_rstudio(
  intrusiveness = getOption("progressr.intrusiveness.gui", 1),
  target = "gui",
  title = function() format(Sys.time(), "Console \%X"),
  ...
)
}
\arguments{
\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{title}{(character or a function) The "name" of the progressor, which
is displayed in front of the progress bar.  If a function, then the name
is created dynamically by calling the function when the progressor is
created.}

\item{\ldots}{Additional arguments passed to \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\value{
A function of class \code{progression_handler} that takes a
\link{progression} condition as its first and only argument.
}
\description{
Progression Handler: Progress Reported in the RStudio Console
}
\section{Requirements}{

This progression handler works only in the RStudio Console.
}

\section{Use this progression handler by default}{

To use this handler by default whenever using the RStudio Console, add
the following to your \file{~/.Rprofile} startup file:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{if (requireNamespace("progressr", quietly = TRUE)) \{
  if (Sys.getenv("RSTUDIO") == "1" && !nzchar(Sys.getenv("RSTUDIO_TERM"))) \{
    options(progressr.handlers = progressr::handler_rstudio)
  \}
\}
}\if{html}{\out{</div>}}
}

\examples{
if (requireNamespace("rstudioapi", quietly = TRUE) && rstudioapi::isAvailable()) {

  handlers("rstudio")
  with_progress({ y <- slow_sum_p(1:10) })
  print(y)
  
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plyr_progress_progressr.R
\name{progress_progressr}
\alias{progress_progressr}
\title{Use Progressr with Plyr Map-Reduce Functions}
\usage{
progress_progressr(...)
}
\arguments{
\item{\ldots}{Not used.}
}
\value{
A named \link[base:list]{base::list} that can be passed as argument \code{.progress}
to any of \pkg{plyr} function accepting that argument.
}
\description{
A "progress bar" for \pkg{plyr}'s \code{.progress} argument.
}
\section{Limitations}{

One can use use \code{\link[doFuture:registerDoFuture]{doFuture::registerDoFuture()}} to run \pkg{plyr} functions
in parallel, e.g. \code{plyr::l_ply(..., .parallel = TRUE)}.  Unfortunately,
using \code{.parallel = TRUE} disables progress updates because, internally,
\pkg{plyr} forces \code{.progress = "none"} whenever \code{.parallel = TRUE}.
Thus, despite the \pkg{future} ecosystem and \pkg{progressr} would support
it, it is not possible to run \pkg{dplyr} in parallel \emph{and} get progress
updates at the same time.
}

\examples{
if (requireNamespace("plyr", quietly=TRUE)) {

  with_progress({
    y <- plyr::llply(1:10, function(x) {
      Sys.sleep(0.1)
      sqrt(x)
    }, .progress = "progressr")
  })
  
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progressor.R
\name{progressor}
\alias{progressor}
\title{Create a Progressor Function that Signals Progress Updates}
\usage{
progressor(
  steps = length(along),
  along = NULL,
  offset = 0L,
  scale = 1L,
  transform = function(steps) scale * steps + offset,
  message = character(0L),
  label = NA_character_,
  trace = FALSE,
  initiate = TRUE,
  finalize = FALSE,
  auto_finish = TRUE,
  on_exit = !identical(envir, globalenv()),
  enable = getOption("progressr.enable", TRUE),
  envir = parent.frame()
)
}
\arguments{
\item{steps}{(integer) Number of progressing steps.}

\item{along}{(vector; alternative) Alternative that sets
\code{steps = length(along)}.}

\item{offset, scale}{(integer; optional) scale and offset applying the transform
\code{steps <- scale * steps + offset}.}

\item{transform}{(function; optional) A function that takes the effective
number of \code{steps} as input and returns another finite and non-negative
number of steps.}

\item{message}{(character vector or a function) If a character vector, then
it is pasted together into a single string using an empty separator.
If a function, then the message is constructed by \code{conditionMessage(p)}
calling this function with the progression condition \code{p} itself as the
first argument.}

\item{label}{(character) A label.}

\item{trace}{(logical) If TRUE, then the call stack is recorded, otherwise
not.}

\item{initiate}{(logical) If TRUE, the progressor will signal a
\link{progression} 'initiate' condition when created.}

\item{finalize}{(logical) If TRUE, the progressor signals a \link{progression}
'shutdown' condition when finalized by the garbage collector.}

\item{auto_finish}{(logical) If TRUE, then the progressor will signal a
\link{progression} 'finish' condition as soon as the last step has been reached.}

\item{on_exit, envir}{(logical) If TRUE, then the created progressor will
signal a \link{progression} 'finish' condition when the calling frame exits.
This is ignored if the calling frame (\code{envir}) is the global environment.}

\item{enable}{(logical) If TRUE, \link{progression} conditions are signaled when
calling the progressor function created by this function.
If FALSE, no \link{progression} conditions is signaled because the progressor
function is an empty function that does nothing.}
}
\value{
A function of class \code{progressor}.
}
\description{
Create a Progressor Function that Signals Progress Updates
}
\details{
A \code{progressor} function can only be created inside a local environment,
e.g. inside a function, within a \code{local()} call, or within a
\code{with_progress()} call.  Notably, it \emph{cannot} be created at the top level,
e.g. immediately at the R prompt or outside a local environment in an
R script.  If attempted, an informative error message is produced, e.g.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{> p <- progressr::progressor(100)
Error in progressr::progressor(100) : 
  A progressor must not be created in the global environment unless
wrapped in a with_progress() or without_progress() call. Alternatively,
create it inside a function or in a local() environment to make sure
there is a finite life span of the progressor
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_pbmcapply.R
\name{handler_pbmcapply}
\alias{handler_pbmcapply}
\title{Progression Handler: Progress Reported via 'pbmcapply' Progress Bars (Text) in the Terminal}
\usage{
handler_pbmcapply(
  char = "=",
  substyle = 3L,
  style = "ETA",
  file = stderr(),
  intrusiveness = getOption("progressr.intrusiveness.terminal", 1),
  target = "terminal",
  ...
)
}
\arguments{
\item{char}{(character) The symbols to form the progress bar for
\code{\link[utils:txtProgressBar]{utils::txtProgressBar()}}.}

\item{substyle}{(integer) The progress-bar substyle according to
\code{\link[pbmcapply:progressBar]{pbmcapply::progressBar()}}.}

\item{style}{(character) The progress-bar style according to}

\item{file}{(connection) A \link[base:connections]{base::connection} to where output should be sent.}

\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{\ldots}{Additional arguments passed to \code{\link[pbmcapply:progressBar]{pbmcapply::progressBar()}}
and \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\description{
A progression handler for \code{\link[pbmcapply:progressBar]{pbmcapply::progressBar()}}.
}
\section{Requirements}{

This progression handler requires the \pkg{pbmcapply} package.
}

\section{Appearance}{

Below are a few examples on how to use and customize this progress handler.
In all cases, we use \code{handlers(global = TRUE)}.
Since \code{style = "txt"} corresponds to using \code{\link[=handler_txtprogressbar]{handler_txtprogressbar()}}
with \code{style = substyle}, the main usage of this handler is with
\code{style = "ETA"} (default) for which \code{substyle} is ignored.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{handlers("pbmcapply")
y <- slow_sum(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_pbmcapply-default.svg}}
}

\examples{
if (requireNamespace("pbmcapply", quietly = TRUE)) {

  handlers("pbmcapply")
  with_progress({ y <- slow_sum(1:10) })
  print(y)
  
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_newline.R
\name{handler_newline}
\alias{handler_newline}
\title{Progression Handler: Progress Reported as a New Line (Text) in the Terminal}
\usage{
handler_newline(
  symbol = "\\n",
  file = stderr(),
  intrusiveness = getOption("progressr.intrusiveness.debug", 0),
  target = "terminal",
  ...
)
}
\arguments{
\item{symbol}{(character string) The character symbol to be outputted,
which by default is the ASCII NL character (\code{'\\n'} = \code{'\\013'}) character.}

\item{file}{(connection) A \link[base:connections]{base::connection} to where output should be sent.}

\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{\ldots}{Additional arguments passed to \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\description{
Progression Handler: Progress Reported as a New Line (Text) in the Terminal
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_ntfy.R
\name{handler_ntfy}
\alias{handler_ntfy}
\title{Progression Handler: Progress Reported via the Ntfy.sh Messaging Service}
\usage{
handler_ntfy(
  intrusiveness = getOption("progressr.intrusiveness.ntfy", 5),
  target = "gui",
  ...,
  title = "Progress update from R"
)
}
\arguments{
\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{title}{title of notification. See \url{https://docs.ntfy.sh/publish/#message-title}}

\item{\ldots}{Additional arguments passed to \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\description{
A progression handler for \code{\link[ntfy:ntfy_send]{ntfy::ntfy_send()}} of the \pkg{ntfy} package,
which sends notifications via the \url{https://ntfy.sh} framework.
}
\section{Requirements}{

This progression handler requires the \pkg{ntfy} package.
}

\examples{
pkg <- "ntfy"
if (requireNamespace(pkg, quietly = TRUE)) {
  ## We need to specify a ntfy.sh topic that progress messages
  ## should be sent to. See help("ntfy_topic", package = "ntfy")
  ## for details
  Sys.setenv(NTFY_TOPIC = "R-my-secret-topic")

  handlers("ntfy")
  with_progress({ y <- slow_sum(1:10) })
  print(y)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_notifier.R
\name{handler_notifier}
\alias{handler_notifier}
\title{Progression Handler: Progress Reported via the Operating-System Notification Framework (GUI, Text)}
\usage{
handler_notifier(
  intrusiveness = getOption("progressr.intrusiveness.notifier", 10),
  target = "gui",
  ...
)
}
\arguments{
\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{\ldots}{Additional arguments passed to \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\value{
A function of class \code{progression_handler} that takes a
\link{progression} condition as its first and only argument.
}
\description{
A progression handler for \code{notify()} of the \pkg{notifier} package.
}
\section{Requirements}{

This progression handler requires the \pkg{notifier} package, which is only
available from \url{https://github.com/gaborcsardi/notifier}.  This can be
installed as \code{remotes::install_github("gaborcsardi/notifier@62d484")}.
}

\examples{
pkg <- "notifier"
if (requireNamespace(pkg, quietly = TRUE)) {

  handlers("notifier")
  with_progress({ y <- slow_sum_p(1:10) })
  print(y)
  
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_txtprogressbar.R
\name{handler_txtprogressbar}
\alias{handler_txtprogressbar}
\title{Progression Handler: Progress Reported as Plain Progress Bars (Text) in the Terminal}
\usage{
handler_txtprogressbar(
  char = "=",
  style = 3L,
  file = stderr(),
  intrusiveness = getOption("progressr.intrusiveness.terminal", 1),
  target = "terminal",
  ...
)
}
\arguments{
\item{char}{(character) The symbols to form the progress bar for
\code{\link[utils:txtProgressBar]{utils::txtProgressBar()}}. Contrary to \code{txtProgressBar()}, this handler
supports also ANSI-colored symbols.}

\item{style}{(integer) The progress-bar style according to
\code{\link[utils:txtProgressBar]{utils::txtProgressBar()}}.}

\item{file}{(connection) A \link[base:connection]{base::connection} to where output should be sent.}

\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{\ldots}{Additional arguments passed to \code{\link[utils:txtProgressBar]{utils::txtProgressBar()}}
and \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\value{
A function of class \code{progression_handler} that takes a
\link{progression} condition as its first and only argument.
}
\description{
A progression handler for \code{\link[utils:txtProgressBar]{utils::txtProgressBar()}}.
}
\section{Appearance}{

Below are a few examples on how to use and customize this progress handler.
In all cases, we use \code{handlers(global = TRUE)}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(progressr)
handlers("txtprogressbar")
y <- slow_sum_p(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_txtprogressbar-default.svg}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(progressr)
handlers(handler_txtprogressbar(style = 1L))
y <- slow_sum_p(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_txtprogressbar-style-1.svg}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(progressr)
handlers(handler_txtprogressbar(style = 3L))
y <- slow_sum_p(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_txtprogressbar-style-3.svg}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(progressr)
handlers(handler_txtprogressbar(char = "#"))
y <- slow_sum_p(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_txtprogressbar-char.svg}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(progressr)
handlers(handler_txtprogressbar(char = "<>"))
y <- slow_sum_p(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_txtprogressbar-char-width-2.svg}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(progressr)
handlers(handler_txtprogressbar(char = cli::col_red(cli::symbol$heart)))
y <- slow_sum_p(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_txtprogressbar-char-ansi.svg}}
}

\examples{
handlers("txtprogressbar")

with_progress({ y <- slow_sum_p(1:10) })
print(y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_void.R
\name{handler_void}
\alias{handler_void}
\title{Progression Handler: No Progress Report}
\usage{
handler_void(intrusiveness = 0, target = "void", enable = FALSE, ...)
}
\arguments{
\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{enable}{(logical) If FALSE, then progress is not reported.}

\item{\ldots}{Additional arguments passed to \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\description{
Progression Handler: No Progress Report
}
\details{
This progression handler gives not output - it is invisible and silent.
}
\examples{
\donttest{\dontrun{
handlers(handler_void())
with_progress(y <- slow_sum(1:100))
print(y)
}}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slow_sum_p.R
\name{slow_sum_p}
\alias{slow_sum_p}
\alias{slow_sum}
\alias{slow_sqrt_p}
\alias{slow_sqrt}
\title{Slowly Calculate Sum or Square Roots of Elements}
\usage{
slow_sum_p(
  x,
  delay = getOption("progressr.demo.delay", 1),
  stdout = getOption("progressr.demo.stdout", FALSE),
  message = getOption("progressr.demo.message", TRUE),
  sticky = getOption("progressr.demo.sticky", TRUE),
  progress = getOption("progressr.demo.progress", TRUE)
)

slow_sum(
  x,
  delay = getOption("progressr.demo.delay", 1),
  stdout = FALSE,
  message = FALSE,
  sticky = FALSE,
  progress = FALSE
)

slow_sqrt_p(
  x,
  delay = getOption("progressr.demo.delay", 0.5),
  stdout = getOption("progressr.demo.stdout", FALSE),
  message = getOption("progressr.demo.message", TRUE),
  sticky = getOption("progressr.demo.sticky", TRUE),
  progress = getOption("progressr.demo.progress", TRUE)
)

slow_sqrt(
  x,
  delay = getOption("progressr.demo.delay", 0.5),
  stdout = FALSE,
  message = FALSE,
  sticky = FALSE,
  progress = FALSE
)
}
\arguments{
\item{x}{Numeric vector to sum or calculate square roots for.}

\item{delay}{Delay in seconds after each addition/calculation.}

\item{stdout}{If TRUE, then a text is outputted to the standard output
per element.}

\item{message}{If TRUE, then a message is outputted per element.}

\item{sticky}{If TRUE, then a "sticky" message is outputted every
ten element.}

\item{progress}{If TRUE, then a progressor is created, otherwise a void
progressor that does nothing is used.}
}
\value{
The sum or square roots of all elements in \code{x}.
}
\description{
Slowly Calculate Sum or Square Roots of Elements
}
\section{Progress updates}{

These functions signal \link{progression} conditions as they progress.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_ascii_alert.R
\name{handler_ascii_alert}
\alias{handler_ascii_alert}
\title{Progression Handler: Progress Reported as ASCII BEL Symbols (Audio or Blink) in the Terminal}
\usage{
handler_ascii_alert(
  symbol = "\\a",
  file = stderr(),
  intrusiveness = getOption("progressr.intrusiveness.audio", 5),
  target = c("terminal", "audio"),
  ...
)
}
\arguments{
\item{symbol}{(character string) The character symbol to be outputted,
which by default is the ASCII BEL character (\code{'\\a'} = \code{'\\007'}) character.}

\item{file}{(connection) A \link[base:connection]{base::connection} to where output should be sent.}

\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{\ldots}{Additional arguments passed to \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\value{
A function of class \code{progression_handler} that takes a
\link{progression} condition as its first and only argument.
}
\description{
A progression handler based on \code{cat("\\a", file=stderr())}.
}
\examples{
handlers("ascii_alert")
with_progress({ y <- slow_sum_p(1:10) })
print(y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_cli.R
\name{handler_cli}
\alias{handler_cli}
\title{Progression Handler: Progress Reported via 'cli' Progress Bars (Text) in the Terminal}
\usage{
handler_cli(
  show_after = 0,
  intrusiveness = getOption("progressr.intrusiveness.terminal", 1),
  target = "terminal",
  type = c("default", "steps", "percent", "time"),
  ...
)
}
\arguments{
\item{show_after}{(numeric) Number of seconds to wait before displaying
the progress bar.}

\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{type}{(character) The type of progress bar to display, which
controls the default \code{format} and \code{format_done} passed to \code{\link[cli:cli_progress_bar]{cli::cli_progress_bar()}}.
If \code{"default"}, the \pkg{cli} default format is used (\code{format = NULL}).
If \code{"steps"}, the progress bar shows the current and total number of
steps using the format string
\code{"{cli::pb_spin} {cli::pb_bar} {cli::pb_current}/{cli::pb_total} {cli::pb_status}"}
(and \code{format_done = "{cli::pb_bar} {cli::pb_current}/{cli::pb_total} {cli::pb_status}"}).
If \code{"percent"}, the progress bar shows the percentage completed using
the format string
\code{"{cli::pb_spin} {cli::pb_bar} {cli::pb_percent} {cli::pb_status}"}
(and \code{format_done = "{cli::pb_bar} {cli::pb_percent} {cli::pb_status}"}).
If \code{"time"}, the progress bar shows the percentage completed, the
current and total number of steps, the estimated time remaining (ETA),
and the total elapsed time using the format string
\code{"[{cli::pb_elapsed}] {cli::pb_spin} {cli::pb_bar} {cli::pb_percent} [{cli::pb_current}/{cli::pb_total}] (ETA: {cli::pb_eta}) {cli::pb_status}"}
(and \code{format_done = "[{cli::pb_elapsed}] {cli::pb_bar} {cli::pb_percent} [{cli::pb_current}/{cli::pb_total}] {cli::pb_status}"}).
For the meaning of these format variables, see
\link[cli:progress-variables]{Progress bar variables} in the \pkg{cli} package.
This argument is ignored if \code{format} is explicitly specified via \code{...}.}

\item{\ldots}{Additional arguments passed to \code{\link[cli:cli_progress_bar]{cli::cli_progress_bar()}}
and \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\value{
A function of class \code{progression_handler} that takes a
\link{progression} condition as its first and only argument.
}
\description{
A progression handler for \code{\link[cli:cli_progress_bar]{cli::cli_progress_bar()}}.
}
\section{Requirements}{

This progression handler requires the \pkg{cli} package.
}

\section{Appearance}{

Below are a few examples on how to use and customize this progress handler.
In all cases, we use \code{handlers(global = TRUE)}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(progressr)
handlers("cli")
y <- slow_sum_p(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_cli-default.svg}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(progressr)
handlers(handler_cli(type = "steps"))
y <- slow_sum_p(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_cli-type-steps.svg}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(progressr)
handlers(handler_cli(type = "percent"))
y <- slow_sum_p(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_cli-type-percent.svg}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(progressr)
handlers(handler_cli(type = "time"))
y <- slow_sum_p(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_cli-type-time.svg}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(progressr)
handlers(handler_cli(format = "\{cli::pb_spin\} \{cli::pb_bar\} \{cli::pb_current\}/\{cli::pb_total\} \{cli::pb_status\}"))
y <- slow_sum_p(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_cli-format-1.svg}}
}

\examples{
if (requireNamespace("cli", quietly = TRUE)) {
  handlers(handler_cli(format = "{cli::pb_spin} {cli::pb_bar} {cli::pb_percent} {cli::pb_status}"))
  with_progress({ y <- slow_sum_p(1:10) })
  print(y)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_rpushbullet.R
\name{handler_rpushbullet}
\alias{handler_rpushbullet}
\title{Progression Handler: Progress Reported via the Pushbullet Messaging Service}
\usage{
handler_rpushbullet(
  intrusiveness = getOption("progressr.intrusiveness.rpushbullet", 5),
  target = "gui",
  ...,
  title = "Progress update from R",
  recipients = NULL,
  email = NULL,
  channel = NULL,
  apikey = NULL,
  devices = NULL
)
}
\arguments{
\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{title}{The title of the note being posted.}

\item{recipients}{A character or numeric vector indicating the
devices this post should go to. If missing, the default device
is looked up from an optional setting, and if none has been set
the push is sent to all devices.}

\item{email}{An alternative way to specify a recipient is to specify
an email address. If both \code{recipients} and \code{email} are
present, \code{recipients} is used.}

\item{channel}{A channel tag used to specify the name of the channel
as the recipient. If either \code{recipients} or \code{email} are present,
they will take precedence over \code{channel}.}

\item{apikey}{The API key used to access the service. It can be
supplied as an argument here, via the global option
\code{rpushbullet.key}, or via the file \code{~/.rpushbullet.json}
which is read at package initialization (and, if found, also sets
the global option).}

\item{devices}{The device to which this post is pushed. It can be
supplied as an argument here, or via the file
\code{~/.rpushbullet.json} which is read at package
initialization.}

\item{\ldots}{Additional arguments passed to \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\description{
A progression handler for \code{\link[RPushbullet:pbPost]{RPushbullet::pbPost()}} of the \pkg{RPushbullet}
package.
}
\section{Requirements}{

This progression handler requires the \pkg{RPushbullet} package, a
Pushbullet account, and configuration according to the instructions
of the \pkg{RPushbullet} package.  It also requires internet access
from the computer where this progress handler is registered.
}

\examples{
pkg <- "RPushbullet"
if (requireNamespace(pkg, quietly = TRUE)) {

  handlers("rpushbullet")
  with_progress({ y <- slow_sum(1:10) })
  print(y)
  
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_pbcol.R
\name{handler_pbcol}
\alias{handler_pbcol}
\title{Progression Handler: Progress Reported as an ANSI Background Color in the Terminal}
\usage{
handler_pbcol(
  adjust = 0,
  pad = 1L,
  complete = function(s) cli::bg_blue(cli::col_white(s)),
  incomplete = function(s) cli::bg_cyan(cli::col_white(s)),
  intrusiveness = getOption("progressr.intrusiveness.terminal", 1),
  target = "terminal",
  ...
)
}
\arguments{
\item{adjust}{(numeric) The adjustment of the progress update,
where \code{adjust = 0} positions the message to the very left, and
\code{adjust = 1} positions the message to the very right.}

\item{pad}{(integer) Amount of padding on each side of the message,
where padding is done by spaces.}

\item{complete, incomplete}{(function) Functions that take "complete" and
"incomplete" strings that comprise the progress bar as input and annotate
them to reflect their two different parts.  The default is to annotation
them with two different background colors and the same foreground color
using the \pkg{cli} package.}

\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{\ldots}{Additional arguments passed to \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\description{
Progression Handler: Progress Reported as an ANSI Background Color in the Terminal
}
\section{Requirements}{

This progression handler requires the \pkg{cli} package.
}

\section{Appearance}{

Below are a few examples on how to use and customize this progress handler.
In all cases, we use \code{handlers(global = TRUE)}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{handlers("pbcol")
y <- slow_sum(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_pbcol-default.svg}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{handlers(handler_pbcol(adjust = 0.5))
y <- slow_sum(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_pbcol-adjust-mid.svg}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{handlers(handler_pbcol(
  adjust = 1,
  complete = function(s) cli::bg_red(cli::col_black(s)),
  incomplete = function(s) cli::bg_cyan(cli::col_black(s))
))
y <- slow_sum(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_pbcol-adjust-right-complete.svg}}
}

\examples{
handlers(handler_pbcol)
with_progress({ y <- slow_sum(1:10) })
print(y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_winprogressbar.R
\name{handler_winprogressbar}
\alias{handler_winprogressbar}
\title{Progression Handler: Progress Reported as a MS Windows Progress Bars in the GUI}
\usage{
handler_winprogressbar(
  intrusiveness = getOption("progressr.intrusiveness.gui", 1),
  target = "gui",
  inputs = list(title = NULL, label = "message"),
  ...
)
}
\arguments{
\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{inputs}{(named list) Specifies from what sources the MS Windows
progress elements 'title' and 'label' should be updated. Valid sources are
\code{"message"}, \code{"sticky_message"} and \code{"non_sticky_message"}, where
\code{"message"} is short for \code{c("non_sticky_message", "sticky_message")}. For
example, \code{inputs = list(title = "sticky_message", label = "message")}
will update the 'title' component from sticky messages only,
whereas the 'label' component is updated using any message.}

\item{\ldots}{Additional arguments passed to \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\description{
A progression handler for \code{winProgressBar()} in the \pkg{utils} package.
}
\section{Requirements}{

This progression handler requires MS Windows.
}

\examples{
\donttest{\dontrun{
handlers(handler_winprogressbar())
with_progress(y <- slow_sum(1:100))
}}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progressr-package.R
\docType{package}
\name{progressr}
\alias{progressr}
\alias{progressr-package}
\title{progressr: A Unifying API for Progress Updates}
\description{
The \pkg{progressr} package provides a minimal, unifying API for scripts
and packages to report progress updates from anywhere including when
using parallel processing.
}
\details{
The package is designed such that \emph{the developer} can to focus on \emph{what}
progress should be reported on without having to worry about \emph{how} to
present it.

The \emph{end user} has full control of \emph{how}, \emph{where}, and \emph{when} to render
these progress updates.  For instance, they can chose to report progress
in the terminal using \code{\link[utils:txtProgressBar]{utils::txtProgressBar()}} (default) or
\code{\link[cli:cli_progress_bar]{cli::cli_progress_bar()}}, via the R graphical user interface (GUI)
using \code{utils::winProgressBar()} or \code{\link[tcltk:tkProgressBar]{tcltk::tkProgressBar()}}, or
via the RStudio GUI using \code{rstudioapi::jobSetProgress()}.
An alternative to above visual rendering, is to report progress as audio
using \code{\link[beepr:beep]{beepr::beep()}}.
It is also possible to use a combination of above progression handlers,
e.g. a progress bar in the terminal together with audio updates.
Besides the existing handlers, it is possible to develop custom
progression handlers.

The \pkg{progressr} package is compatible with \pkg{shiny} applications
and \pkg{knitr} rendering.

The \pkg{progressr} package uses R's condition framework for signaling
progress updated. Because of this, progress can be reported from almost
anywhere in R, e.g. from classical for and while loops, from map-reduce
APIs like the \code{\link[=lapply]{lapply()}} family of functions, \pkg{purrr}, \pkg{plyr}, and
\pkg{foreach}.
The \pkg{progressr} package will also work with parallel processing via
the \pkg{future} framework, e.g. \code{\link[future.apply:future_lapply]{future.apply::future_lapply()}},
\code{\link[furrr:future_map]{furrr::future_map()}}, and \code{\link[foreach:foreach]{foreach::foreach()}} with \pkg{doFuture}.
}
\section{Progression Handlers}{


In the terminal:
\itemize{
\item \link{handler_txtprogressbar} (built-in; default)
\item \link{handler_pbcol} (built-in)
\item \link{handler_ascii_alert} (built-in)
\item \link{handler_pbmcapply} (requires the \pkg{pbmcapply} package)
\item \link{handler_progress} (requires the \pkg{progress} package)
}

In a graphical user interface (GUI):
\itemize{
\item \link{handler_rstudio} (only in the RStudio Console)
\item \link{handler_tkprogressbar} (built-in)
\item \link{handler_winprogressbar} (built-in)
}

As sound:
\itemize{
\item \link{handler_ascii_alert} (built-in)
\item \link{handler_beepr} (requires the \pkg{beepr} package)
}

Via the file system:
\itemize{
\item \link{handler_filesize} (built-in)
}

In Shiny:
\itemize{
\item \link{withProgressShiny} (requires the \pkg{shiny} package)
}

Via notification systems:
\itemize{
\item \link{handler_ntfy} (requires the \pkg{ntfy} package)
\item \link{handler_notifier} (requires the non-CRAN \pkg{notifier} package)
\item \link{handler_rpushbullet} (requires the \pkg{RPushBullet} package)
}

Miscellaneous:
\itemize{
\item \link{handler_debug} (built-in)
\item \link{handler_newline} (built-in)
\item \link{handler_slowdown} (built-in)
\item \link{handler_void} (built-in)
}
}

\examples{
library(progressr)

xs <- 1:5

with_progress({
  p <- progressor(along = xs)
  y <- lapply(xs, function(x) {
    Sys.sleep(0.1)
    p(sprintf("x=\%g", x))
    sqrt(x)
  })
})
}
\seealso{
Useful links:
\itemize{
  \item \url{https://progressr.futureverse.org}
  \item \url{https://github.com/futureverse/progressr}
  \item Report bugs at \url{https://github.com/futureverse/progressr/issues}
}

}
\author{
\strong{Maintainer}: Henrik Bengtsson \email{henrikb@braju.com} (\href{https://orcid.org/0000-0002-7579-5165}{ORCID}) [copyright holder]

}
\keyword{iteration}
\keyword{programming}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_shiny.R
\name{handler_shiny}
\alias{handler_shiny}
\title{Progression Handler: Progress Reported via 'shiny' Widgets (GUI) in the HTML Browser}
\usage{
handler_shiny(
  intrusiveness = getOption("progressr.intrusiveness.gui", 1),
  target = "gui",
  inputs = list(message = NULL, detail = "message"),
  enable = getOption("progressr.enable", TRUE),
  ...
)
}
\arguments{
\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{inputs}{(named list) Specifies from what sources the Shiny progress
elements 'message' and 'detail' should be updated.  Valid sources are
\code{"message"}, \code{"sticky_message"} and \code{"non_sticky_message"}, where
\code{"message"} is short for \code{c("non_sticky_message", "sticky_message")}. For
example, \code{inputs = list(message = "sticky_message", detail = "message")}
will update the Shiny 'message' component from sticky messages only,
whereas the 'detail' component is updated using any message.}

\item{enable}{(logical) If FALSE, then progress is not reported.}

\item{\ldots}{Additional arguments passed to \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\description{
A progression handler for \pkg{shiny} and \code{\link[shiny:withProgress]{shiny::withProgress()}}.
}
\details{
For most Shiny application there is little need to use this Shiny handler
directly.  Instead, it is sufficient to use \code{\link[=withProgressShiny]{withProgressShiny()}}.
}
\section{Requirements}{

This progression handler requires the \pkg{shiny} package.
}

\examples{
\donttest{\dontrun{
handlers(handler_shiny())
with_progress(y <- slow_sum_p(1:100))
}}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/global_progression_handler.R
\name{register_global_progression_handler}
\alias{register_global_progression_handler}
\title{Add or Remove a Global 'progression' Handler}
\usage{
register_global_progression_handler(action = c("add", "remove", "query"))
}
\arguments{
\item{action}{(character string)
If \code{"add"}, a global handler is added.
If \code{"remove"}, it is removed, if it exists.
If \code{"query"}, checks whether a handler is registered or not.}
}
\value{
Returns TRUE if a handler is registered, otherwise FALSE.
If \code{action = "query"}, the value is visible, otherwise invisible.
}
\description{
Add or Remove a Global 'progression' Handler
}
\section{Requirements}{

This function requires R (>= 4.0.0) - the version in which global calling
handlers where introduces.
}

\examples{
\dontshow{if (getRversion() >= "4.0.0" && !is.element("pkgdown", loadedNamespaces()))}
handlers(global = TRUE)

## This renders progress updates for each of the three calls slow_sum()
for (ii in 1:3) {
  xs <- seq_len(ii + 3)
  message(sprintf("\%d. slow_sum()", ii))
  y <- slow_sum(xs, stdout = TRUE, message = TRUE)
  print(y)
}

\dontshow{if (getRversion() >= "4.0.0" && !is.element("pkgdown", loadedNamespaces()))}
handlers(global = FALSE)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_beepr.R
\name{handler_beepr}
\alias{handler_beepr}
\title{Progression Handler: Progress Reported as 'beepr' Sounds (Audio)}
\usage{
handler_beepr(
  initiate = 2L,
  update = 10L,
  finish = 11L,
  interrupt = 9L,
  intrusiveness = getOption("progressr.intrusiveness.audio", 5),
  target = "audio",
  ...
)
}
\arguments{
\item{initiate, update, finish, interrupt}{(integer) Indices of \code{\link[beepr:beep]{beepr::beep()}}
sounds to play when progress starts, is updated, completes, or is
interrupted.  For silence, use \code{NA_integer_}.}

\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{\ldots}{Additional arguments passed to \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\value{
A function of class \code{progression_handler} that takes a
\link{progression} condition as its first and only argument.
}
\description{
A progression handler for \code{\link[beepr:beep]{beepr::beep()}}.
}
\section{Requirements}{

This progression handler requires the \pkg{beepr} package.
}

\examples{
if (requireNamespace("beepr", quietly = TRUE)) {

  handlers("beepr")
  with_progress({ y <- slow_sum_p(1:10) })
  print(y)
  
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/global_progression_handler.R
\name{global_progression_handler}
\alias{global_progression_handler}
\title{A Global Calling Handler For 'progression':s}
\usage{
global_progression_handler(condition)
}
\arguments{
\item{condition}{A logical scalar or a condition object.}
}
\value{
Nothing.
}
\description{
A Global Calling Handler For 'progression':s
}
\section{Requirements}{

This function requires R (>= 4.0.0) - the version in which global calling
handlers where introduces.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_debug.R
\name{handler_debug}
\alias{handler_debug}
\title{Progression Handler: Progress Reported as Debug Information (Text) in the Terminal}
\usage{
handler_debug(
  interval = getOption("progressr.interval", 0),
  intrusiveness = getOption("progressr.intrusiveness.debug", 0),
  target = "terminal",
  uuid = FALSE,
  ...
)
}
\arguments{
\item{interval}{(numeric) The minimum time (in seconds) between
successive progression updates from this handler.}

\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{uuid}{If TRUE, then the progressor UUID and the owner UUID are shown,
otherwise not (default).}

\item{\ldots}{Additional arguments passed to \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\value{
A function of class \code{progression_handler} that takes a
\link{progression} condition as its first and only argument.
}
\description{
Progression Handler: Progress Reported as Debug Information (Text) in the Terminal
}
\section{Appearance}{

Below is how this progress handler renders by default at 0\%, 30\% and 99\%
progress:

With \code{handlers(handler_debug())}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{[21:27:11.236] (0.000s => +0.001s) initiate: 0/100 (+0) '' \{clear=TRUE, enabled=TRUE, status=\}
[21:27:11.237] (0.001s => +0.000s) update: 0/100 (+0) 'Starting' \{clear=TRUE, enabled=TRUE, status=\}
[21:27:14.240] (3.004s => +0.002s) update: 30/100 (+30) 'Importing' \{clear=TRUE, enabled=TRUE, status=\}
[21:27:16.245] (5.009s => +0.001s) update: 100/100 (+70) 'Summarizing' \{clear=TRUE, enabled=TRUE, status=\}
[21:27:16.246] (5.010s => +0.003s) update: 100/100 (+0) 'Summarizing' \{clear=TRIE, enabled=TRUE, status=\}
}\if{html}{\out{</div>}}
}

\examples{
handlers("debug")
with_progress({ y <- slow_sum_p(1:10) })
print(y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{progressr.options}
\alias{progressr.options}
\alias{progressr.clear}
\alias{progressr.debug}
\alias{progressr.demo.delay}
\alias{progressr.demo.stdout}
\alias{progressr.demo.message}
\alias{progressr.demo.sticky}
\alias{progressr.demo.progress}
\alias{progressr.delay_stdout}
\alias{progressr.delay_conditions}
\alias{progressr.enable}
\alias{progressr.enable_after}
\alias{progressr.interrupts}
\alias{progressr.interval}
\alias{progressr.intrusiveness}
\alias{progressr.intrusiveness.audio}
\alias{progressr.intrusiveness.debug}
\alias{progressr.intrusiveness.file}
\alias{progressr.intrusiveness.gui}
\alias{progressr.intrusiveness.notifier}
\alias{progressr.intrusiveness.terminal}
\alias{progressr.handlers}
\alias{progressr.times}
\title{Options and environment variables used by the 'progressr' packages}
\description{
Below are environment variables and \R options that are used by the
\pkg{progressr} package.
Below are all \R options that are currently used by the \pkg{progressr} package.\cr
\cr
\emph{WARNING: Note that the names and the default values of these options may change in future versions of the package.  Please use with care until further notice.}
}
\section{Options for controlling progression reporting}{


\describe{
\item{\option{progressr.handlers}:}{
(function or list of functions)
Zero or more progression handlers that will report on any progression updates.  If empty list, progress updates are ignored.  If NULL, the default (\code{handler_txtprogressbar}) progression handlers is used.  The recommended way to set this option is via \code{\link[=handlers]{handlers()}}. (Default: NULL)
}
}
}

\section{Options for controlling progression handlers}{


\describe{
\item{\option{progressr.clear}:}{
(logical)
If TRUE, any output, typically visual, produced by a reporter will be cleared/removed upon completion, if possible. (Default: TRUE)
}

\item{\option{progressr.enable}:}{
(logical)
If FALSE, then progress is not reported.
(Default: TRUE in interactive mode, otherwise FALSE)
}

\item{\option{progressr.enable_after}:}{
(numeric)
Delay (in seconds) before progression updates are reported.
(Default: \code{0.0})
}

\item{\option{progressr.times}:}{
(numeric)
The maximum number of times a handler should report progression updates. If zero, then progress is not reported.
(Default: \code{+Inf})
}

\item{\option{progressr.interval}:}{
(numeric)
The minimum time (in seconds) between successive progression updates from this handler.
(Default: \code{0.0})
}

\item{\option{progressr.intrusiveness}:}{
(numeric)
A non-negative scalar on how intrusive (disruptive) the reporter to the user. This multiplicative scalar applies to the \emph{interval} and \emph{times} parameters. (Default: \code{1.0})\cr

\describe{
\item{\option{progressr.intrusiveness.audio}:}{(numeric) intrusiveness for auditory progress handlers (Default: \code{5.0})}
\item{\option{progressr.intrusiveness.file}:}{(numeric) intrusiveness for file-based progress handlers (Default: \code{5.0})}
\item{\option{progressr.intrusiveness.gui}:}{(numeric) intrusiveness for graphical-user-interface progress handlers (Default: \code{1.0})}
\item{\option{progressr.intrusiveness.notifier}:}{(numeric) intrusiveness for progress handlers that creates notifications (Default: \code{10.0})}
\item{\option{progressr.intrusiveness.terminal}:}{(numeric) intrusiveness for progress handlers that outputs to the terminal (Default: \code{1.0})}
\item{\option{progressr.intrusiveness.debug}:}{(numeric) intrusiveness for "debug" progress handlers (Default: \code{0.0})}
}
}
}
}

\section{Options for controlling how standard output and conditions are relayed}{


\describe{
\item{\option{progressr.delay_conditions}:}{
(character vector)
condition classes to be captured and relayed at the end after any captured standard output is relayed. (Default: \code{c("condition")})
}

\item{\option{progressr.delay_stdout}:}{
(logical)
If TRUE, standard output is captured and relayed at the end just before any captured conditions are relayed. (Default: TRUE)
}
}
}

\section{Options for controlling interrupts}{


\describe{
\item{\option{progressr.interrupts}:}{
(logical)
Controls whether interrupts should be detected or not.
If FALSE, then interrupts are not detected and progress information
is generated. (Default: \code{TRUE})
}

\item{\option{progressr.delay_stdout}:}{
(logical)
If TRUE, standard output is captured and relayed at the end just before any captured conditions are relayed. (Default: TRUE)
}
}
}

\section{Options for debugging progression updates}{


\describe{
\item{\option{progressr.debug}:}{(logical) If TRUE, extensive debug messages are generated. (Default: FALSE)}
}
}

\section{Options for progressr examples and demos}{


\describe{
\item{\option{progressr.demo.delay}:}{(numeric) Delay (in seconds) between each iteration of \code{\link[=slow_sum_p]{slow_sum_p()}} and \code{\link[=slow_sqrt_p]{slow_sqrt_p()}}. (Default: \code{1.0} and \code{0.5})}

\item{\option{progressr.demo.stdout}:}{(logical) If TRUE, then a text is outputted to the standard output per element. (Default: \code{FALSE})}

\item{\option{progressr.demo.message}:}{(logical) If TRUE, then a message is outputted per element. (Default: \code{TRUE})}

\item{\option{progressr.demo.sticky}:}{(logical) If TRUE, then a "sticky" message is outputted every ten element. (Default: \code{TRUE})}

\item{\option{progressr.demo.progress}:}{(logical) If TRUE, then a progressor is created, otherwise a void progressor that does nothing is used. (Default: \code{TRUE})}
}
}

\section{Environment variables that set R options}{

Some of the above \R \option{progressr.*} options can be set by corresponding
environment variable \env{R_PROGRESSR_*} \emph{when the \pkg{progressr} package
is loaded}.
For example, if \code{R_PROGRESSR_ENABLE = "true"}, then option
\option{progressr.enable} is set to \code{TRUE} (logical).
For example, if \code{R_PROGRESSR_ENABLE_AFTER = "2.0"}, then option
\option{progressr.enable_after} is set to \code{2.0} (numeric).
}

\seealso{
To set \R options when \R starts (even before the \pkg{progressr} package is loaded), see the \link[base]{Startup} help page.  The \href{https://cran.r-project.org/package=startup}{\pkg{startup}} package provides a friendly mechanism for configuring \R at startup.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress_aggregator.R
\name{progress_aggregator}
\alias{progress_aggregator}
\title{Aggregate Progression Conditions}
\usage{
progress_aggregator(progress)
}
\arguments{
\item{progress}{A \link{progressor} function.}
}
\value{
A function of class \code{progress_aggregator}.
}
\description{
Aggregate Progression Conditions
}
\examples{
library(progressr)

message("progress_aggregator() ...")

with_progress({
  progress <- progressor(steps = 4L)
  relay_progress <- progress_aggregator(progress)
  progress()
  relay_progress(slow_sum(1:3))
  relay_progress(slow_sum(1:10))
  progress()
})

message("progress_aggregator() ... done")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_brrr.R
\name{handler_brrr}
\alias{handler_brrr}
\title{Progression Handler: Progress Reported as 'BRRR' Sounds (Audio)}
\usage{
handler_brrr(
  initiate = "takeoff",
  update = "bigsean4",
  finish = "khaled3",
  interrupt = "busta",
  intrusiveness = getOption("progressr.intrusiveness.audio", 5),
  target = "audio",
  ...
)
}
\arguments{
\item{initiate, update, finish, interrupt}{(integer or character) Indices or
names of \code{skrrrahh()} sounds to play when progress starts, is updated,
completes, or is interrupted. For silence, use \code{NA_integer_}.}

\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{\ldots}{Additional arguments passed to \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\value{
A function of class \code{progression_handler} that takes a
\link{progression} condition as its first and only argument.
}
\description{
A progression handler for \code{skrrrahh()} of the \pkg{BRRR} package.
}
\section{Requirements}{

This progression handler requires the \pkg{BRRR} package, which is only
available from \url{https://github.com/brooke-watson/BRRR}.  This can be
installed as \code{remotes::install_github("brooke-watson/BRRR")}.
}

\examples{
pkg <- "BRRR"
if (requireNamespace(pkg, quietly = TRUE)) {

  handlers("brrr")
  with_progress({ y <- slow_sum_p(1:10) })
  print(y)
  
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress.R
\name{progress}
\alias{progress}
\title{Creates and Signals a Progression Condition}
\usage{
progress(..., call = sys.call())
}
\arguments{
\item{call}{(expression) A call expression.}

\item{\ldots}{Arguments pass to \code{\link[=progression]{progression()}}.}
}
\value{
A \link[base:conditions]{base::condition} of class \code{progression}.
}
\description{
\emph{WARNING:} \code{progress()} is defunct - don't use.
}
\seealso{
To create a progression condition, use \code{\link[=progression]{progression()}}.
To signal a progression condition, use \code{\link[base:conditions]{base::signalCondition()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_tutorial.R
\name{handler_tutorial}
\alias{handler_tutorial}
\title{A progressr handler useful for Futureverse tutorials}
\usage{
handler_tutorial(
  ...,
  spec = c("progress", "cli"),
  type = "time",
  clear = FALSE
)
}
\arguments{
\item{spec}{The progress handler to use.}

\item{type}{Controls which predefined handler format string to use.}

\item{clear}{(logical) If TRUE, any output, typically visual, produced
by a reporter will be cleared/removed upon completion, if possible.}

\item{\ldots}{Arguments passed to the selected progress handler.}
}
\value{
A function of class \code{progression_handler} that takes a
\link{progression} condition as its first and only argument.
}
\description{
A progressr handler useful for Futureverse tutorials
}
\details{
The default is to use the \pkg{progress} package. Contrary to the
\pkg{cli} package, the \pkg{progress} package has the capacity to
display \emph{all} progress updates, even when receiving progress updates
at a high rate from, for example, parallel processing.
}
\examples{
# Display progress and keep the progress output when done
handlers("tutorial")

# Display progress and keep every progress reported
handlers("tutorial", "newline")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slow_sum.R
\name{slow_sum}
\alias{slow_sum}
\title{Slowly Calculate Sum of Elements}
\usage{
slow_sum(
  x,
  delay = getOption("progressr.demo.delay", 1),
  stdout = FALSE,
  message = TRUE,
  sticky = TRUE
)
}
\arguments{
\item{x}{Numeric vector to sum}

\item{delay}{Delay in seconds after each addition.}

\item{stdout}{If TRUE, then a text is outputted to the standard output
per element.}

\item{message}{If TRUE, then a message is outputted per element.}

\item{sticky}{If TRUE, then a "sticky" message is outputted every
ten element.}
}
\value{
The sum of all elements in \code{x}.
}
\description{
Slowly Calculate Sum of Elements
}
\section{Progress updates}{

This function signals \link{progression} conditions as it progresses.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handlers.R
\name{handlers}
\alias{handlers}
\title{Control How Progress is Reported}
\usage{
handlers(
  ...,
  append = FALSE,
  on_missing = c("error", "warning", "ignore"),
  default = handler_txtprogressbar,
  global = NULL
)
}
\arguments{
\item{append}{(logical) If FALSE, the specified progression handlers
replace the current ones, otherwise they are appended to them.}

\item{on_missing}{(character) If \code{"error"}, an error is thrown if one of
the progression handlers does not exist.  If \code{"warning"}, a warning
is produced and the missing handler is ignored.  If \code{"ignore"}, the
missing handler is ignored.}

\item{default}{The default progression calling handler to use if none
are set.}

\item{global}{If TRUE, then the global progression handler is enabled.
If FALSE, it is disabled.  If NA, then TRUE is returned if it is enabled,
otherwise FALSE.}

\item{\ldots}{One or more progression handlers.  Alternatively, this
function also accepts a single vector of progression handlers as input.
If this vector is empty, then an empty set of progression handlers will
be set.}
}
\value{
(invisibly) the previous list of progression handlers set.
If no arguments are specified, then the current set of progression
handlers is returned.
If \code{global} is specified, then TRUE is returned if the global progression
handler is enabled, otherwise FALSE.
If both handlers and \code{global} are specified, then the previous list of
progression handlers is returned.
}
\description{
Control How Progress is Reported
}
\details{
This function provides a convenient alternative for getting and setting
option \option{progressr.handlers}.
}
\section{For package developers}{

\strong{IMPORTANT: Setting progression handlers is a privilege that should be
left to the end user. It should not be used by R packages, whose only task
is to \emph{signal} progress updates, not to decide if, when, and how progress
should be reported.}

If you have to set or modify the progression handlers inside a function,
please make sure to undo the settings afterward.  If not, you will break
whatever progression settings the user already has for other purposes
used elsewhere.  To undo your settings, you can do:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{old_handlers <- handlers(c("beepr", "progress"))
on.exit(handlers(old_handlers), add = TRUE)
}\if{html}{\out{</div>}}
}

\section{Configuring progression handling during R startup}{

A convenient place to configure the default progression handler and to
enable global progression reporting by default is in the \file{~/.Rprofile}
startup file.  For example, the following will (i) cause your interactive
R session to use global progression handler by default, and (ii) report
progress via the \pkg{progress} package when in the terminal and via the
RStudio Jobs progress bar when in the RStudio Console.
To configure this, add
the following to your \file{~/.Rprofile} startup file:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{if (interactive() && requireNamespace("progressr", quietly = TRUE)) \{
  ## Enable global progression updates
  if (getRversion() >= 4) progressr::handlers(global = TRUE)

  ## In RStudio Console, or not?
  if (Sys.getenv("RSTUDIO") == "1" && !nzchar(Sys.getenv("RSTUDIO_TERM"))) \{
    options(progressr.handlers = progressr::handler_rstudio)
  \} else \{
    options(progressr.handlers = progressr::handler_progress)
  \}
\}
}\if{html}{\out{</div>}}
}

\examples{
handlers("txtprogressbar")
if (requireNamespace("beepr", quietly = TRUE))
  handlers("beepr", append = TRUE)

with_progress({ y <- slow_sum_p(1:5) })
print(y)


if (getRversion() >= "4.0.0") {
  \dontshow{if (!is.element("pkgdown", loadedNamespaces()))}
  handlers("txtprogressbar", global = TRUE)
  y <- slow_sum_p(1:4)
  z <- slow_sum_p(6:9)
  \dontshow{if (!is.element("pkgdown", loadedNamespaces()))}
  handlers(global = FALSE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/with_progress.R, R/without_progress.R
\name{with_progress}
\alias{with_progress}
\alias{without_progress}
\title{Report on Progress while Evaluating an R Expression}
\usage{
with_progress(
  expr,
  handlers = progressr::handlers(),
  cleanup = TRUE,
  delay_terminal = NULL,
  delay_stdout = NULL,
  delay_conditions = NULL,
  interrupts = getOption("progressr.interrupts", TRUE),
  interval = NULL,
  enable = NULL
)

without_progress(expr)
}
\arguments{
\item{expr}{An \R expression to evaluate.}

\item{handlers}{A progression handler or a list of them.
If NULL or an empty list, progress updates are ignored.}

\item{cleanup}{If TRUE, all progression handlers will be shutdown
at the end regardless of whether the progression is complete or not.}

\item{delay_terminal}{If TRUE, output and conditions that may end up in
the terminal will be delayed.}

\item{delay_stdout}{If TRUE, standard output is captured and relayed
at the end just before any captured conditions are relayed.}

\item{delay_conditions}{A character vector specifying \link[base:condition]{base::condition}
classes to be captured and relayed at the end after any captured
standard output is relayed.}

\item{interrupts}{Controls whether interrupts should be detected or not.
If TRUE and an interrupt is signaled, progress handlers are asked to
report on the current amount of progress when the evaluation was terminated
by the interrupt, e.g. when a user pressed Ctrl-C in an interactive session,
or a batch process was interrupted because it ran out of time.
Note that it's optional for a progress handler to support this and only
some do.}

\item{interval}{(numeric) The minimum time (in seconds) between
successive progression updates from handlers.}

\item{enable}{(logical) If FALSE, then progress is not reported.  The
default is to report progress in interactive mode but not batch mode.
See below for more details.}
}
\value{
Returns the value of the expression.
}
\description{
Report on Progress while Evaluating an R Expression
}
\details{
If you are writing a Shiny app, use the \code{\link[=withProgressShiny]{withProgressShiny()}} function
instead of this one.

If the global progression handler is enabled, it is temporarily disabled
while evaluating the \code{expr} expression.

If a \code{progression} condition causes the progressor to be completed
("reaches 100\%"), then \code{with_progress()} will muffle the \code{progression}
condition preventing it from reaching, say, any global condition
handlers.

\strong{IMPORTANT: This function is meant for end users only.  It should not
be used by R packages, whose only task is to \emph{signal} progress updates,
not to decide if, when, and how progress should be reported.}

\code{without_progress()} evaluates an expression while ignoring all
progress updates.
}
\section{Progression handler functions}{

Formally, progression handlers are calling handlers that are called
when a \link{progression} condition is signaled.  These handlers are functions
that take one argument which is the \link{progression} condition.
}

\section{Progress updates in batch mode}{

When running R from the command line, R runs in a non-interactive mode
(\code{interactive()} returns \code{FALSE}).  The default behavior of
\code{with_progress()} is to \emph{not} report on progress in non-interactive mode.
To have progress being reported also then, set R options
\option{progressr.enable} or environment variable \env{R_PROGRESSR_ENABLE}
to \code{TRUE}.  Alternatively, one can set argument \code{enable=TRUE} when calling
\code{with_progress()}.  For example,

\if{html}{\out{<div class="sourceCode sh">}}\preformatted{$ Rscript -e "library(progressr)" -e "with_progress(slow_sum_p(1:5))"
}\if{html}{\out{</div>}}

will \emph{not} report on progress, whereas:

\if{html}{\out{<div class="sourceCode sh">}}\preformatted{$ export R_PROGRESSR_ENABLE=TRUE
$ Rscript -e "library(progressr)" -e "with_progress(slow_sum_p(1:5))"
}\if{html}{\out{</div>}}

will.
}

\examples{
## The slow_sum_p() example function
slow_sum_p <- progressr::slow_sum_p
print(slow_sum_p)

x <- 1:10

## Without progress updates
y <- slow_sum_p(x)


## Progress reported via txtProgressBar (default)
handlers("txtprogressbar")  ## default
with_progress({
  y <- slow_sum_p(x)
})

## Progress reported via tcltk::tkProgressBar
if (capabilities("tcltk") && requireNamespace("tcltk", quietly = TRUE)) {
  handlers("tkprogressbar")
  with_progress({
    y <- slow_sum_p(x)
  })
}

## Progress reported via progress::progress_bar)
if (requireNamespace("progress", quietly = TRUE)) {
  handlers("progress")
  with_progress({
    y <- slow_sum_p(x)
  })
}

## Progress reported via txtProgressBar and beepr::beep
if (requireNamespace("beepr", quietly = TRUE)) {
  handlers("beepr", "txtprogressbar")
  with_progress({
    y <- slow_sum_p(x)
  })
}

## Progress reported via customized utils::txtProgressBar and beepr::beep,
## if available.
handlers(handler_txtprogressbar(style = 3L))
if (requireNamespace("beepr", quietly = TRUE)) {
  handlers("beepr", append = TRUE)
}

with_progress({
  y <- slow_sum_p(1:30)
})
}
\seealso{
For Shiny apps, use \code{\link[=withProgressShiny]{withProgressShiny()}} instead of this function.
Internally, this function is built around \code{\link[base:withCallingHandlers]{base::withCallingHandlers()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_progress.R
\name{handler_progress}
\alias{handler_progress}
\title{Progression Handler: Progress Reported via 'progress' Progress Bars (Text) in the Terminal}
\usage{
handler_progress(
  format = ":spin [:bar] :percent :message",
  show_after = 0,
  intrusiveness = getOption("progressr.intrusiveness.terminal", 1),
  target = "terminal",
  ...
)
}
\arguments{
\item{format}{(character string) The format of the progress bar.}

\item{show_after}{(numeric) Number of seconds to wait before displaying
the progress bar.}

\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{\ldots}{Additional arguments passed to \code{\link[progress:progress_bar]{progress::progress_bar()}}
and \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\description{
A progression handler for \code{\link[progress:progress_bar]{progress::progress_bar()}}.
}
\section{Requirements}{

This progression handler requires the \pkg{progress} package.
}

\section{Appearance}{

Below are a few examples on how to use and customize this progress handler.
In all cases, we use \code{handlers(global = TRUE)}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{handlers("progress")
y <- slow_sum(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_progress-default.svg}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{handlers(handler_progress(complete = "#"))
y <- slow_sum(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_progress-complete.svg}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{handlers(handler_progress(format = ":spin [:bar] :percent :message"))
y <- slow_sum(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_progress-format-1.svg}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{handlers(handler_progress(format = ":percent [:bar] :eta :message"))
y <- slow_sum(1:25)
}\if{html}{\out{</div>}}

\if{html}{\figure{handler_progress-format-2.svg}}
}

\examples{
if (requireNamespace("progress", quietly = TRUE)) {

  handlers(handler_progress(format = ":spin [:bar] :percent :message"))
  with_progress({ y <- slow_sum(1:10) })
  print(y)
  
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_filesize.R
\name{handler_filesize}
\alias{handler_filesize}
\title{Progression Handler: Progress Reported as the Size of a File on the File System}
\usage{
handler_filesize(
  file = "default.progress",
  intrusiveness = getOption("progressr.intrusiveness.file", 5),
  target = "file",
  enable = getOption("progressr.enable", TRUE),
  ...
)
}
\arguments{
\item{file}{(character) A filename.}

\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{enable}{(logical) If FALSE, then progress is not reported.}

\item{\ldots}{Additional arguments passed to \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\value{
A function of class \code{progression_handler} that takes a
\link{progression} condition as its first and only argument.
}
\description{
Progression Handler: Progress Reported as the Size of a File on the File System
}
\details{
This progression handler reports progress by updating the size of a file
on the file system. This provides a convenient way for an R script running
in batch mode to report on the progress such that the user can peek at the
file size (by default in 0-100 bytes) to assess the amount of the progress
made, e.g. \verb{ls -l -- *.progress}.
If the \file{*.progress} file is accessible via for instance SSH, SFTP,
FTPS, HTTPS, etc., then progress can be assessed from a remote location.
}
\examples{
\donttest{\dontrun{
handlers(handler_filesize(file = "myscript.progress"))
with_progress(y <- slow_sum_p(1:100))
print(y)
}}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_progression_handler.R
\name{make_progression_handler}
\alias{make_progression_handler}
\title{Creates a Progression Calling Handler}
\usage{
make_progression_handler(
  name,
  reporter = list(),
  handler = NULL,
  enable = getOption("progressr.enable", interactive()),
  enable_after = getOption("progressr.enable_after", 0),
  times = getOption("progressr.times", +Inf),
  interval = getOption("progressr.interval", 0),
  intrusiveness = 1,
  clear = getOption("progressr.clear", TRUE),
  target = "terminal",
  ...
)
}
\arguments{
\item{name}{(character) Name of progression handler.}

\item{reporter}{(list) A named list of reporter functions.}

\item{handler}{(function) Function take a \link{progression} condition
as the first argument.}

\item{enable}{(logical) If FALSE, then progress is not reported.}

\item{enable_after}{(numeric) Delay (in seconds) before progression
updates are reported.}

\item{times}{(numeric) The maximum number of times this handler
should report progression updates.
If zero, then progress is not reported.}

\item{interval}{(numeric) The minimum time (in seconds) between
successive progression updates from this handler.}

\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{clear}{(logical) If TRUE, any output, typically visual, produced
by a reporter will be cleared/removed upon completion, if possible.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{\ldots}{Not used.}
}
\value{
A function of class \code{progression_handler} that takes a
\link{progression} condition as its first and only argument.
}
\description{
A progression calling handler is a function that takes a \link[base:conditions]{base::condition}
as its first argument and that can be use together with
\code{\link[base:conditions]{base::withCallingHandlers()}}.  This function helps creating such
progression calling handler functions.
}
\details{
The inner details of progression handlers and how to use this function
are still to be documented.  Until then, see the source code of existing
handlers for how it is used, e.g. \code{progressr::handler_txtprogressbar}.
Please use with care as things might change.
}
\section{Reporter functions}{

The \code{reporter} argument should be a named list of zero or more of the
following functions:
\itemize{
\item \code{initiate}
\item \code{update}
\item \code{finish}
}

These functions are called whenever a \link{progression} condition of type
\code{"initiate"}, \code{"update"}, or \code{"finish"} are received, but only if the
condition is for the progression that is currently handled.
These functions are called with the following arguments (in order):
\itemize{
\item \code{config} - a named list of the configuration of the progression handler:
\code{max_steps} (integer),
\code{interval} (numeric),
\code{enable_after} (numeric),
\code{auto_finish} (logical),
\code{clear} (logical),
\code{target} (character vector)
\item \code{state} - a named list of the current progress state after accounting
for the most recent \code{progression} condition:
\code{step} (integer), \code{message} (character),
\code{delta} (integer),
\code{enabled} (logical),
\code{timestamps} (POSIXct vector)
\item \code{progression} - a \link{progression} condition
\item \ldots - not used (reserved for future needs)
}

In addition to the above functions, the following functions:
\itemize{
\item \code{hide}
\item \code{unhide}
\item \code{reset}
\item \code{interrupt}
}

are called whenever the handler "should" hide or unhide the rendered
progress, or reset it, or when an interrupt is detected.  In these cases,
the \code{progression} argument is of class \code{control_progression}.
}

\examples{
## Create a progression handler that reports on the current progress
## step, the relative change, and the current progress message. This
## is only reported on positive progressions updated
my_handler <- make_progression_handler(name = "my", reporter = list(
  update = function(config, state, progression, ...) {
    if (progression$amount > 0) {
      message(sprintf("step = \%d (+\%g): message = \%s",
                      state$step,
                      progression$amount,
                      sQuote(state$message)))
    }
  }
))

handlers(my_handler)

with_progress({
  y <- slow_sum(1:5)
})

}
\seealso{
\code{\link[base:conditions]{base::withCallingHandlers()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handler_tkprogressbar.R
\name{handler_tkprogressbar}
\alias{handler_tkprogressbar}
\title{Progression Handler: Progress Reported as a Tcl/Tk Progress Bars in the GUI}
\usage{
handler_tkprogressbar(
  intrusiveness = getOption("progressr.intrusiveness.gui", 1),
  target = "terminal",
  inputs = list(title = NULL, label = "message"),
  ...
)
}
\arguments{
\item{intrusiveness}{(numeric) A non-negative scalar on how intrusive
(disruptive) the reporter to the user.}

\item{target}{(character vector) Specifies where progression updates are
rendered.}

\item{inputs}{(named list) Specifies from what sources the MS Windows
progress elements 'title' and 'label' should be updated. Valid sources are
\code{"message"}, \code{"sticky_message"} and \code{"non_sticky_message"}, where
\code{"message"} is short for \code{c("non_sticky_message", "sticky_message")}. For
example, \code{inputs = list(title = "sticky_message", label = "message")}
will update the 'title' component from sticky messages only,
whereas the 'label' component is updated using any message.}

\item{\ldots}{Additional arguments passed to \code{\link[tcltk:tkProgressBar]{tcltk::tkProgressBar()}}
and \code{\link[=make_progression_handler]{make_progression_handler()}}.}
}
\value{
A function of class \code{progression_handler} that takes a
\link{progression} condition as its first and only argument.
}
\description{
A progression handler for \code{\link[tcltk:tkProgressBar]{tcltk::tkProgressBar()}}.
}
\section{Requirements}{

This progression handler requires the \pkg{tcltk} package and that the
current R session supports Tcl/Tk (\code{capabilities("tcltk")}).
}

\examples{
if (capabilities("tcltk") && requireNamespace("tcltk", quietly = TRUE)) {

  handlers("tkprogressbar")
  with_progress({ y <- slow_sum_p(1:10) })
  print(y)
  
}
}
